(*  title     : An encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)

section \<open>Infinite Intervals\<close>

theory InfiniteInterval
 imports
   Interval 
begin
(*
sledgehammer_params [minimize=true,preplay_timeout=10,timeout=60,verbose=true,
                    provers="cvc4 z3 e spass vampire " ]
*)

text \<open>
 An infinite interval is a mapping from the natural numbers to a particular type.
 This is similar as the theory \<open>Omega_Words_Fun\<close> of the Isabelle/HOL distribution.
 The difference is that our version has no empty (no symbols) word. This is needed
 as an finite interval has at least one state. So we have to adapt the  definition of 
  \<open>conc\<close> and \<open>upt\<close>. We also define the usual \<open>isuffix\<close>, \<open>iprefix\<close> and \<open>subinterval\<close> 
 on infinite intervals. 
\<close>

subsection \<open>Definitions\<close>

type_synonym 
 'a infinterval = "nat \<Rightarrow> 'a"

type_synonym
  infiniteindex = "nat infinterval"

definition 
 conc :: "['a interval, 'a infinterval] \<Rightarrow> 'a infinterval" 
where " conc w x = (\<lambda>n. if n \<le> intlen w then nth w n else x (n  - intlen w -1))"


definition
 isuffix :: "[nat, 'a infinterval] \<Rightarrow> 'a infinterval"
 where "isuffix k x = (\<lambda>n. x (k+n))"

definition 
 subinterval :: "'a infinterval \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> 'a interval"
 where
  "subinterval w i j = map w [i..\<le>j]"

definition
 iprefix :: "nat \<Rightarrow> 'a infinterval \<Rightarrow> 'a interval"
 where
  "iprefix n w \<equiv> subinterval w 0 n"

definition infinite_index_sequence :: " nat \<Rightarrow> infiniteindex \<Rightarrow> bool" where
  "infinite_index_sequence x idx \<equiv> (idx 0 = x) \<and> (\<forall> n.  idx n < idx (Suc n))"

subsection \<open>Lemmas\<close>

subsubsection \<open>isuffix\<close>

lemma isuffix_nth:
 "(isuffix k x) n = x (k+n)"
by (simp add: isuffix_def)

lemma isuffix_0:
 "isuffix 0 x = x"
by (simp add: isuffix_def)

lemma isuffix_isuffix:
 "(isuffix m (isuffix n x)) = isuffix (n+m) x"
by (rule ext) (simp add: isuffix_def add.assoc)

subsubsection \<open>iprefix\<close>

lemma iprefix_0:
 "(iprefix 0 x) = \<langle>(x 0)\<rangle>"
by (simp add: iprefix_def subinterval_def)

lemma iprefix_nth:
assumes "k\<le>m"
shows   "    (nth (iprefix m x) k) = (x k)"
using assms
by (simp add: interval_nth_map iprefix_def subinterval_def upt_nth)

lemma iprefix_length:
 "intlen (iprefix n x) = n"
by (simp add:  iprefix_def subinterval_def upt_length)


subsubsection \<open>subinterval\<close>

lemma subinterval_length:
 "intlen (subinterval x i j)  = j -i"
by (simp add: subinterval_def upt_length)

lemma subinterval_nth:
assumes "i+k \<le> j"
shows   "  nth (subinterval x i j) k = x (i+k)"
unfolding subinterval_def
using assms by (simp add: interval_nth_map upt_nth)

lemma iprefix_isuffix:
 "iprefix n (isuffix k x) = subinterval x k (n+k) "
proof -
 have 0: "iprefix n (isuffix k x) = map (\<lambda>n. x (k + n)) [0..\<le>n]"
   by (simp add: iprefix_def isuffix_def subinterval_def)
 have 1: " [k..\<le>n+k] =  (map (\<lambda>i. i+k) [0..\<le>n]) "
 using map_add_upt by simp
 hence 2: "map x [k..\<le>n+k] =  map x (map (\<lambda>i. i+k) [0..\<le>n])"
 by simp
 have 3: "map x (map (\<lambda>i. i+k) [0..\<le>n]) = map (x \<circ> (\<lambda>i. i+k)) [0..\<le>n] "
 by simp
 have 4: "(x \<circ> (\<lambda>i. i+k)) = (\<lambda>n. x (k + n)) " by (metis add.commute comp_apply)
 hence 5: "map (x \<circ> (\<lambda>i. i+k)) [0..\<le>n] = map (\<lambda>n. x (k + n)) [0..\<le>n] "
 by simp
 have 6: " subinterval x k (n+k) = map x [k..\<le>n+k] "
  by (simp add: subinterval_def)
 from 0 2 3 5 6 show ?thesis by auto
qed

lemma subinterval_sub_isuffix:
 assumes "i < j"
 shows   " (subinterval xs (i+k) (j+k)) = (subinterval (isuffix k xs) i j)"
proof -
 have 1: " (subinterval xs (i+k) (j+k)) = 
           iprefix (j-i) (isuffix (i+k) xs) "
  by (simp add: iprefix_isuffix assms less_imp_le_nat)
 have 2: "iprefix (j-i) (isuffix (i+k) xs) =
          iprefix (j-i) (isuffix (i) (isuffix k xs))"
 by (simp add: isuffix_isuffix add.commute)
 have 3: " iprefix (j-i) (isuffix (i) (isuffix k xs)) =
           (subinterval (isuffix k xs) i j)"
   by (simp add: iprefix_isuffix assms less_imp_le_nat)
 from 1 2 3 show ?thesis by auto
qed           

lemma subinterval_sub_isuffix_iidx:
assumes "infinite_index_sequence 0 lsk \<and> n>0"
shows   " (subinterval \<sigma> ((lsk i) +n) ((lsk (Suc i))+n)) =
   (subinterval (isuffix n \<sigma>) (lsk i) (lsk (Suc i)))"
using assms by (simp add: infinite_index_sequence_def subinterval_sub_isuffix)

lemma interval_pref_ipref_3_intlen:
 " intlen (prefix i (iprefix (i+k) xs)) = intlen (iprefix i xs)"
by (simp add: iprefix_length)

lemma interval_pref_ipref_3_nth:
 " (nth (prefix i (iprefix (i+k) xs)) m) = (nth (iprefix i xs) m)"
proof -
obtain nn :: "'a interval \<Rightarrow> 'a interval \<Rightarrow> nat" where
"\<forall>i ia. (i \<noteq> ia \<or> intlen i = intlen ia \<and> 
   (\<forall>n. \<not> n \<le> intlen i \<or> nth i n = nth ia n)) \<and> 
        (i = ia \<or> intlen i \<noteq> intlen ia \<or> nn ia i \<le> intlen i \<and> nth i (nn ia i) \<noteq> nth ia (nn ia i))"
by (metis (no_types) interval_eq_nth_eq) (* 45 ms *)
moreover
{ assume "nth (iprefix i xs) (nn (prefix i (iprefix (i + k) xs)) (iprefix i xs)) \<noteq> 
          nth (prefix i (iprefix (i + k) xs)) (nn (prefix i (iprefix (i + k) xs)) (iprefix i xs))"
  have "\<not> nn (prefix i (iprefix (i + k) xs)) (iprefix i xs) \<le> intlen (iprefix i xs) \<or> 
         nth (iprefix i xs) (nn (prefix i (iprefix (i + k) xs)) (iprefix i xs)) = 
         nth (prefix i (iprefix (i + k) xs)) (nn (prefix i (iprefix (i + k) xs)) (iprefix i xs))"
  by (simp add: iprefix_length iprefix_nth) (* 4 ms *) }
ultimately show ?thesis
by (metis interval_pref_ipref_3_intlen) (* 23 ms *)
qed


lemma interval_pref_ipref_3 [simp]:
    " (prefix i (iprefix (i+k) xs)) = iprefix i xs"
by (meson interval_eq_nth_eq interval_pref_ipref_3_intlen interval_pref_ipref_3_nth)

lemma interval_iprefix_isuffix_swap_intlen:
 "intlen (iprefix ia (isuffix i xs)) = intlen (suffix i (iprefix (ia+i) xs))"
by (simp add: iprefix_length)

lemma interval_iprefix_isuffix_swap_nth:
 assumes "m\<le>ia"
  shows  " (nth (iprefix ia (isuffix i xs)) m) = (nth (suffix i (iprefix (ia+i) xs)) m)"
using assms by (simp add: iprefix_length iprefix_nth isuffix_def)

lemma interval_iprefix_isuffix_swap:
 " iprefix ia (isuffix i xs) = suffix i (iprefix (ia+i) xs)"
 by (simp add: interval_eq_nth_eq interval_iprefix_isuffix_swap_nth iprefix_length)


subsubsection \<open>Conc\<close>

lemma conc_empty_zero:
 "(conc \<langle>s\<rangle> x) 0 = s  "
unfolding conc_def by auto

lemma conc_empty_suc:
 "(conc \<langle>s\<rangle> x) (Suc i) = x i"
unfolding conc_def by auto

lemma conc_conc:
 "conc x (conc y w) = conc (x \<ominus> y) w" (is "?lhs = ?rhs")
proof
 fix n
 have x: "n \<le> intlen x \<longrightarrow> ?lhs n = ?rhs n"
   by (simp add: conc_def interval_intapp_nth)
 have y: "n > intlen x \<and> n \<le> (intlen x) + (intlen y) \<longrightarrow> ?lhs n = ?rhs n"
   by (simp add: conc_def interval_intapp_nth) arith
 have w: "n > (intlen x) + (intlen y) \<longrightarrow> ?lhs n = ?rhs n"
   by (simp add: conc_def interval_intapp_nth) arith
 from x y w show "?lhs n = ?rhs n" using not_less by blast
qed

lemma conc_iprefix_isuffix_help:
 " x a = conc (iprefix n x) (isuffix (Suc n) x) a"
proof (induct n )
case 0
then show ?case 
by (simp add: conc_def iprefix_length iprefix_nth isuffix_nth)
next
case (Suc n)
then show ?case 
  by (simp add: conc_def iprefix_length iprefix_nth)
  (metis isuffix_nth le_SucI not_less_eq_eq ordered_cancel_comm_monoid_diff_class.add_diff_inverse)
qed

lemma conc_iprefix_isuffix:
 " x = conc (iprefix n x) (isuffix (Suc n) x)"

proof 
  (rule ext) 
  show "\<And>xa. x xa = conc (iprefix n x) (isuffix (Suc n) x) xa " 
  by (simp add: conc_iprefix_isuffix_help)
qed


subsubsection \<open>Infinite index sequence\<close>

lemma iidx_1:
 " l = (conc \<langle>(l 0)\<rangle> (\<lambda>x. (l (x+1)))) 
  " (is "?lhs = ?rhs")
proof
 fix n
 have x: "n \<le> intlen \<langle>(l 0)\<rangle> \<longrightarrow> ?lhs n = ?rhs n"
 by (simp add: conc_def interval_intapp_nth)
 have ls: "n > intlen \<langle>(l 0)\<rangle> \<longrightarrow> ?lhs n = ?rhs n"
 by (simp add: conc_def interval_intapp_nth)
 from x ls show "?lhs n = ?rhs n" by auto
qed

lemma iidx_2:
 "infinite_index_sequence 0 (conc \<langle>(l 0)\<rangle> (\<lambda>x. (l (x+1)))) =
  ( (l 0) = 0 \<and> (l 0) < (l 1) \<and> 
  infinite_index_sequence (l 1) (\<lambda>x. (l (x+1))))"
by (auto simp add: infinite_index_sequence_def conc_def)
   (metis Suc_diff_Suc add_diff_cancel_left' gr0I plus_1_eq_Suc zero_less_diff)

lemma iidx_less_plus:
 assumes " infinite_index_sequence n ls"
 shows   " (ls i) < (ls (Suc (i+k))) "
using assms 
by (simp add: infinite_index_sequence_def lift_Suc_mono_less)

lemma iidx_greater:
 assumes "infinite_index_sequence n ls"
 shows   " i>0 \<longrightarrow> n < ls i"
using assms 
proof (induct i)
case 0
then show ?case by simp
next
case (Suc i)
then show ?case by (metis iidx_less_plus infinite_index_sequence_def less_imp_Suc_add)
qed

lemma iidx_3:
 assumes "infinite_index_sequence n ls"
 shows   " infinite_index_sequence 0 ((shiftm n)\<circ>ls)  "
using assms 
by (simp add: infinite_index_sequence_def shiftm_def)
   (metis One_nat_def add_Suc assms diff_less_mono iidx_less_plus
     less_le neq0_conv plus_1_eq_Suc zero_less_diff)

lemma iidx_4:
 " (infinite_index_sequence (x+k) ((shift k)\<circ>ls))=
    infinite_index_sequence x ls "
by (simp add: shift_def infinite_index_sequence_def)


lemma iidx_5:
 assumes  " (infinite_index_sequence k (lsk) \<and> ls = (shiftm k)\<circ>lsk)"
 shows    "infinite_index_sequence 0 ls"
using assms 
by (simp add: infinite_index_sequence_def shiftm_def)    
   (metis add_less_same_cancel1 diff_less_mono lift_Suc_mono_less_iff not_add_less1 not_le_imp_less)

lemma iidx_ext:
 " ((xs :: infiniteindex ) = ys) = (\<forall>i. xs i = ys i)"
by auto

lemma iidx_6:
 "(infinite_index_sequence k lsk \<and> lsk = (shift k)\<circ>ls \<and> infinite_index_sequence 0 ls ) =
  (infinite_index_sequence k lsk \<and> ls = (shiftm k)\<circ>lsk \<and> infinite_index_sequence 0 ls)"
proof (simp add: iidx_ext infinite_index_sequence_def shift_def shiftm_def)
  show "(lsk 0 = k \<and> (\<forall>n. lsk n < lsk (Suc n)) \<and> 
                     (\<forall>i. lsk i = ls i + k) \<and> ls 0 = 0 \<and> (\<forall>n. ls n < ls (Suc n))) =
        (lsk 0 = k \<and> (\<forall>n. lsk n < lsk (Suc n)) \<and> 
                     (\<forall>i. ls i = lsk i - k) \<and> ls 0 = 0 \<and> (\<forall>n. ls n < ls (Suc n)))  "
   proof 
    show "lsk 0 = k \<and> (\<forall>n. lsk n < lsk (Suc n)) \<and> 
                      (\<forall>i. lsk i = ls i + k) \<and> ls 0 = 0 \<and> (\<forall>n. ls n < ls (Suc n)) \<Longrightarrow>
          lsk 0 = k \<and> (\<forall>n. lsk n < lsk (Suc n)) \<and> 
                      (\<forall>i. ls i = lsk i - k) \<and> ls 0 = 0 \<and> (\<forall>n. ls n < ls (Suc n))"
    by auto 
    show "lsk 0 = k \<and> (\<forall>n. lsk n < lsk (Suc n)) \<and> 
                      (\<forall>i. ls i = lsk i - k) \<and> ls 0 = 0 \<and> (\<forall>n. ls n < ls (Suc n)) \<Longrightarrow>
          lsk 0 = k \<and> (\<forall>n. lsk n < lsk (Suc n)) \<and> 
                      (\<forall>i. lsk i = ls i + k) \<and> ls 0 = 0 \<and> (\<forall>n. ls n < ls (Suc n))"
    by (metis add.commute add_diff_inverse_nat add_less_same_cancel1 lift_Suc_mono_less_iff
         not_add_less1) 
   qed
qed

lemma iidx_7:
 " (infinite_index_sequence k lsk \<and> ls= (shiftm k) \<circ> lsk) =
   (infinite_index_sequence k lsk \<and> ls= (shiftm k) \<circ> lsk \<and>
    infinite_index_sequence 0 ls)"
using iidx_5 by blast


lemma iidx_8:
 " ( lsk = (shift k) \<circ> ls \<and> infinite_index_sequence 0 ls) =
   ( lsk = (shift k) \<circ> ls \<and> infinite_index_sequence k lsk \<and>
     infinite_index_sequence 0 ls )"
by (metis Interval.shift_def add_left_imp_eq diff_is_0_eq' interval_idx_shift_mono 
          le_add_diff_inverse2 mono_def iidx_4 rel_simps(46))

lemma iidx_0_a:
assumes " infinite_index_sequence 0 l"
 shows  " (l 0 ) = 0 \<and> (l 0) < (l 1) \<and> infinite_index_sequence (l 1) (\<lambda>x. l(x+1))"
using assms by (simp add: infinite_index_sequence_def) metis

lemma iidx_0_b:
assumes " x =0 "
        " x < (l 0) "
        " infinite_index_sequence (l 0) l "
 shows  "  infinite_index_sequence 0 (conc \<langle>x\<rangle> l)"
using assms diff_Suc_less infinite_index_sequence_def  
by (simp add: conc_def lift_Suc_mono_less)

lemma iidx_0:
 "(\<exists> l. infinite_index_sequence 0 l) =
  (\<exists> ls x. x =0 \<and> x< (ls 0) \<and> infinite_index_sequence (ls 0) ls)"
using infinite_index_sequence_def lessI iidx_0_b by metis


end