(*  Title     : A Shallow Encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)

section \<open>Pi operator\<close>   

theory Pi
imports IntervalFilter  UntilSince 
begin
(*
sledgehammer_params [minimize=true,preplay_timeout=10,timeout=60,verbose=true,
                    provers="
  cvc4 z3 e spass vampire
"]
*)

text \<open>
This theory introduces the Pi operator 
\cite{moszkowski15:_applic_tempor_projec_inter_concur, 
      moszkowski17}. The Pi operator is defined in terms of the filter operator introduced in 
IntervalFilter.thy. We prove the soundness of the rules and axiom system. The until operator from 
UntilSince.thy is used as there is a striking similarity of the expressiveness of the until and 
the Pi operator \cite{moszkowski17}. 
\<close>

subsection \<open>Definitions\<close>



definition sfxfilt :: "'a interval \<Rightarrow> ('a:: world) formula \<Rightarrow> 'a interval interval"
 where "sfxfilt xs f = (filter (\<lambda> ys. f ys) (suffixes xs))"


definition pifilt :: " 'a interval \<Rightarrow> ('a:: world) formula \<Rightarrow> 'a interval"
 where "pifilt xs f = (map (\<lambda>s. nth s 0) (sfxfilt xs f))"


definition pi_d :: "('a:: world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula"
where "pi_d F G \<equiv> \<lambda>s. ( (\<exists> i \<le> intlen s. (suffix i s) \<Turnstile> F) \<and> ( (pifilt s F) \<Turnstile> G ) )"

syntax
 "_pi_d"     :: "[lift,lift] \<Rightarrow> lift"        ("(_ \<Pi> _)" [84,84] 83)

syntax (ASCII)
 "_pi_d"     :: "[lift,lift] \<Rightarrow> lift"        ("(_ PI _)" [84,84] 83)

translations
 "_pi_d"    \<rightleftharpoons> "CONST pi_d"


definition upi_d :: "('a:: world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula"
where "upi_d F G \<equiv> LIFT(\<not>(F \<Pi> (\<not> G)))"


syntax
 "_upi_d"     :: "[lift,lift] \<Rightarrow> lift"        ("(_ \<Pi>\<^sup>u _)" [84,84] 83)

syntax (ASCII)
 "_upi_d"     :: "[lift,lift] \<Rightarrow> lift"        ("(_ UPI _)" [84,84] 83)

translations
 "_upi_d"    \<rightleftharpoons> "CONST upi_d"


subsection \<open>Semantic Lemmas\<close>

lemma sfxfilter_help:
 "(\<exists> ys \<in> set (suffixes xs) . f ys) = (\<exists> i \<le> intlen xs. f (suffix i xs))"
using set_suffixes_sfx by auto

lemma pifiltinit_help:
 "(\<exists> y \<in> set (xs). w \<langle>y\<rangle>) = (\<exists> i \<le> intlen xs. w \<langle>nth xs i\<rangle>)"
by (metis interval_nth_and_set)

lemma sfxfilt_state:
 " sfxfilt \<langle>x\<rangle> f = \<langle>\<langle>x\<rangle>\<rangle>"
by (auto simp:sfxfilt_def)

lemma pifilt_state:
 "pifilt \<langle>x\<rangle> f = \<langle>x\<rangle>"
by (auto simp: pifilt_def sfxfilt_state)

lemma sfxfilt_cons:
shows "(sfxfilt (x\<odot>xs) f) =
  (if (\<exists>i\<le>intlen xs. f (suffix i xs)) then 
               (if f (x\<odot>xs) then 
                   (x\<odot>xs)\<odot> (sfxfilt (xs) f)
                else   (sfxfilt xs f) )
              else \<langle>x\<odot>xs\<rangle>)"
proof -
 have 1: "(sfxfilt (x\<odot>xs) f)  = 
          filter (\<lambda>ys. f ys) (suffixes (x\<odot>xs))
         " 
    by (simp add: sfxfilt_def)
 have 2: "suffixes (x\<odot>xs) = (x\<odot>xs) \<odot> suffixes xs"
   by simp
 have 3: "filter (\<lambda>ys. f ys) (suffixes (x\<odot>xs)) =
          (if (\<exists> ys \<in> set (suffixes xs). f ys) then
            (if f (x\<odot>xs) then (x\<odot>xs)\<odot> (filter (\<lambda>ys. f ys) (suffixes (xs)))  
                          else (filter (\<lambda>ys. f ys) (suffixes (xs))) )
           else \<langle>x\<odot>xs\<rangle>)
         "
   by auto
 have 4: "(if (\<exists> ys \<in> set (suffixes xs). f ys) then
            (if f (x\<odot>xs) then (x\<odot>xs)\<odot> (filter (\<lambda>ys. f ys) (suffixes (xs)))  
                          else (filter (\<lambda>ys. f ys) (suffixes (xs))) )
           else \<langle>x\<odot>xs\<rangle>) =
           (if (\<exists>i\<le>intlen xs. f (suffix i xs)) then 
               (if f (x\<odot>xs) then 
                   (x\<odot>xs)\<odot> (sfxfilt (xs) f)
                else   (sfxfilt xs f) )
              else \<langle>x\<odot>xs\<rangle>)
         " 
    by (auto simp add: sfxfilt_def set_suffixes_sfx)
  show ?thesis by (simp add: "1" "4")   
qed

lemma pifilt_cons:
shows "(pifilt (x\<odot>xs) f) =
  (if (\<exists>i\<le>intlen xs. f (suffix i xs)) then 
               (if f (x\<odot>xs) then 
                   (x)\<odot> (pifilt (xs) f)
                else   (pifilt xs f) )
              else \<langle>x\<rangle>)"
proof -
 have 1: "(pifilt (x\<odot>xs) f) = map (\<lambda>s. (nth s 0)) (sfxfilt (x\<odot>xs) f)"
   by (simp add: pifilt_def) 
 have 2: "map (\<lambda>s. (nth s 0)) (sfxfilt (x\<odot>xs) f) =
          (if (\<exists>i\<le>intlen xs. f (suffix i xs)) then 
               (if f (x\<odot>xs) then 
                   map (\<lambda>s. (nth s 0))((x\<odot>xs)\<odot> (sfxfilt (xs) f))
                else  map (\<lambda>s. (nth s 0)) (sfxfilt xs f) )
              else map (\<lambda>s. (nth s 0)) \<langle>x\<odot>xs\<rangle>)
         "
    using "1" 
    by (metis sfxfilt_cons)
 have 3: " map (\<lambda>s. (nth s 0))((x\<odot>xs)\<odot> (sfxfilt (xs) f)) =
           (x)\<odot> (pifilt (xs) f)
         " 
    by (simp add: pifilt_def)
 have 4: "map (\<lambda>s. (nth s 0)) \<langle>x\<odot>xs\<rangle> = \<langle>x\<rangle>" 
    by simp
 have 5: "map (\<lambda>s. (nth s 0)) (sfxfilt xs f) = (pifilt (xs) f)"
    by (simp add: pifilt_def)
 have 6: "(if (\<exists>i\<le>intlen xs. f (suffix i xs)) then 
               (if f (x\<odot>xs) then 
                   map (\<lambda>s. (nth s 0))((x\<odot>xs)\<odot> (sfxfilt (xs) f))
                else  map (\<lambda>s. (nth s 0)) (sfxfilt xs f) )
              else map (\<lambda>s. (nth s 0)) \<langle>x\<odot>xs\<rangle>) =
           (if (\<exists>i\<le>intlen xs. f (suffix i xs)) then 
               (if f (x\<odot>xs) then 
                   (x)\<odot> (pifilt (xs) f)
                else   (pifilt xs f) )
              else \<langle>x\<rangle>)"
    using "5" by auto
  show ?thesis by (simp add: "1" "2" "6")
qed   

lemma sfxfilt_nth_cons:
 "nth (sfxfilt (x\<odot>xs) f) j =
             (if (\<exists>i\<le>intlen xs. f (suffix i xs)) then 
               (if f (x\<odot>xs) then 
                 (if j =0 then  (x\<odot>xs) else nth (sfxfilt (xs) f) (j-1)) 
                else  nth (sfxfilt (xs) f) j)
              else (x\<odot>xs))"
by (metis Interval.nth.simps(1) add.commute add.left_neutral add_diff_inverse_nat diff_0_eq_0 
    interval_nth_cons_a interval_nth_zero sfxfilt_cons)

lemma pifilt_nth_cons:
 " nth (pifilt (x \<odot> xs) f) i = 
             (if (\<exists>i\<le>intlen xs. f (suffix i xs)) then 
               (if f (x\<odot>xs) then 
                 (if i =0 then  x else nth (pifilt (xs) f) (i-1)) 
                else  nth (pifilt (xs) f) i)
              else x)"
by (metis Interval.nth.simps(1) One_nat_def Suc_pred interval_nth_Suc interval_nth_zero not_gr0
     pifilt_cons)
 
lemma sfxfilt_nth:
assumes "(\<exists> i \<le> intlen \<sigma>. (suffix i \<sigma>) \<Turnstile> f)"
        " i\<le>intlen (sfxfilt \<sigma> f)"
shows   " (nth (sfxfilt \<sigma> f) i) \<Turnstile> f"
using assms in_set_suffixes osfx_suffix sfxfilter_nth
by (simp add: sfxfilt_def, blast)


lemma pifilt_exists:
 assumes "(\<exists>i\<le>intlen \<sigma>. f (suffix i \<sigma>) )"
 shows   "(\<exists>i \<le> intlen(sfxfilt \<sigma> f). (nth (sfxfilt \<sigma> f) i) \<Turnstile> f)"
using assms 
using sfxfilt_nth by blast 

lemma sfxfilt_pifilt_intlen:
assumes "(\<exists> i \<le> intlen \<sigma>. (suffix i \<sigma>) \<Turnstile> f)"
shows   "intlen (pifilt \<sigma> f) = intlen (map (\<lambda>s. nth s 0) (sfxfilt \<sigma> f))  "
using assms by (simp add: pifilt_def)

lemma sfxfilt_pifilt_nth:
assumes "(\<exists> i \<le> intlen \<sigma>. (suffix i \<sigma>) \<Turnstile> f)"
        "j\<le> intlen (pifilt \<sigma> f)"
shows   " nth (pifilt \<sigma> f) j = nth (map (\<lambda>s. nth s 0) (sfxfilt \<sigma> f)) j  "
using assms
by (simp add: pifilt_def)

lemma sfxfilt_pifilt:
assumes "(\<exists> i \<le> intlen \<sigma>. (suffix i \<sigma>) \<Turnstile> f)"
shows   "(map (\<lambda>s. nth s 0) (sfxfilt \<sigma> f)) = pifilt \<sigma> f"
using assms by (simp add: pifilt_def)

lemma sfxfilt_intlen_bound:
assumes "(\<exists>i\<le>intlen xs. f (suffix i xs))"
shows   "intlen (sfxfilt xs f) \<le> intlen xs "
using assms 
by (simp add: sfxfilt_def)
   (metis IntervalFilter.length_filter_le intlen_suffixes)

lemma pifilt_intlen_bound:
assumes "(\<exists>i\<le>intlen \<sigma>. f (suffix i \<sigma>))"
shows   "intlen (pifilt \<sigma> f) \<le> intlen \<sigma> "
using assms by (simp add: pifilt_def sfxfilt_intlen_bound)

lemma sfxfilt_intlen_nth_bound:
assumes "(\<exists>i\<le>intlen \<sigma>. f (suffix i \<sigma>))"
        " j\<le> intlen (sfxfilt \<sigma> f)"
shows   "intlen (nth (sfxfilt \<sigma> f) j) \<le> intlen \<sigma> "
using assms 
by (simp add: sfxfilt_def sfxfilter_help sfxfilter_nth_bound)

lemma sfxfilt_pifilt_nth_suffix:
assumes "(\<exists> i \<le> intlen \<sigma>. (suffix i \<sigma>) \<Turnstile> f)"
        "j\<le> intlen (pifilt \<sigma> f)"
shows   "  nth (sfxfilt \<sigma> f) j =  suffix (intlen \<sigma> - (intlen( nth (sfxfilt \<sigma> f) j))) \<sigma>  "
proof -
 have 1: "nth (sfxfilt \<sigma> f) j = nth (filter f (suffixes \<sigma>)) j" 
   by (simp add: sfxfilt_def)
 have 2: "suffix (intlen \<sigma> - intlen (nth (filter f (suffixes \<sigma>)) j)) \<sigma> =
          suffix (intlen \<sigma> - (intlen(nth (sfxfilt \<sigma> f) j))) \<sigma>" 
   by (simp add: sfxfilt_def)
 have 3: "j \<le> intlen (filter f (suffixes \<sigma>))" 
   using assms by (simp add: pifilt_def sfxfilt_def)
 have 4: "(\<exists>i\<le>intlen \<sigma>. f (suffix i \<sigma>))" 
    using assms by auto
 have 5: "(\<exists> ys \<in> set (suffixes \<sigma>). f ys)"
   using "4" 
   using in_set_suffixes osfx_suffix by blast 
 have 6: "nth (filter f (suffixes \<sigma>)) j =
          suffix (intlen \<sigma> - intlen (nth (filter f (suffixes \<sigma>)) j)) \<sigma>
         " 
   using "3" "5" sfxfilter_nth_suffix[of "\<sigma>" "f" "j"] by blast
 show ?thesis using "1" "6" by auto
qed

lemma pifilt_nth:
 assumes "(\<exists>i\<le>intlen \<sigma>. f (suffix i \<sigma>))"
         "i \<le> intlen (pifilt \<sigma> f)"
shows    "(\<exists> k \<le> intlen \<sigma>. nth (pifilt \<sigma> f) i = nth \<sigma> k)"
using assms sfxfilt_pifilt_nth_suffix[of "\<sigma>" "f" "i"] 
by (simp add: pifilt_def) 
   (metis diff_le_self interval_intfirst_suffix interval_intlen_gr_zero interval_nth_map 
    interval_suffix_zero)

lemma sfxfilt_intlen_imp:
 assumes "(\<exists>i\<le>intlen \<sigma>. f (suffix i \<sigma>) \<and> g (suffix i \<sigma>))"
 shows   " intlen (sfxfilt \<sigma> (LIFT(f \<and> g))) \<le> intlen (sfxfilt \<sigma> f)"
proof -
 have 1: "intlen (sfxfilt \<sigma> (LIFT(f \<and> g))) = 
          intlen (filter (\<lambda>ys. f ys \<and> g ys) (suffixes \<sigma>))" 
    by (simp add: sfxfilt_def)
 have 2: "intlen (filter f (suffixes \<sigma>)) = intlen (sfxfilt \<sigma> f) "
    by (simp add:  sfxfilt_def)
 have 3: "\<exists> x \<in> set (suffixes \<sigma>). f x \<and> g x" 
    using assms by auto
 have 4: "intlen (filter (\<lambda>x. f x \<and> g x) (suffixes \<sigma>)) \<le> intlen (filter f (suffixes \<sigma>))" 
    using "3" filter_intlen_imp[of "suffixes \<sigma>" "f" "g"] by auto
 show ?thesis using "1" "2" "4" by auto
qed

lemma pifilt_intlen_imp:
 assumes "(\<exists>i\<le>intlen \<sigma>. f (suffix i \<sigma>) \<and> g (suffix i \<sigma>))"
 shows   " intlen (pifilt \<sigma> (LIFT(f \<and> g))) \<le> intlen (pifilt \<sigma> f)"
using assms
by (simp add: sfxfilt_intlen_imp pifilt_def)

lemma interval_set_sfxfilt [simp]:
assumes "(\<exists> i \<le> intlen xs. f (suffix i xs))"
shows "(set (sfxfilt xs f)) = {ys. ys \<in> set (suffixes xs) \<and> f (ys)}"
using assms 
proof -
 have 1: "set (sfxfilt xs f) = set (filter f (suffixes xs))"
    by (simp add: sfxfilt_def) 
 have 2: "\<exists> x \<in> set (suffixes xs). f x" 
    using assms using in_set_suffixes osfx_suffix by blast
 have 3: "set (filter f (suffixes xs)) = 
          {ys. ys \<in> set (suffixes xs) \<and> f ys}
         " 
   using "2" set_filter[of "suffixes xs" "f"] by auto
 show ?thesis by (simp add: "1" "3")
qed 


lemma interval_subset_sfxfilt [simp]:
assumes "(\<exists> i \<le> intlen xs. f (suffix i xs))"
shows "(set (sfxfilt xs f)) \<le> (set (sfxfilt xs (LIFT(f \<or> g))))"
proof -
 have 1: "\<exists> x \<in> set (suffixes xs). f x"
   using assms using in_set_suffixes osfx_suffix by blast 
 have 2: "set (sfxfilt xs f) = set (filter f (suffixes xs))"
     by (simp add: sfxfilt_def)
 have 3: " set (sfxfilt xs (LIFT(f \<or> g))) = set (filter (\<lambda>x. f x \<or> g x) (suffixes xs))"
   by (simp add: sfxfilt_def)
 have 4: "set (filter f (suffixes xs)) \<le> set (filter (\<lambda>x. f x \<or> g x) (suffixes xs))"
   using "1"  subset_filter[of "suffixes xs" "f" "g"] by auto
 show ?thesis using "2" "3" "4" by blast
qed


lemma interval_set_pifilt [simp]:
assumes "(\<exists> i \<le> intlen xs. f (suffix i xs))"
shows "(set (pifilt xs f)) = {(nth ys 0)| ys. ys \<in> set(suffixes xs) \<and> f (ys)}"
proof -
 have 1: "(set (pifilt xs f)) = (set (map (\<lambda>s. nth s 0) (sfxfilt xs f)))"
    by (simp add: pifilt_def)
 have 2: "(set (map (\<lambda>s. nth s 0) (sfxfilt xs f))) =
          {(nth ys 0) | ys. ys \<in> set (sfxfilt xs f)}"
    by (induction xs) auto
 have 3: "{(nth ys 0) | ys. ys \<in> set (sfxfilt xs f)} =
          {(nth ys 0)| ys. ys \<in> set(suffixes xs) \<and> f (ys)}
         " 
    using assms by auto
 show ?thesis using "1" "2" "3" by auto
qed

lemma interval_nth_sfxfilt_in_set:
"x \<in> set (sfxfilt \<sigma> (LIFT(f \<or> g))) =
 (\<exists> i \<le> intlen (sfxfilt \<sigma> (LIFT(f \<or> g))). x =(nth (sfxfilt \<sigma> (LIFT(f \<or> g))) i))"
by (metis interval_nth_and_set)


lemma sfxfilt_nth_or:
assumes "(\<exists> i \<le> intlen \<sigma>. (suffix i \<sigma>) \<Turnstile> f)"
shows   "(\<exists> i \<le> intlen (sfxfilt \<sigma> (LIFT(f \<or> g))). (nth (sfxfilt \<sigma> (LIFT(f \<or> g)) ) i) \<Turnstile> f)"
proof -
 have 1: "\<exists> x \<in> set (suffixes \<sigma>). f x" 
   using assms by auto
 have 2: "sfxfilt \<sigma> (LIFT(f \<or> g)) = filter (\<lambda> x. f x \<or> g x) (suffixes \<sigma>) "
   by (simp add: sfxfilt_def)
 have 3: "\<exists> x \<in> set(filter (\<lambda> x. f x \<or> g x) (suffixes \<sigma>)). f x "
   using "1" filter_nth_or[of "suffixes \<sigma>" "f" "g"] by auto
 from "2" "3" show ?thesis by (metis interval_nth_sfxfilt_in_set)
qed

lemma NotPiFalse:
 "\<sigma> \<Turnstile> \<not> ((#False) \<Pi> f)"
by (simp add: pi_d_def)

lemma pifilt_true:
 " pifilt \<sigma> (LIFT(#True)) = \<sigma> "
by (simp add: pifilt_def sfxfilt_def IntervalFilter.filter_True)

lemma pifilt_init_state:
 "(pifilt \<langle>x\<rangle> (LIFT(init w))) = \<langle>x\<rangle>"
by ( auto simp add: pifilt_def sfxfilt_def)

lemma  pifilt_init_cons:
 "(pifilt (x\<odot>xs) (LIFT(init w))) =
  (if (\<exists>i\<le>intlen xs. w \<langle>nth xs i\<rangle>)  then
           (if  w \<langle>x\<rangle> then x \<odot> (pifilt xs (LIFT(init w)))
                           else (pifilt xs (LIFT(init w))))
         else \<langle>x\<rangle>)"
by (simp add: pifilt_def sfxfilt_def init_defs)
   (metis interval_nth_and_set interval_nth_map intlen_suffixes map_first_suffixes)

lemma PiStatesem: 
 " (\<sigma> \<Turnstile> (init w) \<Pi> f) =
   ((\<exists>i\<le>intlen \<sigma>. w \<langle>nth \<sigma> i\<rangle>) \<and> f (pifilt \<sigma> (LIFT(init w))))"
by (simp add: pi_d_def init_defs) 


subsection \<open>Soundness of Axioms\<close>

subsubsection \<open>PiK\<close>

lemma PiKsem:
 "\<sigma> \<Turnstile> f1 \<Pi>\<^sup>u ( f \<longrightarrow> g) \<longrightarrow> (  f1 \<Pi>\<^sup>u f \<longrightarrow> f1 \<Pi>\<^sup>u g ) "
by (simp add: upi_d_def init_defs pi_d_def) auto

lemma PiK:
 "\<turnstile> f1 \<Pi>\<^sup>u (f \<longrightarrow> g) \<longrightarrow> (  f1 \<Pi>\<^sup>u f \<longrightarrow> f1 \<Pi>\<^sup>u g )"
using PiKsem Valid_def by blast 

subsubsection \<open>PiDc\<close>

lemma PiDcsem:
 " \<sigma> \<Turnstile> f \<Pi> g \<longrightarrow> f \<Pi>\<^sup>u g"
by (simp add: upi_d_def init_defs pi_d_def)

lemma PiDc:
 "\<turnstile> f \<Pi> g \<longrightarrow> f \<Pi>\<^sup>u g"
using PiDcsem Valid_def by blast

subsubsection \<open>PiN\<close>

lemma PiN:
 assumes "\<turnstile> g"
 shows   "\<turnstile> f \<Pi>\<^sup>u g"
using assms by (simp add: Valid_def pi_d_def upi_d_def)

subsubsection \<open>PiTrueEqvDiamond\<close>

lemma PiTrueEqvDiamond:
 "\<turnstile> f \<Pi> #True = \<diamond> f"
by (simp add: Valid_def pi_d_def sometimes_defs)

subsubsection \<open>PiOr\<close>

lemma PiOr:
 "\<turnstile> f \<Pi> (g1 \<or> g2) = (f \<Pi> g1 \<or> f \<Pi> g2)"
by (simp add: Valid_def pi_d_def) blast

subsubsection \<open>UPiFalseEqvBoxNot:\<close>

lemma UPiFalseEqvBoxNot:
 "\<turnstile> f \<Pi>\<^sup>u #False = \<box> (\<not>f)"
by (simp add: Valid_def upi_d_def pi_d_def always_defs)

subsubsection \<open>BoxEqvImpPiEqv\<close>

lemma BoxEqvImpPiEqvsem:
assumes "(\<sigma> \<Turnstile> \<box> (f1 = f2))"
shows   "(\<sigma> \<Turnstile>(f1 \<Pi> g = f2 \<Pi> g))"
proof -
 show   "(\<sigma> \<Turnstile>(f1 \<Pi> g = f2 \<Pi> g))" 
 proof -
  have 1: "\<forall>n\<le>intlen \<sigma>. f1 (suffix n \<sigma>) = f2 (suffix n \<sigma>)" 
   using assms by (simp add: always_defs)
  have 2: "(\<sigma> \<Turnstile>(f1 \<Pi> g)) = ((\<exists>i\<le>intlen \<sigma>. f1 (suffix i \<sigma>)) \<and> g (pifilt \<sigma> f1))"
   by (simp add: pi_d_def)
  have 3: "(\<exists>i\<le>intlen \<sigma>. f1 (suffix i \<sigma>)) = (\<exists>i\<le>intlen \<sigma>. f2 (suffix i \<sigma>))"
    using "1" by blast  
  have 4: " (sfxfilt \<sigma> f1) =  (sfxfilt \<sigma> f2)"
     using "1" 
   proof (induct \<sigma>)
   case (St x)
   then show ?case by (simp add: sfxfilt_state)
   next
   case (Cons x1a \<sigma>)
   then show ?case 
     proof -
       have 41: "(sfxfilt (x1a \<odot> \<sigma>) f1) = 
         (if (\<exists>i\<le>intlen \<sigma>. (suffix i \<sigma>) \<Turnstile> f1) then
           (if ((x1a\<odot>\<sigma>) \<Turnstile> f1) then (x1a\<odot>\<sigma>) \<odot> (sfxfilt \<sigma> f1)
                             else (sfxfilt \<sigma> f1))
         else \<langle>x1a\<odot>\<sigma>\<rangle>)"
        using sfxfilt_cons by blast
       have 42: " sfxfilt \<sigma> f1 = sfxfilt \<sigma> f2"
        by (metis Cons.hyps Cons.prems Suc_le_mono interval_suffix_suc intlen.simps(2) 
            plus_1_eq_Suc)
       have 43: "((x1a\<odot>\<sigma>) \<Turnstile> f1) = ((x1a\<odot>\<sigma>) \<Turnstile> f2)"
        using Cons.prems by auto
       have 44: "(\<exists>i\<le>intlen \<sigma>. (suffix i \<sigma>) \<Turnstile> f1) =
                 (\<exists>i\<le>intlen \<sigma>. (suffix i \<sigma>) \<Turnstile> f2)"
        by (metis  Suc_le_mono interval_suffix_suc intlen.simps(2) local.Cons(2) plus_1_eq_Suc) 
       have 45: "(if (\<exists>i\<le>intlen \<sigma>. (suffix i \<sigma>) \<Turnstile> f1) then
           (if ((x1a\<odot>\<sigma>) \<Turnstile> f1) then (x1a\<odot>\<sigma>) \<odot> (sfxfilt \<sigma> f1)
                             else (sfxfilt \<sigma> f1))
         else \<langle>x1a\<odot>\<sigma>\<rangle>) =
             (if (\<exists>i\<le>intlen \<sigma>. (suffix i \<sigma>) \<Turnstile> f2) then
           (if ((x1a\<odot>\<sigma>) \<Turnstile> f2) then (x1a\<odot>\<sigma>) \<odot> (sfxfilt \<sigma> f2)
                             else (sfxfilt \<sigma> f2))
         else \<langle>x1a\<odot>\<sigma>\<rangle>)" 
        using "42" "43" "44" by auto
       have 46: "(if (\<exists>i\<le>intlen \<sigma>. (suffix i \<sigma>) \<Turnstile> f2) then
           (if ((x1a\<odot>\<sigma>) \<Turnstile> f2) then (x1a\<odot>\<sigma>) \<odot> (sfxfilt \<sigma> f2)
                             else (sfxfilt \<sigma> f2))
         else \<langle>x1a\<odot>\<sigma>\<rangle>) = (sfxfilt (x1a \<odot> \<sigma>) f2)"
        by (simp add: sfxfilt_cons)
       show ?thesis 
       using "41" "45" "46" by presburger 
      qed
   qed
  have 47: "(pifilt \<sigma> f1) =  (pifilt \<sigma> f2)" 
     by (simp add: "4" pifilt_def)
  have 5: "((\<exists>i\<le>intlen \<sigma>. f1 (suffix i \<sigma>)) \<and> g (pifilt \<sigma> f1)) =
           ((\<exists>i\<le>intlen \<sigma>. f2 (suffix i \<sigma>)) \<and> g (pifilt \<sigma> f2))"
   by (simp add: "3" "47")
  show ?thesis by (simp add: "5" pi_d_def)
  qed    
qed

lemma BoxEqvImpPiEqv:
 "\<turnstile> \<box> (f1 = f2) \<longrightarrow> (f1 \<Pi> g = f2 \<Pi> g)"
using BoxEqvImpPiEqvsem by (simp add: Valid_def,auto)      


subsubsection \<open>PiDiamondImpDiamond\<close>

lemma PiDiamondImpDiamondsem:
 "\<sigma> \<Turnstile> f \<Pi> (\<diamond> (init w)) \<longrightarrow> \<diamond> (init w)"
using pifilt_nth by (simp add: Valid_def pi_d_def sometimes_defs init_defs) fastforce

lemma PiDiamondImp:
 "\<turnstile> f \<Pi> (\<diamond> (init w)) \<longrightarrow> \<diamond> (init w)"
 using PiDiamondImpDiamondsem Valid_def by blast


subsubsection \<open>PiAssoc\<close>

                        
lemma PiAssocsem1:
assumes " i \<le> intlen \<sigma>"
        "f (suffix i \<sigma>)"
        "ia \<le> intlen (pifilt \<sigma> f)"
        "w \<langle>nth (pifilt \<sigma> f) ia\<rangle>" 
shows   "\<exists>i\<le>intlen \<sigma>. f (suffix i \<sigma>) \<and> w \<langle>nth (suffix i \<sigma>) 0\<rangle>"
proof -
 have 1: "(nth (pifilt \<sigma> f) ia) = (nth (map (\<lambda>s. nth s 0) (sfxfilt \<sigma> f)) ia)"
   using assms(1) assms(2) assms(3) sfxfilt_pifilt_nth by blast 
 have 2: "(nth (map (\<lambda>s. nth s 0) (sfxfilt \<sigma> f)) ia) =
          (\<lambda>s. nth s 0) (nth (sfxfilt \<sigma> f) ia)" 
  using interval_nth_map  by auto
 have 3: "f  (nth (sfxfilt \<sigma> f) ia)"
    using sfxfilt_nth 
    by (metis assms(1) assms(2) assms(3) interval_intlen_map pifilt_def)
 have 4: "nth (sfxfilt \<sigma> f) ia = suffix (intlen \<sigma> - intlen (nth (sfxfilt \<sigma> f) ia)) \<sigma>"
   using sfxfilt_pifilt_nth_suffix  assms(1) assms(2) assms(3) by blast
 have 5: "w \<langle>nth (suffix (intlen \<sigma> - intlen (nth (sfxfilt \<sigma> f) ia)) \<sigma>) 0\<rangle>  " 
   using "1" "2" "4" assms(4) by auto
 show ?thesis 
 by (metis "3" "4" "5" diff_le_self)
qed
  
lemma PiAssocsem2:
 assumes "i \<le> intlen \<sigma>"
         "f (suffix i \<sigma>)"
         " w \<langle>Interval.nth \<sigma> i\<rangle>"
shows "\<exists>j\<le>intlen (pifilt \<sigma> f). w \<langle>nth (pifilt \<sigma> f) j\<rangle>"
proof -
 have 1: "\<exists>j\<le>intlen (sfxfilt \<sigma> f). f (nth (sfxfilt \<sigma> f) j)"
     using assms pifilt_exists by blast
 have 2: " (LIFT (init w)) (suffix i \<sigma>)"
    by (simp add: assms init_defs)
 have 3: "\<exists>j\<le>intlen (sfxfilt \<sigma> (LIFT(init w))). (LIFT(init w)) (nth (sfxfilt \<sigma> (LIFT(init w))) j)"
   using pifilt_exists "2" assms by blast
 have 4: "(LIFT (f \<and> init w)) (suffix i \<sigma>)"
    by (simp add: assms init_defs)
 have 5: "\<exists>j\<le>intlen (sfxfilt \<sigma> (LIFT(f \<and> init w))). 
           (LIFT(f \<and> init w)) (nth (sfxfilt \<sigma> (LIFT(f \<and> init w))) j) "
   using pifilt_exists "4" assms by blast
 have 6: "\<exists>i\<le>intlen \<sigma>. suffix i \<sigma> \<Turnstile> f \<and> init w"
   using "4" assms by blast
 have 7: "\<exists>j\<le>intlen (sfxfilt \<sigma> (LIFT((f \<and> init w) \<or> (f \<and> \<not> (init w)))  )). 
           (LIFT(f \<and> init w)) (nth (sfxfilt \<sigma> (LIFT((f \<and> init w) \<or> (f \<and> \<not>(init w))) )) j)" 
  using "6" sfxfilt_nth_or[of "\<sigma>" "LIFT(f \<and> init w)" "LIFT(f \<and> \<not>(init w))"] 
      by auto
 have 8: "\<And>\<sigma> . (\<sigma> \<Turnstile> ((f \<and> init w) \<or> (f \<and> \<not> (init w)))) = (\<sigma> \<Turnstile> f)"
    by auto  
 have 9: "(sfxfilt \<sigma> (LIFT((f \<and> init w) \<or> (f \<and> \<not> (init w)))  )) = 
          (sfxfilt \<sigma> f)"
      using "8" by (simp add: sfxfilt_def)
 have 10: "\<exists>j\<le>intlen (sfxfilt \<sigma> f). 
           (LIFT(f \<and> init w)) (nth (sfxfilt \<sigma> f) j)"
   using "7" "9" by auto 
 have 11: "intlen (sfxfilt \<sigma> f) = intlen (pifilt \<sigma> f)" 
     by (simp add: pifilt_def)
 have 12: "\<exists>j\<le>intlen (pifilt \<sigma> f). 
           (LIFT(init w)) (nth (sfxfilt \<sigma> f) j)" 
    using "10" "11" by auto
 from 12 11 show ?thesis 
   by (simp add: init_defs)
      (metis interval_nth_map pifilt_def)
qed    
    
lemma PiAssocsema:
 "((\<exists>i\<le>intlen \<sigma>. f (suffix i \<sigma>)) \<and>
     (\<exists>i\<le>intlen (pifilt \<sigma> f). w \<langle>nth (suffix i (pifilt \<sigma> f)) 0\<rangle>)) =
  (\<exists>i\<le>intlen \<sigma>. f (suffix i \<sigma>) \<and> w \<langle>nth (suffix i \<sigma>) 0\<rangle>) "
using PiAssocsem1 PiAssocsem2 by fastforce

lemma PiAssocsemb:
 "((\<exists>i\<le>intlen \<sigma>. f (suffix i \<sigma>)) \<and>
     (\<exists>i\<le>intlen (pifilt \<sigma> f). (LIFT(init w)) (suffix i (pifilt \<sigma> f)) )) =
  (\<exists>i\<le>intlen \<sigma>. (LIFT(f \<and> init w)) (suffix i \<sigma>)) "
using PiAssocsem1 PiAssocsem2 
by (simp add: init_defs) fastforce

lemma pifilt_state_help:
 "(\<exists> x \<in> set (suffixes xs) . (LIFT(init w)) x) = (\<exists> x \<in> set xs. w (\<langle>x\<rangle>))"
proof (auto simp add: init_defs)
 show "\<And>x. osfx x xs \<Longrightarrow> w \<langle>Interval.nth x 0\<rangle> \<Longrightarrow> \<exists>x\<in>interval.set xs. w \<langle>x\<rangle>" 
 using in_set_suffixes interval_sfx_1 by blast
 show "\<And>x. x \<in> interval.set xs \<Longrightarrow> w \<langle>x\<rangle> \<Longrightarrow> \<exists>x\<in>interval.set (suffixes xs). w \<langle>Interval.nth x 0\<rangle>"
 by (metis in_set_suffixes interval_intfirst_suffix interval_nth_and_set interval_nth_zero_intfirst
      osfx_suffix)
qed


lemma pifilt_init:
 assumes "(\<exists>i\<le>intlen xs. (LIFT(init w)) (suffix i xs))"
 shows   "(pifilt xs (LIFT(init w))) = filter (\<lambda>y. w (\<langle>y\<rangle>)) xs"
using assms
proof (induct xs)
case (St x)
then show ?case 
by (simp add: pifilt_init_state)
next
case (Cons x1a xs)
then show ?case 
  proof -
  have 1: "pifilt (x1a\<odot>xs) (LIFT(init w)) =
           map (\<lambda>xs . (nth xs 0)) (sfxfilt (x1a\<odot>xs) (LIFT(init w)))
          " 
       using sfxfilt_pifilt by (simp add: pifilt_def)
  have 2: "sfxfilt (x1a\<odot>xs) (LIFT(init w)) =
           (filter (\<lambda> ys. (LIFT(init w)) ys) (suffixes (x1a \<odot>xs)))
          " 
       using sfxfilt_def by blast
  have 3: "suffixes (x1a \<odot>xs) = (x1a\<odot>xs)\<odot> (suffixes xs) "
       by simp
  have 4: "(filter (\<lambda> ys. (LIFT(init w)) ys) (suffixes (x1a \<odot>xs))) =
           (if (\<exists> x \<in> set (suffixes xs) . (LIFT(init w)) x) then
                (if (LIFT(init w)) (x1a\<odot>xs) then (x1a\<odot>xs)\<odot> (filter (LIFT(init w)) (suffixes xs)) 
                 else (filter (LIFT(init w)) (suffixes xs)))
            else \<langle>x1a \<odot> xs\<rangle>)
          "  
       by simp
  have 5: " map (\<lambda>xs . (nth xs 0)) (filter (\<lambda> ys. (LIFT(init w)) ys) (suffixes (x1a \<odot>xs))) =
            (if (\<exists> x \<in> set (suffixes xs) . (LIFT(init w)) x) then
                (if (LIFT(init w)) (x1a\<odot>xs) 
                 then (x1a) \<odot> map (\<lambda>xs . (nth xs 0)) (filter (LIFT(init w)) (suffixes xs)) 
                 else map (\<lambda>xs . (nth xs 0)) (filter (LIFT(init w)) (suffixes xs)))
             else \<langle>x1a\<rangle>)
          " 
       by auto
  have 6: "filter (\<lambda>y. w \<langle>y\<rangle>) (x1a \<odot> xs) =
           (if (\<exists> x \<in> set xs. w (\<langle>x\<rangle>)) then 
              (if w (\<langle>x1a\<rangle>) then x1a\<odot> (filter (\<lambda>y. w \<langle>y\<rangle>) xs) else (filter (\<lambda>y. w \<langle>y\<rangle>) xs))
            else \<langle>x1a\<rangle>)
          "
        by simp
  have 61: "(\<exists> x \<in> set (suffixes xs) . (LIFT(init w)) x) =
             (\<exists> x \<in> set xs. w (\<langle>x\<rangle>))
           " 
        by (auto simp: init_defs interval_sfx_1)
           (metis init_defs interval_prefix_zero_intfirst pifilt_state_help)
  have 62: "(LIFT(init w)) (x1a\<odot>xs) = w (\<langle>x1a\<rangle>)" 
    by (simp add: init_defs)
  have 63: " (\<exists> x \<in> set xs. w (\<langle>x\<rangle>)) \<longrightarrow> 
             map (\<lambda>xs . (nth xs 0)) (filter (LIFT(init w)) (suffixes xs)) =
             (filter (\<lambda>y. w \<langle>y\<rangle>) xs)" 
       by (auto simp add: interval_nth_and_set ) 
         (metis Cons.hyps interval.distinct(1) pifilt_cons pifilt_def pifilt_init_cons sfxfilt_def)
  have 7: "(if (\<exists> x \<in> set (suffixes xs) . (LIFT(init w)) x) then
                (if (LIFT(init w)) (x1a\<odot>xs) 
                 then (x1a) \<odot> map (\<lambda>xs . (nth xs 0)) (filter (LIFT(init w)) (suffixes xs)) 
                 else map (\<lambda>xs . (nth xs 0)) (filter (LIFT(init w)) (suffixes xs)))
            else \<langle>x1a\<rangle>) =
           (if (\<exists> x \<in> set xs. w (\<langle>x\<rangle>)) then 
              (if w (\<langle>x1a\<rangle>) then x1a\<odot> (filter (\<lambda>y. w \<langle>y\<rangle>) xs) else (filter (\<lambda>y. w \<langle>y\<rangle>) xs))
            else \<langle>x1a\<rangle>)
          " 
      by (simp add: "61" "62" "63")
     show ?thesis using "1" "2" "5" "7" by auto 
   qed       
qed

lemma pifilt_pifilt :
assumes "(\<exists>i\<le>intlen xs. f (suffix i xs))"
        "(\<exists>i\<le>intlen (pifilt xs f). w \<langle>nth (suffix i (pifilt xs f)) 0\<rangle>)"
shows "(pifilt (pifilt xs f) (LIFT(init w)))  = pifilt xs (LIFT(f \<and> init w))"
proof -
 have 1: "\<exists>i\<le>intlen (pifilt xs f). (LIFT(init w)) (suffix i (pifilt xs f))"
      using assms by (simp add: init_defs)
 have 2: "(pifilt (pifilt xs f) (LIFT(init w))) =
          filter (\<lambda>y. w (\<langle>y\<rangle>)) (pifilt xs f)
         " 
      using "1" pifilt_init[of "(pifilt xs f)" "w" ] by auto 
 have 3: "(pifilt xs f) =
          map (\<lambda>s. nth s 0) (sfxfilt xs f)" 
      by (simp add: assms sfxfilt_pifilt)
 have 4: "(sfxfilt xs f) = filter (\<lambda> ys. f ys) (suffixes xs)"
      using sfxfilt_def by blast
 have 5: " (pifilt xs f) = map (\<lambda>s. nth s 0) (filter (\<lambda> ys. f ys) (suffixes xs))" 
      by (simp add: "3" "4")
 have 6: "filter (\<lambda>y. w (\<langle>y\<rangle>)) (pifilt xs f)  =
          filter (\<lambda>y. w (\<langle>y\<rangle>)) (map (\<lambda>s. nth s 0) (filter (\<lambda> ys. f ys) (suffixes xs)))
         " 
   using "5" by simp
 have 7: "filter (\<lambda>y. w (\<langle>y\<rangle>)) (map (\<lambda>s. nth s 0) (filter (\<lambda> ys. f ys) (suffixes xs))) =
      map (\<lambda>s. nth s 0) (filter ((\<lambda>y. w (\<langle>y\<rangle>))\<circ>(\<lambda>s. nth s 0)) (filter (\<lambda> ys. f ys) (suffixes xs))) " 
   using assms by (metis "3" "4" filter_map in_set_suffixes interval_sfx_1 osfx_suffix)
 have 8: "\<exists>x\<in>interval.set (filter (\<lambda> ys. f ys) (suffixes xs)). ((\<lambda>y. w (\<langle>y\<rangle>))\<circ>(\<lambda>s. nth s 0)) x" 
      using assms "3" "4"  
      by simp_all
         (metis  interval_intlen_map interval_nth_map interval_nth_zero_intfirst nth_set)
 have 9: "\<exists>x\<in>interval.set (suffixes xs). (\<lambda> ys. f ys) x" 
      using assms in_set_suffixes osfx_suffix by blast
 have 10: "\<exists>x\<in>interval.set (suffixes xs). ((\<lambda>y. w (\<langle>y\<rangle>))\<circ>(\<lambda>s. nth s 0)) x \<and> (\<lambda> ys. f ys) x" 
      using "8" "9" by auto
 have 11: "filter ((\<lambda>y. w (\<langle>y\<rangle>))\<circ>(\<lambda>s. nth s 0)) (filter (\<lambda> ys. f ys) (suffixes xs)) =
          filter (\<lambda> zs. ((\<lambda>y. w (\<langle>y\<rangle>))\<circ>(\<lambda>s. nth s 0)) zs \<and> (\<lambda> ys. f ys) zs) (suffixes xs)
         " 
      using filter_filter[of "(\<lambda> ys. f ys)" "(suffixes xs)" "((\<lambda>y. w (\<langle>y\<rangle>))\<circ>(\<lambda>s. nth s 0))" ]  
           "8" "10" by blast
 have 12: "\<exists>i\<le>intlen xs. (\<lambda> zs. ((\<lambda>y. w (\<langle>y\<rangle>))\<circ>(\<lambda>s. nth s 0)) zs \<and> (\<lambda> ys. f ys) zs) (suffix i xs) " 
      by (metis PiAssocsem1 assms comp_apply interval_intfirst_suffix interval_suffix_zero le0)
 have 13: " (filter (\<lambda> zs. ((\<lambda>y. w (\<langle>y\<rangle>))\<circ>(\<lambda>s. nth s 0)) zs \<and> (\<lambda> ys. f ys) zs) (suffixes xs))
            = ( sfxfilt xs (\<lambda> zs. ((\<lambda>y. w (\<langle>y\<rangle>))\<circ>(\<lambda>s. nth s 0)) zs \<and> (\<lambda> ys. f ys) zs)) 
          " 
      by (simp add: sfxfilt_def)
 have 14: "\<exists>i\<le>intlen xs. (\<lambda> zs. ((\<lambda>y. w (\<langle>y\<rangle>))\<circ>(\<lambda>s. nth s 0)) zs \<and> (\<lambda> ys. f ys) zs) (suffix i xs)"
      using "12" by blast
 have 15: "map (\<lambda>s. nth s 0) 
               (( sfxfilt xs (\<lambda> zs. ((\<lambda>y. w (\<langle>y\<rangle>))\<circ>(\<lambda>s. nth s 0)) zs \<and> (\<lambda> ys. f ys) zs)) ) = 
           pifilt xs (\<lambda> zs. ((\<lambda>y. w (\<langle>y\<rangle>))\<circ>(\<lambda>s. nth s 0)) zs \<and> (\<lambda> ys. f ys) zs)
          "  
      using "14" by (simp add: sfxfilt_pifilt)
 have 16: "\<And>xs . (\<lambda> zs. ((\<lambda>y. w (\<langle>y\<rangle>))\<circ>(\<lambda>s. nth s 0)) zs \<and> (\<lambda> ys. f ys) zs) xs =
            (LIFT(f \<and> init w)) xs
          " 
      by (auto simp add: init_defs)
 have 17: "pifilt xs (\<lambda> zs. ((\<lambda>y. w (\<langle>y\<rangle>))\<circ>(\<lambda>s. nth s 0)) zs \<and> (\<lambda> ys. f ys) zs) =
           pifilt xs (LIFT(f \<and> init w))
          " 
      using "16" by presburger
 show ?thesis 
 using "11" "13" "15" "17" "2" "3" "4" "7" by auto    
qed


lemma PiAssocsem:
 "\<sigma> \<Turnstile> f \<Pi> ((init w) \<Pi>  g) =   (f \<and> (init w)) \<Pi> g"
proof (auto simp add: pi_d_def init_defs)
 show "\<And>i ia.
       g (pifilt (pifilt \<sigma> f) (LIFT(init w))) \<Longrightarrow>
       i \<le> intlen \<sigma> \<Longrightarrow>
       f (suffix i \<sigma>) \<Longrightarrow>
       ia \<le> intlen (pifilt \<sigma> f) \<Longrightarrow> 
       w \<langle>Interval.nth (pifilt \<sigma> f) ia\<rangle> \<Longrightarrow> 
       \<exists>i\<le>intlen \<sigma>. f (suffix i \<sigma>) \<and> w \<langle>Interval.nth \<sigma> i\<rangle>" 
  using PiAssocsem1 by fastforce
 show "\<And>i ia.
       g (pifilt (pifilt \<sigma> f) (LIFT(init w))) \<Longrightarrow>
       i \<le> intlen \<sigma> \<Longrightarrow> 
       f (suffix i \<sigma>) \<Longrightarrow> 
       ia \<le> intlen (pifilt \<sigma> f) \<Longrightarrow> 
       w \<langle>Interval.nth (pifilt \<sigma> f) ia\<rangle> \<Longrightarrow> 
       g (pifilt \<sigma> (LIFT(f \<and> init w)))" 
    by (metis interval_intfirst_suffix interval_nth_zero_intfirst pifilt_pifilt)
 show "\<And>i. g (pifilt \<sigma> (LIFT(f \<and> init w))) \<Longrightarrow>
         i \<le> intlen \<sigma> \<Longrightarrow> f (suffix i \<sigma>) \<Longrightarrow> 
         w \<langle>Interval.nth \<sigma> i\<rangle> \<Longrightarrow> 
         \<exists>i\<le>intlen (pifilt \<sigma> f). w \<langle>Interval.nth (pifilt \<sigma> f) i\<rangle>" 
    by (metis PiAssocsem2)
 show "\<And>i. g (pifilt \<sigma> (LIFT(f \<and> init w))) \<Longrightarrow> 
        i \<le> intlen \<sigma> \<Longrightarrow> 
        f (suffix i \<sigma>) \<Longrightarrow> 
        w \<langle>Interval.nth \<sigma> i\<rangle> \<Longrightarrow> 
        g (pifilt (pifilt \<sigma> f) (LIFT(init w))) " 
    by (metis PiAssocsem2 interval_intfirst_suffix interval_nth_zero_intfirst pifilt_pifilt)
qed

lemma  PiAssoc:
 "\<turnstile> f \<Pi> ((init w) \<Pi>  g) =   (f \<and> (init w)) \<Pi> g"
 using PiAssocsem Valid_def by blast

subsubsection \<open>PiNotEqvDiamondAndNotPi\<close>

lemma PiNotEqvDiamondAndNotPisem:
 "\<sigma> \<Turnstile> f \<Pi> (\<not> g) = (\<diamond>f \<and> \<not>(f \<Pi> g))"
by (simp add: pi_d_def sometimes_defs ) blast


lemma PiNotEqvDiamondAndNotPi:
 "\<turnstile> f \<Pi> (\<not> g) = (\<diamond>f \<and> \<not>(f \<Pi> g))"
using PiNotEqvDiamondAndNotPisem Valid_def by blast

subsubsection \<open>PiChopDist\<close>

lemma set_fuse:
 assumes "intlast xs = intfirst ys"
 shows   " set (fuse xs ys) = set xs \<union> set ys"
using assms
proof (induction xs arbitrary: ys)
case (St x)
then show ?case 
by (metis fuse_St interval_fuse_rightneutral interval_intapp_assoc interval_set_intapp 
    opfx_code(1) opfx_def sup.idem)
next
case (Cons x1a xs)
then show ?case by simp
qed

lemma filter_chop:
assumes "intlast xs = intfirst ys \<and> P (intlast xs)"
        "(\<exists> x \<in> set (fuse xs ys). P x)"
        "(\<exists> x \<in> set xs. P x)"
        "(\<exists> x \<in> set ys. P x)"
shows "filter P (fuse xs ys) = fuse (filter P xs)  (filter P ys)"
using assms 
proof (induction xs arbitrary: ys)
case (St x)
then show ?case 
by simp
next
case (Cons x1a xs)
then show ?case 
   proof (cases "(\<exists> x \<in> set xs. P x)")
   case True
   then show ?thesis 
      using Cons.IH Cons.prems(1) Cons.prems(4) set_fuse by fastforce
   next
   case False
   then show ?thesis 
     using Cons.prems(1) nth_set order_refl by force
   qed
qed

lemma filter_chop1:
assumes " n\<le> intlen xs \<and> P (intlast (prefix n xs))"
        "(\<exists> x \<in> set xs. P x)"
        "(\<exists> x \<in> set (prefix n xs). P x)"
        "(\<exists> x \<in> set (suffix n xs). P x)"
shows "filter P xs = fuse (filter P (prefix n xs))  (filter P (suffix n xs))"
proof -
 have 1: "(\<exists> x \<in> set xs. P x) = 
           (\<exists> x \<in> set (fuse (prefix n xs) (suffix n xs)). P x)
         " 
     by (simp add: assms(1) interval_fuse_prefix_suffix)
 have 2: "intlast (prefix n xs) = intfirst (suffix n xs)"
     using Interval.interval_intlast_intfirst by blast
 have 3: "xs = fuse (prefix n xs) (suffix n xs)"
   by (simp add: assms(1) interval_fuse_prefix_suffix) 
 have 4: "filter P (fuse (prefix n xs) (suffix n xs)) =
          fuse (filter P (prefix n xs))  (filter P (suffix n xs))" 
  using assms 2 3 filter_chop[of "(prefix n xs)" "(suffix n xs)" "P"] 
  by auto  
 show ?thesis using "3" "4" by auto
qed


lemma filter_chop1_prefix:
assumes " n\<le> intlen xs"
        " P (intlast (prefix n xs))"
        "(\<exists> x \<in> set xs. P x)"
        "(\<exists> x \<in> set (prefix n xs). P x)"
        "(\<exists> x \<in> set (suffix n xs). P x)"
shows "prefix (intlen (filter P (prefix n xs))) (filter P xs) = 
       (filter P (prefix n xs))"
proof -
 have 2: "filter P xs = fuse (filter P (prefix n xs))  (filter P (suffix n xs))" 
   using assms filter_chop1 by blast
 have 3: " intlast (filter P (prefix n xs)) = intfirst (filter P (suffix n xs))"
   using assms by (metis Interval.interval_intlast_intfirst filter_intfirst filter_intlast)
 have 4: "prefix (intlen (filter P (prefix n xs))) 
                 (fuse (filter P (prefix n xs))  (filter P (suffix n xs)) ) = 
         (filter P (prefix n xs))
         " 
     using interval_prefix_fuse using "3" by blast
 show ?thesis by (simp add: "2" "4")
qed

lemma filter_chop1_suffix:
assumes " n\<le> intlen xs "
        " P (intlast (prefix n xs))"
        "(\<exists> x \<in> set xs. P x)"
        "(\<exists> x \<in> set (prefix n xs). P x)"
        "(\<exists> x \<in> set (suffix n xs). P x)"
shows "suffix (intlen (filter P (prefix n xs))) (filter P xs) = 
       (filter P (suffix n xs))"
proof -
 have 1: "filter P xs = fuse (filter P (prefix n xs))  (filter P (suffix n xs))" 
   using assms filter_chop1 by blast
 have 3: " intlast (filter P (prefix n xs)) = intfirst (filter P (suffix n xs))"
   using assms by (metis Interval.interval_intlast_intfirst filter_intfirst filter_intlast)
 have 4: "suffix (intlen (filter P (prefix n xs)))  
                 (fuse (filter P (prefix n xs))  (filter P (suffix n xs))) =
          (filter P (suffix n xs))
         " 
     using interval_suffix_fuse using "3" by blast
  show ?thesis by (simp add: "1" "4")
qed


lemma PiChopDistsema:
 assumes "(\<sigma> \<Turnstile> (init w) \<Pi> (g;h))"
 shows  " (\<sigma> \<Turnstile> ((init w) \<Pi> g);((init w) \<and> ((init w) \<Pi> h)))"
proof -
 have 1: "(\<sigma> \<Turnstile> (init w) \<Pi> (g;h))" 
   using assms by auto
 have 2: "((\<exists>i\<le>intlen \<sigma>. (LIFT(init w)) (suffix i \<sigma>)) \<and>
           ((pifilt \<sigma> (LIFT(init w))) \<Turnstile> g;h)
          )"
    using "1" by (simp add: pi_d_def)
 have 3: "(\<exists>i\<le>intlen \<sigma>. (LIFT(init w)) (suffix i \<sigma>))"  
   using "2" by auto
 have 4: "((pifilt \<sigma> (LIFT(init w))) \<Turnstile> g;h)"
   using "2" by auto 
 have 5: "filter (\<lambda>y. w (\<langle>y\<rangle>)) \<sigma> \<Turnstile> g;h"
    using pifilt_init 
    using "2" by fastforce 
 have 6: "\<exists>n\<le>intlen (filter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>).
            g (prefix n (filter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>)) \<and> 
            h (suffix n (filter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>))"
  using "5" by (simp add: chop_defs)  
 obtain n where 7: "n\<le>intlen (filter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>) \<and>
         g (prefix n (filter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>)) \<and> 
            h (suffix n (filter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>))"
   using "6" by auto
 have 8: "\<exists>i\<le>intlen \<sigma>. w \<langle>nth \<sigma> i\<rangle>"
   using "3" by (auto simp add: init_defs) 
 have 9: "\<exists> x \<in> set \<sigma>. w \<langle>x\<rangle>"
    using "8" nth_set by blast
 have 10: "(prefix n (filter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>)) = 
           (filter (\<lambda>y. w \<langle>y\<rangle>) (prefix ((nth (nfilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma> 0) n) ) \<sigma>) )"
     by (simp add: "7" "9" filter_nfilter_prefix_1) 
 have 11: "(suffix n (filter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>)) =
           (filter (\<lambda>y. w \<langle>y\<rangle>) (suffix ((nth (nfilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma> 0) n) ) \<sigma>) )"
   by (simp add: "7" "9" filter_nfilter_suffix_1)
 have 12: "g (filter (\<lambda>y. w \<langle>y\<rangle>) (prefix ((nth (nfilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma> 0) n) ) \<sigma>) )"
   using "10" "7" by auto
 have 13: "h (filter (\<lambda>y. w \<langle>y\<rangle>) (suffix ((nth (nfilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma> 0) n) ) \<sigma>) )" 
    using "11" "7" by auto
 have 14: "((nth (nfilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma> 0) n) ) \<le> intlen \<sigma>" 
   by (metis "7" "9" add_cancel_right_left nfilter_intlen nfilter_upper_bound)
 have 15: "w \<langle>nth (suffix ((nth (nfilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma> 0) n) ) \<sigma>) 0\<rangle>" 
    by (metis (no_types, lifting) "14" "7" "9" filter_nth_aa interval_intfirst_suffix 
        interval_suffix_zero nfilter_filter nfilter_intlen nfilter_nth_n_zero zero_le)
 have 16: "(\<exists>i\<le>intlen (prefix  ((nth (nfilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma> 0) n) ) \<sigma>). 
            w \<langle>nth (suffix i (prefix  ((nth (nfilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma> 0) n) ) \<sigma>)) 0\<rangle>)" 
   using "14" "15" by auto
 have 17: "(\<exists>i\<le>intlen (suffix ((nth (nfilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma> 0) n) ) \<sigma>).
           w \<langle>nth (suffix (i + ((nth (nfilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma> 0) n) )) \<sigma>) 0\<rangle>)"
    using "15" by auto
 have 18: "(\<exists>n\<le>intlen \<sigma>.
       (\<exists>i\<le>intlen (prefix n \<sigma>). w \<langle>nth (suffix i (prefix n \<sigma>)) 0\<rangle>) \<and>
       g (filter (\<lambda>y. w (\<langle>y\<rangle>)) (prefix n \<sigma>)) \<and>
       w \<langle>nth (suffix n \<sigma>) 0\<rangle> \<and>
       (\<exists>i\<le>intlen (suffix n \<sigma>). w \<langle>nth (suffix (i + n) \<sigma>) 0\<rangle>) \<and> 
        h (filter (\<lambda>y. w (\<langle>y\<rangle>)) (suffix n \<sigma>)) )" 
     using "12" "13" "14" "15" "16" "17" by blast
 have 19: "(\<exists>n\<le>intlen \<sigma>.
       (\<exists>i\<le>intlen (prefix n \<sigma>). w \<langle>nth (suffix i (prefix n \<sigma>)) 0\<rangle>) \<and>
       g (pifilt (prefix n \<sigma>) (LIFT(init w))) \<and>
       w \<langle>nth (suffix n \<sigma>) 0\<rangle> \<and>
       (\<exists>i\<le>intlen (suffix n \<sigma>). w \<langle>nth (suffix (i + n) \<sigma>) 0\<rangle>) \<and> 
        h (pifilt (suffix n \<sigma>) (LIFT(init w))))" 
   by (metis "18" init_defs interval_prefix_zero_intfirst interval_suffix_suffix pifilt_init)
 have 20: "((\<exists>n\<le>intlen \<sigma>.
       (\<exists>i\<le>intlen (prefix n \<sigma>). (LIFT(init w)) (suffix i (prefix n \<sigma>))) \<and>
       g (pifilt (prefix n \<sigma>) (LIFT(init w))) \<and>
       (LIFT(init w)) (suffix n \<sigma>) \<and> (\<exists>i\<le>intlen (suffix n \<sigma>). (LIFT(init w)) (suffix (i + n) \<sigma>)) 
         \<and> h (pifilt (suffix n \<sigma>) (LIFT(init w)))))" 
    by (metis "19" init_defs interval_prefix_zero_intfirst)
 have 21: "(\<sigma> \<Turnstile> ((init w) \<Pi> g);((init w) \<and> ((init w) \<Pi> h)))"
      using "20" by (simp add: chop_defs pi_d_def) 
show ?thesis 
using "21" by auto
qed       

      
lemma PiChopDistsemb:
 assumes " (\<sigma> \<Turnstile> ((init w) \<Pi> g);((init w) \<and> ((init w) \<Pi> h)))"
 shows  "(\<sigma> \<Turnstile> (init w) \<Pi> (g;h))"
proof -  
 have 1: "(\<sigma> \<Turnstile> ((init w) \<Pi> g);((init w) \<and> ((init w) \<Pi> h)))" 
   using assms by auto
 have 2: "\<exists> n \<le> intlen \<sigma>. 
            ((prefix n \<sigma>) \<Turnstile> ((init w) \<Pi> g)) \<and>
            ((suffix n \<sigma>) \<Turnstile> ((init w) \<and> ((init w) \<Pi> h)))
         " 
    using assms chop_defs by blast
 obtain n where 3: "n \<le> intlen \<sigma> \<and> ((prefix n \<sigma>) \<Turnstile> ((init w) \<Pi> g)) \<and>
            ((suffix n \<sigma>) \<Turnstile> ((init w) \<and> ((init w) \<Pi> h)))" 
   using "2" by auto
 have 4: "((\<exists>i\<le>intlen (prefix n \<sigma>). (LIFT(init w)) (suffix i (prefix n \<sigma>))) \<and>
           ((pifilt (prefix n \<sigma>) (LIFT(init w))) \<Turnstile> g)
          )" 
    by (meson "3" pi_d_def)
 have 5: "(\<exists>i\<le>intlen (prefix n \<sigma>). (LIFT(init w)) (suffix i (prefix n \<sigma>)))" 
   using "4" by auto
 have 6: "g (pifilt (prefix n \<sigma>) (LIFT(init w)))" 
   using "4" by auto
 have 7: "g (filter (\<lambda>y. w (\<langle>y\<rangle>)) (prefix n \<sigma>)) "
    using "5" "6" pifilt_init by (metis)
 have 8: "((\<exists>i\<le>intlen (suffix n \<sigma>). (LIFT(init w)) (suffix i (suffix n \<sigma>))) \<and>
           ((pifilt (suffix n \<sigma>) (LIFT(init w))) \<Turnstile> h)
          )" 
   by (metis "3" intensional_rews(3) pi_d_def)
 have 9: "(\<exists>i\<le>intlen (suffix n \<sigma>). (LIFT(init w)) (suffix i (suffix n \<sigma>)))"
    using "8" by auto
 have 10: "h (pifilt (suffix n \<sigma>) (LIFT(init w)))"  
   using "8" by auto
 have 11: "h (filter (\<lambda>y. w (\<langle>y\<rangle>)) (suffix n \<sigma>))  "
    using "10" "9" pifilt_init by metis 
 have 12: "(\<lambda>y. w (\<langle>y\<rangle>)) (intlast (prefix n \<sigma>))"
    by (metis "3" init_defs intensional_rews(3) interval_intfirst_suffix interval_intlast_prefix 
        interval_nth_zero_intfirst interval_prefix_zero_intfirst)
 have 13: "\<exists> x \<in> set \<sigma>. (\<lambda>y. w (\<langle>y\<rangle>)) x"
   using "12" "3" nth_set using interval_intlast_prefix by fastforce 
 have 14: "\<exists> x \<in> set (prefix n \<sigma>) . (\<lambda>y. w (\<langle>y\<rangle>)) x" 
   using "12" nth_set by (metis interval_nth_intlen_intlast order_refl)
 have 15: "\<exists> x \<in> set (suffix n \<sigma>) . (\<lambda>y. w (\<langle>y\<rangle>)) x" 
   by (metis "12" "3" interval_intfirst_suffix interval_intlast_prefix interval_intlen_gr_zero 
       interval_nth_zero_intfirst nth_set)
 have 16: "(filter (\<lambda>y. w (\<langle>y\<rangle>)) (prefix n \<sigma>)) =
           prefix (intlen (filter (\<lambda>y. w (\<langle>y\<rangle>)) (prefix n \<sigma>))) (filter (\<lambda>y. w (\<langle>y\<rangle>)) \<sigma>)
          " 
   using "12" "13" "14" "15" "3" 
          filter_chop1_prefix[of "n" "\<sigma>" "(\<lambda>y. w \<langle>y\<rangle>)"] by auto
 have 17: "(filter (\<lambda>y. w (\<langle>y\<rangle>)) (suffix n \<sigma>)) =
           suffix (intlen (filter (\<lambda>y. w (\<langle>y\<rangle>)) (prefix n \<sigma>))) (filter (\<lambda>y. w (\<langle>y\<rangle>)) \<sigma>)
          " 
   using "12" "13" "14" "15" "3" 
          filter_chop1_suffix[of "n" "\<sigma>" "(\<lambda>y. w \<langle>y\<rangle>)"] by auto
 have 18: "\<exists>n\<le>intlen (filter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>).
            g (prefix n (filter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>)) \<and> 
            h (suffix n (filter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>))"
        by (metis "11" "16" "17" "7" interval_pref_intlen_bound)
 have 19: "filter (\<lambda>y. w (\<langle>y\<rangle>)) \<sigma> \<Turnstile> g;h"  
   by (simp add: "18" chop_defs)
 have 20: "((pifilt \<sigma> (LIFT(init w))) \<Turnstile> g;h)" 
   by (metis (mono_tags, lifting) "18" "3" Interval.interval_intlast_intfirst intensional_rews(3) 
       interval_chop_fuse interval_fuse_prefix_suffix pifilt_init)
 have 21: "(\<exists>i\<le>intlen \<sigma>. (LIFT(init w)) (suffix i \<sigma>))" 
   using "3" by auto
 show ?thesis 
 by (simp add: "20" "21" pi_d_def)
qed

lemma PiChopDistsem:
 "\<sigma> \<Turnstile> (init w) \<Pi> (g;h) = ((init w) \<Pi> g);((init w) \<and> ((init w) \<Pi> h))"
using PiChopDistsema PiChopDistsemb unl_lift2 by blast

lemma PiChopDist:
 "\<turnstile> (init w) \<Pi> (g;h) = ((init w) \<Pi> g);((init w) \<and> ((init w) \<Pi> h))"
using PiChopDistsem Valid_def by blast

subsubsection \<open>PiProp\<close>

lemma Pistate:
 "(\<sigma> \<Turnstile> (init w) \<Pi> f) = 
  ((\<exists> x \<in> set \<sigma>. w \<langle>x\<rangle>) \<and>  ( (filter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>) \<Turnstile> f) )"
proof -
 have 1: "(\<sigma> \<Turnstile> (init w) \<Pi> f) =
          ( (\<exists> i\<le> intlen \<sigma>.  w \<langle>nth \<sigma> i\<rangle>) \<and> ((pifilt \<sigma> (LIFT(init w))) \<Turnstile> f) )"
  by (auto simp add: pi_d_def init_defs)
 have 2: "(\<exists>i\<le>intlen \<sigma>. (LIFT(init w)) (suffix i \<sigma>)) =
          (\<exists> i\<le> intlen \<sigma>.  w \<langle>nth \<sigma> i\<rangle>)"
  by (auto simp add: init_defs) 
 have 3: "(\<exists> i\<le> intlen \<sigma>.  w \<langle>nth \<sigma> i\<rangle>) \<longrightarrow>
           (pifilt \<sigma> (LIFT(init w))) = (filter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>)" 
   using pifilt_init using "2" by blast
 have 4: "(\<exists> i\<le> intlen \<sigma>.  w \<langle>nth \<sigma> i\<rangle>) =(\<exists> x \<in> set \<sigma>. w \<langle>x\<rangle>) "
    using interval_nth_and_set by force
 show ?thesis 
 using "1" "3" "4" by auto
qed


lemma PiPropsem1a:
 "(\<sigma> \<Turnstile> f \<Pi> $p) = 
   ((\<exists>i\<le>intlen \<sigma>. f (suffix i \<sigma>)) \<and> p (nth \<sigma> (nth (nfilter f (suffixes \<sigma>) 0) 0)) )
  "
using interval_nth_map[of "(\<lambda>s. nth s 0)" "(filter f (suffixes \<sigma>))" "0"] 
using nfilter_filter[of "suffixes \<sigma>" "f" "0" "0" ] 
by (simp add: pi_d_def current_val_d_def pifilt_def sfxfilt_def )
   (metis in_set_suffixes interval_nth_map map_first_suffixes osfx_suffix)

lemma PiPropsem2a:
 " (\<sigma> \<Turnstile> (\<not> f) \<U> (f \<and> $p)) =
   (\<exists>k\<le>intlen \<sigma>. f (suffix k \<sigma>) \<and> p (nth \<sigma> k) \<and> (\<forall>j<k. \<not> f (suffix j \<sigma>)))"
by (simp add: until_d_def current_val_d_def)

lemma PiPropsem3a:
 assumes "(\<sigma> \<Turnstile> f \<Pi> $p)"
 shows   " (\<sigma> \<Turnstile> (\<not> f) \<U> (f \<and> $p))"
proof -
 have 1: "((\<exists>i\<le>intlen \<sigma>. f (suffix i \<sigma>)) \<and> p (nth \<sigma> (nth (nfilter f (suffixes \<sigma>) 0) 0) ) )"
    using assms  PiPropsem1a by auto
 have 2: "\<exists> x \<in> set(suffixes \<sigma>). f x" 
   using "1" in_set_suffixes osfx_suffix by blast
 have 3: "\<forall> x \<in> set(nfilter f (suffixes \<sigma>) 0). f  (nth (suffixes \<sigma>) x)"
   by (simp add: "2")
 have 4: "f (suffix (nth (nfilter f (suffixes \<sigma>) 0) 0)  \<sigma>) " 
   by (metis "2" "3" add.left_neutral interval_intlen_gr_zero nfilter_upper_bound nth_set 
       nth_suffixes)
 have 5: "(\<forall>j< (nth (nfilter f (suffixes \<sigma>) 0) 0). \<not> f (suffix j \<sigma>)) " 
   using nfilter_not_before[of "suffixes \<sigma>" "f" ] 2 
   proof -
   have f1: "\<forall>n.\<not> n < Interval.nth (nfilter f (suffixes \<sigma>) 0) 0 \<or> \<not> f (Interval.nth (suffixes \<sigma>) n)"
   using \<open>\<And>i. \<lbrakk>\<exists>x\<in>set (suffixes \<sigma>). f x; i < nth (nfilter f (suffixes \<sigma>) 0) 0\<rbrakk> \<Longrightarrow>
        \<not> f (nth (suffixes \<sigma>) i)\<close> \<open>\<exists>x\<in>set (suffixes \<sigma>). f x\<close> by blast 
   obtain ii :: "'a interval" where
   f2: "ii \<in> set (suffixes \<sigma>) \<and> f ii"
   using \<open>\<exists>x\<in>set (suffixes \<sigma>). f x\<close> by blast 
   obtain nn :: "'a interval interval \<Rightarrow> 'a interval \<Rightarrow> nat" where
   "\<forall>x0 x1. (\<exists>v2\<le>intlen x0. nth x0 v2 = x1) = (nn x0 x1 \<le> intlen x0 \<and> nth x0 (nn x0 x1) = x1)"
   by moura 
   then have "nn (suffixes \<sigma>) ii \<le> intlen (suffixes \<sigma>) \<and> nth (suffixes \<sigma>) (nn (suffixes \<sigma>) ii) = ii"
   using f2 by (meson interval_nth_and_set) 
   then have "nth (nfilter f (suffixes \<sigma>) 0) 0 \<le> intlen (suffixes \<sigma>)"
   using f2 f1 by (metis (no_types) dual_order.trans le_less_linear) 
   then show ?thesis
   using f1 by (metis (no_types) dual_order.trans less_or_eq_imp_le nth_suffixes) 
   qed 
 have 6: "(\<exists>k\<le>intlen \<sigma>. f (suffix k \<sigma>) \<and> p (nth \<sigma> k) \<and> (\<forall>j<k. \<not> f (suffix j \<sigma>)))"
   by (metis "1" "4" "5" interval_suf_first neqE) 
 show ?thesis using 6 PiPropsem2a by metis
qed

lemma PiPropsem3b:
 assumes "(\<sigma> \<Turnstile> (\<not> f) \<U> (f \<and> $p))"
 shows   " (\<sigma> \<Turnstile> f \<Pi> $p)"
proof -
 have 1: "(\<exists>k\<le>intlen \<sigma>. f (suffix k \<sigma>) \<and> p (nth \<sigma> k) \<and> (\<forall>j<k. \<not> f (suffix j \<sigma>)))"
  using assms  PiPropsem2a by auto
 obtain k where 2: "k\<le>intlen \<sigma> \<and> f (suffix k \<sigma>) \<and> p (nth \<sigma> k) \<and> (\<forall>j<k. \<not> f (suffix j \<sigma>))"
   using 1 by auto
 have 3: "(\<exists>i\<le>intlen \<sigma>. f (suffix i \<sigma>))"
   using "2" by blast  
 have 31: "\<exists>x \<in> set(suffixes \<sigma>). f x"
   using "2" by auto 
 have 32: "f (suffix (nth (nfilter f (suffixes \<sigma>) 0) 0) \<sigma>)"
   using nfilter_holds[of "suffixes \<sigma>" "f" "0"]  nfilter_not_before[of "suffixes \<sigma>" "f"]
    by (metis "31" diff_zero interval_intlen_gr_zero nfilter_upper_bound nth_set nth_suffixes 
       ordered_cancel_comm_monoid_diff_class.add_diff_inverse)
 have 4: "p (nth \<sigma> (nth (nfilter f (suffixes \<sigma>) 0) 0) )"
   by (metis "1" "31" "32" intlen_suffixes linorder_neqE_nat nfilter_not_before nth_suffixes)
 show ?thesis using 4 3 by (simp add: PiPropsem1a)
qed

lemma PiPropsema:
 "\<sigma> \<Turnstile> f \<Pi> $p = (\<not> f) \<U> (f \<and> $p)"
using PiPropsem3a PiPropsem3b unl_lift2 by blast


lemma PiProp:
 "\<turnstile> f \<Pi> $p = (\<not> f) \<U> (f \<and> $p)"
using PiPropsema Valid_def by blast

subsubsection \<open>PiNext\<close>



lemma PiNextsem1:
 "(\<sigma> \<Turnstile> f \<Pi> (\<circle> g)) =
  ((\<exists>i\<le>intlen \<sigma>. f (suffix i \<sigma>)) \<and>
    0 < intlen (filter f (suffixes \<sigma>)) \<and>
    g (suffix (Suc 0) (map (\<lambda>s. nth s 0) (filter f (suffixes \<sigma>)))))"
by (simp add: pi_d_def next_defs pifilt_def sfxfilt_def)  

lemma PiNextsem2:
 "(\<sigma> \<Turnstile> (\<not> f) \<U> (f \<and> \<circle>(f \<Pi> g))) =
  (\<exists>k\<le>intlen \<sigma>.
       f (suffix k \<sigma>) \<and>
       k < intlen \<sigma> \<and>
       (\<exists>i\<le>intlen \<sigma> - Suc k. f (suffix (Suc (i + k)) \<sigma>)) \<and>
       g (map (\<lambda>s. nth s 0) (filter f (suffixes (suffix (Suc k) \<sigma>)))) \<and> (\<forall>j<k. \<not> f (suffix j \<sigma>)))
  "
by (simp add: until_d_def next_defs pi_d_def pifilt_def sfxfilt_def) 


lemma PiNextsem3:
 assumes "(\<sigma> \<Turnstile> f \<Pi> (\<circle> g))" 
 shows    "(\<sigma> \<Turnstile> (\<not> f) \<U> (f \<and> \<circle>(f \<Pi> g)))" 
proof -
 have 1: "((\<exists>i\<le>intlen \<sigma>. f (suffix i \<sigma>)) \<and>
    0 < intlen (filter f (suffixes \<sigma>)) \<and>
    g (suffix (Suc 0) (map (\<lambda>s. nth s 0) (filter f (suffixes \<sigma>)))))" 
   using assms PiNextsem1 by auto
 have 2: "\<exists> x \<in> set(suffixes \<sigma>). f x" 
   using "1" in_set_suffixes osfx_suffix by blast
 have 3: "\<forall> x \<in> set(nfilter f (suffixes \<sigma>) 0). f  (nth (suffixes \<sigma>) x)"
   by (simp add: "2")
 have 4: "f (suffix (nth (nfilter f (suffixes \<sigma>) 0) 0)  \<sigma>) "
  by (metis "2" add.left_neutral interval_intlen_gr_zero nfilter_filter nfilter_nth_n_zero 
      nfilter_upper_bound nth_suffixes sfxfilter_nth)
 have 41: "(nth (nfilter f (suffixes \<sigma>) 0) 1) \<in> set(nfilter f (suffixes \<sigma>) 0)"
   by (metis "1" "2" One_nat_def Suc_leI nfilter_intlen nth_set)
 have 42: "  (nth (nfilter f (suffixes \<sigma>) 0) 1) \<le> intlen (suffixes \<sigma>) "
    by (metis "1" "2" One_nat_def Suc_leI add_cancel_right_left nfilter_intlen nfilter_upper_bound) 
 have 5: "f (suffix (nth (nfilter f (suffixes \<sigma>) 0) 1)  \<sigma>)" 
    using "3" "41" "42" nth_suffixes by fastforce
 have 6: "(nth (nfilter f (suffixes \<sigma>) 0) 0) < (nth (nfilter f (suffixes \<sigma>) 0) 1)" 
   by (simp add: "1" "2" idx_nfilter_mono nfilter_intlen)
 have 7: " Suc (nth (nfilter f (suffixes \<sigma>) 0) 0) \<le> intlen \<sigma>" 
   by (metis "1" "3" IntervalFilter.length_filter_le Suc_diff_Suc diff_is_0_eq' 
       filter_nfilter_suffix interval_intlen_gr_zero interval_suffix_length intlen_suffixes 
       not_less_eq_eq nth_set)
 have 8: "0 < intlen (nfilter f (suffixes \<sigma>) 0)" 
   by (simp add: "1" "2" nfilter_intlen)
 have 9: "(nth (nfilter f (suffixes \<sigma>) 0) 1) \<le> intlen \<sigma>" 
    using nfilter_upper_bound[of "suffixes \<sigma>" "f" "1" "0"] 
    by (simp add: "2" "8" Suc_leI)  
 have 10: "(Suc (nth (nfilter f (suffixes \<sigma>) 0) 0)) \<le> (nth (nfilter f (suffixes \<sigma>) 0) 1)"
    using "6" Suc_leI by blast 
 have 11: "(nth (nfilter f (suffixes \<sigma>) 0) 1) = 
            (nth (nfilter f (suffixes \<sigma>) 0) 1) - (Suc (nth (nfilter f (suffixes \<sigma>) 0) 0)) +
              (Suc (nth (nfilter f (suffixes \<sigma>) 0) 0)) 
          " 
        using "10" by auto
  have 12: "(nth (nfilter f (suffixes \<sigma>) 0) 1) - (Suc (nth (nfilter f (suffixes \<sigma>) 0) 0)) \<le>
            intlen \<sigma> - Suc (nth (nfilter f (suffixes \<sigma>) 0) 0)"
     using "9" diff_le_mono by blast 
 have 13: "\<exists> i \<le> intlen \<sigma> - Suc (nth (nfilter f (suffixes \<sigma>) 0) 0). 
             (nth (nfilter f (suffixes \<sigma>) 0) 1) = (Suc (i + (nth (nfilter f (suffixes \<sigma>) 0) 0)))" 
  using 11 12 by auto
 have 14: "(\<exists>i\<le>intlen \<sigma> - Suc (nth (nfilter f (suffixes \<sigma>) 0) 0). 
            f (suffix (Suc (i + (nth (nfilter f (suffixes \<sigma>) 0) 0))) \<sigma>))" 
     using "13" "5" by auto 
 have 15: "(suffix (Suc 0) (map (\<lambda>s. nth s 0) (filter f (suffixes \<sigma>)))) =
           (map (\<lambda>s. nth s 0) 
                (filter f (suffixes (suffix (Suc (nth (nfilter f (suffixes \<sigma>) 0) 0)) \<sigma>))))" 
   using 2 8 by (simp add: "1" Suc_leI filter_suffixes_map)
 have 16: "g (map (\<lambda>s .nth s 0) 
                  (filter f (suffixes (suffix (Suc (nth (nfilter f (suffixes \<sigma>) 0) 0)) \<sigma>))))  "
     using "1" "15" by auto    
 have 17: "\<forall> j< (nth (nfilter f (suffixes \<sigma>) 0) 0). \<not> f (suffix j \<sigma>)"
   by (metis "7" Suc_leD Suc_leI in_set_suffixes intlen_suffixes le_trans nfilter_not_before
        nth_suffixes osfx_suffix) 
 have 18: "(\<exists>k\<le>intlen \<sigma>.
       f (suffix k \<sigma>) \<and>
       k < intlen \<sigma> \<and>
       (\<exists>i\<le>intlen \<sigma> - Suc k. f (suffix (Suc (i + k)) \<sigma>)) \<and>
       g (map (\<lambda>s. nth s 0) (filter f (suffixes (suffix (Suc k) \<sigma>)))) \<and> (\<forall>j<k. \<not> f (suffix j \<sigma>)))"
  using "14" "16" "17" "4" "7" Suc_leD Suc_le_lessD by blast 
 show ?thesis using 18 by (simp add: PiNextsem2)
qed


lemma PiNextsem4:
 assumes "(\<sigma> \<Turnstile> (\<not> f) \<U> (f \<and> \<circle>(f \<Pi> g)))" 
 shows    "(\<sigma> \<Turnstile> f \<Pi> (\<circle> g))"
proof -
 have 1: "(\<exists>k\<le>intlen \<sigma>.
       f (suffix k \<sigma>) \<and>
       k < intlen \<sigma> \<and>
       (\<exists>i\<le>intlen \<sigma> - Suc k. f (suffix (Suc (i + k)) \<sigma>)) \<and>
       g (map (\<lambda>s. nth s 0) (filter f (suffixes (suffix (Suc k) \<sigma>)))) \<and> (\<forall>j<k. \<not> f (suffix j \<sigma>)))"
   using assms   by (simp add: PiNextsem2)
 obtain k where 2: "k\<le>intlen \<sigma> \<and> f (suffix k \<sigma>) \<and> k < intlen \<sigma> \<and> (\<exists>i\<le>intlen \<sigma> - Suc k. 
                    f (suffix (Suc (i + k)) \<sigma>)) \<and>
       g (map (\<lambda>s. nth s 0) (filter f (suffixes (suffix (Suc k) \<sigma>)))) \<and> (\<forall>j<k. \<not> f (suffix j \<sigma>))"
   using 1 by auto 
 have 3: "\<exists> x \<in> set (suffixes \<sigma>). f x"
   using "1" in_set_suffixes osfx_suffix by blast 
 have 4: "\<forall> x \<in> set(nfilter f (suffixes \<sigma>) 0). f  (nth (suffixes \<sigma>) x)"   
   by (simp add: "3")
 have 5: "f (suffix (nth (nfilter f (suffixes \<sigma>) 0) 0)  \<sigma>) "
  by (metis "3" "4" add.left_neutral interval_intlen_gr_zero nfilter_upper_bound nth_set 
      nth_suffixes)
 have 6: "0 < intlen (filter f (suffixes \<sigma>))"
  using filter_length_zero_conv_a[of "suffixes \<sigma>" "f"] 
  by (metis "2" "3" Nat.le_diff_conv2 Suc_leI Suc_n_not_le_n add_Suc_right intlen_suffixes le_add2 
      neq0_conv nth_suffixes)
 have 61: "(nth (nfilter f (suffixes \<sigma>) 0) 1) \<in> set(nfilter f (suffixes \<sigma>) 0)"
   by (metis "3" "6" One_nat_def Suc_leI nfilter_intlen nth_set) 
 have 62: "(nth (nfilter f (suffixes \<sigma>) 0) 1) \<le> intlen (suffixes \<sigma>)"
   by (metis "3" "6" One_nat_def Suc_leI add_cancel_right_left nfilter_intlen nfilter_upper_bound) 
 have 7: "f (suffix (nth (nfilter f (suffixes \<sigma>) 0) 1)  \<sigma>)"
   using "4" "61" "62" nth_suffixes by fastforce
 have 8: "(\<exists>i\<le>intlen \<sigma>. f (suffix i \<sigma>))" 
    using "1" by blast
 have 9: "g (suffix (Suc 0) (map (\<lambda>s. nth s 0) (filter f (suffixes \<sigma>))))"
  by (metis "2" "3" "5" "6" Suc_leI intlen_suffixes linorder_neqE_nat 
      nfilter_not_before nth_suffixes filter_suffixes_map) 
 have 10: "((\<exists>i\<le>intlen \<sigma>. f (suffix i \<sigma>)) \<and>
           0 < intlen (filter f (suffixes \<sigma>)) \<and>
           g (suffix (Suc 0) (map (\<lambda>s. nth s 0) (filter f (suffixes \<sigma>)))))" 
  using "6" "8" "9" by blast
 show ?thesis using 10 
 by (simp add: PiNextsem1)
qed


lemma PiNextsem:
 " (\<sigma> \<Turnstile> f \<Pi> (\<circle> g) = (\<not> f) \<U> (f \<and> \<circle>(f \<Pi> g)))"
using PiNextsem3 PiNextsem4 
using unl_lift2 by blast


lemma PiNext:
 "\<turnstile> f \<Pi> (\<circle> g) = (\<not> f) \<U> (f \<and> \<circle>(f \<Pi> g))"
using PiNextsem Valid_def by blast

subsubsection \<open>PiUntil\<close>


lemma PiUntilDistsem1:
  "(\<sigma> \<Turnstile> f \<Pi> (g \<U> h)) =
   ((\<exists>i\<le>intlen \<sigma>. f (suffix i \<sigma>)) \<and>
    (\<exists>k\<le>intlen (filter f (suffixes \<sigma>)).
        h (suffix k (map (\<lambda>s. nth s 0) (filter f (suffixes \<sigma>)))) \<and>
        (\<forall>j<k. g (suffix j (map (\<lambda>s. nth s 0) (filter f (suffixes \<sigma>)))))))"
by (simp add: pi_d_def pifilt_def sfxfilt_def until_d_def)


lemma PiUntilDistsem2:
 " (\<sigma> \<Turnstile> ( f \<Pi> g ) \<U> ( f \<Pi> h ) ) =
   (\<exists>k\<le>intlen \<sigma>.
       (\<exists>i\<le>intlen \<sigma> - k. f (suffix (i + k) \<sigma>)) \<and>
       h (map (\<lambda>s. nth s 0) (filter f (suffixes (suffix k \<sigma>)))) \<and>
       (\<forall>j<k. (\<exists>i\<le>intlen \<sigma> - j. f (suffix (i + j) \<sigma>)) \<and> 
              g (map (\<lambda>s. nth s 0) (filter f (suffixes (suffix j \<sigma>))))))"
by (simp add: until_d_def pi_d_def pifilt_def sfxfilt_def)


lemma cover:
 assumes "(\<exists>i<k. ff(i) <(j::nat) \<and> j\<le> ff(Suc i))"
         "(\<forall>i<k. ff(i)<ff(Suc i))"
 shows   "ff(0)<j \<and> j \<le>ff(k)  "
using assms 
proof (induct k arbitrary:j)
case 0
then show ?case by simp
next
case (Suc k)
then show ?case 
  proof -
   have 1: "(\<exists>i<k. ff(i) <(j::nat) \<and> j\<le> ff(Suc i)) \<or> (ff (k) <j \<and> j\<le>ff(Suc k))"
     using Suc.prems(1) less_SucE by blast
   have 2: "(\<forall>i<k. ff(i)<ff(Suc i))" 
     using Suc.prems(2) by auto
   have 3: "ff k < ff (Suc k)"
    by (simp add: Suc.prems(2)) 
   have 4: "(ff(0)<j \<and> j \<le>ff(k)) \<or>  (ff (k) <j \<and> j\<le>ff(Suc k))"
      using "1" "2" Suc.hyps by blast 
   have 41: "ff(0)<j" 
     by (metis "2" "4" Suc.hyps less_antisym less_le_trans less_or_eq_imp_le zero_less_Suc)
   have 42: "j \<le>ff(Suc k)"
     using "3" "4" by linarith 
   have 5: "ff(0)<j \<and> j \<le>ff(Suc k) " 
     by (simp add: "41" "42")
   show ?thesis 
   by (simp add: "5")
  qed    
qed


lemma cover_a:
 assumes "(\<forall> j.  
          (j\<le> ff 0) \<or> (\<exists>i<k. ff(i) <(j::nat) \<and> j\<le> ff(Suc i)) 
            \<longrightarrow> gg j)
          "
         "(\<forall>i<k. ff(i)<ff(Suc i))"
 shows   "(\<forall> j< ff k. gg j)  "
proof -
 have 1: "(\<forall> j< ff 0. gg j)" 
   by (simp add: assms(1))
 have 2: "(\<forall> j. ff 0 < j \<and> j\<le> ff k \<longrightarrow> gg j)" 
   proof 
     fix j
     show "ff 0 < j \<and> j \<le> ff k \<longrightarrow> gg j"  
      using assms 
       proof (induct k arbitrary: j)
       case 0
       then show ?case by simp
       next
       case (Suc k)
       then show ?case 
         proof -
          have 21: "(\<forall> j.  
          (j\<le> ff 0) \<or> (\<exists>i<k. ff(i) <(j::nat) \<and> j\<le> ff(Suc i))
             \<or> (ff k < j \<and> j \<le> ff (Suc k)    ) 
            \<longrightarrow> gg j)" 
             using Suc.prems(1) less_SucI by blast
          have 22: "(\<forall>i<k. ff(i)<ff(Suc i))" 
            using Suc.prems(2) by auto
          have 23: "ff k < ff (Suc k)"
             by (simp add: Suc.prems(2))
          have 24: "(\<forall> j.  
                    (j\<le> ff 0) \<or> (ff 0 <j \<and> j \<le> ff k) 
                     \<or> (ff k < j \<and> j \<le> ff (Suc k)    ) 
                   \<longrightarrow> gg j)"  
            using "21" "22" Suc.hyps by blast
          have 25: "ff 0 < j \<and> j \<le> ff (Suc k) \<longrightarrow> gg j" 
                 using "21" "22" Suc.hyps not_le by blast
          show ?thesis using "25" by blast
         qed
       qed
    qed
  show ?thesis 
  by (metis "2" assms(1) less_or_eq_imp_le linorder_neqE_nat)
qed


lemma PiUntilDistsem3:
 assumes "(\<sigma> \<Turnstile> f \<Pi> (g \<U> h))"
 shows  " (\<sigma> \<Turnstile> ( f \<Pi> g ) \<U> ( f \<Pi> h ) )"
proof -
  have 1: "((\<exists>i\<le>intlen \<sigma>. f (suffix i \<sigma>)) \<and>
    (\<exists>k\<le>intlen (filter f (suffixes \<sigma>)).
        h (suffix k (map (\<lambda>s. nth s 0) (filter f (suffixes \<sigma>)))) \<and>
        (\<forall>j<k. g (suffix j (map (\<lambda>s. nth s 0) (filter f (suffixes \<sigma>)))))))"
   using assms PiUntilDistsem1 by blast
  have 2: "\<exists> x \<in> set(suffixes \<sigma>). f x"
     using "1" in_set_suffixes osfx_suffix by blast
  have 3: "\<forall> x \<in> set(nfilter f (suffixes \<sigma>) 0). f  (nth (suffixes \<sigma>) x)"
   by (simp add: "2")
  have 4: "(\<exists>k\<le>intlen (filter f (suffixes \<sigma>)).
        h (suffix k (map (\<lambda>s. nth s 0) (filter f (suffixes \<sigma>)))) \<and>
        (\<forall>j<k. g (suffix j (map (\<lambda>s. nth s 0) (filter f (suffixes \<sigma>))))))"
    using 1 by auto
  obtain k where 5: "k\<le>intlen (filter f (suffixes \<sigma>)) \<and> 
                     h (suffix k (map (\<lambda>s. nth s 0) (filter f (suffixes \<sigma>)))) \<and>
                     (\<forall>j<k. g (suffix j (map (\<lambda>s. nth s 0) (filter f (suffixes \<sigma>)))))"
    using 4 by auto
  have 6: "f (suffix (nth (nfilter f (suffixes \<sigma>) 0) k)  \<sigma>) "   
    by (metis (no_types, lifting) "2" "3" "5" add.left_neutral nfilter_intlen nfilter_upper_bound 
        nth_set nth_suffixes)
  have 7: "k=0 \<longrightarrow> (\<exists>i\<le>intlen \<sigma> - k. f (suffix (i + k) \<sigma>)) \<and>
       h (map (\<lambda>s. nth s 0) (filter f (suffixes (suffix k \<sigma>)))) \<and>
       (\<forall>j<k. (\<exists>i\<le>intlen \<sigma> - j. f (suffix (i + j) \<sigma>)) \<and> 
              g (map (\<lambda>s. nth s 0) (filter f (suffixes (suffix j \<sigma>)))))"
   using "1" "5" by auto
  have 71:" k>0 \<longrightarrow> (nth (nfilter f (suffixes \<sigma>) 0) (k -1)) \<in> set(nfilter f (suffixes \<sigma>) 0)"
    by (metis "2" "5" diff_le_self le_trans nfilter_intlen nth_set) 
  have 8: "k>0 \<longrightarrow> f (suffix (nth (nfilter f (suffixes \<sigma>) 0) (k -1))  \<sigma>)" 
   by (metis "2" "3" "71" add.left_neutral interval_nth_and_set nfilter_upper_bound nth_suffixes)
  have 9: "k>0 \<longrightarrow> (nth (nfilter f (suffixes \<sigma>) 0) (k -1)) < (nth (nfilter f (suffixes \<sigma>) 0) k)"    
     by (metis "2" "5" One_nat_def Suc_diff_Suc Suc_le_lessD diff_zero idx_nfilter_mono 
         nfilter_intlen)
  have 10: "k>0 \<longrightarrow>  (nth (nfilter f (suffixes \<sigma>) 0) (k -1)) \<le> intlen \<sigma>"
    using  nfilter_upper_bound[of "suffixes \<sigma>" "f" "k-1" "0"] 
    by (simp add: "2" "5" Suc_leD nfilter_intlen) 
    have 11: "k>0 \<longrightarrow> (nth (nfilter f (suffixes \<sigma>) 0) k) \<le> intlen \<sigma>" 
    using nfilter_upper_bound[of "suffixes \<sigma>" "f" "k" "0"] 
    by (simp add: "2" "5" nfilter_intlen)
  have 12: "k>0 \<longrightarrow> 
             h (map (\<lambda>s. nth s 0) 
                    (filter f (suffixes (suffix (nth (nfilter f (suffixes \<sigma>) 0) k) \<sigma>))))" 
     by (metis "11" "5" "6" filter_nfilter_suffix intlen_suffixes map_suffix nth_suffixes 
         suffix_suffixes)  
  have 121: "k>0 \<longrightarrow> 
             h (map (\<lambda>s. nth s 0) 
                    (filter f (suffixes (suffix (Suc (nth (nfilter f (suffixes \<sigma>) 0) (k-1))) \<sigma>))))" 
     by (metis "2" "5" Suc_diff_1 filter_suffixes_map)
  have 13: "k>0 \<longrightarrow> (\<exists>i\<le>intlen \<sigma> - (nth (nfilter f (suffixes \<sigma>) 0) k). 
                      f (suffix (i + (nth (nfilter f (suffixes \<sigma>) 0) k)) \<sigma>))"
        using "6" by auto     
  have 131: "k>0 \<longrightarrow> (\<exists>i\<le>intlen \<sigma> - (Suc (nth (nfilter f (suffixes \<sigma>) 0) (k-1))). 
                       f (suffix (i + (Suc (nth (nfilter f (suffixes \<sigma>) 0) (k-1)))) \<sigma>))"
    using "11" "6" "9" diff_le_mono by fastforce 
  have 14: "k>0 \<longrightarrow> (\<forall> j< (nth (nfilter f (suffixes \<sigma>) 0) k). 
                      (\<exists>i\<le>intlen \<sigma> - j. f (suffix (i + j) \<sigma>)))" 
      using "11" "6" diff_le_mono by fastforce
  have 141: "k>0 \<longrightarrow> (\<forall> j< (Suc (nth (nfilter f (suffixes \<sigma>) 0) (k-1))). 
                        (\<exists>i\<le>intlen \<sigma> - j. f (suffix (i + j) \<sigma>)))" 
    using "14" "9" by auto
  have 15: "(\<forall>j<k. g (suffix j (map (\<lambda>s. nth s 0) (filter f (suffixes \<sigma>)))))" 
     using "5" by blast
  have 151: "(\<forall>j<k. g (map (\<lambda>s. nth s 0) (suffix j  (filter f (suffixes \<sigma>)))))" 
   by (simp add: "5" less_le_trans less_or_eq_imp_le map_suffix)
  have 152: "(\<forall>j<k. (suffix j  (filter f (suffixes \<sigma>))) = 
                    (filter f (suffix (nth (nfilter f (suffixes \<sigma>) 0) j) (suffixes \<sigma>))))"
     by (meson "2" "5" filter_nfilter_suffix_1 le_trans less_imp_le_nat)
  have 16: "(\<forall>j<k. g  (map (\<lambda>s. nth s 0) 
                           (filter f (suffix (nth (nfilter f (suffixes \<sigma>) 0) j) (suffixes \<sigma>))))) "
  using 151 152 filter_nfilter_suffix by simp
  have 1610: " (nth (nfilter f (suffixes \<sigma>) 0) k) \<le> intlen(suffixes \<sigma>)" 
    by (metis "2" "5" diff_zero interval_intlen_gr_zero nfilter_intlen 
        nfilter_upper_bound ordered_cancel_comm_monoid_diff_class.add_diff_inverse)
  have 1611: "(\<forall>j<k.  (nth (nfilter f (suffixes \<sigma>) 0) j) <   (nth (nfilter f (suffixes \<sigma>) 0) k)) "
   by (simp add: "2" "5" idx_nfilter_gr nfilter_intlen) 
  have 1612: "(\<forall>j<k.  (nth (nfilter f (suffixes \<sigma>) 0) j) \<le> intlen(suffixes \<sigma>))"
    using "1610" "1611" by auto
  have 161: "(\<forall>j<k. ( (suffix (nth (nfilter f (suffixes \<sigma>) 0) j) (suffixes \<sigma>))) =
                    ( (suffixes (suffix (nth (nfilter f (suffixes \<sigma>) 0) j) \<sigma>))))"
    using suffix_suffixes using "1612" by blast
  have 17: "(\<forall>j<k. g  (map (\<lambda>s. nth s 0) 
                      (filter f (suffixes (suffix (nth (nfilter f (suffixes \<sigma>) 0) j) \<sigma>)))))" 
       using "16" "161" by auto
  have 18: "k>0 \<longrightarrow>  
            g  (map (\<lambda>s. nth s 0) 
               (filter f (suffixes (suffix (nth (nfilter f (suffixes \<sigma>) 0) (k-1)) \<sigma>))))" 
   using 17 by simp
  have 19: "k>0 \<longrightarrow> 
            g  (map (\<lambda>s. nth s 0) 
                    (filter f (suffixes (suffix (nth (nfilter f (suffixes \<sigma>) 0) 0) \<sigma>))))"
     using "17" by blast 
  have 20: "k>0 \<longrightarrow> (filter f (suffixes (suffix (nth (nfilter f (suffixes \<sigma>) 0) 0) \<sigma>))) =
                    (filter f (suffix (nth (nfilter f (suffixes \<sigma>) 0) 0) (suffixes \<sigma>)))" 
       using "161" by auto 
  have 21: "k>0 \<longrightarrow> (\<forall> j\<le> (nth (nfilter f (suffixes \<sigma>) 0 ) 0). 
                        ( (suffixes (suffix j \<sigma>))) =
                        ( (suffix j (suffixes \<sigma>)))) "
       using "1612" le_trans suffix_suffixes by fastforce
  have 22: "k>0 \<longrightarrow> (\<forall> j\<le> (nth (nfilter f (suffixes \<sigma>) 0 ) 0). 
                        (filter f (suffixes (suffix j \<sigma>))) =
                        (filter f (suffix j (suffixes \<sigma>))))" 
        using "21" by auto
  have 23: " k>0 \<longrightarrow> 
              (\<forall> j\<le> (nth (nfilter f (suffixes \<sigma>) 0 ) 0).  
                (map (\<lambda>s. nth s 0) 
                     (filter f (suffixes (suffix (nth (nfilter f (suffixes \<sigma>) 0) 0) \<sigma>)))) =
                (map (\<lambda>s. nth s 0) (filter f (suffixes (suffix j \<sigma>))))
               ) "
     by (simp add: "2" "21" filter_suffixes_map_help_0)
  have 24: "k>0 \<longrightarrow> 
              (\<forall> j\<le> (nth (nfilter f (suffixes \<sigma>) 0 ) 0).  
                 g (map (\<lambda>s. nth s 0) 
                        (filter f (suffixes (suffix (nth (nfilter f (suffixes \<sigma>) 0) 0) \<sigma>)))) =
                 g (map (\<lambda>s. nth s 0) (filter f (suffixes (suffix j \<sigma>))))
              )" 
    using "23" by auto
  have 241: "k>0 \<longrightarrow> (\<forall> j\<le> (nth (nfilter f (suffixes \<sigma>) 0 ) 0).  
                      g (map (\<lambda>s. nth s 0) (filter f (suffixes (suffix j \<sigma>)))) ) "
     using "19" "24" by blast 
  have 25: "k>0 \<longrightarrow> (\<forall> i < k -1. 
                      (\<forall> l. l\<le>  (nth (nfilter f (suffixes \<sigma>) 0) (Suc i) ) \<and>
                            (nth (nfilter f (suffixes \<sigma>) 0) i ) < l \<longrightarrow>  
                       (filter f (suffix (nth(nfilter f (suffixes \<sigma>) 0) (Suc i)) (suffixes \<sigma>)) =
                       (filter f (suffix l (suffixes  \<sigma>)))     ) )  ) "
       proof 
        assume "k>0"
        show "(\<forall> i < k -1. 
                      (\<forall> l. l\<le>  (nth (nfilter f (suffixes \<sigma>) 0) (Suc i) ) \<and>
                            (nth (nfilter f (suffixes \<sigma>) 0) i ) < l \<longrightarrow>  
                       (filter f (suffix (nth(nfilter f (suffixes \<sigma>) 0) (Suc i)) (suffixes \<sigma>)) =
                       (filter f (suffix l (suffixes  \<sigma>)))     ) )  )"
        proof 
         fix i
         show "i < k - 1 \<longrightarrow>
                 (\<forall>l. l \<le> nth (nfilter f (suffixes \<sigma>) 0) (Suc i) \<and> 
                      nth (nfilter f (suffixes \<sigma>) 0) i < l \<longrightarrow>
                      filter f (suffix (nth (nfilter f (suffixes \<sigma>) 0) (Suc i)) (suffixes \<sigma>)) =
                      filter f (suffix l (suffixes \<sigma>))) "
         proof -
          have 251: "k=1 \<longrightarrow> i < k - 1 \<longrightarrow>
                     (\<forall>l. l \<le> nth (nfilter f (suffixes \<sigma>) 0) (Suc i) \<and> 
                          nth (nfilter f (suffixes \<sigma>) 0) i < l \<longrightarrow>
                        filter f (suffix (nth (nfilter f (suffixes \<sigma>) 0) (Suc i)) (suffixes \<sigma>)) =
                        filter f (suffix l (suffixes \<sigma>)))" 
             by auto
          have 252: "k>1 \<longrightarrow> i < k - 1 \<longrightarrow>
                      (\<forall>l. l \<le> nth (nfilter f (suffixes \<sigma>) 0) (Suc i) \<and> 
                           nth (nfilter f (suffixes \<sigma>) 0) i < l \<longrightarrow>
                          filter f (suffix (nth (nfilter f (suffixes \<sigma>) 0) (Suc i)) (suffixes \<sigma>)) =
                          filter f (suffix l (suffixes \<sigma>)))"
            proof
             assume "k>1"
             show "i < k - 1 \<longrightarrow> 
                          (\<forall>l. l \<le> nth (nfilter f (suffixes \<sigma>) 0) (Suc i) \<and> 
                                nth (nfilter f (suffixes \<sigma>) 0) i < l \<longrightarrow>
                           filter f (suffix (nth (nfilter f (suffixes \<sigma>) 0) (Suc i)) (suffixes \<sigma>)) =
                           filter f (suffix l (suffixes \<sigma>))
                          )"   
             proof
               assume "i < k - 1"
               show "(\<forall>l. l \<le> nth (nfilter f (suffixes \<sigma>) 0) (Suc i) \<and> 
                          nth (nfilter f (suffixes \<sigma>) 0) i < l \<longrightarrow>
                       filter f (suffix (nth (nfilter f (suffixes \<sigma>) 0) (Suc i)) (suffixes \<sigma>)) =
                       filter f (suffix l (suffixes \<sigma>)))"
               proof
               fix l         
                show "l \<le> nth (nfilter f (suffixes \<sigma>) 0) (Suc i) \<and> 
                      nth (nfilter f (suffixes \<sigma>) 0) i < l \<longrightarrow>
                       filter f (suffix (nth (nfilter f (suffixes \<sigma>) 0) (Suc i)) (suffixes \<sigma>)) =
                       filter f (suffix l (suffixes \<sigma>))"
                 using  filter_suffixes_map_help_j[of "l" "f" "suffixes \<sigma>" "i"] 
                  using "2" "5" \<open>i < k - 1\<close> by linarith  
                qed
            qed
          qed
        show ?thesis 
        by (simp add: "252")
      qed
     qed
   qed  
  have 261: "k>0 \<longrightarrow> (\<forall> i < k -1. 
                      (\<forall> l. l\<le>  (nth (nfilter f (suffixes \<sigma>) 0) (Suc i) ) \<and>
                            (nth (nfilter f (suffixes \<sigma>) 0) i ) < l \<longrightarrow> 
                            l \<le> intlen (suffixes \<sigma>) ))
            "
       by (metis "1612" Suc_diff_1 Suc_mono le_eq_less_or_eq less_le_trans)
 have 262: "k>0 \<longrightarrow> (\<forall> i < k -1. 
                      (\<forall> l. l\<le>  (nth (nfilter f (suffixes \<sigma>) 0) (Suc i) ) \<and>
                            (nth (nfilter f (suffixes \<sigma>) 0) i ) < l \<longrightarrow> 
                            (suffix l (suffixes  \<sigma>)) = (suffixes (suffix l \<sigma>)) ))"
     using "261" suffix_suffixes by blast 
  have 26: "k>0 \<longrightarrow> (\<forall> i < k -1. 
                      (\<forall> l. l\<le>  (nth (nfilter f (suffixes \<sigma>) 0) (Suc i) ) \<and>
                            (nth (nfilter f (suffixes \<sigma>) 0) i ) < l \<longrightarrow>  
                       (map (\<lambda>s. nth s 0) 
                         (filter f (suffix (nth(nfilter f (suffixes \<sigma>) 0) (Suc i)) (suffixes \<sigma>))))=
                       (map (\<lambda>s. nth s 0) (filter f (suffixes  (suffix l \<sigma>)))     ) )  ) "
     using "25" "262" by auto
  have 27: "k>0 \<longrightarrow> (\<forall> i < k -1. 
                      (\<forall> l. l\<le>  (nth (nfilter f (suffixes \<sigma>) 0) (Suc i) ) \<and>
                            (nth (nfilter f (suffixes \<sigma>) 0) i ) < l \<longrightarrow>  
                      g (map (\<lambda>s. nth s 0) 
                         (filter f 
                            (suffix (nth(nfilter f (suffixes \<sigma>) 0) (Suc i)) (suffixes \<sigma>))))))"
     by (simp add: "16")
  have 28:  "k>0 \<longrightarrow> (\<forall> i < k -1. 
                      (\<forall> l. l\<le>  (nth (nfilter f (suffixes \<sigma>) 0) (Suc i) ) \<and>
                            (nth (nfilter f (suffixes \<sigma>) 0) i ) < l \<longrightarrow>  
                      g (map (\<lambda>s. nth s 0) (filter f (suffixes  (suffix l \<sigma>)))     )))  "
    using "25" "262" "27" by auto
  have 281: "k>0 \<longrightarrow> 
             (\<forall>j. 
                (j \<le> (nth (nfilter f (suffixes \<sigma>) 0) 0 ) \<or>
                (\<exists> i. i < k -1 \<and> 
                     j\<le>  (nth (nfilter f (suffixes \<sigma>) 0) (Suc i) ) \<and>
                     (nth (nfilter f (suffixes \<sigma>) 0) i ) < j)  ) 
                     \<longrightarrow>  g (map (\<lambda>s. nth s 0) (filter f (suffixes  (suffix j \<sigma>)))     )) "
       using 241 28 by blast
  have 282: "k>0 \<longrightarrow>  (\<forall>i<k - 1. 
                        nth (nfilter f (suffixes \<sigma>) 0) i < nth (nfilter f (suffixes \<sigma>) 0) (Suc i))"
     by (metis "2" "5" Suc_diff_1 Suc_leI idx_nfilter_gr le_SucI le_trans lessI nfilter_intlen)
  have 29: "k>0 \<longrightarrow> (\<forall>j< (Suc (nth (nfilter f (suffixes \<sigma>) 0) (k-1))).  
                        g (map (\<lambda>s. nth s 0) (filter f (suffixes (suffix j \<sigma>)))))"     
     using 281 282 cover_a[of "\<lambda> x. nth (nfilter f (suffixes \<sigma>) 0) x " "k-1" 
                              "(\<lambda> j. g (map (\<lambda>s. nth s 0) (filter f (suffixes (suffix j \<sigma>)))))"]
     using "18" less_antisym by blast
  have 30: "k>0 \<longrightarrow> (\<exists>k\<le>intlen \<sigma>.
       (\<exists>i\<le>intlen \<sigma> - k. f (suffix (i + k) \<sigma>)) \<and>
       h (map (\<lambda>s. nth s 0) (filter f (suffixes (suffix k \<sigma>)))) \<and>
       (\<forall>j<k. (\<exists>i\<le>intlen \<sigma> - j. f (suffix (i + j) \<sigma>)) \<and> 
              g (map (\<lambda>s. nth s 0) (filter f (suffixes (suffix j \<sigma>))))))" 
    using 29 121 131 141 
    by (meson "11" "9" Suc_leI less_le_trans)
 have 31: "(\<exists>k\<le>intlen \<sigma>.
       (\<exists>i\<le>intlen \<sigma> - k. f (suffix (i + k) \<sigma>)) \<and>
       h (map (\<lambda>s. nth s 0) (filter f (suffixes (suffix k \<sigma>)))) \<and>
       (\<forall>j<k. (\<exists>i\<le>intlen \<sigma> - j. f (suffix (i + j) \<sigma>)) \<and> 
              g (map (\<lambda>s. nth s 0) (filter f (suffixes (suffix j \<sigma>))))))" 
  using 30 7 by blast
  show ?thesis using 31
  by (simp add: PiUntilDistsem2)
qed 



lemma PiUntilDistsem4:
 assumes "(\<sigma> \<Turnstile> ( f \<Pi> g ) \<U> ( f \<Pi> h ) )"
 shows  "(\<sigma> \<Turnstile> f \<Pi> (g \<U> h)) "
proof -
 have 1: "(\<exists>k\<le>intlen \<sigma>.
       (\<exists>i\<le>intlen \<sigma> - k. f (suffix (i + k) \<sigma>)) \<and>
       h (map (\<lambda>s. nth s 0) (filter f (suffixes (suffix k \<sigma>)))) \<and>
       (\<forall>j<k. (\<exists>i\<le>intlen \<sigma> - j. f (suffix (i + j) \<sigma>)) \<and> 
              g (map (\<lambda>s. nth s 0) (filter f (suffixes (suffix j \<sigma>))))))" 
  using assms by (simp add: PiUntilDistsem2)
 obtain k where 2: "k\<le>intlen \<sigma> \<and> (\<exists>i\<le>intlen \<sigma> - k. f (suffix (i + k) \<sigma>)) \<and>
     h (map (\<lambda>s. nth s 0) (filter f (suffixes (suffix k \<sigma>)))) \<and>
     (\<forall>j<k. (\<exists>i\<le>intlen \<sigma> - j. f (suffix (i + j) \<sigma>)) \<and> 
        g (map (\<lambda>s. nth s 0) (filter f (suffixes (suffix j \<sigma>)))))" 
   using 1 by auto
 have 3: "(\<exists>i\<le>intlen \<sigma> - k. f (suffix (i + k) \<sigma>))"
    using 2 by auto
 have 4: "k \<le> intlen \<sigma>"
   using 2 by auto
 obtain i where 5: "i\<le>intlen \<sigma> - k \<and> f (suffix (i + k) \<sigma>)"
   using 3 by auto 
 have 6: " i+k \<le> intlen \<sigma>"
    using 4 5 Nat.le_diff_conv2 by blast
 have 61: "\<exists>x \<in> set (suffixes (suffix k \<sigma>)). f x"
   by (metis "4" "5" in_set_suffixes interval_suffix_length_good interval_suffix_suffix osfx_suffix)
 have 7: "(\<exists>i\<le>intlen \<sigma>. f (suffix i \<sigma>))"
   using "5" "6" by blast
 have 71: "\<exists>x \<in> set (suffixes \<sigma>). f x"
   using "5" "6" in_set_suffixes osfx_suffix by blast 
 have 72: "\<exists> x \<in> set(nfilter f (suffixes \<sigma>) 0). f (nth (suffixes \<sigma>) x) " 
   by (metis "5" "6" cancel_comm_monoid_add_class.diff_cancel intlen_suffixes le_refl 
       nfilter_holds_not_a nth_set nth_suffixes 
       ordered_cancel_comm_monoid_diff_class.add_diff_inverse)
 have 73: "f (suffix (nth (nfilter f (suffixes (suffix k \<sigma>)) 0) 0)  (suffix k \<sigma>)) "
      using nfilter_holds[of "suffixes (suffix k \<sigma>)" "f" "0"] 
      by (metis "61" diff_zero interval_intlen_gr_zero nfilter_upper_bound nth_set nth_suffixes 
          ordered_cancel_comm_monoid_diff_class.add_diff_inverse)  
 have 74: "f (suffix ((nth (nfilter f (suffixes (suffix k \<sigma>)) 0) 0)+k)  \<sigma>) "
     using "73" by auto 
 have 75: "f (suffix k (suffix (nth (nfilter f (suffixes (suffix k \<sigma>)) 0) 0) \<sigma>))  "
    by (metis "73" add.commute interval_suffix_suffix) 
 have 8: "h (map (\<lambda>s. nth s 0) (filter f (suffixes (suffix k \<sigma>))))"  
    using 2 by auto
 have 9: "(filter f (suffixes (suffix k \<sigma>))) = (filter f (suffix k (suffixes \<sigma>)))"
   by (simp add: "4" suffix_suffixes)
 have 10: "h (map (\<lambda>s. nth s 0) 
             (suffix (intlen(filter f (suffixes \<sigma>)) - intlen (filter f (suffixes (suffix k \<sigma>))))
                     (filter f (suffixes \<sigma>))))"
   using "2" "61" sfxfilter_suffix_suffix by fastforce 
 have 11: "h (suffix (intlen(filter f (suffixes \<sigma>)) - intlen (filter f (suffixes (suffix k \<sigma>))))
                     (map (\<lambda>s. nth s 0) (filter f (suffixes \<sigma>))))" 
  by (metis "10" diff_le_self map_suffix)
 have 12: "(\<forall>j<k. (\<exists>i\<le>intlen \<sigma> - j. f (suffix (i + j) \<sigma>)) \<and> 
              g (map (\<lambda>s. nth s 0) (filter f (suffixes (suffix j \<sigma>)))))"
  using "2" by blast   
 have 13: "(\<forall>j<k. (\<exists>x \<in> set(suffixes (suffix j \<sigma>)). f x) \<and> 
              g (map (\<lambda>s. nth s 0) (filter f (suffixes (suffix j \<sigma>)))))"
    by (metis "2" interval_suffix_length interval_suffix_suffix intlen_suffixes nth_set nth_suffixes) 
 have 14: "(\<forall>j<k. (\<exists>x \<in> set(suffixes (suffix j \<sigma>)). f x) \<and> 
              g (map (\<lambda>s. nth s 0) 
                 (suffix (intlen(filter f (suffixes \<sigma>)) - intlen (filter f (suffixes (suffix j \<sigma>))))
                     (filter f (suffixes \<sigma>)))))"
    using 13 sfxfilter_suffix_suffix 
    by (metis (no_types, lifting) "2" dual_order.strict_iff_order less_le_trans)
 have 15: "(\<forall>j<k. (\<exists>x \<in> set(suffixes (suffix j \<sigma>)). f x) \<and>
              g (suffix (intlen(filter f (suffixes \<sigma>)) - intlen (filter f (suffixes (suffix j \<sigma>))))
                     (map (\<lambda>s. nth s 0)  (filter f (suffixes \<sigma>)))))" 
  by (metis "14" diff_le_self map_suffix)
 have 150: "k=0 \<longrightarrow>(\<forall>jj<(intlen(filter f (suffixes \<sigma>)) - intlen (filter f (suffixes (suffix k \<sigma>)))). 
             g (suffix jj (map (\<lambda>s. nth s 0) (filter f (suffixes \<sigma>))))) "
     by auto
 have 151: "(\<forall>jj<(intlen(filter f (suffixes \<sigma>)) - intlen (filter f (suffixes (suffix k \<sigma>)))). 
               (suffix jj (filter f (suffixes \<sigma>)))  =
               filter f (suffixes (suffix ((intlen \<sigma>) - intlen (nth (filter f (suffixes \<sigma>)) jj)) \<sigma>))
             )"
      by (simp add: "71" sfxfilter_suffix_suffix_a)
 have 153: "(\<forall>jj<(intlen(filter f (suffixes \<sigma>)) - intlen (filter f (suffixes (suffix k \<sigma>)))).
            g (map (\<lambda>s. nth s 0) 
                 (filter f 
                   (suffixes (suffix ((intlen \<sigma>) - intlen (nth (filter f (suffixes \<sigma>)) jj))   \<sigma>))))
            )" 
       using "13" sfx_suffix_upperbound by blast 
 have 1530: "(\<forall>jj<(intlen(filter f (suffixes \<sigma>)) - intlen (filter f (suffixes (suffix k \<sigma>)))).
              g (map (\<lambda>s. nth s 0) (suffix jj  (filter f (suffixes \<sigma>)))))"
        by (simp add: "151" "153")
 have 1531: "(\<forall>jj<(intlen(filter f (suffixes \<sigma>)) - intlen (filter f (suffixes (suffix k \<sigma>)))).
              g  (suffix jj  (map (\<lambda>s. nth s 0) (filter f (suffixes \<sigma>)))))"
     by (metis "1530" diff_le_self interval_suffix_length interval_suffix_length_code less_le_trans 
         less_numeral_extra(3) map_suffix zero_less_diff)
 have 154: "((\<exists>i\<le>intlen \<sigma>. f (suffix i \<sigma>)) \<and>
    (\<exists>k\<le>intlen (filter f (suffixes \<sigma>)).
        h (suffix k (map (\<lambda>s. nth s 0) (filter f (suffixes \<sigma>)))) \<and>
        (\<forall>j<k. g (suffix j (map (\<lambda>s. nth s 0) (filter f (suffixes \<sigma>)))))))" 
    using "11" "1531" "7" diff_le_self by blast
 show ?thesis 
 by (simp add: "154" PiUntilDistsem1)
qed

lemma  PiUntilDistsem:
 " \<sigma> \<Turnstile>  f \<Pi> (g \<U> h) = ( f \<Pi> g ) \<U> ( f \<Pi> h ) "
using PiUntilDistsem3  PiUntilDistsem4 using unl_lift2 by blast

lemma  PiUntilDist:
 " \<turnstile>  f \<Pi> (g \<U> h) = ( f \<Pi> g ) \<U> ( f \<Pi> h ) "
using PiUntilDistsem Valid_def by blast

subsubsection \<open>PiChopstar\<close>


lemma wnextboxnotstatesem:
assumes " k \<le> intlen \<sigma>"
 shows  " (\<forall> j \<le> intlen \<sigma>. k<j \<longrightarrow> \<not> (\<lambda>y. w \<langle>y\<rangle>) (nth \<sigma> j)) =
             (LIFT(wnext (\<box> (init (\<not> w))))) (suffix k \<sigma>)  
             "
using assms  
 proof  (auto simp add: always_defs init_defs wnext_defs)
  show "\<And>j. j \<le> intlen \<sigma> \<Longrightarrow> k < j \<Longrightarrow> w \<langle>nth \<sigma> j\<rangle> \<Longrightarrow> 
        \<forall>n\<le>intlen \<sigma> - Suc k. \<not> w \<langle>nth \<sigma> (Suc (n + k))\<rangle> \<Longrightarrow> False"  
    proof (cases k)
     show "\<And>j. j \<le> intlen \<sigma> \<Longrightarrow> k < j \<Longrightarrow> w \<langle>nth \<sigma> j\<rangle> \<Longrightarrow> 
            \<forall>n\<le>intlen \<sigma> - Suc k. \<not> w \<langle>nth \<sigma> (Suc (n + k))\<rangle> \<Longrightarrow> k = 0 \<Longrightarrow> False" 
      by simp
         (metis Suc_le_mono Suc_pred less_le_trans)
     show "\<And>j nat. j \<le> intlen \<sigma> \<Longrightarrow> k < j \<Longrightarrow> w \<langle>nth \<sigma> j\<rangle> \<Longrightarrow> 
           \<forall>n\<le>intlen \<sigma> - Suc k. \<not> w \<langle>nth \<sigma> (Suc (n + k))\<rangle> \<Longrightarrow> k = Suc nat \<Longrightarrow> False "  
      proof -
      fix j :: nat and nat :: nat
      assume a1: "k < j"
      assume a2: "\<forall>n\<le>intlen \<sigma> - Suc k. \<not> w \<langle>Interval.nth \<sigma> (Suc (n + k))\<rangle>"
      assume a3: "j \<le> intlen \<sigma>"
      assume a4: "w \<langle>nth \<sigma> j\<rangle>"
      have "Suc (k + (j - Suc k)) = j"
      using a1 by simp 
      then show False
      using a4 a3 a2 by (metis diff_add_inverse2 diff_le_mono 
                         linordered_semidom_class.add_diff_inverse not_add_less2) (* 20 ms *)
      qed
   qed
qed

lemma NotStateUntilStateAndsem:
 "(\<sigma> \<Turnstile> (init (\<not> w)) \<U> ((init w) \<and> f) ) =
   (\<exists>k\<le>intlen \<sigma>. w \<langle>nth \<sigma> k\<rangle> \<and> f (suffix k \<sigma>) \<and> (\<forall>j<k. \<not> w \<langle>nth \<sigma> j\<rangle>))
  "
by (auto simp add: until_d_def init_defs) 


lemma StateUntilEqvWPrevChopsem:
  "\<sigma> \<Turnstile> (init w) \<U> f = (wprev (\<box> (init w)));f"
by (auto simp add:  min.absorb1 until_d_def wprev_defs always_defs init_defs chop_defs)


lemma StateUntilEqvWPrevChop:
 "\<turnstile> (init w) \<U> f = (wprev (\<box> (init w)));f"
using StateUntilEqvWPrevChopsem Valid_def by blast

lemma UntilChopDist:
 " \<turnstile> (init w) \<U> (g;h) = ( (init w) \<U> g) ;h"
by (metis ChopAssoc StateUntilEqvWPrevChop inteq_reflection)



lemma PiEmptysem:
 " \<sigma> \<Turnstile> (init w) \<Pi> empty = (init (\<not> w)) \<U> ((init w) \<and> wnext (\<box> (init (\<not> w))))"
proof -
have 1: "(\<sigma> \<Turnstile> (init w) \<Pi> empty) = 
         ((\<exists>x\<in>interval.set \<sigma>. w \<langle>x\<rangle>) \<and> intlen (IntervalFilter.filter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>) = 0)"
 by (simp add: init_defs empty_defs Pistate)
have 2: "((\<exists>x\<in>interval.set \<sigma>. w \<langle>x\<rangle>) \<and> intlen (IntervalFilter.filter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>) = 0) =
        (\<exists> k \<le> intlen \<sigma> . (\<lambda>y. w \<langle>y\<rangle>) (nth \<sigma> k) \<and> 
             (\<forall> j. j<k \<longrightarrow> \<not> (\<lambda>y. w \<langle>y\<rangle>) (nth \<sigma> j)) \<and>
             (\<forall> j \<le> intlen \<sigma>. k<j \<longrightarrow> \<not> (\<lambda>y. w \<langle>y\<rangle>) (nth \<sigma> j)))
        " 
   by (simp add: filter_length_zero_conv_2)
have 3: "(\<exists> k \<le> intlen \<sigma> . (\<lambda>y. w \<langle>y\<rangle>) (nth \<sigma> k) \<and> 
             (\<forall> j. j<k \<longrightarrow> \<not> (\<lambda>y. w \<langle>y\<rangle>) (nth \<sigma> j)) \<and>
             (\<forall> j \<le> intlen \<sigma>. k<j \<longrightarrow> \<not> (\<lambda>y. w \<langle>y\<rangle>) (nth \<sigma> j))) =
           (\<exists>k\<le>intlen \<sigma>.
       w \<langle>nth \<sigma> k\<rangle> \<and>
       (LIFT(wnext (\<box> (init (\<not> w))))) (suffix k \<sigma>) \<and>
       (\<forall>j<k. \<not> w \<langle>nth \<sigma> j\<rangle>))
        " using wnextboxnotstatesem
     by metis
have 4: " (\<exists>k\<le>intlen \<sigma>.
       w \<langle>nth \<sigma> k\<rangle> \<and>
       (LIFT(wnext (\<box> (init (\<not> w))))) (suffix k \<sigma>) \<and>
       (\<forall>j<k. \<not> w \<langle>nth \<sigma> j\<rangle>)) =
         (\<sigma> \<Turnstile> (init (\<not> w)) \<U> ((init w) \<and> wnext (\<box> (init (\<not> w)))))
        "
  by (simp add: NotStateUntilStateAndsem) 
from 1 2 3 4 show ?thesis by auto
qed
  

lemma PiEmpty:
 "\<turnstile> (init w) \<Pi> empty = (init (\<not> w)) \<U> ((init w) \<and> wnext (\<box> (init (\<not> w))))"
using PiEmptysem Valid_def by blast


lemma StatePiBoxStatesem:
 "\<sigma> \<Turnstile> (init w) \<Pi> f = (init w) \<Pi> (f \<and> \<box> (init w))"
proof -
 have 1: "(\<sigma> \<Turnstile> (init w) \<Pi> f) =
          ((\<exists> x \<in> set \<sigma>. w \<langle>x\<rangle>) \<and>  ( (filter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>) \<Turnstile>  f))
        " 
   by (metis Pistate)
 have 2: "(\<sigma> \<Turnstile> (init w) \<Pi> (f \<and> \<box> (init w))) = 
          ((\<exists> x \<in> set \<sigma>. w \<langle>x\<rangle>) \<and>  ( (filter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>) \<Turnstile> f \<and> \<box> (init w)))
         " 
   by (metis Pistate)
 have 3: "( (filter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>) \<Turnstile> f \<and> \<box> (init w))
         = (f (filter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>) \<and>
    (\<forall>n\<le>intlen (filter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>). w \<langle>nth (filter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>) n\<rangle>))" 
   by (simp add: always_defs init_defs)
 have 4: "((\<exists> x \<in> set \<sigma>. w \<langle>x\<rangle>) \<and> (f (filter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>) \<and>
    (\<forall>n\<le>intlen (filter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>). w \<langle>nth (filter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>) n\<rangle>))) =
        ((\<exists> x \<in> set \<sigma>. w \<langle>x\<rangle>) \<and> f (filter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>))
         "
    by (meson filter_nth_aa)
 show ?thesis using "1" "2" "3" "4" by auto
qed

lemma StatePiBoxState:
 "\<turnstile> (init w) \<Pi> f = (init w) \<Pi> (f \<and> \<box> (init w))"
using StatePiBoxStatesem Valid_def by blast

lemma StatePiUntil1:
 "\<turnstile> ( (init (\<not> w)) \<U> ( (init w) \<and> (init w) \<Pi> f )) =
           (wprev (\<box> (init (\<not> w))));( (init w) \<and> (init w) \<Pi> f )  "
using StateUntilEqvWPrevChop by blast
  
lemma StatePiUntilsem2:
 "(\<sigma> \<Turnstile> (wprev (\<box> (init (\<not> w))));( (init w) \<and> (init w) \<Pi> f )) =
          (\<sigma> \<Turnstile> ((wprev (\<box> (init (\<not> w)))); ((init w) \<and> empty)) ; ((init w) \<and> (init w) \<Pi> f ) )"
by (auto simp add: chop_defs init_defs empty_defs min.absorb1)
   fastforce

lemma StatePiUntil2:
 "\<turnstile> (wprev (\<box> (init (\<not> w))));( (init w) \<and> (init w) \<Pi> f ) =
          ( ((wprev (\<box> (init (\<not> w)))); ((init w) \<and> empty)) ; ((init w) \<and> (init w) \<Pi> f ) )"
by (simp add: StatePiUntilsem2 Valid_def)

lemma StatePiUntil3:
 "\<turnstile> ((wprev (\<box> (init (\<not> w)))); ((init w) \<and> empty)) =
          ( ((init (\<not> w)) \<U> ((init w) \<and> wnext (\<box> (init (\<not> w))))) ; ((init w) \<and> empty)) " 
proof -
 have 1: " \<turnstile> (wprev (\<box> (init (\<not> w)))); ((init w) \<and> empty) =
             (init (\<not> w)) \<U> ((init w) \<and> empty) "
        by (meson Prop11 StateUntilEqvWPrevChop)
 have 2: "\<turnstile> ((init w) \<and> empty) = ((init w) \<and> wnext (\<box> (init (\<not> w))));((init w) \<and> empty)"
    by (auto simp add: min.absorb1 Valid_def init_defs empty_defs wnext_defs always_defs chop_defs) 
       (metis Suc_pred eq_imp_le interval_intlen_gr_zero le_neq_trans)
 show ?thesis by (metis "1" "2" UntilChopDist inteq_reflection)
qed


lemma StatePiUntilsem4:
"(\<sigma> \<Turnstile> ((init (\<not> w)) \<U> ((init w) \<and> wnext (\<box> (init (\<not> w))))) ; ((init w) \<and> empty)) =
          (\<sigma> \<Turnstile> ((init w) \<Pi> empty); ((init w) \<and> empty)) " 
 by (metis PiEmpty inteq_reflection)

lemma StatePiUntil4:
 "\<turnstile> ((init (\<not> w)) \<U> ((init w) \<and> wnext (\<box> (init (\<not> w))))) ; ((init w) \<and> empty) =
          ( ((init w) \<Pi> empty); ((init w) \<and> empty))"
by (simp add: StatePiUntilsem4 Valid_def)

lemma StatePiUntilsem:
 "\<sigma> \<Turnstile> (init w) \<Pi> f = (init (\<not> w)) \<U> ( (init w) \<and> (init w) \<Pi> f )"
proof -
 have 2: "(\<sigma> \<Turnstile> (init (\<not> w)) \<U> ( (init w) \<and> (init w) \<Pi> f )) =
          (\<sigma> \<Turnstile> (wprev (\<box> (init (\<not> w))));( (init w) \<and> (init w) \<Pi> f ))
         " 
   using StateUntilEqvWPrevChopsem[of "LIFT(\<not> w)" "LIFT((init w) \<and> (init w) \<Pi> f)" "\<sigma>"]
     by simp
 have 7: "(\<sigma> \<Turnstile> (wprev (\<box> (init (\<not> w))));( (init w) \<and> (init w) \<Pi> f )) = 
          (\<sigma> \<Turnstile> (((init w) \<Pi> empty); ((init w) \<and> empty));( (init w) \<and> (init w) \<Pi> f ))  
         "
     by (metis PiEmpty StatePiUntil2 StatePiUntil3 inteq_reflection) 
 have 8: "(\<sigma> \<Turnstile> (((init w) \<Pi> empty); ((init w) \<and> empty));( (init w) \<and> (init w) \<Pi> f )) =
          (\<sigma> \<Turnstile> (((init w) \<Pi> empty));( (init w) \<and> (init w) \<Pi> f ))
         " 
   by (auto simp add: chop_defs init_defs empty_defs min.absorb1)
      fastforce
 have 9: "(\<sigma> \<Turnstile> (((init w) \<Pi> empty));( (init w) \<and> (init w) \<Pi> f )) =
          (\<sigma> \<Turnstile> (init w) \<Pi> (empty;f))
         "    
    using PiChopDistsema PiChopDistsemb by blast
 have 10: "(\<sigma> \<Turnstile> (init w) \<Pi> (empty;f)) = (\<sigma> \<Turnstile> (init w) \<Pi> f)"
     by (metis chop_defs empty_defs interval_prefix_length_good interval_suffix_zero pi_d_def 
         zero_order(1)) 
 show ?thesis 
 by (simp add: "10" "2" "7" "8" "9")
qed

lemma StatePiUntil:
 "\<turnstile> (init w) \<Pi> f = (init (\<not> w)) \<U> ( (init w) \<and> (init w) \<Pi> f )"
using StatePiUntilsem by blast


lemma StateAndPiEmpty:
 "\<turnstile>  ((init w) \<and> (init w) \<Pi> empty) = (w \<and> empty) ; (wnext (\<box> (init (\<not> w))))"
proof -
 have 1: "\<turnstile>  ((init w) \<and> (init w) \<Pi> empty) = 
             ((init w) \<and> (init (\<not> w)) \<U> ( (init w) \<and> (wnext (\<box> (init (\<not> w))))) ) "
   using PiEmpty by fastforce 
 have 2: "\<turnstile> ((init w) \<and> (init (\<not> w)) \<U> ( (init w) \<and> (wnext (\<box> (init (\<not> w))))) )
             = ((init w) \<and> (wnext (\<box> (init (\<not> w))))) "
    by (auto simp add: until_d_def Valid_def init_defs)
      force
 have 3: "\<turnstile> ((init w) \<and> (wnext (\<box> (init (\<not> w))))) = (w \<and> empty) ; (wnext (\<box> (init (\<not> w))))"
   by (metis InitAndEmptyEqvAndEmpty StateAndEmptyChop inteq_reflection) 
 show ?thesis 
 using "1" "2" "3" by fastforce
qed

lemma PiPowerExpandsem:
 " (\<sigma> \<Turnstile> (\<exists>k. (init w) \<Pi> (power f k))) = 
    
     (\<sigma> \<Turnstile> (init w) \<Pi> empty \<or> (\<exists>k. (init w) \<Pi> (f;(power f (k))))) "
proof -
 have 1: "(\<sigma> \<Turnstile> (\<exists>k. (init w) \<Pi> (power f k))) =
          (\<exists> k. (\<sigma> \<Turnstile> (init w) \<Pi> (power f k)))  "
     by simp 
 have 2: "(\<exists> k. (\<sigma> \<Turnstile> (init w) \<Pi> (power f k))) =
          ( (\<sigma> \<Turnstile> (init w) \<Pi> (power f 0)) \<or>     (\<exists> k. 1 \<le> k \<and> (\<sigma> \<Turnstile> (init w) \<Pi> (power f (k))))) "
         by (metis One_nat_def diff_Suc_1 le_SucE le_add1 plus_1_eq_Suc)
 have 3: "(\<sigma> \<Turnstile> (init w) \<Pi> (power f 0)) = (\<sigma> \<Turnstile> (init w) \<Pi> empty)" 
   by simp
 have 4: "(\<exists> k. 1 \<le> k \<and> (\<sigma> \<Turnstile> (init w) \<Pi> (power f (k)))) =
          (\<exists> k.  (\<sigma> \<Turnstile> (init w) \<Pi> (power f (Suc k))))" 
     by (metis le_add1 ordered_cancel_comm_monoid_diff_class.add_diff_inverse plus_1_eq_Suc)
 have 5: "(\<exists> k.  (\<sigma> \<Turnstile> (init w) \<Pi> (power f (Suc k)))) =
          (\<sigma> \<Turnstile> (\<exists>k. (init w) \<Pi> (power f (Suc k))))
         " 
      by simp
 have 6: " ((\<sigma> \<Turnstile> (init w) \<Pi> empty) \<or> (\<sigma> \<Turnstile> (\<exists>k. (init w) \<Pi> (power f (Suc k))))) = 
           (\<sigma> \<Turnstile> (init w) \<Pi> empty \<or> (\<exists>k. (init w) \<Pi> (f;power f (k))) )  "
    by auto
 show ?thesis 
 using "1" "2" "3" "4" "5" "6" by blast
qed


lemma PiPowerExpandsem1:
 " \<forall> \<sigma>. \<sigma> \<Turnstile>  (\<exists>k. (init w) \<Pi> (power f k)) = 
     ( (init w) \<Pi> empty   \<or> (\<exists>k. (init w) \<Pi> (power f (Suc k))  ) )     "
proof 
  fix \<sigma>
  show "\<sigma> \<Turnstile>  (\<exists>k. (init w) \<Pi> (power f k)) = 
              ( (init w) \<Pi> empty   \<or> (\<exists>k. (init w) \<Pi> (power f (Suc k))  ) )" 
   proof -
     have 1: "(\<sigma> \<Turnstile>  (\<exists>k. (init w) \<Pi> (power f k)) = 
              ( (init w) \<Pi> empty   \<or> (\<exists>k. (init w) \<Pi> (power f (Suc k))  ) )) 
               = ( (\<sigma> \<Turnstile>  (\<exists>k. (init w) \<Pi> (power f k))) = 
                   (\<sigma> \<Turnstile> (init w) \<Pi> empty   \<or> (\<exists>k. (init w) \<Pi> (power f (Suc k))  ) ) )
              " 
       by auto
     have 2: "( (\<sigma> \<Turnstile>  (\<exists>k. (init w) \<Pi> (power f k))) = 
                   (\<sigma> \<Turnstile> (init w) \<Pi> empty   \<or> (\<exists>k. (init w) \<Pi> (power f (Suc k))  ) ) )" 
        using PiPowerExpandsem[of "w" "f" "\<sigma>"]  by simp
     show ?thesis 
     using "1" "2" by blast
   qed
qed

lemma PiPowerExpand:
 " \<turnstile>  (\<exists>k. (init w) \<Pi> (power f k)) = 
     ( (init w) \<Pi> empty   \<or> (\<exists>k. (init w) \<Pi> (power f (Suc k))  ) )     "
using PiPowerExpandsem1[of "w" "f" ] by (auto simp add: Valid_def PiPowerExpandsem1)


lemma exists_expand_sem:
 "(\<sigma> \<Turnstile> (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) k))) = 
  ((\<sigma> \<Turnstile> (power ((init w) \<Pi> f \<and> fin w) 0)) \<or> 
   (\<sigma> \<Turnstile> (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) (Suc k)))))"
by (metis (no_types, lifting) not0_implies_Suc unl_Rex)  

lemma exists_expand:
 "\<turnstile>  (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) k)) = 
     ((power ((init w) \<Pi> f \<and> fin w) 0) \<or> (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) (Suc k))))  "
using exists_expand_sem Valid_def by fastforce



subsubsection \<open>TruePiEqv\<close>

lemma TruePiEqvsem:
 "\<sigma> \<Turnstile> #True \<Pi> f = f"
by (simp add: pi_d_def pifilt_true) auto

lemma TruePiEqv:
 "\<turnstile> (#True) \<Pi> f = f"
using  TruePiEqvsem by  (auto simp add: Valid_def)

subsubsection \<open>BoxImpEqvPi\<close>

lemma BoxImpEqvPi:
 "\<turnstile> \<box> f \<longrightarrow> g = f \<Pi> g "
by (simp add: Valid_def always_defs  pi_d_def pifilt_def sfxfilt_def) 
   (metis filter_True interval_intlen_gr_zero interval_nth_and_set intlen_suffixes
     map_first_suffixes nth_suffixes)


subsubsection \<open>PiEqvDiamondUPi\<close>

lemma PiEqvDiamondUPi:
 "\<turnstile> f \<Pi> g = ( \<diamond> f \<and> f \<Pi>\<^sup>u g)"
by (simp add: Valid_def upi_d_def sometimes_defs pi_d_def,blast)

subsubsection \<open>PiEqvUntilPi\<close>

lemma PiEqvUntilPi:
 "\<turnstile> (init w) \<Pi> g = (init (\<not> w)) \<U> ((init w) \<Pi> g)"
by (metis StatePiUntil UntilUntilsem Valid_def inteq_reflection)



subsubsection \<open>UPiEqvBoxOrPi\<close>

lemma UPiEqvBoxOrPi:
 "\<turnstile> f \<Pi>\<^sup>u g = (\<box> (\<not> f) \<or> f \<Pi> g)"
by (simp add: Valid_def upi_d_def always_defs pi_d_def,blast)

subsection \<open>Theorems\<close>

lemma UPiImpRule:
 assumes "\<turnstile> g1 \<longrightarrow> g2"
 shows   "\<turnstile> f \<Pi>\<^sup>u g1 \<longrightarrow> f \<Pi>\<^sup>u g2"
using assms 
by (meson MP PiK PiN)

lemma UPiEqvRule:
 assumes "\<turnstile> g1 = g2"
 shows   "\<turnstile> f \<Pi>\<^sup>u g1 = f \<Pi>\<^sup>u g2"
proof -
 have 1: "\<turnstile>  g1 \<longrightarrow> g2"
   using assms by (simp add: int_iffD1)
 have 2: "\<turnstile> f \<Pi>\<^sup>u g1 \<longrightarrow> f \<Pi>\<^sup>u g2"
   using "1" UPiImpRule by blast
 have 3: "\<turnstile> g2 \<longrightarrow> g1"
   using assms by (simp add: int_iffD2)
 have 4: "\<turnstile> f \<Pi>\<^sup>u g2 \<longrightarrow> f \<Pi>\<^sup>u g1"
   using "3" UPiImpRule by blast
 from 3 4 show ?thesis 
 by (simp add: "2" int_iffI)
qed

lemma PiEqvNotUPiNot:
 "\<turnstile>  f \<Pi> g = (\<not> (f \<Pi>\<^sup>u (\<not> g)))"
by (simp add: upi_d_def)

lemma NotPiEqvNotUPi:
 "\<turnstile>  f \<Pi> (\<not> g) = (\<not> (f \<Pi>\<^sup>u g))"
by (simp add: upi_d_def)

lemma UPiEqvNotPiNot:
 "\<turnstile> f \<Pi>\<^sup>u g = (\<not> (f \<Pi> (\<not> g)))"
by (simp add: upi_d_def)

lemma NotUPiEqvNotPi:
 "\<turnstile> f \<Pi>\<^sup>u (\<not>g) = (\<not> (f \<Pi> g))"
by (simp add: upi_d_def)

lemma PiImpRule:
 assumes "\<turnstile> g1 \<longrightarrow> g2"
 shows   "\<turnstile> f \<Pi> g1 \<longrightarrow> f \<Pi> g2"
proof -
 have 1: "\<turnstile> \<not> g2 \<longrightarrow> \<not> g1"
  by (simp add: assms) 
 have 2: "\<turnstile> f \<Pi>\<^sup>u (\<not> g2) \<longrightarrow> f \<Pi>\<^sup>u (\<not>g1)"
   using 1 UPiImpRule by blast
 have 3: "\<turnstile> \<not>(f \<Pi>\<^sup>u (\<not>g1)) \<longrightarrow> \<not>(f \<Pi>\<^sup>u (\<not>g2))"
  using 2 by fastforce
 from 3 show ?thesis using PiEqvNotUPiNot by fastforce
qed

lemma PiEqvRule:
 assumes "\<turnstile> g1 = g2"
 shows   "\<turnstile> f \<Pi> g1 = f \<Pi> g2"
proof -
 have 1: "\<turnstile> g1 \<longrightarrow> g2"
  using assms by (simp add: int_iffD1)
 have 2: "\<turnstile> f \<Pi> g1 \<longrightarrow> f \<Pi> g2"
   using "1" PiImpRule by blast
 have 3: "\<turnstile> g2 \<longrightarrow> g1 "
   using assms by (simp add: int_iffD2)
 have 4:"\<turnstile> f \<Pi> g2 \<longrightarrow> f \<Pi> g1"
  using "3" PiImpRule by blast
 from 2 4 show ?thesis by (simp add: int_iffI)
qed


lemma UPiAndPiImpPiAnd:
 "\<turnstile> f1 \<Pi>\<^sup>u f \<and> f1 \<Pi> (\<not> g) \<longrightarrow> f1 \<Pi> (f \<and> \<not>g)"
proof -
 have 1: "\<turnstile> (\<not>(f \<longrightarrow> g)) = (f \<and> \<not>g)"
  by fastforce
 have 2: "\<turnstile> (\<not> (f1 \<Pi>\<^sup>u (f \<longrightarrow>g))) = f1 \<Pi> (\<not>(f \<longrightarrow> g))"
   by (simp add: NotPiEqvNotUPi int_iffD1 int_iffD2 int_iffI)
 have 3: "\<turnstile>  \<not>(  f1 \<Pi>\<^sup>u f \<longrightarrow> f1 \<Pi>\<^sup>u g ) \<longrightarrow> \<not> (f1 \<Pi>\<^sup>u (f \<longrightarrow>g))"
   by (simp add: PiK) 
 have 4: "\<turnstile> (\<not>(  f1 \<Pi>\<^sup>u f \<longrightarrow> f1 \<Pi>\<^sup>u g ) ) = (f1 \<Pi>\<^sup>u f \<and> f1 \<Pi> (\<not> g))"
   using NotPiEqvNotUPi[of "f1" "g"] by fastforce  
 have 5:"\<turnstile> f1 \<Pi> (\<not>(f \<longrightarrow> g)) = f1 \<Pi> (f \<and> \<not>g)" 
    using 1 by (simp add: PiEqvRule)
 from 1 2 3 4 5 show ?thesis by fastforce
qed
 
lemma UPiAndPiImpPiAndA:
 "\<turnstile> f1 \<Pi>\<^sup>u f \<and> f1 \<Pi> g \<longrightarrow> f1 \<Pi> (f \<and> g)"
using UPiAndPiImpPiAnd[of "f1" "f" "LIFT(\<not>g)"] by fastforce


lemma PiAndPiImpPiAnd:
 "\<turnstile> f1 \<Pi> f \<and> f1 \<Pi> g \<longrightarrow> f1 \<Pi> (f \<and> g)"
proof -
 have 1: "\<turnstile>  f1 \<Pi>\<^sup>u f \<and> f1 \<Pi> g \<longrightarrow> f1 \<Pi> (f \<and> g) "
   using UPiAndPiImpPiAndA by fastforce
 have 2: "\<turnstile> f1 \<Pi> f \<longrightarrow> f1 \<Pi>\<^sup>u f"
  using PiDc by blast 
 from 1 2 show ?thesis by fastforce
qed


lemma PiAnd:
 "\<turnstile> f \<Pi> (g1 \<and> g2) = (f \<Pi> g1 \<and> f \<Pi> g2)"
proof - 
 have 1: "\<turnstile> f \<Pi> (g1 \<and> g2) \<longrightarrow> f \<Pi> g1"
   by (meson PiImpRule Prop12 int_iffD1 lift_and_com)
 have 2: "\<turnstile> f \<Pi> (g1 \<and> g2) \<longrightarrow> f \<Pi> g2"
   by (meson PiImpRule Prop12 int_iffD1 lift_and_com)
 have 3: "\<turnstile> f \<Pi> (g1 \<and> g2) \<longrightarrow> f \<Pi> g1 \<and> f \<Pi> g2" 
   using 1 2 by fastforce
 have 4: "\<turnstile> f \<Pi> g1 \<and> f \<Pi> g2 \<longrightarrow> f \<Pi> (g1 \<and> g2)"
   by (simp add: PiAndPiImpPiAnd)
 from  3 4 show ?thesis by fastforce
qed

 
  

lemma UPiAnd:
 "\<turnstile> f \<Pi>\<^sup>u (g1 \<and> g2) = ( f \<Pi>\<^sup>u g1 \<and> f \<Pi>\<^sup>u g2 )"
proof -
 have 1: "\<turnstile> f \<Pi> (\<not>g1 \<or> \<not> g2) = (f \<Pi> (\<not>g1) \<or> f \<Pi> (\<not> g2)) "
  by (simp add: PiOr)
 have 2: "\<turnstile> (\<not>(f \<Pi> (\<not>g1 \<or> \<not> g2))) = (\<not>(f \<Pi> (\<not>g1) \<or> f \<Pi> (\<not> g2)))"
   using 1 by fastforce
 have 3: "\<turnstile> (\<not>(f \<Pi> (\<not>g1 \<or> \<not> g2))) = f \<Pi>\<^sup>u (\<not>(\<not>g1 \<or> \<not> g2))"
   by (meson NotUPiEqvNotPi Prop11) 
 have 4: "\<turnstile> (\<not>(\<not>g1 \<or> \<not> g2)) = (g1 \<and> g2)"
   by fastforce
 have 5: "\<turnstile> f \<Pi>\<^sup>u (\<not>(\<not>g1 \<or> \<not> g2)) = f \<Pi>\<^sup>u (g1 \<and> g2)"
  using 4 by (simp add: UPiEqvRule)
 have 6: "\<turnstile> (\<not>(f \<Pi> (\<not>g1) \<or> f \<Pi> (\<not> g2))) = (\<not>(f \<Pi> (\<not>g1)) \<and> \<not>(f \<Pi>(\<not>g2)))"
   by fastforce
 have 7: "\<turnstile>\<not>(f \<Pi> (\<not>g1)) = f \<Pi>\<^sup>u g1"
   by (simp add: NotPiEqvNotUPi)  
 have 8: "\<turnstile> \<not>(f \<Pi> (\<not>g2)) = f \<Pi>\<^sup>u g2 "
   by (simp add: NotPiEqvNotUPi)  
 have 9: "\<turnstile> (\<not>(f \<Pi> (\<not>g1)) \<and> \<not>(f \<Pi>(\<not>g2))) = (f \<Pi>\<^sup>u g1 \<and> f \<Pi>\<^sup>u g2)"
   using 6 7 8 by fastforce
 from 2 3 5 6 9 show ?thesis by fastforce
qed 


lemma UpiAndImp:
 "\<turnstile> f \<Pi>\<^sup>u (g1 \<longrightarrow> g2) \<and> f \<Pi> g1 \<longrightarrow> f \<Pi> g2"
proof -
 have 2: "\<turnstile> f \<Pi>\<^sup>u ((\<not>g2) \<longrightarrow> (\<not>g1)) \<longrightarrow> ( f \<Pi>\<^sup>u (\<not> g2) \<longrightarrow> f \<Pi>\<^sup>u (\<not> g1) )"
   using PiK by blast
 have 3: "\<turnstile> (f \<Pi>\<^sup>u (\<not>g2) \<longrightarrow> f \<Pi>\<^sup>u (\<not>g1)) = ((\<not> (f \<Pi>\<^sup>u (\<not>g1))) \<longrightarrow> (\<not>(f \<Pi>\<^sup>u (\<not>g2))))"
   by auto
 have 4: "\<turnstile> (\<not> (f \<Pi>\<^sup>u (\<not>g2))) = f \<Pi> g2" 
    by (simp add: upi_d_def)
 have 5: "\<turnstile> (\<not> (f \<Pi>\<^sup>u (\<not>g1))) = f \<Pi> g1" 
    by (simp add: upi_d_def)
 have 6: "\<turnstile> f \<Pi>\<^sup>u ((\<not>g2) \<longrightarrow> (\<not>g1)) = f \<Pi>\<^sup>u (g1 \<longrightarrow> g2)"
    by simp
 have 7: "\<turnstile> (f \<Pi>\<^sup>u (g1 \<longrightarrow> g2) \<and> f \<Pi> g1 \<longrightarrow> f \<Pi> g2) =
            (f \<Pi>\<^sup>u (g1 \<longrightarrow> g2)  \<longrightarrow> (f \<Pi> g1 \<longrightarrow> f \<Pi> g2))  "
   by auto
 show ?thesis 
 using "2" "4" "5" by fastforce
qed

lemma BoxImpUPiBox:
 "\<turnstile> \<box> (init w) \<longrightarrow> f \<Pi>\<^sup>u (\<box> (init w))"
proof -
 have 1: "\<turnstile> f \<Pi> (\<diamond> (init (\<not>w))) \<longrightarrow> \<diamond> (init (\<not>w))"
     by (simp add: PiDiamondImp) 
 have 2: "\<turnstile> \<not>  \<diamond> (init (\<not>w)) \<longrightarrow> \<not> (f \<Pi> (\<diamond> (init (\<not>w))))" 
   using "1" by auto
 have 3: "\<turnstile> (\<not>  \<diamond> (init (\<not>w))) = \<box> (init w)" 
   by (metis "2" Initprop(2) Prop10 always_d_def inteq_reflection)
 have 4: "\<turnstile> (\<not> (f \<Pi> (\<diamond> (init (\<not>w))))) = f \<Pi>\<^sup>u  (\<box> (init w))" 
   by (simp add: upi_d_def)
      (metis "3" int_simps(4) inteq_reflection) 
 show ?thesis 
 using "2" "3" "4" by fastforce
qed

lemma WPrevPi:
 "\<turnstile> (init w) \<Pi> f = (wprev (\<box> (init (\<not> w)))); ((init w) \<and> (init w) \<Pi> f)"
   using StatePiUntil StatePiUntil1 by fastforce


lemma PiChopstarhelp2a:
 "\<turnstile>  (w \<and> empty);(power ( (f;(w \<and> empty)) \<and> more) k) =
            (power (((w \<and> empty);f) \<and> more) k ) ;(w \<and> empty)"
proof (induction k)
case 0
then show ?case 
by (metis ChopEmpty EmptyChop inteq_reflection pow_0)
next
case (Suc k)
then show ?case 
  proof -
   have 1: "\<turnstile> (w \<and> empty);power (f;(w \<and> empty) \<and> more) (Suc k)  =
               (w \<and> empty);((f;(w \<and> empty) \<and> more); power (f;(w \<and> empty) \<and> more) k) "
     by simp
   have 11: "\<turnstile> (f \<and> more);(w \<and> empty) = (fin w \<and> (f \<and> more)) "
      by (metis FinEqvTrueChopAndEmpty TrueChopAndEmptyEqvChopAndEmpty inteq_reflection)
   have 12: "\<turnstile> (f;(w \<and> empty)) = (fin w \<and> f)" 
          by (meson AndFinEqvChopAndEmpty Prop04 lift_and_com)
   have 13: "\<turnstile> (f;(w \<and> empty) \<and> more) = ((fin w \<and> f) \<and> more)" 
       using "12" by auto
   have 14: "\<turnstile> ((fin w \<and> f) \<and> more) = (fin w \<and> (f \<and> more))"
      by fastforce
   have 2: "\<turnstile> (f;(w \<and> empty) \<and> more) = (f \<and> more);(w \<and> empty)"
         using "11" "13" by fastforce
   have 21: "\<turnstile> ((f;(w \<and> empty) \<and> more); power (f;(w \<and> empty) \<and> more) k) =
               ((f \<and> more); (power ((w \<and> empty);f \<and> more) k;(w \<and> empty)))"
      by (metis "2" ChopAssocB Suc.IH int_eq)
   have 3: "\<turnstile> (w \<and> empty);((f;(w \<and> empty) \<and> more); power (f;(w \<and> empty) \<and> more) k) =
              (w \<and> empty);((f \<and> more); (power ((w \<and> empty);f \<and> more) k;(w \<and> empty)))  " 
       by (simp add: "21" RightChopEqvChop)
   have 4: "\<turnstile> (w \<and> empty);((f \<and> more); (power ((w \<and> empty);f \<and> more) k;(w \<and> empty))) =
              ((w \<and> empty);(f \<and> more)); ((power ((w \<and> empty);f \<and> more) k;(w \<and> empty)))"       
       using ChopAssoc by blast
   have 5: "\<turnstile> (w \<and> empty);(f \<and> more) = ((w \<and> empty);f \<and> more) "
       by (auto simp add: Valid_def chop_defs empty_defs more_defs)
   have 6: "\<turnstile> ((w \<and> empty);(f \<and> more)); ((power ((w \<and> empty);f \<and> more) k;(w \<and> empty))) =
             ((w \<and> empty);f \<and> more); ((power ((w \<and> empty);f \<and> more) k;(w \<and> empty))) "
       using "5" LeftChopEqvChop by blast
   have 7: "\<turnstile> ((w \<and> empty);f \<and> more); ((power ((w \<and> empty);f \<and> more) k;(w \<and> empty))) =
              ((power ((w \<and> empty);f \<and> more) (Suc k);(w \<and> empty)))" 
     by (simp add: ChopAssoc)
   show ?thesis
   by (metis "1" "3" "4" "5" "7" int_eq)
   qed
qed

lemma PiChopstarhelp2:
 "\<turnstile> (w \<and> empty);(f;(w \<and> empty))\<^sup>\<star> = ((w \<and> empty);f)\<^sup>\<star>;(w \<and> empty)"
proof -
 have 1: "\<turnstile> (w \<and> empty);(f;(w \<and> empty))\<^sup>\<star> = (w \<and> empty);(\<exists>k. power ( (f;(w \<and> empty)) \<and> more) k)  " 
   by (simp add: chopstar_d_def powerstar_d_def)
 have 2: "\<turnstile> (w \<and> empty);(\<exists>k. power ( (f;(w \<and> empty)) \<and> more) k) =
             (\<exists>k. (w \<and> empty);(power ( (f;(w \<and> empty)) \<and> more) k))     "
    using ChopExist by fastforce 
 have 3: "\<turnstile> (\<exists>k. (w \<and> empty);(power ( (f;(w \<and> empty)) \<and> more) k)) =
            (\<exists>k. (power ((w \<and> empty);f \<and> more) k);(w \<and> empty)) 
            " 
    by (simp add: ExEqvRule PiChopstarhelp2a)
 have 4: "\<turnstile> (\<exists>k. (power ((w \<and> empty);f \<and> more) k);(w \<and> empty)) =
            (\<exists>k. (power ((w \<and> empty);f \<and> more) k));(w \<and> empty)"
    using ExistChop by fastforce
 have 5: "\<turnstile> (\<exists>k. (power ((w \<and> empty);f \<and> more) k));(w \<and> empty) =
            ((w \<and> empty);f)\<^sup>\<star>;(w \<and> empty)" 
   by (simp add: chopstar_d_def powerstar_d_def)
 show ?thesis 
 using "1" "2" "3" "4" "5" by fastforce
qed  

lemma PiPowerSuca:
 "\<turnstile> (init w) \<Pi> (power f (Suc k)) = ((init w) \<Pi> f);((init w) \<and> (init w) \<Pi> (power f k))"
by (simp add: PiChopDist)

lemma PiPowerSucb:
 "\<turnstile> (init w) \<Pi> (power f (Suc k)) = ((init w) \<Pi> f \<and> fin w);((init w) \<and>  (init w) \<Pi> (power f k))"
by (metis (no_types, lifting) AndFinChopEqvStateAndChop ChopAssoc ChopImpDiamond FinChopEqvDiamond
    FinEqvTrueChopAndEmpty InitAndEmptyEqvAndEmpty PiPowerSuca Prop10 StateAndEmptyChop 
   inteq_reflection)


lemma PiPowerSucc:
  "\<turnstile> (init w) \<Pi> (power f (Suc k)) = 
     (power ((init w) \<Pi> f \<and> fin w) (Suc k));((init w) \<and> (init w) \<Pi> empty)"
proof (induction k)
case 0
then show ?case 
by (metis ChopEmpty PiPowerSucb inteq_reflection pow_0 pow_Suc)
next
case (Suc k)
then show ?case 
by (metis AndFinChopEqvStateAndChop AndFinEqvChopAndEmpty ChopAssocB InitAndEmptyEqvAndEmpty 
   PiPowerSuca inteq_reflection pow_Suc)
qed 

lemma PiPowerSucd:
  "\<turnstile> (init w) \<Pi> (power f (Suc k)) = (power ((init w) \<Pi> f \<and> fin w) (Suc k));((init w) \<Pi> empty)"
proof (induction k)
case 0
then show ?case 
by (metis (no_types, lifting) AndFinChopEqvStateAndChop AndFinEqvChopAndEmpty 
    InitAndEmptyEqvAndEmpty PiPowerSuca int_eq pow_0 pow_Suc)
next
case (Suc k)
then show ?case 
by (metis (no_types, lifting) ChopAssoc PiPowerSucc inteq_reflection pow_Suc)
qed

lemma PiChopstar:
 "\<turnstile> (init w) \<Pi> (f\<^sup>\<star>) = (init (\<not> w)) \<U> ((init w) \<and> (((init w) \<Pi> f) \<and> fin w)\<^sup>\<star>;wnext(\<box> (init (\<not> w))))"
proof -
 have 1: "\<turnstile> (init w) \<Pi> (f\<^sup>\<star>) = (init w) \<Pi> (\<exists>k. power f k)"
   by (metis ChopstarEqvPowerstar PiEqvRule powerstar_d_def) 
 have 2: "\<turnstile> (init w) \<Pi> (\<exists>k. power f k) = (\<exists>k. (init w) \<Pi> (power f k))"
   by (simp add: Valid_def pi_d_def) 
 have 3: "\<turnstile> (\<exists>k. (init w) \<Pi> (power f k)) = 
            ( (init w) \<Pi> empty \<or> (\<exists>k. (init w) \<Pi> (power f (Suc k))))"
  using PiPowerExpand by auto
 have 4: "\<turnstile> (\<exists>k. (init w) \<Pi> (power f (Suc k))) =
            (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) (Suc k));((init w) \<Pi> empty))" 
   by (meson ExEqvRule PiPowerSucd)
 have 5: "\<turnstile> (init w) \<Pi> empty = empty ; ((init w) \<Pi> empty)"
   by (simp add: EmptyChop int_iffD1 int_iffD2 int_iffI)
 have 6: "\<turnstile> (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) (Suc k));((init w) \<Pi> empty)) =
            (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) (Suc k)));((init w) \<Pi> empty)"
    by (simp add: Semantics.ExistChop) 
 have 7: "\<turnstile> ( (init w) \<Pi> empty \<or> (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) (Suc k));((init w) \<Pi> empty)))=
            ( empty; ((init w) \<Pi> empty) \<or> 
              (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) (Suc k)));((init w) \<Pi> empty))"
         using "5" "6" by fastforce
 have 8: "\<turnstile> ( empty; ((init w) \<Pi> empty) \<or> 
              (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) (Suc k)));((init w) \<Pi> empty)) =
            ( empty \<or> (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) (Suc k))));((init w) \<Pi> empty)"
     by (meson OrChopEqv Prop11)
 have 9: "\<turnstile> empty = (power ((init w) \<Pi> f \<and> fin w) 0)"
       by simp 
 have 10: "\<turnstile> ( empty \<or> (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) (Suc k)))) = 
             ( (power ((init w) \<Pi> f \<and> fin w) 0) \<or> (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) (Suc k))))"
   by simp
 have 11: "\<turnstile> ( (power ((init w) \<Pi> f \<and> fin w) 0) \<or> (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) (Suc k)))) =
             (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) k))"
     using exists_expand[of "w" "f"] by fastforce   
 have 12: "\<turnstile> ( (init w) \<Pi> empty \<or> 
               (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) (Suc k));((init w) \<Pi> empty))) =
            (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) ( k))); ((init w) \<Pi> empty)"
    by (metis "11" "7" "8" "9" inteq_reflection)
 have 13: "\<turnstile> (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) ( k))); ((init w) \<Pi> empty) =
             ((init w) \<Pi> f \<and> fin w)\<^sup>\<star> ;  ((init w) \<Pi> empty)"
   by (metis ChopstarEqvPowerstar LeftChopEqvChop inteq_reflection powerstar_d_def)
 have 14: "\<turnstile> ((init w) \<Pi> f \<and> fin w)\<^sup>\<star> ;  ((init w) \<Pi> empty) =
             ( ((init w) \<Pi> empty) \<or> 
               (((init w) \<Pi> f \<and> fin w);((init w) \<Pi> f \<and> fin w)\<^sup>\<star>);((init w) \<Pi> empty))"
    using CSChopEqvOrChopPlusChop by blast
 have 15: "\<turnstile> ((init w) \<Pi> empty) =  (init (\<not> w)) \<U> ( (init w) \<and> wnext (\<box> (init (\<not> w))))"
   by (simp add: PiEmpty) 
 have 16:"\<turnstile> (((init w) \<Pi> f \<and> fin w);((init w) \<Pi> f \<and> fin w)\<^sup>\<star>);((init w) \<Pi> empty) =
            (((init w) \<Pi> f \<and> fin w);((init w) \<Pi> f \<and> fin w)\<^sup>\<star>); wnext (\<box> (init (\<not> w)))"
   by (metis (no_types, lifting) AndFinEqvChopAndEmpty ChopAssoc InitAndEmptyEqvAndEmpty 
         PiChopstarhelp2 StateAndEmptyChop StateAndPiEmpty inteq_reflection)
 have 17: "\<turnstile> ((init w) \<Pi> f \<and> fin w) = ((init w) \<Pi> f);((init w) \<and> empty)"
   by (metis AndFinEqvChopAndEmpty InitAndEmptyEqvAndEmpty inteq_reflection)
 have 18: "\<turnstile> ((init w) \<Pi> f) = wprev(\<box> (init (\<not> w)));((init w) \<and> (init w) \<Pi> f)"
   by (simp add: WPrevPi)
 have 19: "\<turnstile> wprev(\<box> (init (\<not> w)));((init w) \<and> (init w) \<Pi> f) = 
             wprev(\<box> (init (\<not> w)));(((init w) \<and> empty) ; ((init w) \<Pi> f))"
   by (metis RightChopEqvChop StateAndEmptyChop inteq_reflection)
 have 20: "\<turnstile> wprev(\<box> (init (\<not> w)));(((init w) \<and> empty) ; ((init w) \<Pi> f)) =
             (init (\<not> w)) \<U> ( (init w) \<and> ((init w) \<Pi> f)) "
   using "18" "19" StatePiUntil by fastforce   
 have 21: "\<turnstile> (((init w) \<Pi> f \<and> fin w);((init w) \<Pi> f \<and> fin w)\<^sup>\<star>); wnext (\<box> (init (\<not> w))) =
             (init (\<not> w)) \<U> (
                       (init w) \<and> 
                       ((((init w) \<Pi> f) \<and> fin w);((init w) \<Pi> f \<and> fin w)\<^sup>\<star>);wnext (\<box> (init (\<not> w)))
                             )"
    by (metis "17" "18" "19" "20" StateAndChop UntilChopDist inteq_reflection)   
 have 22: "\<turnstile> ((init (\<not> w)) \<U> ( (init w) \<and> wnext (\<box> (init (\<not> w)))) \<or>
             (init (\<not> w)) \<U> ( 
                    (init w) \<and> 
                    ((((init w) \<Pi> f) \<and> fin w);((init w) \<Pi> f \<and> fin w)\<^sup>\<star>);wnext (\<box> (init (\<not> w)))
                            ) ) =
              (init (\<not> w)) \<U>  ( 
         ((init w) \<and> wnext (\<box> (init (\<not> w)))) \<or>
         ( (init w) \<and> ((((init w) \<Pi> f) \<and> fin w);((init w) \<Pi> f \<and> fin w)\<^sup>\<star>);wnext (\<box> (init (\<not> w))))
                               ) "
       using UntilOrDist by fastforce 
 have 23: "\<turnstile> ( 
        ((init w) \<and> wnext (\<box> (init (\<not> w)))) \<or>
        ( (init w) \<and> ((((init w) \<Pi> f) \<and> fin w);((init w) \<Pi> f \<and> fin w)\<^sup>\<star>);wnext (\<box> (init (\<not> w))))
             ) =
             ( (init w) \<and> (((init w) \<Pi> f) \<and> fin w)\<^sup>\<star>;wnext(\<box> (init (\<not> w))))"
   by (metis (no_types, lifting) CSChopEqvOrChopPlusChop ChopOrEqv StateAndEmptyChop 
       inteq_reflection) 
 have 24: "\<turnstile> (init w) \<Pi> (f\<^sup>\<star>) = ( (init w) \<Pi> empty \<or> (\<exists>k. (init w) \<Pi> (power f (Suc k))))" 
   using "1" "2" "3" by fastforce
 have 25: "\<turnstile> ( (init w) \<Pi> empty \<or> (\<exists>k. (init w) \<Pi> (power f (Suc k)))) =
             ( (init w) \<Pi> empty \<or> (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) (Suc k));((init w) \<Pi> empty)))" 
   using "4" by fastforce
 have 26: "\<turnstile> ( (init w) \<Pi> empty \<or> 
               (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) (Suc k));((init w) \<Pi> empty))) =
             ((init w) \<Pi> f \<and> fin w)\<^sup>\<star> ;  ((init w) \<Pi> empty)"
    using "12" "13" by fastforce
 have 27:"\<turnstile> ((init w) \<Pi> f \<and> fin w)\<^sup>\<star> ;  ((init w) \<Pi> empty) =
            (init (\<not> w)) \<U> ((init w) \<and> (((init w) \<Pi> f) \<and> fin w)\<^sup>\<star>;wnext(\<box> (init (\<not> w))))"
    by (metis "14" "15" "16" "21" "22" "23" inteq_reflection)
 from 24 25 26 27 show ?thesis by fastforce
qed 








end