(*  Title     : A Shallow Encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)

section \<open>Projection operator\<close>   

theory Projection
imports Fuse TimeReversal
begin
(*
sledgehammer_params [minimize=true,preplay_timeout=10,timeout=60,verbose=true,
                    provers="
  cvc4 z3 e spass vampire
"]
*)
text \<open>
This theory introduces the projection operator \cite{moszkowski95:_compos}. 
The projection operator is defined and we prove the soundness of the rules and axiom system. 
\<close>

subsection \<open>Definitions\<close>

primrec filt ::  "'a interval \<Rightarrow> index \<Rightarrow> 'a interval"
 where    "filt \<sigma> \<langle>l\<rangle>     = \<langle>nth \<sigma> l\<rangle>"
        | "filt \<sigma> (x\<odot>ls) = (nth \<sigma> x) \<odot> filt \<sigma> ls"
   
primrec lsum :: "'a interval interval \<Rightarrow> nat \<Rightarrow> index"
 where  "lsum \<langle>xs\<rangle> a = \<langle>a+(intlen xs)\<rangle>"
      | "lsum (xs \<odot> xxs) a = (a+(intlen xs)) \<odot> (lsum xxs (a+(intlen xs)))"

definition addzero :: "index \<Rightarrow> index"
 where "addzero ls = (if intlen ls = 0 then 
        (if intfirst ls = 0 then ls else 0 \<odot> ls) else 0 \<odot> ls)"

definition powerinterval :: "('a::world) formula \<Rightarrow> 'a interval \<Rightarrow> index \<Rightarrow> bool"
 where "powerinterval F \<sigma> l = (\<forall> i. i<intlen l \<longrightarrow> ((sub (nth l i) (nth l (Suc i)) \<sigma>) \<Turnstile> F))"

definition cpl :: "('a::world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a interval \<Rightarrow> index"
 where "cpl f g \<sigma> = (\<some> l. index_sequence 0 l \<and> powerinterval f \<sigma> l \<and> 
                          (nth l (intlen l)) = intlen \<sigma> \<and> ((filt \<sigma> l) \<Turnstile> g))"

primrec lcpl :: " ('a ::world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a interval \<Rightarrow> index \<Rightarrow> nat interval interval"
 where  " lcpl  f g \<sigma> \<langle>x\<rangle> = \<langle>(map (shift x) (cpl f g (sub x x \<sigma>)))\<rangle>"
      | " lcpl  f g \<sigma> (x \<odot> xs) =  
            (case xs of \<langle>y\<rangle> \<Rightarrow> \<langle>(map (shift x) (cpl f g (sub x y \<sigma>)))\<rangle> 
             | y \<odot> ys \<Rightarrow> (map (shift x) (cpl f g (sub x y \<sigma>))) \<odot> (lcpl f g \<sigma> xs)) " 


definition projection_d :: "('a:: world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula"
where "projection_d F G \<equiv> \<lambda>s. (\<exists> l. index_sequence 0 l  \<and> (nth l (intlen l)) = (intlen s) \<and>
                                  powerinterval F s l \<and> ((filt s l) \<Turnstile> G)
                          )"

syntax
 "_projection_d"     :: "[lift,lift] \<Rightarrow> lift"        ("(_ \<triangle> _)" [84,84] 83)

syntax (ASCII)
 "_projection_d"     :: "[lift,lift] \<Rightarrow> lift"        ("(_ proj _)" [84,84] 83)

translations
 "_projection_d"    \<rightleftharpoons> "CONST projection_d"

definition uprojection_d :: "('a:: world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula"
where "uprojection_d F G \<equiv> LIFT(\<not>(F \<triangle> (\<not> G)))"


syntax
 "_uprojection_d"     :: "[lift,lift] \<Rightarrow> lift"        ("(_ \<nabla> _)" [84,84] 83)

syntax (ASCII)
 "_uprojection_d"     :: "[lift,lift] \<Rightarrow> lift"        ("(_ uproj _)" [84,84] 83)

translations
 "_uprojection_d"    \<rightleftharpoons> "CONST uprojection_d"

definition dp_d :: "('a:: world) formula \<Rightarrow> 'a formula"
where "dp_d F \<equiv> LIFT(#True \<triangle> F)"

definition bp_d :: "('a:: world) formula \<Rightarrow> 'a formula"
where "bp_d F \<equiv> LIFT(#True \<nabla> F)"

syntax
 "_dp_d"     :: "lift \<Rightarrow> lift"        ("(dp _)" [88] 87)
 "_bp_d"     :: "lift \<Rightarrow> lift"        ("(bp _)" [88] 87)

syntax (ASCII)
 "_dp_d"     :: "lift \<Rightarrow> lift"        ("(dp _)" [88] 87)
 "_bp_d"     :: "lift \<Rightarrow> lift"        ("(bp _)" [88] 87)

translations
 "_dp_d"    \<rightleftharpoons> "CONST dp_d"
 "_bp_d"    \<rightleftharpoons> "CONST bp_d"


subsection \<open>Lemmas\<close>

subsubsection \<open>filt Lemmas\<close>

lemma filt_intlen:
 "intlen(filt s0 l) = intlen l"
by (induct l, simp, simp)

lemma filt_nth:
 assumes " i \<le> intlen (filt s0 l) "
 shows   " (nth (filt s0 l) i) = (nth s0 (nth l i)) "
using assms
proof (induct l arbitrary: i)
case (St x)
then show ?case by simp
next
case (Cons x1a l)
then show ?case by simp (metis Nitpick.case_nat_unfold Suc_eq_plus1 le_diff_conv)
qed



lemma filt_expand: 
 " (s1 = (filt s0 l)) =
   ( intlen s1 = intlen l \<and>
     (\<forall> i \<le> intlen s1. (nth s1 i) = (nth s0 (nth l i))))"
by (metis filt_intlen filt_nth interval_eq_nth_eq )

lemma filt_fuse:
   " filt xs (fuse l1 l2) = (fuse (filt xs l1) (filt xs l2))"
by (induct l1 arbitrary: l2 xs) simp_all


lemma fuse_filt_intlen:
assumes "index_sequence 0 l"
        "(nth l (intlen l)) = intlen xs"
shows  "intlen (filt (fuse (prefix (nth l n) xs) (suffix (nth l n) xs)) l) =
        intlen (fuse (filt (prefix (nth l n) xs) (prefix n l)) 
               (filt (suffix (nth l n) xs) (map (shiftm (nth l n)) (suffix n l) )))"
proof -
 have 1: "intlen (filt (fuse (prefix (nth l n) xs) (suffix (nth l n) xs)) l) = intlen l "
    by (simp add: filt_intlen)
 have 2: "intlen (fuse (filt (prefix (nth l n) xs) (prefix n l)) 
                       (filt (suffix (nth l n) xs) (map (shiftm (nth l n)) (suffix n l) ))) =
          (intlen((filt (prefix (nth l n) xs) (prefix n l))) +
           intlen(filt (suffix (nth l n) xs) (map (shiftm (nth l n)) (suffix n l) ))) "
    using interval_fuse_intlen_a by blast
 have 3: "intlen((filt (prefix (nth l n) xs) (prefix n l))) = intlen(prefix n l) "
    using filt_intlen by blast
 have 4: "intlen(filt (suffix (nth l n) xs) (map (shiftm (nth l n)) (suffix n l) )) =
          intlen(suffix n l)"
    by (simp add: filt_intlen)
 have 5: "intlen(prefix n l) + intlen(suffix n l) = intlen l"
   by (simp)
 from 1 2 3 4 5 show ?thesis by auto
qed
    

lemma fuse_filt_nth_a:
 assumes "index_sequence 0 l"
         "(nth l (intlen l)) = intlen xs"
         "i \<le> intlen l"
         "n \<le> intlen l"
 shows   " nth (filt (fuse (prefix (nth l n) xs) (suffix (nth l n) xs)) l) i =
           nth xs (nth l i)"
proof -
  have 1: "filt (fuse (prefix (nth l n) xs) (suffix (nth l n) xs)) l =
           filt xs l "
     using assms by (metis interval_fuse_prefix_suffix interval_idx_less_last_1 le_neq_implies_less
        less_imp_le_nat not_less)
  have 2: "nth (filt xs l) i = nth xs (nth l i)"
     using assms by (metis filt_nth filt_intlen  )
  from 1 2 show ?thesis by auto
qed


lemma fuse_filt_nth_b:
 assumes "index_sequence 0 l"
         "(nth l (intlen l)) = intlen xs"
         "i \<le> intlen l"
         "n \<le> intlen l"
 shows   "nth (fuse (filt (prefix (nth l n) xs) (prefix n l)) 
                  (filt (suffix (nth l n) xs) (map (shiftm (nth l n)) (suffix n l) ))) i =
          nth xs (nth l i)"
proof -
 have 1: "i \<le> intlen(fuse (filt (prefix (nth l n) xs) (prefix n l)) 
                  (filt (suffix (nth l n) xs) (map (shiftm (nth l n)) (suffix n l) ))) "
    using assms 
    by (metis filt_intlen interval_fuse_intlen_a interval_fuse_prefix_suffix_intlen 
       interval_intlen_map) 
 have 2: " intlast(filt (prefix (nth l n) xs) (prefix n l)) = nth xs (nth l n) "
  using assms filt_nth[of _ "(prefix (nth l n) xs)" "(prefix n l)" ] 
  by (metis filt_intlen interval_idx_less_equal interval_intlast_prefix interval_nth_intlen_intlast 
      interval_prefix_length_good order_refl) 
 have 3: "intfirst(filt (suffix (nth l n) xs) (map (shiftm (nth l n)) (suffix n l) )) =
          nth (filt (suffix (nth l n) xs) (map (shiftm (nth l n)) (suffix n l) )) 0"
    by simp
 have 4: "nth (filt (suffix (nth l n) xs) (map (shiftm (nth l n)) (suffix n l) )) 0 =
          nth (suffix (nth l n) xs) (nth (map (shiftm (nth l n)) (suffix n l) ) 0)"
         using filt_nth by blast
 have 5: "nth (suffix (nth l n) xs) (nth (map (shiftm (nth l n)) (suffix n l) ) 0) =
          nth  (suffix (nth l n) xs) (nth l (n+0) - (nth l n))"
    by (simp add: assms interval_nth_map shiftm_def)   
 have 6: "nth  (suffix (nth l n) xs) (nth l (n+0) - (nth l n)) = 
          nth  (suffix (nth l n) xs) 0 "
    by simp 
 have 7: "nth  (suffix (nth l n) xs) 0 = nth xs (nth l n) "
    using assms by (metis Nat.add_0_right interval_nth_suffix le0  ) 
 have 8: "intlast(filt (prefix (nth l n) xs) (prefix n l)) =
          intfirst(filt (suffix (nth l n) xs) (map (shiftm (nth l n)) (suffix n l) )) "
   using "2" "4" "5" "7" by auto
 have 10: "nth (fuse (filt (prefix (nth l n) xs) (prefix n l)) 
                  (filt (suffix (nth l n) xs) (map (shiftm (nth l n)) (suffix n l) ))) i =
          (if i \<le> intlen (filt (prefix (nth l n) xs) (prefix n l)) 
           then nth (filt (prefix (nth l n) xs) (prefix n l)) i
           else nth (filt (suffix (nth l n) xs) (map (shiftm (nth l n)) (suffix n l) )) 
                    (i - intlen (filt (prefix (nth l n) xs) (prefix n l)))) "
    using "1" "8" interval_fuse_nth by auto
 have 11: "intlen (filt (prefix (nth l n) xs) (prefix n l)) = n "
   using assms by (metis filt_intlen interval_prefix_length_good) 
 have 12: "i\<le>n \<longrightarrow> nth (filt (prefix (nth l n) xs) (prefix n l)) i = 
                   nth (prefix (nth l n) xs) (nth (prefix n l ) i)"
    by (simp add: "11" filt_nth)
 have 13: "i\<le>n \<longrightarrow> nth (prefix (nth l n) xs) (nth (prefix n l ) i) =
                   nth (prefix (nth l n) xs) (nth l i) "
    by (simp add: assms)
 have 15: " i\<le>n \<longrightarrow> nth (prefix (nth l n) xs) (nth l i) = nth xs (nth l i)"
         using assms(1) assms(2) assms(4) interval_idx_less_equal interval_nth_prefix by blast
 have 16: "nth (filt (suffix (nth l n) xs) (map (shiftm (nth l n)) (suffix n l) )) 
                    (i - intlen (filt (prefix (nth l n) xs) (prefix n l)))  =
           nth (filt (suffix (nth l n) xs) (map (shiftm (nth l n)) (suffix n l) )) 
                    (i - n)"
    by (simp add: "11") 
 have 17: "nth (filt (suffix (nth l n) xs) (map (shiftm (nth l n)) (suffix n l) )) (i - n) =
           nth (suffix (nth l n) xs) (nth (map (shiftm (nth l n)) (suffix n l) ) (i -n)) "
    using assms filt_nth[of "i-n" "(suffix (nth l n) xs)" "(map (shiftm (nth l n)) (suffix n l) )" ]
      by (simp add: filt_intlen) 
 have 18: "i> n \<longrightarrow> nth (map (shiftm (nth l n)) (suffix n l) ) (i -n) =
           nth l (n+(i-n)) - (nth l n)"
    using assms 
    by (simp add: interval_idx_shiftm_suffix_nth)
 have 19: "i>n \<longrightarrow> nth (suffix (nth l n) xs) (nth (map (shiftm (nth l n)) (suffix n l) ) (i -n)) 
                   = nth (suffix (nth l n) xs) ((nth l i) - (nth l n))" 
    by (simp add: "18")    
 have 20 : "i>n \<longrightarrow> nth (suffix (nth l n) xs) ((nth l i) - (nth l n)) = 
                    nth xs ((nth l n) + ((nth l i) - (nth l n)))"
    using assms by (metis diff_le_mono interval_idx_less_last_1 interval_nth_suffix 
        le_neq_implies_less less_imp_le_nat nat_le_linear)
 have 22: "i>n \<longrightarrow> (nth l n) + ((nth l i) - (nth l n)) =  (nth l i)"
    using assms using interval_idx_less_equal by fastforce
 have 23: "i>n \<longrightarrow> nth xs ((nth l n) + ((nth l i) - (nth l n))) = nth xs (nth l i) "
    by (simp add: "22")
 from 10 show ?thesis by (simp add: "11" "12" "13" "15" "17" "19" "20" "23")
qed

lemma fuse_filt_nth:
assumes "index_sequence 0 l"
        "(nth l (intlen l)) = intlen xs"
        "i \<le> intlen l"
        "n \<le> intlen l"
shows  "nth (filt (fuse (prefix (nth l n) xs) (suffix (nth l n) xs)) l) i =
        nth (fuse (filt (prefix (nth l n) xs) (prefix n l)) 
                  (filt (suffix (nth l n) xs) (map (shiftm (nth l n)) (suffix n l) ))) i"
using assms fuse_filt_nth_a[of "l" "xs" "i" "n"]   
            fuse_filt_nth_b[of "l" "xs" "i" "n"] by simp

lemma fuse_filt:
assumes "index_sequence 0 l"
        "(nth l (intlen l)) = intlen xs"
        "n \<le> intlen l"
shows  "filt (fuse (prefix (nth l n) xs) (suffix (nth l n) xs)) l =
        fuse (filt (prefix (nth l n) xs) (prefix n l)) 
             (filt (suffix (nth l n) xs) (map (shiftm (nth l n)) (suffix n l) ))"
using assms fuse_filt_intlen fuse_filt_nth interval_eq_nth_eq by (metis filt_intlen)


lemma fuse_filt_a:
 assumes "index_sequence 0 l1"
         "index_sequence (intlast l1) l2"
         "intlast l2 = intlen xs"
 shows   " filt (fuse (prefix (intlast l1) xs) (suffix (intfirst l2) xs)) (fuse l1 l2) =
           fuse (filt (prefix (intlast l1) xs) l1) 
                (filt (suffix (intfirst l2) xs) (map (shiftm (intfirst l2)) l2)) "
proof -
 have 1: "intlast l1 = intfirst l2" 
  using assms by (metis index_sequence_def interval_nth_zero_intfirst)
 have 2: "index_sequence 0 (fuse l1 l2)"
   using assms by (metis index_sequence_def interval_idx_fuse interval_nth_zero_intfirst)
 have 3: "intlast(fuse l1 l2) = intlen xs"
   using assms by (metis "1" add.left_neutral interval_fuse_nth_a interval_fuse_intlen_a 
       interval_nth_intlen_intlast le_add2)  
 have 4: "intlen l1 \<le> intlen (fuse l1 l2) " 
       by (simp add: interval_fuse_intlen_a)
 have 5: "filt (fuse (prefix (nth (fuse l1 l2) (intlen l1)) xs)
                     (suffix (nth (fuse l1 l2) (intlen l1)) xs))
               (fuse l1 l2)
          =
         fuse (filt (prefix (nth (fuse l1 l2) (intlen l1)) xs) (prefix (intlen l1) (fuse l1 l2))) 
              (filt (suffix (nth (fuse l1 l2) (intlen l1)) xs) 
                    (map (shiftm (nth (fuse l1 l2) (intlen l1))) (suffix (intlen l1) (fuse l1 l2))))"
   using "2" "3" "4" fuse_filt by auto
 have 6: "filt (fuse (prefix (intlast l1) xs) (suffix (intfirst l2) xs)) (fuse l1 l2) =
          filt (fuse (prefix (nth (fuse l1 l2) (intlen l1)) xs) 
                     (suffix (nth (fuse l1 l2) (intlen l1)) xs)) 
               (fuse l1 l2)" 
    using "1" "4" interval_intlast_prefix interval_prefix_fuse by fastforce
 have 7: "(filt (prefix (intlast l1) xs) l1) =
          (filt (prefix (nth (fuse l1 l2) (intlen l1)) xs) (prefix (intlen l1) (fuse l1 l2)))" 
    using "1" "4" interval_intlast_prefix interval_prefix_fuse by fastforce
 have 8: "(filt (suffix (intfirst l2) xs) (map (shiftm (intfirst l2)) l2)) =
          (filt (suffix (nth (fuse l1 l2) (intlen l1)) xs) 
                (map (shiftm (nth (fuse l1 l2) (intlen l1))) (suffix (intlen l1) (fuse l1 l2)) ))"
    using "1" "4" interval_intlast_prefix interval_prefix_fuse interval_suffix_fuse by metis
 show ?thesis using "5" "6" "7" "8" by auto
qed  

lemma filt_prefix:
assumes "n \<le> intlen l"
shows   "  prefix n (filt \<sigma> l) = filt \<sigma> (prefix n l) "
proof -
 have 1: "intlen (prefix n (filt \<sigma> l)) = intlen ( filt \<sigma> (prefix n l))"
   by (simp add: assms filt_intlen)
 have 2: "\<forall> i\<le>intlen(prefix n (filt \<sigma> l)). 
           (nth (prefix n (filt \<sigma> l)) i) = (nth ( filt \<sigma> (prefix n l)) i)"
   by (simp add: assms filt_nth filt_intlen le_trans)
 show ?thesis using "1" "2" interval_eq_nth_eq by blast
qed


lemma filt_prefix_idx:
assumes "n \<le> intlen \<sigma>"
        "index_sequence 0 l"
        "nth l (intlen l) = n "
shows  " filt \<sigma> l = filt (prefix n \<sigma>) l"
proof -
 have 1: " intlen(filt \<sigma> l) = intlen (filt (prefix n \<sigma>) l) " 
   by (simp add: filt_intlen)
 have 2: "\<forall>i\<le>intlen(filt \<sigma> l). (nth (filt \<sigma> l) i) = (nth (filt (prefix n \<sigma>) l) i)"
   by (metis assms filt_nth filt_intlen interval_idx_less_equal 
        interval_nth_prefix interval_prefix_length_good order_refl) 
 show ?thesis 
 by (simp add: "1" "2" interval_eq_nth_eq)
qed


lemma filt_suffix:
assumes "n \<le> intlen l"
shows  "  suffix n (filt \<sigma> l) = filt \<sigma> (suffix n l) "
proof -
 have 1: "intlen (suffix n (filt \<sigma> l)) = intlen (filt \<sigma> (suffix n l))"
   by (simp add: assms filt_intlen)
 have 2: "\<forall>i\<le>intlen (suffix n (filt \<sigma> l)). 
   (nth (suffix n (filt \<sigma> l)) i) = (nth (filt \<sigma> (suffix n l)) i) "
   by (simp add: assms filt_nth filt_intlen 
       ordered_cancel_comm_monoid_diff_class.le_diff_conv2)
 show ?thesis using "1" "2" interval_eq_nth_eq by blast
qed


lemma filt_suffix_idx_intlen:
assumes "n \<le> intlen \<sigma>"
        "index_sequence 0 l"
        "nth l (intlen l) = intlen \<sigma> - n"
shows " intlen (filt \<sigma> (map (shift n) l)) = intlen (filt (suffix n \<sigma>) l)"
using assms by (simp add: filt_intlen)

lemma filt_suffix_idx_nth:
assumes "n \<le> intlen \<sigma>"
        "index_sequence 0 l"
        "i\<le>intlen l"
        "nth l (intlen l) = intlen \<sigma> - n"
shows " nth (filt \<sigma> (map (shift n) l)) i = nth (filt (suffix n \<sigma>) l) i"
proof -
 have 1: " nth (filt \<sigma> (map (shift n) l)) i = nth \<sigma> (nth (map (shift n) l) i)  "
 by (simp add: assms filt_nth filt_intlen) 
 have 2: "nth \<sigma> (nth (map (shift n) l) i) = (nth \<sigma> ((nth l i) + n))"
 by (simp add: Interval.shift_def interval_nth_map) 
 have 3: " nth (filt (suffix n \<sigma>) l) i = nth (suffix n \<sigma>) (nth l i)  "  
   by (simp add: assms filt_nth filt_intlen)
 have 4: "nth (suffix n \<sigma>) (nth l i) = (nth \<sigma> ((nth l i) + n)) " 
    by (metis add.commute assms eq_iff interval_idx_less_equal interval_nth_suffix 
        interval_suffix_length_good)
 show ?thesis by (simp add: "1" "2" "3" "4")
qed


lemma filt_suffix_idx:
assumes "n \<le> intlen \<sigma>"
        "index_sequence 0 l"
        "nth l (intlen l) = intlen \<sigma> - n"
shows " filt \<sigma> (map (shift n) l) = filt (suffix n \<sigma>) l"
using assms  filt_suffix_idx_intlen filt_suffix_idx_nth interval_eq_nth_eq
by (simp add: filt_suffix_idx_nth interval_eq_nth_eq filt_intlen)

lemma filt_filt:
 " (nth (filt (filt xxs l1) l2) k) = 
   (nth xxs  (nth l1 (nth l2 k)))"
by (metis filt_expand interval_intlen_map interval_nth_map)

lemma filt_filt_map:
 "(filt (filt xxs l1) l2) = (filt xxs (map (\<lambda> x. (nth l1 x)) l2))"
by (metis filt_expand interval_intlen_map interval_nth_map)

lemma filt_map:
 " filt xs l = map (\<lambda>x. nth xs x) l"
 by (metis filt_expand interval_intlen_map interval_nth_map)

lemma filt_map_filt:
 "(filt (filt xxs l1) l2) = filt xxs (filt l1 l2)"
by (metis filt_filt_map filt_map)

lemma filt_sub:
 assumes "k \<le> n"
         "n \<le> intlen l"
 shows   " (sub k n (filt \<sigma> l)) = (filt \<sigma> (sub k n l))"
using assms 
by (simp add: sub_def filt_prefix filt_suffix)

lemma filt_lfuse_map:
 "  (filt \<sigma> (lfuse (xxs)) )=
     (lfuse (map (\<lambda> xs . (filt \<sigma> xs)) xxs))    "
proof (induct xxs)
case (St x)
then show ?case by simp
next
case (Cons x1a xxs)
then show ?case by (simp add: filt_fuse)
qed



subsubsection \<open>powerinterval lemmas\<close>

lemma powerinterval_splita0:
 assumes " index_sequence 0 l"
         " n \<le> intlen l "
         " nth l (intlen l) = intlen \<sigma>"
         " i < intlen(prefix n l)"
 shows   "  (sub (nth (prefix n l) i) (nth (prefix n l) (Suc i)) (prefix (nth l n) \<sigma>)) =
            (sub (nth l i) (nth l (Suc i)) \<sigma>) 
              " 
proof -
 have 01: "(nth l n) \<le> intlen \<sigma>" 
   by (metis assms(1) assms(2) assms(3) interval_idx_less_last_1 le_less)
 have 02: "(nth (prefix n l) i) \<le> (nth (prefix n l) (Suc i))"
   using assms(1) assms(3) assms(4) interval_idx_expand by fastforce
 have 03: "(nth (prefix n l) (Suc i)) \<le> (nth l n)" 
    using assms(1) assms(2) assms(3) assms(4) interval_idx_less_equal by fastforce
 have 1: "intlen (sub (nth (prefix n l) i) (nth (prefix n l) (Suc i)) (prefix (nth l n) \<sigma>)) =
          intlen(sub (nth l i) (nth l (Suc i)) (\<sigma>)) 
         " using interval_intlen_sub 
         using "01" "02" "03" assms(4) by auto
 have 2: " (\<forall> j \<le> intlen (sub (nth (prefix n l) i) (nth (prefix n l) (Suc i)) (prefix (nth l n) \<sigma>)).
               (nth (sub (nth (prefix n l) i) (nth (prefix n l) (Suc i)) (prefix (nth l n) \<sigma>)) j) =
               (nth (sub (nth l i) (nth l (Suc i)) (\<sigma>)) j)) 
               "
   proof 
    fix j
    show "j \<le> intlen (sub (nth (prefix n l) i) (nth (prefix n l) (Suc i)) (prefix (nth l n) \<sigma>)) \<longrightarrow>
               (nth (sub (nth (prefix n l) i) (nth (prefix n l) (Suc i)) (prefix (nth l n) \<sigma>)) j) =
               (nth (sub (nth l i) (nth l (Suc i)) (\<sigma>)) j)
         "
    proof -
     have 21: "intlen (sub (nth (prefix n l) i) (nth (prefix n l) (Suc i)) (prefix (nth l n) \<sigma>)) = 
               (nth l (Suc i)) - (nth l i) "
        using "1" assms(1) assms(3) assms(4) interval_idx_expand by fastforce
     have 22: "(nth (prefix n l) i) = (nth l i) "
      using assms by auto
     have 23: "(nth (prefix n l) (Suc i)) = (nth l (Suc i)) "
        using assms by auto
     have 24: "j  \<le> (nth l (Suc i)) - (nth l i) \<longrightarrow> 
               nth (sub (nth l i) (nth l (Suc i)) \<sigma>) j = (nth \<sigma> ((nth l i)+j)) "
        using assms interval_idx_expand interval_nth_sub less_le_trans by fastforce
     have 25: "j  \<le> (nth l (Suc i)) - (nth l i) \<longrightarrow> (nth l (Suc i)) \<le> intlen (prefix (nth l n) \<sigma>) "
         using assms(1) assms(2) assms(3) assms(4) interval_idx_expand interval_idx_less_equal 
         by fastforce
     have 26: "j \<le> (nth l (Suc i)) - (nth l i) \<longrightarrow> 
               (nth (sub (nth (prefix n l) i) (nth (prefix n l) (Suc i)) (prefix (nth l n) \<sigma>)) j)  =
               (nth (prefix (nth l n) \<sigma>) ((nth l i)+j)) "
          using "25" assms interval_idx_expand interval_nth_sub less_le_trans by fastforce
     have 27: "(nth l (Suc i))  \<le>(nth l n)" 
        using assms(1) assms(2) assms(3) assms(4) interval_idx_less_equal by fastforce
     have 28: "j \<le> (nth l (Suc i)) - (nth l i) \<longrightarrow> (nth l i)+j\<le>(nth l n)" 
         using "27" assms(1) assms(3) assms(4) interval_idx_expand by fastforce
     have 29: "j \<le> (nth l (Suc i)) - (nth l i) \<longrightarrow> 
                (nth (prefix (nth l n) \<sigma>) ((nth l i)+j)) = (nth \<sigma> ((nth l i)+j))"
         using assms interval_nth_prefix using "28" by blast
     show ?thesis 
     using "21" "22" "23" "24" "26" "29" by auto
   qed
  qed
  show ?thesis 
  using "1" "2" interval_eq_nth_eq by blast
qed

lemma powerinterval_splita:
 assumes " index_sequence 0 l"
         " n \<le> intlen l"
         " nth l (intlen l) = intlen \<sigma>"
         " powerinterval f \<sigma> l "
 shows   " powerinterval f (prefix (nth l n) \<sigma>) (prefix n l)"
proof -
 have 0: "(\<forall> i. i<intlen l \<longrightarrow> ( (sub (nth l i) (nth l (Suc i)) \<sigma>) \<Turnstile> f)) "
   using assms by (simp add: powerinterval_def)
 have 1: " powerinterval f (prefix (nth l n) \<sigma>) (prefix n l) =
           (\<forall>i. i < intlen(prefix n l) \<longrightarrow>
                ((sub (nth (prefix n l) i) (nth (prefix n l) (Suc i)) (prefix (nth l n) \<sigma>)) \<Turnstile> f)) "
       using powerinterval_def by blast
 have 2: "(\<forall>i. i < intlen(prefix n l) \<longrightarrow>
                ((sub (nth (prefix n l) i) (nth (prefix n l) (Suc i)) (prefix (nth l n) \<sigma>)) \<Turnstile> f)) "
  proof 
    fix i
    show "i < intlen(prefix n l) \<longrightarrow>
                ((sub (nth (prefix n l) i) (nth (prefix n l) (Suc i)) (prefix (nth l n) \<sigma>)) \<Turnstile> f) " 
    proof -
     have 21: " i<n \<longrightarrow> ((sub (nth l i) (nth l (Suc i)) \<sigma>) \<Turnstile> f)  " 
       using "0" assms less_le_trans by blast
     have 22: "i<intlen(prefix n l) \<longrightarrow> 
               ((sub (nth (prefix n l) i) (nth (prefix n l) (Suc i)) \<sigma>) \<Turnstile> f) "
       by (simp add: "21" assms)
     have 23: "i<intlen(prefix n l) \<longrightarrow>  
               ((sub (nth (prefix n l) i) (nth (prefix n l) (Suc i)) (prefix (nth l n) \<sigma>)) \<Turnstile> f)"
      using "22" assms powerinterval_splita0 by fastforce     
     show ?thesis using "23" by blast
    qed
   qed
  show ?thesis using "1" "2" by blast
qed

lemma powerinterval_splitb0:
 assumes " index_sequence 0 l "
         " n \<le> intlen l"
         "  nth l (intlen l) = intlen \<sigma>"
         "  i < (intlen l -n) "
 shows   "(sub (nth (((map (shiftm (nth l n))  (suffix n l)))) i) 
                   (nth (((map (shiftm (nth l n))  (suffix n l)))) (Suc i)) 
                   (suffix (nth l n) \<sigma>)) =
          (sub (nth l (i+n)) (nth l ((Suc i)+n)) \<sigma>) "
proof -
 have 1: "intlen(((map (shiftm (nth l n))  (suffix n l)))) = (intlen l) - n "
      by (simp add: assms) 
 have 2: " i<(intlen l -n) \<longrightarrow> 
               (nth (((map (shiftm (nth l n))  (suffix n l)))) i) =
                (nth l (n + i)) - (nth l n) "
      using assms interval_idx_shiftm_suffix_nth by force
 have 3: "i<(intlen l -n) \<longrightarrow> 
              (nth (((map (shiftm (nth l n))  (suffix n l)))) (Suc i)) =
              (nth l (n + (Suc i))) - (nth l n) "
       using assms interval_idx_shiftm_suffix_nth by fastforce
 have 4: "i<(intlen l -n) \<longrightarrow>
              (sub (nth (((map (shiftm (nth l n))  (suffix n l)))) i) 
                   (nth (((map (shiftm (nth l n))  (suffix n l)))) (Suc i)) 
                   (suffix (nth l n) \<sigma>)) =
                (sub ( (nth l (n + i)) - (nth l n) )
                     ( (nth l (n + (Suc i))) - (nth l n) )
                   (suffix (nth l n) \<sigma>)) " 
          by (simp add: "2" "3")
 have 5: " i<(intlen l -n) \<longrightarrow> (nth l (n + i))  < (nth l (n + (Suc i)))"
  using assms index_sequence_def by auto
 have 6: "i<(intlen l -n) \<longrightarrow> (nth l n) \<le> (nth l (n + i))" 
    using assms by (metis add.commute interval_idx_less_equal le_add1 less_diff_conv less_imp_le_nat)
 have 7: "i<(intlen l -n) \<longrightarrow> (nth l n) \<le> (nth l (n + (Suc i)))" 
    using "5" "6" by linarith
 have 8: " i<(intlen l -n) \<longrightarrow> (nth l (n + i)) - (nth l n) < (nth l (n + (Suc i))) - (nth l n)" 
   using "5" "6" diff_less_mono by blast
 have 9: "i<(intlen l -n) \<longrightarrow> (nth l (n + (Suc i))) - (nth l n) \<le> intlen \<sigma> - (nth l n)"
   by (metis Suc_leI add.commute assms diff_le_mono interval_idx_less_last_1 le_eq_less_or_eq 
             less_diff_conv ordered_cancel_comm_monoid_diff_class.add_diff_inverse)
 have 10: "i<(intlen l -n) \<longrightarrow> 
              (sub ( (nth l (n + i)) - (nth l n) )
                     ( (nth l (n + (Suc i))) - (nth l n) )
                   (suffix (nth l n) \<sigma>)) = 
               (sub (nth l (i+n)) (nth l ((Suc i)+n)) \<sigma>)  "
   using interval_sub_suffix 
   by (metis "6" "7" "8" "9" add.commute le_add_diff_inverse2)
 show ?thesis 
 using "2" "3" "10" assms by auto   
qed

lemma powerinterval_splitb:
 assumes " index_sequence 0 l"
         "  n \<le> intlen l"
         "  nth l (intlen l) = intlen \<sigma> "
         " powerinterval f \<sigma> l "
 shows   " powerinterval f (suffix (nth l n) \<sigma>) ((map (shiftm (nth l n))  (suffix n l)))"
proof -
 have 0: "(\<forall> i. i<intlen l \<longrightarrow> ( (sub (nth l i) (nth l (Suc i)) \<sigma>) \<Turnstile> f)) "
   using assms by (simp add: powerinterval_def)
 have 1: "powerinterval f (suffix (nth l n) \<sigma>) ((map (shiftm (nth l n))  (suffix n l))) =
         (\<forall> i. i<intlen(((map (shiftm (nth l n))  (suffix n l)))) \<longrightarrow> 
            ( (sub (nth (((map (shiftm (nth l n))  (suffix n l)))) i) 
                   (nth (((map (shiftm (nth l n))  (suffix n l)))) (Suc i)) 
                   (suffix (nth l n) \<sigma>)) \<Turnstile> f))   "
     by (simp add: powerinterval_def)
 have 2: "(\<forall> i. i<intlen(((map (shiftm (nth l n))  (suffix n l)))) \<longrightarrow> 
            ( (sub (nth (((map (shiftm (nth l n))  (suffix n l)))) i) 
                   (nth (((map (shiftm (nth l n))  (suffix n l)))) (Suc i)) 
                   (suffix (nth l n) \<sigma>)) \<Turnstile> f)) "
  proof 
   fix i
   show "i<intlen(((map (shiftm (nth l n))  (suffix n l)))) \<longrightarrow> 
            ( (sub (nth (((map (shiftm (nth l n))  (suffix n l)))) i) 
                   (nth (((map (shiftm (nth l n))  (suffix n l)))) (Suc i)) 
                   (suffix (nth l n) \<sigma>)) \<Turnstile> f) "
   proof -
    have 21: " i< (intlen l) -n \<longrightarrow> 
               ((sub (nth l (i+n)) (nth l ((Suc i)+n)) \<sigma>) \<Turnstile> f)"
        by (simp add: "0")
    show ?thesis 
    using "21" assms powerinterval_splitb0 by fastforce
   qed
  qed
 show ?thesis using "1" "2" by blast
qed
       
lemma powerinterval_split:
 assumes " index_sequence 0 l"
         "  n \<le> intlen l"
         "  nth l (intlen l) = intlen \<sigma>"
 shows   " powerinterval f \<sigma> l =
           (powerinterval f (prefix (nth l n) \<sigma>) (prefix n l) \<and>
            powerinterval f (suffix (nth l n) \<sigma>) ((map (shiftm (nth l n))  (suffix n l))))"
proof -
 have 1: "powerinterval f \<sigma> l \<Longrightarrow> 
          powerinterval f (prefix (nth l n) \<sigma>) (prefix n l) "
   by (simp add: assms powerinterval_splita)
 have 2: "powerinterval f \<sigma> l \<Longrightarrow> 
          powerinterval f (suffix (nth l n) \<sigma>) ((map (shiftm (nth l n))  (suffix n l))) "
   by (simp add: assms powerinterval_splitb)
 have 3: "powerinterval f (prefix (nth l n) \<sigma>) (prefix n l) =
          (\<forall>i. i<n \<longrightarrow> ( (sub (nth l i) (nth l (Suc i)) \<sigma>) \<Turnstile> f)) "
  by (metis assms interval_prefix_length_good powerinterval_def powerinterval_splita0)
 have 4: "powerinterval f (suffix (nth l n) \<sigma>) ((map (shiftm (nth l n))  (suffix n l))) =
         (\<forall> i. i<(intlen l) -n \<longrightarrow> ((sub (nth l (i+n)) (nth l ((Suc i)+n)) \<sigma>) \<Turnstile> f)) "
   by (simp add: assms powerinterval_def powerinterval_splitb0)
 have 5: "(\<forall> i. i<(intlen l) -n \<longrightarrow> ((sub (nth l (i+n)) (nth l ((Suc i)+n)) \<sigma>) \<Turnstile> f)) =
          (\<forall> i. n \<le> i \<and> i<(intlen l) \<longrightarrow> ((sub (nth l i) (nth l (Suc i)) \<sigma>) \<Turnstile> f)) "
     by (metis le_add2 le_add_diff_inverse2 less_diff_conv plus_nat.simps(2))
 have 5: "(powerinterval f (prefix (nth l n) \<sigma>) (prefix n l) \<and>
            powerinterval f (suffix (nth l n) \<sigma>) ((map (shiftm (nth l n))  (suffix n l)))) \<Longrightarrow>
           powerinterval f \<sigma> l "
  using "3" "4" "5" assms powerinterval_def 
  by (metis not_less_eq not_less_less_Suc_eq order.order_iff_strict)
 show ?thesis 
 using "1" "2" "5" by blast  
qed

lemma powerinterval_fuse:
 assumes " index_sequence 0 l1"
         " index_sequence 0 l2"
         " intlast l1 = cp"
         "  cp \<le> intlen \<sigma> "
         " intlast l2  = intlen \<sigma> - cp"
 shows   " powerinterval f \<sigma> (fuse l1 (map (shift cp) l2)) =
           (powerinterval f (prefix cp \<sigma>) l1 \<and> 
            powerinterval f (suffix cp \<sigma>) l2 )"
proof -
 have 1: "index_sequence 0 (fuse l1 (map (shift cp) l2)) "
  using assms interval_idx_fuse[of "l1" "(map (shift cp) l2)"]   
  by (metis index_sequence_def interval_idx_link interval_nth_zero_intfirst) 
 have 2: "cp = (nth l1 (intlen l1))"
  using assms interval_nth_intlen_intlast by blast
 have 3: "intlen l1 \<le> intlen (fuse l1 (map (shift cp) l2))"
   by (simp add: interval_fuse_intlen_a)
 have 4: "nth (fuse l1 (map (shift cp) l2)) (intlen (fuse l1 (map (shift cp) l2))) = intlen \<sigma>" 
    by (metis assms interval_idx_fuse_intlen interval_intlen_gr_zero interval_nth_intlen_intlast)
 have 5: "cp = nth (fuse l1 (map (shift cp) l2)) (intlen l1)"
    by (metis "3" assms interval_fuse_nth interval_idx_fuse_intfirst_intlast 
        interval_nth_intlen_intlast order_refl) 
 have 6: "(map (shiftm cp) (map (shift cp) l2)) = l2 "
  using assms by (metis interval_idx_link interval_lsk_ls)
 have 7: "intlast l1 = intfirst (map (shift cp) l2)"
   by (metis assms interval_idx_fuse_intfirst_intlast interval_nth_intlen_intlast)
  show ?thesis 
  using "1" "3" "4" "5" "6" "7" interval_prefix_fuse interval_suffix_fuse powerinterval_split 
   by fastforce
qed


lemma powerinterval_idx:
 "(powerinterval (LIFT(f \<and> more)) \<sigma> l \<and> (nth l 0) = 0 \<and> (nth l (intlen l)) = intlen \<sigma>) =
  (index_sequence 0 l \<and> (nth l (intlen l)) = intlen \<sigma> \<and> powerinterval f \<sigma> l)"
proof (auto simp add: powerinterval_def index_sequence_def more_defs)
 show "\<And>n. \<forall>i<intlen l.
            f (sub (Interval.nth l i) (Interval.nth l (Suc i)) \<sigma>) \<and>
            0 < intlen (sub (Interval.nth l i) (Interval.nth l (Suc i)) \<sigma>) \<Longrightarrow>
         Interval.nth l 0 = 0 \<Longrightarrow>
         Interval.nth l (intlen l) = intlen \<sigma> \<Longrightarrow> n < intlen l \<Longrightarrow> 
         Interval.nth l n < Interval.nth l (Suc n)"
   by (simp add: Interval.sub_def)
 show "\<And>i. Interval.nth l 0 = 0 \<Longrightarrow>
         \<forall>n<intlen l. Interval.nth l n < Interval.nth l (Suc n) \<Longrightarrow>
         Interval.nth l (intlen l) = intlen \<sigma> \<Longrightarrow>
         \<forall>i<intlen l. f (sub (Interval.nth l i) (Interval.nth l (Suc i)) \<sigma>) \<Longrightarrow>
         i < intlen l \<Longrightarrow> 0 < intlen (sub (Interval.nth l i) (Interval.nth l (Suc i)) \<sigma>) " 
   by (simp add: Interval.sub_def) 
       (metis index_sequence_def interval_idx_less_last_1)
qed

subsubsection \<open>cpl lemmas\<close>

lemma cpl_expand:
assumes "(\<exists> l. index_sequence 0 l \<and> powerinterval f \<sigma> l \<and> (nth l (intlen l)) = intlen \<sigma> \<and>
             ((filt \<sigma> l) \<Turnstile> g  )) "
shows  " index_sequence 0 (cpl f g \<sigma>) \<and> powerinterval f \<sigma> (cpl f g \<sigma>) \<and>
           (nth (cpl f g \<sigma>) (intlen (cpl f g \<sigma>))) = intlen \<sigma> \<and> ((filt \<sigma> (cpl f g \<sigma>)) \<Turnstile> g)"
proof -
 have 0: "cpl f g \<sigma> = (\<some> l. index_sequence 0 l \<and> powerinterval f \<sigma> l \<and> 
                         (nth l (intlen l)) = intlen \<sigma> \<and> ((filt \<sigma> l) \<Turnstile> g)) "
   using cpl_def by force
 have 1: "(\<exists> l. index_sequence 0 l \<and> powerinterval f \<sigma> l \<and> (nth l (intlen l)) = intlen \<sigma> \<and>
            ((filt \<sigma> l) \<Turnstile> g))"
      using assms by auto
 have 2: "index_sequence 0 (cpl f g \<sigma>) \<and> powerinterval f \<sigma> (cpl f g \<sigma>) \<and>
           (nth (cpl f g \<sigma>) (intlen (cpl f g \<sigma>))) = intlen \<sigma> \<and> ((filt \<sigma> (cpl f g \<sigma>)) \<Turnstile> g)"
  using "0" "1" 
   someI_ex[of "\<lambda>l. index_sequence 0 l \<and> powerinterval f \<sigma> l \<and> (nth l (intlen l)) = intlen \<sigma> \<and>
            ((filt \<sigma> l) \<Turnstile> g)"]  by simp
 show ?thesis 
 using "2" by blast
qed  

lemma cpl_projection:
 "(\<sigma> \<Turnstile> f \<triangle> g) =  
  ( index_sequence 0 (cpl f g \<sigma>) \<and> powerinterval f \<sigma> (cpl f g \<sigma>) \<and>
           (nth (cpl f g \<sigma>) (intlen (cpl f g \<sigma>))) = intlen \<sigma> \<and> g (filt \<sigma> (cpl f g \<sigma>)) )"
using cpl_expand by (simp add: projection_d_def, blast)

lemma cpl_empty:
 assumes "intlen \<sigma> = 0 \<and> (\<sigma> \<Turnstile> f \<triangle> g)"
 shows   "(cpl f g \<sigma>) = \<langle>0\<rangle>"
 using assms cpl_projection interval_idx_less_last_1 interval_st_intlen by fastforce   

lemma cpl_empty_a:
 assumes "intlen \<sigma> = 0"
         "(cpl f g \<sigma>) = \<langle>0\<rangle>"
         "g(filt \<sigma> \<langle>0\<rangle>)"
 shows   "(\<sigma> \<Turnstile> f \<triangle> g)"
proof -
 have 1: "index_sequence 0 \<langle>0\<rangle> "
   by (simp add: index_sequence_def)
 have 2: "powerinterval f \<sigma> \<langle>0\<rangle> "
   by (simp add: powerinterval_def)
 have 3: "(nth \<langle>0\<rangle> (intlen \<langle>0\<rangle>)) = intlen \<sigma> "
    by (simp add: assms)
 have 4: " g(filt \<sigma> \<langle>0\<rangle>)"
    using assms by blast
 from 1 2 3 4 show ?thesis 
 by (simp add: assms cpl_projection)
qed  


lemma cpl_more:
 assumes "intlen \<sigma> >0"
         "(\<sigma> \<Turnstile> f \<triangle> g)"
 shows   "intlen(cpl f g \<sigma>)>0"
by (metis assms cpl_projection gr0I index_sequence_def)

lemma cpl_more_than_first:
 assumes "intlen \<sigma> >0"
         "(\<sigma> \<Turnstile> f \<triangle> g)"
 shows   "(nth (cpl f g \<sigma>) 0) = 0"
using assms cpl_projection index_sequence_def by auto 

lemma cpl_more_than_last:
 assumes "intlen \<sigma> >0"
         "(\<sigma> \<Turnstile> f \<triangle> g)"
 shows   "(nth (cpl f g \<sigma>) (intlen (cpl f g \<sigma>))) = intlen \<sigma>"
using assms cpl_projection by blast

lemma cpl_sub_more:
 assumes "n < k"
         "k \<le> intlen \<sigma>"
         "((sub n k \<sigma>) \<Turnstile> f \<triangle> g)"
 shows   " intlen(cpl f g (sub n k \<sigma>)) > 0"
using assms 
by (simp add: cpl_more)

lemma cpl_bounds:
 assumes "n < k"
         "k \<le> intlen \<sigma>"
         "((sub n k \<sigma>) \<Turnstile> f \<triangle> g)"
         "i < intlen (cpl f g (sub n k \<sigma>))" 
 shows   "0 \<le> (nth (cpl f g (sub n k \<sigma>)) i) \<and> (nth (cpl f g (sub n k \<sigma>)) i) \<le> k -n"
using assms 
by (metis cpl_projection interval_idx_less_last_1 interval_intlen_sub le0 less_imp_le_nat)

lemma cpl_map_bounds:
 assumes "n < k"
         "k \<le> intlen \<sigma>"
         "((sub n k \<sigma>) \<Turnstile> f \<triangle> g)"
         "i < intlen (map (shift n) (cpl f g (sub n k \<sigma>))) " 
 shows   "n \<le> (nth (map (shift n) (cpl f g (sub n k \<sigma>))) i) \<and> 
               (nth (map (shift n) (cpl f g (sub n k \<sigma>))) i) \<le> k "
using assms 
by (metis Interval.shift_def Nat.le_diff_conv2 cpl_bounds interval_intlen_map 
    interval_nth_map le_add2 less_imp_le_nat)

lemma cpl_intfirst:
 assumes "(sub x1a (intfirst l) \<sigma>) \<Turnstile> f \<triangle> g"
 shows   "intfirst((map (shift x1a) (cpl f g (sub x1a (intfirst l) \<sigma>))))  = x1a"
proof -
 have 1: "( index_sequence 0 (cpl f g (sub x1a (intfirst l) \<sigma>)) \<and> 
            powerinterval f (sub x1a (intfirst l) \<sigma>) (cpl f g (sub x1a (intfirst l) \<sigma>)) \<and>
           (nth (cpl f g (sub x1a (intfirst l) \<sigma>)) (intlen (cpl f g (sub x1a (intfirst l) \<sigma>)))) = 
           intlen (sub x1a (intfirst l) \<sigma>) \<and> 
           g (filt (sub x1a (intfirst l) \<sigma>) (cpl f g (sub x1a (intfirst l) \<sigma>))) ) " 
          using cpl_projection assms by auto
 have 2: "index_sequence 0 (cpl f g (sub x1a (intfirst l) \<sigma>)) "
   using "1" by auto
 have 3: " (nth (cpl f g (sub x1a (intfirst l) \<sigma>)) 0) =0"
    using 2 index_sequence_def by blast
 show ?thesis 
 by (metis "3" Interval.shift_def add.left_neutral interval_nth_map interval_nth_zero_intfirst)
qed

lemma cpl_intfirst_same:
 assumes "(sub x1a x1a \<sigma>) \<Turnstile> f \<triangle> g"
 shows   "intfirst((map (shift x1a) (cpl f g (sub x1a x1a \<sigma>))))  = x1a"
proof -
 have 1: " intfirst (\<langle> x1a \<rangle>) = x1a"  
     by auto
 from 1 cpl_intfirst show ?thesis by (metis assms)
qed

lemma cpl_intlast:
 assumes " ((sub x (intfirst l) \<sigma>) \<Turnstile> f \<triangle> g) \<and> x < intfirst l \<and> intfirst l \<le> intlen \<sigma> "
 shows   "intlast((map (shift x) (cpl f g (sub x (intfirst l) \<sigma>))))  = intfirst l"
 proof -
  have 01: "( index_sequence 0 (cpl f g (sub x (intfirst l) \<sigma>)) \<and> 
            powerinterval f (sub x (intfirst l) \<sigma>) (cpl f g (sub x (intfirst l) \<sigma>)) \<and>
           (nth (cpl f g (sub x (intfirst l) \<sigma>)) (intlen (cpl f g (sub x (intfirst l) \<sigma>)))) = 
           intlen (sub x (intfirst l) \<sigma>) \<and> 
           g (filt (sub x (intfirst l) \<sigma>) (cpl f g (sub x (intfirst l) \<sigma>))) ) "
     using cpl_projection assms by (simp add: cpl_projection)
  have 02: "(nth (cpl f g (sub x (intfirst l) \<sigma>)) (intlen (cpl f g (sub x (intfirst l) \<sigma>)))) = 
             intlen (sub x (intfirst l) \<sigma>) "
        using 01 by auto
  have 03: "intlast(cpl f g (sub x (intfirst l) \<sigma>)) = 
               (nth (cpl f g (sub x (intfirst l) \<sigma>)) (intlen (cpl f g (sub x (intfirst l) \<sigma>))))"
        by simp
  have 04: "x < intfirst l "
    using assms by blast
  have 05: "intlen (sub x (intfirst l) \<sigma>) = (intfirst l) -x "
   using assms interval_intlen_sub less_imp_le_nat by blast
  have 06: "intlast((map (shift x) (cpl f g (sub x (intfirst l) \<sigma>)))) =
             ((intfirst l) -x)+x "
    by (metis "02" "05" Interval.shift_def interval_intlen_map interval_nth_intlen_intlast 
       interval_nth_map)
  show ?thesis using "04" "06" by auto
qed

lemma cpl_intlast_i:
 assumes "((sub (nth l i) (nth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)"
         "(nth l i) < (nth l (Suc i))"
         "(nth l (Suc i)) \<le> intlen \<sigma> "
 shows   "intlast((map (shift (nth l i)) (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>))))  = 
          (nth l (Suc i))"
 proof -
  have 01: "( index_sequence 0 (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)) \<and> 
            powerinterval f (sub (nth l i) (nth l (Suc i)) \<sigma>) 
                            (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)) \<and>
           (nth (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)) 
                (intlen (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)))) = 
               intlen (sub (nth l i) (nth l (Suc i)) \<sigma>) \<and> 
           g (filt (sub (nth l i) (nth l (Suc i)) \<sigma>) (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>))) )"
     using cpl_projection assms by (simp add: cpl_projection)
  have 02: "(nth (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)) 
                 (intlen (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)))) = 
             intlen (sub (nth l i) (nth l (Suc i)) \<sigma>) "
        using 01 by auto
  have 03: "intlast(cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)) = 
               (nth (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)) 
                    (intlen (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>))))"
        by simp
  have 04: "(nth l i) < (nth l (Suc i)) "
        by (simp add: assms)
  have 05: "intlen (sub (nth l i) (nth l (Suc i)) \<sigma>) = (nth l (Suc i)) -(nth l i) "
        by (simp add:  assms less_imp_le_nat)
  have 06: "intlast((map (shift (nth l i)) (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)))) =
             ((nth l (Suc i)) -(nth l i))+(nth l i) "
    by (metis "02" "05" Interval.shift_def interval_intlen_map interval_nth_intlen_intlast 
        interval_nth_map)
  show ?thesis using "04" "06" by auto
qed


subsubsection \<open>lcpl lemmas\<close>


lemma lcpl_nth:
 assumes "index_sequence (nth l 0) l"
         " i < intlen l"
 shows   " (nth (lcpl f g \<sigma> l) i) = 
           (map (shift (nth l i)) (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)))"
using assms
proof (induct l arbitrary: i)
case (St x)
then show ?case by simp
next
case (Cons x1a ls)
then show ?case 
   proof (cases i)
   case 0
   then show ?thesis 
     proof (cases ls)
     case (St x1)
     then show ?thesis by (simp add: "0")
     next
     case (Cons x21 x22)
     then show ?thesis by (simp add: "0")
     qed
   next
   case (Suc nat)
   then show ?thesis 
     proof (cases ls)
     case (St x1)
     then show ?thesis 
     by (metis Cons.prems(2) One_nat_def Suc intlen.simps(1) intlen.simps(2) leD le_add1
          plus_1_eq_Suc)
     next
     case (Cons x21 x22)
     then show ?thesis 
     using Cons.hyps Cons.prems(1) Cons.prems(2) Suc interval_idx_cons by auto
     qed
   qed
qed


lemma lcpl_intlen:
 assumes "index_sequence (nth l 0) l"
         " intlen l > 0 " 
 shows   " intlen(lcpl f g \<sigma> l) = intlen l -1"
using assms 
 proof 
  (induct l)
  case (St x)
  then show ?case by simp
  next
  case (Cons x1a l)
  then show ?case 
   proof (cases l)
   case (St x1)
   then show ?thesis by simp
   next
   case (Cons x21 x22)
   then show ?thesis using Cons.hyps Cons.prems(1) interval_idx_cons by auto
   qed
qed

lemma lcpl_intlen_zero:
 assumes "index_sequence (nth l 0) l"
         "intlen l = 0" 
 shows   "intlen(lcpl f g \<sigma> l) = 0"
using assms 
by (metis interval_suffix_intlen interval_suffix_zero intlen.simps(1) lcpl.simps(1))

lemma lcpl_last:
 assumes "index_sequence (nth l 0) l"
         "(nth l (intlen l)) = intlen \<sigma>"
         " intlen l >0 "
 shows   " intlast (lcpl f g \<sigma> l) = 
             (map (shift (nth l (intlen l-1))) 
                 (cpl f g (sub (nth l (intlen l -1)) (nth l (intlen l)) \<sigma>)))"
proof -
 have 1: "intlast (lcpl f g \<sigma> l) = (nth (lcpl f g \<sigma> l) (intlen (lcpl f g \<sigma> l))) "
   by simp
 have 2: "(nth (lcpl f g \<sigma> l) (intlen (lcpl f g \<sigma> l))) =
          (map (shift (nth l (intlen (lcpl f g \<sigma> l)))) 
               (cpl f g (sub (nth l (intlen (lcpl f g \<sigma> l))) 
                             (nth l (Suc (intlen (lcpl f g \<sigma> l)))) 
                             \<sigma>))) "
      using assms lcpl_nth 
      by (metis One_nat_def Suc_pred diff_le_self lcpl_intlen le_eq_less_or_eq n_not_Suc_n )
 have 3: "(map (shift (nth l (intlen (lcpl f g \<sigma> l)))) 
               (cpl f g (sub (nth l (intlen (lcpl f g \<sigma> l))) 
                             (nth l (Suc (intlen (lcpl f g \<sigma> l)))) 
                             \<sigma>))) =
          (map (shift (nth l (intlen l-1))) 
               (cpl f g (sub (nth l (intlen l-1)) (nth l (Suc (intlen l-1))) \<sigma>))) "
        using assms by (metis lcpl_intlen )
 show ?thesis by (simp add: "2" "3" assms(3))
qed

lemma lcpl_last_last:
 assumes "index_sequence (nth l 0) l"
         "(nth l (intlen l)) = intlen \<sigma>"
         "((sub (nth l (intlen l -1) ) (nth l (intlen l)) \<sigma>) \<Turnstile> f \<triangle> g)"
         " intlen l > 0 "
 shows   " intlast (intlast (lcpl f g \<sigma> l)) = intlen \<sigma> "
by (metis One_nat_def Suc_pred assms cpl_intlast_i diff_less interval_idx_less_last_1 lcpl_last 
le_eq_less_or_eq zero_less_one)


lemma lcpl_zero_zero:
  assumes "index_sequence (nth l 0) l"
          " (nth l (intlen l)) = intlen \<sigma>"
          " intlen l = 0"
 shows   "(nth (lcpl f g \<sigma> l) 0) = 
          (map (shift (nth l 0)) (cpl f g (sub (nth l 0) (nth l 0) \<sigma>)))"
using assms by (metis Interval.nth.simps(1) interval_suffix_intlen interval_suffix_zero lcpl.simps(1))
  
lemma lcpl_intfirst:
 assumes "index_sequence (nth l 0) l"
         "(nth l (intlen l)) = intlen \<sigma> "
         "(\<forall> i<intlen l. (sub (nth l i) (nth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)  "
         " intlen l > 0"
 shows   " intfirst(intfirst((lcpl f g \<sigma> l))) = intfirst l"
proof -
 have 01: "(intfirst((lcpl f g \<sigma> l))) = 
           (map (shift (nth l 0)) (cpl f g (sub (nth l 0) (nth l (Suc 0)) \<sigma>)))" 
      using assms by (metis interval_nth_zero_intfirst lcpl_nth  )
 have 02: "intlen l > 0 \<longrightarrow> 
           intfirst((map (shift (nth l 0)) (cpl f g (sub (nth l 0) (nth l (Suc 0)) \<sigma>)))) =
            (nth l 0)" 
            using assms by (metis Suc_leI cpl_intfirst interval_intlast_intfirst interval_intlast_prefix)
 show ?thesis 
 using "01" "02" interval_nth_zero_intfirst by (simp add: assms(4))
qed

 
lemma lcpl_lfuse_lastfirst:
 assumes "index_sequence (nth l 0) l"
         "(nth l (intlen l)) = intlen \<sigma>"
         "((intlen l = 0 \<and> ((sub (nth l 0) (nth l 0) \<sigma>) \<Turnstile> f \<triangle> g)) \<or>
           (intlen l > 0 \<and> (\<forall> i<intlen l. (sub (nth l i) (nth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)))"
 shows   "lastfirst (lcpl f g \<sigma> l)"
using assms 
proof 
  (induct l)
  case (St x)
  then show ?case by simp
  next
  case (Cons x1a ls)
  then show ?case 
   proof -
    have 1: "intlen ls = 0 \<longrightarrow> lastfirst (lcpl f g \<sigma> (x1a \<odot> ls))     "
          using Cons.prems by (metis One_nat_def add_diff_cancel_left' interval_st_intlen 
              intlen.simps(2) lastfirst.simps(1) lcpl_intlen plus_1_eq_Suc zero_less_one)
    have 2: "intlen ls > 0 \<longrightarrow> lastfirst ( lcpl f g \<sigma> (x1a \<odot> ls) ) = 
               lastfirst((map (shift x1a) (cpl f g (sub x1a (intfirst ls) \<sigma>))) \<odot> (lcpl f g \<sigma> ls))"
    by (metis (no_types, lifting) interval.simps(6) interval_intlen_cons_1 interval_nth_zero 
        interval_nth_zero_intfirst lcpl.simps(2))
    have 3: "intlen ls > 0 \<longrightarrow> 
             lastfirst((map (shift x1a) (cpl f g (sub x1a (intfirst ls) \<sigma>))) \<odot> (lcpl f g \<sigma> ls)) =
              ( intlast( (map (shift x1a) (cpl f g (sub x1a (intfirst ls) \<sigma>))) ) = 
               intfirst(intfirst((lcpl f g \<sigma> ls))) \<and> lastfirst((lcpl f g \<sigma> ls)) )"
      using lastfirst.simps(2) by blast
   have 4: "x1a < (intfirst ls)"
      using Cons.prems by (metis  index_sequence_def interval_nth_Suc interval_nth_zero 
                interval_nth_zero_intfirst intlen.simps(2) plus_1_eq_Suc zero_less_Suc) 
    have 5: " (intfirst ls) \<le> intlen \<sigma>"
        using Cons.prems by (metis Suc_lessI eq_iff interval_idx_less_last_1 
            interval_nth_Suc interval_nth_zero_intfirst intlen.simps(2) less_imp_le_nat 
            plus_1_eq_Suc zero_less_Suc)
    have 6: "intlen ls >0 \<longrightarrow> 
              intlast( (map (shift x1a) (cpl f g (sub x1a (intfirst ls) \<sigma>))) ) = intfirst ls"
      using Cons.prems by (metis "4" "5" cpl_intlast interval_nth_Suc interval_nth_zero 
          interval_nth_zero_intfirst less_le_not_le)
    have 7: "index_sequence (nth ls 0) ls"
       using assms Cons.prems index_sequence_def by auto
    have 8: "nth ls (intlen ls) = intlen \<sigma> "
        using Cons.prems by auto
    have 9: "( (\<forall> i<intlen ( ls). (sub (nth ( ls) i) (nth ( ls) (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g))"
       using Cons.prems by auto   
    have 10: "intlen ls >0 \<longrightarrow> intfirst(intfirst((lcpl f g \<sigma> ls))) = intfirst ls  "
      using "7" "8" "9" 
      by (metis cpl_intfirst interval_intlen_cons_1 interval_nth_Suc interval_nth_zero_intfirst 
          lcpl_nth)
    have 11: "intlen ls >0 \<longrightarrow> 
                lastfirst((lcpl f g \<sigma> ls))"
      using "7" "8" "9" Cons.hyps by blast
   show ?thesis using "1" "10" "11" "2" "6" by auto
  qed
qed

lemma lcpl_lfuse_intlen:
 assumes "index_sequence (nth l 0) l"
         "(nth l (intlen l)) = intlen \<sigma> " 
         "((intlen l =0 \<and> ((sub (nth l 0 ) (nth l 0 ) \<sigma> \<Turnstile> f \<triangle> g))) \<or>
           (intlen l >0 \<and> (\<forall> i<intlen l. (sub (nth l i) (nth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)))"
 shows   "(intlen l =0 \<longrightarrow> intlen(lfuse (lcpl f g \<sigma> l)) = 0) \<and>
          (intlen l >0 \<longrightarrow> 
            intlen(lfuse (lcpl f g \<sigma> l)) = (\<Sum>k=0..intlen l-1. intlen (nth (lcpl f g \<sigma> l) k)))"
proof -
 have 1: "lastfirst (lcpl f g \<sigma> l)" 
  using assms lcpl_lfuse_lastfirst by blast
 have 2: "intlen l = 0 \<longrightarrow> 
           lfuse (lcpl f g \<sigma> l) =  
           ((map (shift (nth l 0)) (cpl f g (sub (nth l 0) (nth l 0) \<sigma>))))"
      by (metis interval_suffix_intlen interval_suffix_zero lcpl.simps(1) lfuse_St)
 have 3: "intlen l = 0 \<longrightarrow> 
           intlen ((map (shift (nth l 0)) (cpl f g (sub (nth l 0) (nth l 0) \<sigma>)))) = 0"
    using assms cpl_empty interval_intlen_sub by fastforce
 have 4: "intlen l = 0 \<longrightarrow> intlen(lfuse (lcpl f g \<sigma> l)) = 0"
    using "2" "3" by auto 
 have 5: "intlen l > 0 \<longrightarrow>
            intlen(lfuse (lcpl f g \<sigma> l)) = (\<Sum>k=0..intlen l-1. intlen (nth (lcpl f g \<sigma> l) k)) "
   using interval_lfuse_intlen 
   by (metis assms lcpl_intlen lcpl_lfuse_lastfirst)
 from 4 5 show ?thesis by simp
qed

lemma lcpl_lfuse_idx:
 assumes "index_sequence 0 l"
         " (nth l (intlen l)) = intlen \<sigma>" 
         "(\<forall> i<intlen l. (sub (nth l i) (nth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)"
         "intlen l > 0 "
 shows   "index_sequence (intfirst (lfuse  (lcpl f g \<sigma> l)))  (lfuse ( (lcpl f g \<sigma> l)))"
proof -
 have 0: "intlen \<sigma> >0 \<longrightarrow> intlen l >0 "
  using assms gr_zeroI index_sequence_def by fastforce
 have 2: "intlen \<sigma> >0 \<longrightarrow> lastfirst (lcpl f g \<sigma> l)"
           using "0" assms index_sequence_def lcpl_lfuse_lastfirst by blast
 have 3: " (\<forall>i<intlen l. index_sequence 0 (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>))) "
     using assms cpl_projection by auto
 have 4: " (\<forall>i<intlen l. 
              index_sequence (nth l i) 
                             (map (shift (nth l i)) (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>))))"
    using "3" interval_idx_link by blast
 have 5: "intlen \<sigma> >0 \<longrightarrow>
           (\<forall>i<intlen l. index_sequence (nth l i) (nth (lcpl f g \<sigma> l) i) ) "
    using assms by (metis "4" index_sequence_def lcpl_nth)
 have 6: " intlen \<sigma> >0 \<longrightarrow>
          (\<forall> i<intlen l. intfirst (nth (lcpl f g \<sigma> l) i) = (nth l i)) "
    by (metis "5" index_sequence_def interval_nth_zero_intfirst)
 have 7: " intlen \<sigma> >0 \<longrightarrow>
            (\<forall>i<intlen l. 
              index_sequence (intfirst (nth (lcpl f g \<sigma> l) i)) (nth (lcpl f g \<sigma> l) i) ) "
     using "6" "5" by simp
 have 8: "intlen \<sigma> >0 \<longrightarrow>
           index_sequence (intfirst (lfuse  (lcpl f g \<sigma> l)))  (lfuse ( (lcpl f g \<sigma> l))) "
     using assms 
     by (metis (mono_tags, lifting) "2" "7" Suc_diff_1 index_sequence_def interval_idx_lfuse 
         lcpl_intlen le_imp_less_Suc)
  from 8  show ?thesis 
  by (metis assms(1) assms(2) assms(4) index_sequence_def interval_idx_less_last_1)
qed

lemma lcpl_intlen_nth_gr_zero:
 assumes "index_sequence 0 l"
         "(nth l (intlen l)) = intlen \<sigma>"
         "(\<forall> i<intlen l. (sub (nth l i) (nth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)"
         " intlen \<sigma> >0"
 shows  " (\<forall> j \<le> intlen (lcpl f g \<sigma> l). intlen(nth (lcpl f g \<sigma> l) j) > 0) "
proof 
   fix j
   show "j \<le> intlen (lcpl f g \<sigma> l) \<longrightarrow> 0 < intlen (Interval.nth (lcpl f g \<sigma> l) j) "
   proof -
   have 1: "intlen \<sigma>>0 \<longrightarrow> lastfirst (lcpl f g \<sigma> l)"
      by (metis assms index_sequence_def lcpl_lfuse_lastfirst neq0_conv)
   have 2: "intlen \<sigma>>0 \<longrightarrow> intlen l >0"
      using assms gr_zeroI index_sequence_def by fastforce
   have 3: "intlen \<sigma>>0 \<longrightarrow> j \<le> intlen (lcpl f g \<sigma> l) \<longrightarrow> 
           (nth (lcpl f g \<sigma> l) j) = 
             (map (shift (nth l j)) (cpl f g (sub (nth l j) (nth l (Suc j)) \<sigma>)))" 
     using assms 
     by (metis "2" One_nat_def Suc_pred index_sequence_def lcpl_intlen lcpl_nth less_Suc_eq_le)
   have 4: "intlen \<sigma>>0 \<longrightarrow> j \<le> intlen (lcpl f g \<sigma> l) \<longrightarrow>
            intlen (map (shift (nth l j)) (cpl f g (sub (nth l j) (nth l (Suc j)) \<sigma>))) > 0 "
     by (metis (no_types, lifting) "2" One_nat_def Suc_pred add.commute assms cpl_sub_more gr0I 
         index_sequence_def interval_idx_expand interval_intlen_map lcpl_intlen 
         le_imp_less_Suc plus_1_eq_Suc)
   show ?thesis 
   using  "3" "4" by (simp add: assms(4))
   qed
qed  


lemma lcpl_intlast_nth:
 assumes "index_sequence 0 l"
         "(nth l (intlen l)) = intlen \<sigma>"
         "(\<forall> i<intlen l. (sub (nth l i) (nth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)"
         " intlen \<sigma> > 0 " 
         " j \<le> intlen (lcpl f g \<sigma> l) "
 shows  "intlast(nth (lcpl f g \<sigma> l) j) = (nth l (Suc j))"
proof -
 have 0: "intlen \<sigma> >0 \<longrightarrow> intlen l >0"
    using assms gr_zeroI index_sequence_def by fastforce
 have 1: "intlen \<sigma>>0 \<longrightarrow> 
          j \<le> intlen (lcpl f g \<sigma> l)   \<longrightarrow>
           nth (lcpl f g \<sigma> l) j = 
            (map (shift (nth l j)) (cpl f g (sub (nth l j) (nth l (Suc j)) \<sigma>)))
          " 
    using assms lcpl_nth[of "l" "j" "f" "g" "\<sigma>"] 
    proof -
    show ?thesis
    by (metis (no_types) "0" Suc_diff_1 
        \<open>\<lbrakk>index_sequence (nth l 0) l; j < intlen l\<rbrakk> \<Longrightarrow> 
          nth (lcpl f g \<sigma> l) j = 
          map (shift (Interval.nth l j)) (cpl f g (sub (nth l j) (Interval.nth l (Suc j)) \<sigma>))\<close> 
        \<open>index_sequence 0 l\<close> index_sequence_def lcpl_intlen le_imp_less_Suc) 
    qed 
 have 2: "intlen \<sigma>>0 \<longrightarrow> 
          j \<le> intlen (lcpl f g \<sigma> l)   \<longrightarrow> (nth l (Suc j)) \<le> intlen \<sigma>"  
     using assms 
     by (metis "0" Suc_diff_1 Suc_eq_plus1 index_sequence_def interval_idx_expand lcpl_intlen leD 
         not_less_eq)
 have 2: "intlen \<sigma>>0 \<longrightarrow> 
          j \<le> intlen (lcpl f g \<sigma> l)   \<longrightarrow>
          intlast (map (shift (nth l j)) (cpl f g (sub (nth l j) (nth l (Suc j)) \<sigma>))) = 
           (nth l (Suc j))"
    using assms cpl_intlast_i[of "f" "g" "l" "j" "\<sigma>"]  
    by (metis (no_types, hide_lams) "0" "1" "2" One_nat_def Suc_le_lessD Suc_pred 
     index_sequence_def lcpl_intlen not_less_eq_eq)
 show ?thesis using "1" "2" by (simp add: assms(4) assms(5)) 
qed  



lemma lcpl_lfuse_filt_power_help:
 assumes "index_sequence 0 l"
         "(nth l (intlen l)) = intlen \<sigma>" 
         "(\<forall> i<intlen l. (sub (nth l i) (nth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)"
         " intlen \<sigma> > 0 " 
 shows   " (\<forall> i<intlen l. g (filt \<sigma> (nth (lcpl f g \<sigma> l) i)) )"
proof -
 have 1: "(\<forall> i<intlen l. g (filt (sub (nth l i) (nth l (Suc i)) \<sigma>) 
                                 (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)) )) "
    using assms cpl_projection by blast 
 have 2: "intlen \<sigma> >0 \<longrightarrow> 
           (\<forall> i<intlen l. 
            (filt \<sigma> (nth (lcpl f g \<sigma> l) i)) =
            (filt \<sigma> (map (shift (nth l i)) (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)))))
          "
    using assms by (metis index_sequence_def lcpl_nth) 
 have 3: "intlen \<sigma> >0 \<longrightarrow> 
           (\<forall> i<intlen l. 
              intlen (filt \<sigma> (nth (lcpl f g \<sigma> l) i)) =
              intlen (filt (sub (nth l i) (nth l (Suc i)) \<sigma>) 
                                 (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)) ) 
           ) 
          " 
        by (simp add: "2" filt_intlen)
 have 4: "intlen \<sigma> >0 \<longrightarrow> 
           (\<forall> i<intlen l.
            (\<forall> j\<le> intlen (filt \<sigma> (nth (lcpl f g \<sigma> l) i)).
               (nth (filt \<sigma> (map (shift (nth l i)) (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)))) j) 
                =
                (nth \<sigma> ((nth (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)) j) +(nth l i) ))
            )
          )"  
        using interval_nth_map shift_def filt_map by metis
 have 5: "intlen \<sigma> >0 \<longrightarrow> 
           (\<forall> i<intlen l.
            (\<forall> j\<le> intlen (filt \<sigma> (nth (lcpl f g \<sigma> l) i)).
               (nth (filt (sub (nth l i) (nth l (Suc i)) \<sigma>) 
                                 (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)) ) j) =
               (nth (sub (nth l i) (nth l (Suc i)) \<sigma>) 
                    (nth (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)) j)) 
            ))"
       by (simp add: filt_map interval_nth_map)  
 have 6: "intlen \<sigma> >0 \<longrightarrow> 
           (\<forall> i<intlen l.
               (nth l (Suc i)) \<le> intlen \<sigma> 
           )"
      using assms interval_idx_expand by auto 
 have 7: "intlen \<sigma> >0 \<longrightarrow> 
           (\<forall> i<intlen l.
                (nth l i) \<le> (nth l (Suc i))
           ) "
       using assms index_sequence_def less_imp_le by blast 
 have 8: "intlen \<sigma> >0 \<longrightarrow> 
           (\<forall> i<intlen l.
            (\<forall> j\<le> intlen (filt \<sigma> (nth (lcpl f g \<sigma> l) i)). 
                (nth (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)) j) \<le> (nth l (Suc i)) - (nth l i)
            ))"
        by (metis (no_types, lifting) "3" "6" assms(1) assms(3) cpl_bounds cpl_projection 
            filt_intlen index_sequence_def interval_intlen_sub le_eq_less_or_eq)
 have 9: "intlen \<sigma> >0 \<longrightarrow> 
           (\<forall> i<intlen l.
            (\<forall> j\<le> intlen (filt \<sigma> (nth (lcpl f g \<sigma> l) i)).
              (nth (sub (nth l i) (nth l (Suc i)) \<sigma>) 
                    (nth (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)) j)) =
              (nth \<sigma> ((nth  (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)) j) + (nth l i)))
            ))
         "   using "6" 
         by (simp add:  "7" "8" add.commute)
  have 10: "intlen \<sigma> >0 \<longrightarrow> 
           (\<forall> i<intlen l.
            (\<forall> j\<le> intlen (filt \<sigma> (nth (lcpl f g \<sigma> l) i)).
               (filt (sub (nth l i) (nth l (Suc i)) \<sigma>) 
                                 (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)) ) =
              (filt \<sigma> (nth (lcpl f g \<sigma> l) i)) 
             ))
               " 
            by (simp add: filt_expand) 
                (metis "2" "3" "4" "9" filt_intlen filt_map interval_nth_map)
  show ?thesis using "1" "10" assms(4) by fastforce
qed



subsubsection \<open>lsum lemmas\<close>

lemma lsum_state:
 "lsum \<langle>xs\<rangle> a = \<langle>a+intlen xs\<rangle>"
by simp

lemma lsum_addzero_state:
 " addzero (lsum \<langle>xs\<rangle> 0) =  (if intlen xs = 0 then \<langle>0\<rangle> else  \<langle>0, intlen xs \<rangle>)"
by (simp add: addzero_def)

lemma lsum_addzero_cons:
 " addzero (lsum (xs\<odot>xxs) 0) = 0 \<odot> (lsum (xs\<odot>xxs) 0)"
by (simp add: addzero_def)

lemma lsum_intfirst:
 "intfirst (lsum xxs a) = a+intlen(intfirst xxs)"
by (case_tac xxs) simp_all

lemma lsum_intlen:
 "intlen (lsum xxs a) = intlen xxs"
by (induct xxs arbitrary: a) simp_all
  
lemma lsum_addzero_intfirst:
 "intfirst (addzero (lsum xxs 0)) = 0"
by (simp add: addzero_def lsum_intfirst)

lemma lsum_addzero_intlen:
 "(intlen xxs = 0 \<and> intlen(intfirst xxs) = 0 \<longrightarrow> 
   intlen (addzero (lsum xxs 0)) =  0)
  \<and>
  (intlen xxs = 0 \<and> intlen(intfirst xxs) > 0 \<longrightarrow> 
   intlen (addzero (lsum xxs 0)) =  1)
  \<and>
  (intlen xxs > 0  \<longrightarrow>
   intlen (addzero (lsum xxs 0)) =  (intlen xxs) +1)
 "
by (simp add: addzero_def)
   (metis add.left_neutral interval_nth_zero_intfirst lsum_intfirst lsum_intlen)

lemma lsum_nth_help:
assumes  "i>0"
         " i \<le> intlen xxs +1"
shows " (\<Sum>k = 0..(i-1). intlen (Interval.nth (xxs) k)) =
            (\<Sum>k = 1..i. intlen (Interval.nth (xxs) (k-1))) "
using assms 
proof 
  (induct i)
  case 0
  then show ?case by blast
  next
  case (Suc i)
  then show ?case 
  proof simp_all
  assume a1: "0 < i \<Longrightarrow> (\<Sum>k = 0..i - Suc 0. intlen (Interval.nth xxs k)) = 
                        (\<Sum>k = Suc 0..i. intlen (Interval.nth xxs (k - Suc 0)))"
  have f2: "\<forall>f. sum f {1::nat..0} = (0::nat)"
  by auto 
  have f3: "\<forall>n f. sum f {n::nat..n} = (f n::nat)"
  by simp 
  have f4: "\<forall>f n. (sum f {0::nat..n - 1} + (f n::nat) = sum f {0..n} \<or> 0 = n) \<or> \<not> 0 \<le> n"
  by (metis (no_types) One_nat_def Suc_pred le_eq_less_or_eq sum.atLeast0_atMost_Suc) (* 18 ms *)
  have f5: "\<forall>n. (0::nat) \<le> n"
  by blast 
  have "\<forall>n. (0::nat) + n = n"
  by linarith 
  then show "(\<Sum>n = 0..i. intlen (Interval.nth xxs n)) = 
             (\<Sum>n = Suc 0..i. intlen (Interval.nth xxs (n - Suc 0))) + intlen (Interval.nth xxs i)"
  using f5 f4 f3 f2 a1 by (metis One_nat_def le_eq_less_or_eq) (* 103 ms *)
  qed
qed
  

lemma lsum_nth:
 assumes "i\<le> intlen xxs" 
 shows "nth (lsum xxs a) i = a+(\<Sum>k::nat= 0..i. intlen(nth xxs k))"
using assms
proof 
 (induct xxs arbitrary: a i)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xxs)
 then show ?case 
  proof -
   have 1: "nth (lsum (x1a \<odot> xxs) a) i = nth ((a+intlen x1a)\<odot> (lsum xxs (a+intlen x1a))) i"
    by simp
   have 2: "i \<le> intlen xxs +1"
    using Cons.prems by auto  
   have 3: "i =0 \<longrightarrow> 
            nth ((a+intlen x1a)\<odot> (lsum xxs (a+intlen x1a))) i = (a+intlen x1a)"
      by simp 
   have 4: "a+intlen x1a = a + (\<Sum>k = 0..0. intlen (Interval.nth (x1a \<odot> xxs) k))" 
    by simp
   have 5: "i>0 \<and> i \<le> intlen xxs +1 \<longrightarrow> 
            nth ((a+intlen x1a)\<odot> (lsum xxs (a+intlen x1a))) i = 
             nth (lsum xxs (a+intlen x1a)) (i-1) "
        by (metis One_nat_def Suc_pred interval_nth_Suc)
   have 6: "i>0 \<and> i \<le> intlen xxs +1 \<longrightarrow> 
            nth (lsum xxs (a+intlen x1a)) (i-1) = 
             (a+intlen x1a) + (\<Sum>k = 0..(i-1). intlen (Interval.nth (xxs) k)) " 
      using Cons.hyps le_diff_conv by blast
   have 7: "i>0 \<and> i \<le> intlen xxs +1 \<longrightarrow> 
            (\<Sum>k = 0..(i-1). intlen (Interval.nth (xxs) k)) =
            (\<Sum>k = 1..i. intlen (Interval.nth (xxs) (k-1))) "
       using lsum_nth_help by blast
   have 8: "i>0 \<and> i \<le> intlen xxs +1 \<longrightarrow>  
             (\<Sum>k = 1..i. intlen (Interval.nth (xxs) (k-1))) = 
             (\<Sum>k = 1..i. intlen (Interval.nth (x1a \<odot> xxs) (k)))
             "
     by (metis (no_types, lifting) atLeastAtMost_iff interval_nth_Suc
         ordered_cancel_comm_monoid_diff_class.add_diff_inverse plus_1_eq_Suc sum.cong) 
   have 9: "i>0 \<and> i \<le> intlen xxs +1 \<longrightarrow> 
            (a+intlen x1a) + (\<Sum>k = 1..i. intlen (Interval.nth (x1a \<odot> xxs) (k))) =
             a + (\<Sum>k = 0..i. intlen (Interval.nth (x1a \<odot> xxs) (k)))" 
          by (simp add: sum.atLeast_Suc_atMost)
   show ?thesis 
   by (metis "1" "2" "3" "4" "5" "6" "7" "8" "9" not_gr_zero)
  qed
qed


lemma lsum_addzero_nth:
 assumes "i\<le> intlen (addzero (lsum xxs 0))" 
 shows "(intlen xxs = 0 \<and> intlen(intfirst xxs) = 0 \<longrightarrow>
          nth (addzero (lsum xxs 0)) i = (nth (lsum xxs 0) i) )
        \<and>
        (intlen xxs = 0 \<and> intlen(intfirst xxs) > 0 \<longrightarrow>
          nth (addzero (lsum xxs 0)) i = (nth (0\<odot>(lsum xxs 0)) i) )
        \<and>
        (intlen xxs > 0 \<longrightarrow>
          nth (addzero (lsum xxs 0)) i = (nth (0\<odot>(lsum xxs 0)) i) )
       "
using assms
by (metis add.left_neutral addzero_def less_numeral_extra(3) lsum_intfirst lsum_intlen)


lemma lsum_intlast:
 " intlast (lsum xxs a) = a+(\<Sum>k::nat= 0..(intlen xxs). intlen(nth xxs k))"
by (metis interval_nth_intlen_intlast le_refl lsum_intlen lsum_nth)


lemma lsum_addzero_intlast:
 "intlast (addzero (lsum xxs 0)) = intlast(lsum xxs 0)"
by (simp add: addzero_def)


lemma lsum_nth_leq_Suc:
 assumes "i < intlen xxs"
         " (\<forall> j \<le> intlen xxs. intlen(nth xxs j) > 0)"
 shows " nth (lsum xxs a) i < nth (lsum xxs a) (Suc i)"
proof -
 have 1: "nth (lsum xxs a) i = a+(\<Sum>k::nat= 0..i. intlen(nth xxs k)) "
     using assms less_imp_le_nat lsum_nth by blast 
 have 2: "nth (lsum xxs a) (Suc i) = a+(\<Sum>k::nat= 0..(Suc i). intlen(nth xxs k))" 
   using assms by (simp add: lsum_nth Suc_leI)
 have 3: "(\<Sum>k::nat= 0..(Suc i). intlen(nth xxs k)) =
          (\<Sum>k::nat= 0..i. intlen(nth xxs k)) + intlen(nth xxs (Suc i))  " 
      using sum.atLeast0_atMost_Suc by blast
 have 4: " intlen(nth xxs (Suc i)) > 0"
    using assms by auto
 show ?thesis using "1" "2" "3" "4" by linarith 
qed 

lemma lsum_addzero_nth_leq_Suc:
 assumes "i < intlen(addzero (lsum xxs 0))"
         "(\<forall> j \<le> intlen xxs. intlen(nth xxs j) > 0)"
 shows " nth (addzero (lsum xxs 0)) i < nth (addzero (lsum xxs 0)) (Suc i)"
proof (cases i)
case 0
then show ?thesis 
 by (metis add.left_neutral addzero_def assms(2) dual_order.order_iff_strict gr_zeroI
     interval_nth_Suc interval_nth_zero_intfirst lsum_addzero_intfirst lsum_intfirst)
next
case (Suc nat)
then show ?thesis 
 by (metis Suc_less_eq addzero_def assms(1) assms(2) interval_nth_Suc intlen.simps(2) lsum_intlen 
     lsum_nth_leq_Suc plus_1_eq_Suc)
qed

lemma lsum_idx:
assumes "(\<forall> j \<le> intlen xxs. intlen(nth xxs j) > 0)" 
shows "index_sequence (nth (lsum xxs a) 0) (lsum xxs a)"
by (simp add: assms index_sequence_def lsum_intlen lsum_nth_leq_Suc)

lemma lsum_addzero_idx:
assumes "(\<forall> j \<le> intlen xxs. intlen(nth xxs j) > 0)" 
shows "index_sequence 0 (addzero (lsum xxs 0))"
by (metis interval_idx_expand1 add.left_neutral addzero_def assms interval_nth_zero_intfirst le0
           lsum_idx lsum_intfirst)


lemma filt_lfuse_lsum_a:
assumes "lastfirst (xs \<odot>xxs)"
        "(\<forall> j \<le> intlen xxs. intlen(nth xxs j) > 0)"
        "intlen xs > 0"
shows "(filt (fuse xs (lfuse xxs)) (lsum xxs (intlen xs))) =
          (filt (lfuse xxs) ( (lsum xxs 0)))"
using assms 
proof 
  (induct xxs arbitrary: xs)
  case (St x)
  then show ?case 
   proof -
    have 1: "filt (fuse xs (lfuse \<langle>x\<rangle>)) (lsum \<langle>x\<rangle> (intlen xs)) =
             filt (fuse xs x) (lsum \<langle>x\<rangle> (intlen xs))"
      by simp
    have 2: "filt (fuse xs x) (lsum \<langle>x\<rangle> (intlen xs)) =
             filt (fuse xs x) \<langle>intlen xs + intlen x\<rangle>"
      by simp
    have 3: "filt (fuse xs x) (\<langle>intlen xs + intlen x\<rangle>) =
             \<langle>(nth (fuse xs x) (intlen xs + intlen x))\<rangle> "
      by simp
    have 4: "\<langle>(nth (fuse xs x) (intlen xs + intlen x))\<rangle> =
             \<langle>(nth x (intlen x))\<rangle>"
    using St.prems interval_fuse_nth_a by auto
    have 5: "filt (lfuse \<langle>x\<rangle>) (lsum \<langle>x\<rangle> 0) =
             filt (x) (\<langle>intlen x\<rangle>) "
          by simp
    have 6: "filt (x) (\<langle>intlen x\<rangle>) = \<langle>(nth x (intlen x))\<rangle>"
            by auto  
    show ?thesis 
    by (simp add: "4")
   qed
  next
  case (Cons x1a xxs)
  then show ?case 
   proof -
    have 01: "filt (fuse xs (lfuse (x1a \<odot> xxs))) (lsum (x1a \<odot> xxs) (intlen xs)) =
              filt (fuse xs (fuse x1a (lfuse xxs))) (lsum (x1a \<odot> xxs) (intlen xs)) " 
      by simp
    have 02: "filt (fuse xs (fuse x1a (lfuse xxs))) 
                   (lsum (x1a \<odot> xxs) (intlen xs)) =
              filt (fuse xs (fuse x1a (lfuse xxs))) 
                   ((intlen xs + intlen x1a) \<odot>  lsum (xxs) (intlen xs +intlen x1a))" 
           by simp
    have 03: "filt (fuse xs (fuse x1a (lfuse xxs))) 
                   ((intlen xs + intlen x1a) \<odot>  lsum (xxs) (intlen xs +intlen x1a)) =
              (nth (fuse xs (fuse x1a (lfuse xxs))) (intlen xs + intlen x1a)) \<odot>
              filt (fuse xs (fuse x1a (lfuse xxs))) (lsum (xxs) (intlen xs +intlen x1a)) "
        by simp
    have 04: "(nth (fuse xs (fuse x1a (lfuse xxs))) (intlen xs + intlen x1a)) =
              (nth (fuse xs x1a) (intlen xs +intlen x1a)) "
        proof -
        have f1: "Interval.nth xs (intlen xs) = Interval.nth x1a 0"
        using Cons.prems(1) by force 
        have f2: "lastfirst (x1a \<odot> xxs)"
        using Cons.prems(1) lastfirst.simps(2) by blast
        then have "Interval.nth x1a (intlen x1a) = Interval.nth (Interval.nth xxs 0) 0"
        by simp 
        then show ?thesis
        using f2 f1 
          by (metis Cons.prems(1) add.right_neutral interval_fuse_intlen_a 
               interval_fuse_nth_a interval_intfirst_lfuse_intfirst interval_nth_intlen_intlast 
              interval_nth_zero interval_nth_zero_intfirst le0 le_add1 lfuse_Cons) 
        qed
    have 05: "(nth (fuse xs x1a) (intlen xs +intlen x1a)) = (nth x1a (intlen x1a)) "
       using Cons.prems(1) interval_fuse_nth_a by force
    have 06: " filt (fuse xs (fuse x1a (lfuse xxs))) (lsum (xxs) (intlen xs +intlen x1a)) =
               filt (fuse (fuse xs x1a) (lfuse xxs)) (lsum (xxs) (intlen(fuse xs x1a)))"
       by (metis Cons.prems(1) interval_FusionAssoc interval_fuse_intlen_a interval_intfirst_lfuse 
           interval_intfirst_lfuse_intfirst lastfirst.simps(2))
    have 07: "lastfirst ( (fuse xs x1a)\<odot> xxs ) "
     by (metis "05" Cons.prems(1) interval_fuse_intlen_a interval_nth_intlen_intlast lastfirst.simps(2))
    have 08: " intlen (fuse xs x1a) > 0"
     by (simp add: Cons.prems interval_fuse_intlen_a)
    have 09: "filt (fuse (fuse xs x1a) (lfuse xxs)) (lsum (xxs) (intlen(fuse xs x1a))) =
              filt (lfuse xxs) (lsum (xxs) 0)"
        by (metis "07" "08" Cons.prems(2) Suc_leI interval_nth_Suc intlen.simps(2) less_Suc_eq_le 
           local.Cons(1) plus_1_eq_Suc)
    have 10: "filt (fuse xs (fuse x1a (lfuse xxs))) 
                   ((intlen xs + intlen x1a) \<odot>  lsum (xxs) (intlen xs +intlen x1a)) =
              (nth x1a (intlen x1a)) \<odot> filt (lfuse xxs) (lsum (xxs) 0)" 
     by (simp add: "04" "05" "06" "09")
    have 11: "filt (lfuse (x1a \<odot> xxs)) (lsum (x1a \<odot> xxs) 0) =
              filt (fuse x1a (lfuse xxs)) ((intlen x1a) \<odot> (lsum xxs (intlen x1a)))  "
     by simp
    have 12: "filt (fuse x1a (lfuse xxs)) ((intlen x1a) \<odot> (lsum xxs (intlen x1a))) =
              (nth (fuse x1a (lfuse xxs)) (intlen x1a)) \<odot> 
              (filt (fuse x1a (lfuse xxs)) (lsum xxs (intlen x1a)))"
        by simp
    have 13: "(nth (fuse x1a (lfuse xxs)) (intlen x1a)) = (nth x1a (intlen x1a))"
     by (metis Cons.prems(1) eq_iff interval_fuse_intlen_a interval_fuse_nth 
         interval_intfirst_lfuse_intfirst lastfirst.simps(2) le_add1)
    have 14: "(filt (fuse x1a (lfuse xxs)) (lsum xxs (intlen x1a))) =
              filt (lfuse xxs) (lsum xxs 0)"
      using Cons.prems(1) Cons.prems(2) interval_nth_zero intlen.simps(2) lastfirst.simps(2) 
        local.Cons(1) by fastforce
    have 15: "filt (fuse x1a (lfuse xxs)) ((intlen x1a) \<odot> (lsum xxs (intlen x1a))) = 
              (nth x1a (intlen x1a)) \<odot> (filt (lfuse xxs) (lsum xxs 0))"
     by (simp add: "13" "14") 
     show ?thesis using "10" "15" by auto
   qed
qed

lemma filt_lfuse_lsum:
assumes "lastfirst (xs \<odot>xxs)"
        " (\<forall> j \<le> intlen xxs. intlen(nth xxs j) > 0)"
        " intlen xs > 0"
shows " (filt (lfuse (xs\<odot>xxs)) (addzero (lsum (xs\<odot>xxs) 0))) = 
        (intfirst xs)\<odot> (intlast xs)\<odot> (filt (lfuse xxs) (lsum xxs 0))"
proof -
 have 1: " lfuse (xs \<odot> xxs) = fuse xs (lfuse xxs) "
   by simp
 have 2: "(filt (lfuse (xs\<odot>xxs)) (addzero (lsum (xs\<odot>xxs) 0))) =
          (filt (fuse xs (lfuse xxs)) (addzero (lsum (xs\<odot>xxs) 0)))  "
  using "1" by simp
 have 3: "addzero (lsum (xs\<odot>xxs) 0) =   
         0\<odot>(intlen xs) \<odot> (lsum xxs (intlen xs)) "
   using lsum_addzero_cons by auto
 have 4: "(filt (fuse xs (lfuse xxs)) (addzero (lsum (xs\<odot>xxs) 0))) =
          (filt (fuse xs (lfuse xxs)) (0\<odot>(intlen xs) \<odot> (lsum xxs (intlen xs)))) "
   using "3" by auto 
 have 5: " (filt (fuse xs (lfuse xxs)) (0\<odot>(intlen xs) \<odot> (lsum xxs (intlen xs)))) =
            (nth (fuse xs (lfuse xxs)) 0) \<odot> 
            (nth (fuse xs (lfuse xxs)) (intlen xs)) \<odot>
            (filt (fuse xs (lfuse xxs)) (lsum xxs (intlen xs)))  "
   by simp
 have 6: "(nth (fuse xs (lfuse xxs)) 0) = (nth xs 0)"
   using assms by (metis interval_fuse_nth interval_intfirst_lfuse_intfirst interval_intlen_gr_zero 
       lastfirst.simps(2))
 have 7: "(nth (fuse xs (lfuse xxs)) (intlen xs)) = (nth xs (intlen xs)) "
   using assms by (metis interval_fuse_intlen_a interval_fuse_nth interval_intfirst_lfuse_intfirst 
       lastfirst.simps(2) le_add1 order_refl)   
 have 8: "index_sequence 0 (addzero (lsum  xxs 0))"
   using assms lsum_addzero_idx by blast  
 have 9: "(filt (fuse xs (lfuse xxs)) (lsum xxs (intlen xs))) =
          (filt (lfuse xxs) (lsum xxs 0))"
     using assms filt_lfuse_lsum_a by blast
   show ?thesis 
   using "3" "6" "7" "9" interval_nth_intlen_intlast interval_nth_zero_intfirst by force
qed   
 
lemma filt_lfuse_lsum_1:
assumes "lastfirst (xs \<odot>xxs)" 
        " (\<forall> j \<le> intlen xxs. intlen(nth xxs j) > 0)"
        " intlen xs > 0"
shows " (filt (lfuse (xs\<odot>xxs)) ((lsum (xs\<odot>xxs) 0))) = 
         (intlast xs)\<odot> (filt (lfuse xxs) (lsum xxs 0))"
using assms by (simp, 
                metis assms(1) eq_iff filt_lfuse_lsum_a interval_fuse_intlen_a 
                interval_fuse_nth interval_intfirst_lfuse_intfirst lastfirst.simps(2) le_add1) 

lemma filt_lfuse_lsum_2:
assumes "lastfirst (xxs)"
        " (\<forall> j \<le> intlen xxs. intlen(nth xxs j) > 0)"
        " j \<le> intlen xxs "
shows   " (nth (filt (lfuse (xxs)) ((lsum (xxs) 0))) j) = intlast(nth xxs j)"
using assms 
 proof (induct xxs arbitrary: j)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xxs)
 then show ?case  
  proof -
    have 1: "j= 0 \<and> j \<le> intlen (x1a \<odot> xxs) \<longrightarrow>
         nth (filt (lfuse (x1a \<odot> xxs)) (lsum (x1a \<odot> xxs) 0)) j = intlast (nth (x1a \<odot> xxs) j)   "
      using filt_lfuse_lsum_1 using Cons.prems by fastforce
    have 2: "j> 0 \<and> j \<le> intlen (x1a \<odot> xxs) \<longrightarrow>
         nth (filt (lfuse (x1a \<odot> xxs)) (lsum (x1a \<odot> xxs) 0)) j = intlast (nth (x1a \<odot> xxs) j) "
     by (metis Cons.hyps Cons.prems(2) Suc_le_mono Suc_pred filt_lfuse_lsum_1 
         interval_intlen_gr_zero interval_nth_Suc interval_nth_zero intlen.simps(2) 
         lastfirst.simps(2) local.Cons(2) plus_1_eq_Suc) 
   show ?thesis using "1" "2" not_gr_zero using Cons.prems(3) by blast
  qed
qed

lemma filt_lfuse_lsum_3:
assumes "lastfirst (xxs)"
        " (\<forall> j \<le> intlen xxs. intlen(nth xxs j) > 0)"
        " j \<le> intlen (addzero (lsum xxs 0)) " 
shows   " ( j=0 \<longrightarrow> (nth (filt (lfuse (xxs)) (addzero(lsum (xxs) 0))) j) = intfirst(intfirst xxs))
          \<and>
          (j>0 \<longrightarrow> (nth (filt (lfuse (xxs)) (addzero(lsum (xxs) 0))) j) = intlast(nth xxs (j-1)))
        "
using assms 
proof -
 have 1: "j \<le> intlen (addzero (lsum xxs 0)) \<and> j=0 \<longrightarrow>
           (nth (filt (lfuse (xxs)) (addzero(lsum (xxs) 0))) j) = intfirst(intfirst xxs)  " 
     using assms by (metis filt_intlen filt_nth interval_lastfirst_lfuse interval_nth_zero_intfirst 
          lsum_addzero_intfirst)
 have 2: "j \<le> intlen (addzero (lsum xxs 0)) \<and> j>0 \<longrightarrow>
        (nth (filt (lfuse xxs) (addzero(lsum (xxs) 0))) j) = 
         (nth (lfuse xxs) (nth  (addzero(lsum (xxs) 0)) j))         
         " 
     by (simp add: filt_intlen filt_nth)
 have 3: "j \<le> intlen (addzero (lsum xxs 0)) \<and> j>0 \<longrightarrow> 
           nth  (addzero(lsum (xxs) 0)) j = nth (lsum xxs 0) (j-1)
         "    
    by (metis One_nat_def Suc_pred interval_nth_Suc leD lsum_addzero_intlen lsum_addzero_nth neq0_conv)
 have 4: "j \<le> intlen (addzero (lsum xxs 0)) \<and> j>0 \<longrightarrow> 
          (nth (lfuse xxs) (nth (lsum xxs 0) (j-1))) = intlast(nth xxs (j -1)) " 
   by (metis assms diff_is_0_eq' filt_intlen filt_lfuse_lsum_2 filt_nth interval_nth_zero_intfirst 
       le0 le_diff_conv lsum_addzero_intlen lsum_intlen neq0_conv)
 show ?thesis 
 using "1" "2" "3" "4" by (simp add: assms(3))
qed

   

lemma filt_lfuse_lsum_4:
 assumes "lastfirst (xxs)"
          "(\<forall> j \<le> intlen xxs. intlen(nth xxs j) > 0)"
 shows   " (nth (addzero (lsum xxs 0)) (intlen (addzero (lsum xxs 0)))) \<le> intlen (lfuse xxs)"
using assms
by (metis add_cancel_right_left eq_iff interval_intlast_prefix interval_lfuse_intlen 
    interval_prefix_intlen lsum_addzero_intlast lsum_intlen lsum_nth) 

lemma filt_lfuse_lsum_5:
 assumes "lastfirst (xxs)"
         " (\<forall> j \<le> intlen xxs. intlen(nth xxs j) > 0)"
         " i\<le>intlen (addzero (lsum xxs 0)) "
 shows   " (nth (addzero (lsum xxs 0)) i) \<le> intlen (lfuse xxs)"
using assms filt_lfuse_lsum_4[of "xxs"] lsum_addzero_idx[of "xxs"]
proof -
have f1: "Interval.nth (addzero (lsum xxs 0)) (intlen (addzero (lsum xxs 0))) \<le> intlen (lfuse xxs)"
using \<open>\<forall>j\<le>intlen xxs. 0 < intlen (Interval.nth xxs j)\<close> 
     \<open>\<lbrakk>lastfirst xxs; \<forall>j\<le>intlen xxs. 0 < intlen (Interval.nth xxs j)\<rbrakk> \<Longrightarrow> 
      Interval.nth (addzero (lsum xxs 0)) (intlen (addzero (lsum xxs 0))) \<le> intlen (lfuse xxs)\<close> 
      \<open>lastfirst xxs\<close> by blast (* 0.4 ms *)
have "\<not> i < intlen (addzero (lsum xxs 0)) \<or> 
       nth (addzero (lsum xxs 0)) i < nth (addzero (lsum xxs 0)) (intlen (addzero (lsum xxs 0)))"
using \<open>\<forall>j\<le>intlen xxs. 0 < intlen (Interval.nth xxs j) \<Longrightarrow> index_sequence 0 (addzero (lsum xxs 0))\<close> 
      \<open>\<forall>j\<le>intlen xxs. 0 < intlen (Interval.nth xxs j)\<close> interval_idx_less_last_1 by blast (* 1 ms *)
then show ?thesis
using f1 \<open>i \<le> intlen (addzero (lsum xxs 0))\<close> by force (* 6 ms *)
qed

lemma lfuse_intlen_b:
 assumes "lastfirst xxs"
  shows  "( \<forall>i. 1 \<le> i \<and> i \<le>intlen (xxs) \<longrightarrow> 
                  (\<forall>j\<le>intlen (nth (xxs) (i)).
                     nth ((lsum xxs 0) ) (i-1) + j \<le> intlen (lfuse xxs)))"
proof -
 have 0: "(\<forall>i. 1 \<le> i \<and> i \<le>intlen (xxs) \<longrightarrow>  
            (i-1) \<le> intlen xxs 
          ) "
     by linarith
 have 1: "(\<forall>i. 1 \<le> i \<and> i \<le>intlen (xxs) \<longrightarrow> 
                   nth ((lsum xxs 0) ) (i-1) = 
                    (\<Sum>k::nat= 0..(i-1). intlen(nth xxs k))
            ) "
         by (metis "0" add.left_neutral lsum_nth)
 have 2: "intlen (lfuse xxs) = (\<Sum>k::nat= 0..(intlen xxs). intlen(nth xxs k))" 
     using assms interval_lfuse_intlen by blast
 have 3: "(\<forall>i. 1 \<le> i \<and> i \<le>intlen (xxs) \<longrightarrow> 
            (\<forall>j\<le>intlen (nth (xxs) (i)).
              (\<Sum>k::nat= 0..(i-1). intlen(nth xxs k)) + j \<le> 
               (\<Sum>k::nat= 0..(i). intlen(nth xxs k)) 
            ))
         " 
     by (metis (no_types, lifting) add_le_cancel_left le_add_diff_inverse plus_1_eq_Suc 
      sum.atLeast0_atMost_Suc)
 have 4: "(\<forall>i. 1 \<le> i \<and> i \<le>intlen (xxs) \<longrightarrow> 
            (\<Sum>k::nat= 0..(i). intlen(nth xxs k)) \<le> 
             (\<Sum>k::nat= 0..(intlen xxs). intlen(nth xxs k))
          )
        " 
        by (metis add.commute diff_is_0_eq diff_zero le_add1 le_add_diff_inverse2 not_less_eq_eq 
            sum.ub_add_nat)  
 show ?thesis 
 using "1" "2" "3" "4" by fastforce
qed


lemma lsum_shift:
 assumes "lastfirst xxs"
         " (\<forall> j \<le> intlen xxs. intlen(nth xxs j) > 0)"
         " i\<le>intlen xxs "  
 shows   "  nth (lsum xxs a) i = a+ nth (lsum xxs 0) i"
using assms by (simp add: lsum_nth)

lemma lsum_lfuse_nth_lsum_nth:
assumes "lastfirst xxs"
        " (\<forall> j \<le> intlen xxs. intlen(nth xxs j) > 0)"
shows    " (\<forall> i\<le>intlen xxs. 
                 (\<forall> j\<le> intlen(nth xxs i). 
                  ( (nth (lfuse xxs) ((nth (addzero (lsum xxs 0)) i) +j) )) =
                  ( (nth (nth xxs i) j))  ))"
using assms
proof 
 (induct xxs)
 case (St x)
 then show ?case 
 by (metis Interval.nth.simps(1) add_cancel_right_left interval_nth_zero_intfirst intlen.simps(1) 
     le_zero_eq lfuse_St lsum_addzero_intfirst)
 next
 case (Cons x1a xxs)
 then show ?case 
  proof -
   have 1: "(\<forall>i\<le>intlen (x1a \<odot> xxs).
             (\<forall>j\<le>intlen (nth (x1a \<odot> xxs) i).
               nth (lfuse (x1a \<odot> xxs)) (nth (addzero (lsum (x1a \<odot> xxs) 0)) i + j) = 
               nth (nth (x1a \<odot> xxs) i) j)) 
           = 
           ( (\<forall>j\<le>intlen (nth (x1a \<odot> xxs) 0).
                  nth (lfuse (x1a \<odot> xxs)) (nth (addzero (lsum (x1a \<odot> xxs) 0)) 0 + j) = 
                  nth (nth (x1a \<odot> xxs) 0) j) \<and> 
             (\<forall>i. 1 \<le> i \<and> i\<le>intlen (x1a \<odot> xxs) \<longrightarrow> 
                  (\<forall>j\<le>intlen (nth (x1a \<odot> xxs) i).
                    nth (lfuse (x1a \<odot> xxs)) (nth (addzero (lsum (x1a \<odot> xxs) 0)) i + j) = 
                    nth (nth (x1a \<odot> xxs) i) j)) )          
         "
    by (metis One_nat_def Suc_leI interval_intlen_gr_zero not_gr_zero)
   have 2: "(\<forall>j\<le>intlen (nth (x1a \<odot> xxs) 0).
                  nth (lfuse (x1a \<odot> xxs)) (nth (addzero (lsum (x1a \<odot> xxs) 0)) 0 + j) = 
                  nth (nth (x1a \<odot> xxs) 0) j) 
            =
            (\<forall>j\<le>intlen (x1a).
                  nth (lfuse (x1a \<odot> xxs)) ( j) = nth (x1a) j)
           "
       by (metis add_cancel_right_left interval_nth_zero interval_nth_zero_intfirst 
           lsum_addzero_intfirst)
   have 3: "intlast x1a = intfirst(intfirst xxs)"
      using Cons.prems lastfirst.simps(2) by blast
   have 4: "(\<forall>j\<le>intlen (x1a).
                  nth (lfuse (x1a \<odot> xxs)) ( j) = nth (x1a) j) " 
     by (metis Cons.prems(1) interval_fuse_intlen_a interval_fuse_nth 
         interval_intfirst_lfuse_intfirst lastfirst.simps(2) le_add1 le_trans  lfuse_Cons)
   have 41: "(\<forall>j\<le>intlen (nth (x1a \<odot> xxs) 0).
                  nth (lfuse (x1a \<odot> xxs)) (nth (addzero (lsum (x1a \<odot> xxs) 0)) 0 + j) = 
                  nth (nth (x1a \<odot> xxs) 0) j)" 
     using "2" "4" by blast
   have 5: "(\<forall>i. 1 \<le> i \<and> i\<le>intlen (x1a \<odot> xxs) \<longrightarrow> 
             (\<forall>j\<le>intlen (nth (x1a \<odot> xxs) i).
                    nth (lfuse (x1a \<odot> xxs)) (nth (addzero (lsum (x1a \<odot> xxs) 0)) i + j) = 
                    nth (nth (x1a \<odot> xxs) i) j)) =
            (\<forall>i. 0 \<le> i-1 \<and> i-1 \<le>intlen (xxs) \<longrightarrow> 
             (\<forall>j\<le>intlen (nth (x1a \<odot> xxs) i).
                    nth (lfuse (x1a \<odot> xxs)) (nth (addzero (lsum (x1a \<odot> xxs) 0)) i + j) = 
                    nth (nth (x1a \<odot> xxs) i) j))" 
         using "1" "2" "4" le_diff_conv by auto
   have 6: "(\<forall>i. 0 \<le> i-1 \<and> i-1 \<le>intlen (xxs) \<longrightarrow> 
              (\<forall>j\<le>intlen (nth (x1a \<odot> xxs) i).
                    nth (lfuse (x1a \<odot> xxs)) (nth (addzero (lsum (x1a \<odot> xxs) 0)) i + j) = 
                    nth (nth (x1a \<odot> xxs) i) j)) =
            (\<forall>i \<le>intlen (xxs).
              (\<forall>j\<le>intlen (nth (x1a \<odot> xxs) (Suc i)).
                    nth (lfuse (x1a \<odot> xxs)) (nth (addzero (lsum (x1a \<odot> xxs) 0)) (Suc i) + j) = 
                    nth (nth (x1a \<odot> xxs) (Suc i)) j))" (is "?L = ?R")
            proof 
             show "?L \<Longrightarrow> ?R" 
              using "2" "4" by simp
             show "?R \<Longrightarrow> ?L" 
              using "2" "4" by (metis One_nat_def Suc_pred gr0I)
            qed
   have 7: "(\<forall>i \<le>intlen (xxs). 
             (\<forall>j\<le>intlen (nth (x1a \<odot> xxs) (Suc i)).
                    nth (lfuse (x1a \<odot> xxs)) (nth (addzero (lsum (x1a \<odot> xxs) 0)) (Suc i) + j) =
                    nth (nth (x1a \<odot> xxs) (Suc i)) j)) =
            (\<forall>i \<le>intlen (xxs). 
             (\<forall>j\<le>intlen (nth (xxs) (i)).
                    nth (lfuse (x1a \<odot> xxs)) (nth (addzero (lsum (x1a \<odot> xxs) 0)) (Suc i) + j) =
                    nth (nth (xxs) (i)) j))" 
            by simp
   have 8: "(\<forall>i \<le>intlen (xxs).
             (\<forall>j\<le>intlen (nth (xxs) (i)).
                    nth (lfuse (x1a \<odot> xxs)) (nth (addzero (lsum (x1a \<odot> xxs) 0)) (Suc i) + j) = 
                    nth (nth (xxs) (i)) j)) =
            (\<forall>i \<le>intlen (xxs).
             (\<forall>j\<le>intlen (nth (xxs) (i)).
                    nth (fuse x1a (lfuse xxs)) (nth ( (lsum (x1a \<odot> xxs) 0)) (i) + j) = 
                    nth (nth (xxs) (i)) j))"
       by (metis interval_nth_Suc lfuse_Cons lsum_addzero_cons)
   have 9: "(\<forall>i \<le>intlen (xxs).
             (\<forall>j\<le>intlen (nth (xxs) (i)).
                    nth (fuse x1a (lfuse xxs)) (nth ( (lsum (x1a \<odot> xxs) 0)) (i) + j) = 
                    nth (nth (xxs) (i)) j))
              =
            (\<forall>i \<le>intlen (xxs). 
             (\<forall>j\<le>intlen (nth (xxs) (i)).
                    nth (fuse x1a (lfuse xxs)) (nth ((intlen(x1a)\<odot>(lsum xxs) (intlen x1a)) ) i +j) =
                    nth (nth (xxs) (i)) j))
            " 
    by simp
  have 11: "(\<forall>i \<le>intlen (xxs). 
             (\<forall>j\<le>intlen (nth (xxs) (i)).
               nth (fuse x1a (lfuse xxs)) (nth ((intlen(x1a)\<odot> (lsum xxs) (intlen x1a)) ) (i) + j) = 
               nth (nth (xxs) (i)) j)) 
              =
            ((\<forall>i. i=0 \<longrightarrow> 
              (\<forall>j\<le>intlen (nth (xxs) (i)).
                nth (fuse x1a (lfuse xxs)) (nth ((intlen(x1a)\<odot> (lsum xxs) (intlen x1a)) ) (i) + j) = 
                nth (nth (xxs) (i)) j))
               \<and> 
             (\<forall>i. 1 \<le> i \<and> i \<le>intlen (xxs) \<longrightarrow> 
              (\<forall>j\<le>intlen (nth (xxs) (i)).
                nth (fuse x1a (lfuse xxs)) (nth ((intlen(x1a)\<odot> (lsum xxs) (intlen x1a)) ) (i) + j) = 
                nth (nth (xxs) (i)) j)) ) 
           "
         by (metis Suc_leI add.commute add_less_same_cancel1  interval_nth_zero le0 le_eq_less_or_eq 
             less_one not_gr_zero plus_1_eq_Suc)
  have 12: "(\<forall>i. i=0 \<longrightarrow> 
             (\<forall>j\<le>intlen (nth (xxs) (i)).
               nth (fuse x1a (lfuse xxs)) (nth ((intlen(x1a)\<odot> (lsum xxs) (intlen x1a)) ) (i) + j) = 
               nth (nth (xxs) (i)) j)) =
            (\<forall>j\<le>intlen (nth (xxs) (0)).
               nth (fuse x1a (lfuse xxs)) (nth ((intlen(x1a)\<odot> (lsum xxs) (intlen x1a)) ) (0) + j) = 
               nth (nth (xxs) (0)) j)
          "
   by blast
  have 13: "(\<forall>j\<le>intlen (nth (xxs) (0)).
              nth (fuse x1a (lfuse xxs)) (nth ((intlen(x1a)\<odot> (lsum xxs) (intlen x1a)) ) (0) + j) = 
              nth (nth (xxs) (0)) j) =
            (\<forall>j\<le>intlen (nth (xxs) (0)).
              nth (fuse x1a (lfuse xxs)) (intlen(x1a) + j) = 
              nth (nth (xxs) (0)) j)
           " 
    by auto
  have 14: "(\<forall>j\<le>intlen (nth (xxs) (0)).
              nth (fuse x1a (lfuse xxs)) (intlen(x1a) + j) = 
              nth (nth (xxs) (0)) j) 
            = 
            (\<forall>j\<le>intlen (nth (xxs) (0)).
              nth ( (lfuse xxs)) (j) = 
              nth (nth (xxs) (0)) j)
           " 
    using Cons.prems interval_fuse_nth_a interval_intfirst_lfuse_intfirst interval_lfuse_intlen_a 
    by (metis lastfirst.simps(2) le0)
  have 15: "(\<forall>i. 1 \<le> i \<and> i \<le>intlen (xxs) \<longrightarrow> 
             (\<forall>j\<le>intlen (nth (xxs) (i)).
               nth (fuse x1a (lfuse xxs)) (nth ((intlen(x1a)\<odot> (lsum xxs) (intlen x1a)) ) (i) + j) = 
               nth (nth (xxs) (i)) j)) =
            (\<forall>i. 1 \<le> i \<and> i \<le>intlen (xxs) \<longrightarrow> 
             (\<forall>j\<le>intlen (nth (xxs) (i)).
               nth (fuse x1a (lfuse xxs)) (nth ( (lsum xxs (intlen x1a)) ) (i-1) + j) = 
               nth (nth (xxs) (i)) j))
          " 
    by (metis interval_nth_Suc ordered_cancel_comm_monoid_diff_class.add_diff_inverse plus_1_eq_Suc)
  have 16: "(\<forall>i. 1 \<le> i \<and> i \<le>intlen (xxs) \<longrightarrow>
               nth ( (lsum xxs (intlen x1a)) ) (i-1) = 
               intlen x1a + nth (lsum xxs 0)  (i-1))
            " 
       by (simp add: le_diff_conv lsum_nth)
  have 17: "(\<forall>i. 1 \<le> i \<and> i \<le>intlen (xxs) \<longrightarrow> 
              (\<forall>j\<le>intlen (nth (xxs) (i)).
                nth (fuse x1a (lfuse xxs)) (nth ( (lsum xxs (intlen x1a)) ) (i-1) + j) = 
                nth (nth (xxs) (i)) j))
            =
            (\<forall>i. 1 \<le> i \<and> i \<le>intlen (xxs) \<longrightarrow> 
             (\<forall>j\<le>intlen (nth (xxs) (i)).
               nth (fuse x1a (lfuse xxs)) (intlen x1a + nth ((lsum xxs 0) ) (i-1) + j) =
               nth (nth (xxs) (i)) j))"
      using "16" by auto
  have 18: "(\<forall>i. 1 \<le> i \<and> i \<le>intlen (xxs) \<longrightarrow> 
             (\<forall>j\<le>intlen (nth (xxs) (i)).
               nth ((lsum xxs 0) ) (i-1) + j \<le> intlen (lfuse xxs)))"
     using Cons.prems lastfirst.simps(2) lfuse_intlen_b by blast
  have 19: "(\<forall>i. 1 \<le> i \<and> i \<le>intlen (xxs) \<longrightarrow> 
             (\<forall>j\<le>intlen (nth (xxs) (i)). 
                nth (fuse x1a (lfuse xxs)) (intlen x1a + nth ((lsum xxs 0) ) (i-1) +j) =
                nth ( (lfuse xxs)) ( nth ((lsum xxs 0) ) (i-1) + j))) 
              " 
    by (metis Cons.prems(1) ab_semigroup_add_class.add_ac(1) interval_fuse_nth_a 
        interval_intfirst_lfuse_intfirst lastfirst.simps(2) lfuse_intlen_b)
  have 20: "(\<forall>i. 1 \<le> i \<and> i \<le>intlen (xxs) \<longrightarrow> 
             (\<forall>j\<le>intlen (nth (xxs) (i)).
               nth (fuse x1a (lfuse xxs)) (intlen x1a + nth ((lsum xxs 0) ) (i-1) + j) = 
               nth (nth (xxs) (i)) j)) =
            (\<forall>i. 1 \<le> i \<and> i \<le>intlen (xxs) \<longrightarrow> 
             (\<forall>j\<le>intlen (nth (xxs) (i)).
               nth ( (lfuse xxs)) ( nth ((lsum xxs 0) ) (i-1) + j) = 
               nth (nth (xxs) (i)) j))
              "  
     using 19 by auto
  have 201: "((\<forall>i. i=0 \<longrightarrow> 
              (\<forall>j\<le>intlen (nth (xxs) (i)).
                nth (fuse x1a (lfuse xxs)) (nth ((intlen(x1a)\<odot> (lsum xxs) (intlen x1a)) ) (i) + j) = 
                nth (nth (xxs) (i)) j))
               \<and> 
             (\<forall>i. 1 \<le> i \<and> i \<le>intlen (xxs) \<longrightarrow> 
              (\<forall>j\<le>intlen (nth (xxs) (i)).
                nth (fuse x1a (lfuse xxs)) (nth ((intlen(x1a)\<odot> (lsum xxs) (intlen x1a)) ) (i) + j) = 
                nth (nth (xxs) (i)) j)) ) =
              ((\<forall>j\<le>intlen (nth (xxs) (0)).
                  nth ( (lfuse xxs)) (j) = 
                  nth (nth (xxs) (0)) j) \<and>
               (\<forall>i. 1 \<le> i \<and> i \<le>intlen (xxs) \<longrightarrow> 
                (\<forall>j\<le>intlen (nth (xxs) (i)).
                  nth ( (lfuse xxs)) ( nth ((lsum xxs 0) ) (i-1) + j) = 
                  nth (nth (xxs) (i)) j)) ) "
        using "14" "15" "17" "20" by auto
  have 21: "lastfirst xxs "
    using Cons.prems lastfirst.simps(2) by blast
  have 22: " (\<forall> j \<le> intlen xxs. intlen(nth xxs j) > 0)"
    using Cons.prems by auto
  have 23: "(\<forall>i\<le>intlen xxs.
             (\<forall>j\<le>intlen (nth xxs i).
               nth (lfuse xxs) (nth (addzero (lsum xxs 0)) i + j) = 
               nth (nth xxs i) j)) "       
     using "21" "22" Cons.hyps by blast
  have 24: "((\<forall>j\<le>intlen (nth xxs 0).
               nth (lfuse xxs) (nth (addzero (lsum xxs 0)) 0 + j) = 
               nth (Interval.nth xxs 0) j) \<and>
             (\<forall>i. 1 \<le> i \<and> i \<le>intlen xxs \<longrightarrow>
              (\<forall>j\<le>intlen (nth xxs i).
                nth (lfuse xxs) (nth (addzero (lsum xxs 0)) i + j) = 
                nth (nth xxs i) j))  )"      
        using "23" by blast
   have 25: "(\<forall>j\<le>intlen (nth (xxs) (0)).
                  nth ( (lfuse xxs)) (j) = 
                  nth (nth (xxs) (0)) j)" 
     by (metis "24" add_cancel_right_left interval_nth_zero_intfirst lsum_addzero_intfirst)
   have 26: "(\<forall>i. 1 \<le> i \<and> i \<le>intlen (xxs) \<longrightarrow> 
                  nth (addzero (lsum xxs 0)) i = nth ((lsum xxs 0) ) (i-1)
             )"
       by (metis addzero_def interval_nth_Suc less_eq_Suc_le less_le_trans lsum_intlen 
           not_less_eq_eq ordered_cancel_comm_monoid_diff_class.add_diff_inverse plus_1_eq_Suc 
           zero_less_one) 
   have 27: "(\<forall>i. 1 \<le> i \<and> i \<le>intlen (xxs) \<longrightarrow> 
                (\<forall>j\<le>intlen (nth (xxs) (i)).
                  nth ( (lfuse xxs)) ( nth ((lsum xxs 0) ) (i-1) + j) = 
                  nth (nth (xxs) (i)) j)) =
             (\<forall>i. 1 \<le> i \<and> i \<le>intlen xxs \<longrightarrow>
              (\<forall>j\<le>intlen (nth xxs i).
                nth (lfuse xxs) (nth (addzero (lsum xxs 0)) i + j) = 
                nth (nth xxs i) j)) " 
     by (simp add: "26")
  show ?thesis 
  using "11" "201" "24" "25" "27" "6" "8" by auto
qed    

qed




lemma lcpl_lsum_less_th_equal:
 assumes "index_sequence 0 l"
         " (nth l (intlen l)) = intlen \<sigma>"
         " (\<forall> i<intlen l. (sub (nth l i) (nth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)"
         " intlen \<sigma> > 0 "
         " i < intlen (addzero (lsum (lcpl f g \<sigma> l) 0)) " 
 shows   " (nth (addzero (lsum (lcpl f g \<sigma> l) 0)) (Suc i)) \<le>
           intlen( (lfuse (lcpl f g \<sigma> l)))"
using assms 
by (metis (no_types, lifting) Suc_leI filt_lfuse_lsum_5 index_sequence_def 
    lcpl_intlen_nth_gr_zero lcpl_lfuse_lastfirst neq0_conv)
 



lemma lcpl_lsum_intlen:
 assumes "index_sequence 0 l"
         " (nth l (intlen l)) = intlen \<sigma>"
         " (\<forall> i<intlen l. (sub (nth l i) (nth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)"
         " intlen \<sigma> >0 " 
 shows   "intlen (addzero (lsum ((lcpl f g \<sigma> l)) 0)) = intlen l"
proof -
 have 1: "intlen \<sigma> >0 \<longrightarrow> intlen (lcpl f g \<sigma> l) = intlen l -1"
   using assms index_sequence_def lcpl_intlen by fastforce
 have 2: "intlen \<sigma>>0 \<longrightarrow> intlen l >0"
   using assms gr_zeroI index_sequence_def by fastforce
 have 3: "intlen \<sigma> > 0 \<longrightarrow> intlen (intfirst (lcpl f g \<sigma> l)) > 0"  
    by (metis assms interval_intlen_gr_zero interval_nth_zero_intfirst lcpl_intlen_nth_gr_zero)
 have 4: "intlen \<sigma> >0 \<longrightarrow> intlen (lcpl f g \<sigma> l) = 0 \<longrightarrow> 
           intlen (addzero (lsum ((lcpl f g \<sigma> l)) 0)) = intlen l " 
     using "1" "2" "3" lsum_addzero_intlen by fastforce
 have 5: "intlen \<sigma> >0 \<longrightarrow> intlen (lcpl f g \<sigma> l) > 0 \<longrightarrow> 
        intlen (addzero (lsum ((lcpl f g \<sigma> l)) 0)) = intlen l" 
    by (simp add: "1" lsum_addzero_intlen)
  show ?thesis using "4" "5" using assms(4) by blast
qed

lemma lcpl_lsum_nth:
 assumes "index_sequence 0 l"
         " (nth l (intlen l)) = intlen \<sigma> "
         "(\<forall> i<intlen l. (sub (nth l i) (nth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)"
         " intlen \<sigma>>0 " 
         " j \<le> intlen (addzero (lsum (lcpl f g \<sigma> l) 0)) "
 shows  " ( j=0 \<longrightarrow> 
          (nth (filt (lfuse ((lcpl f g \<sigma> l))) (addzero(lsum ((lcpl f g \<sigma> l)) 0))) j) =
           intfirst(intfirst (lcpl f g \<sigma> l)) )
          \<and>
          (j>0 \<longrightarrow> (nth (filt (lfuse ((lcpl f g \<sigma> l))) (addzero(lsum ((lcpl f g \<sigma> l)) 0))) j) =
            intlast(nth (lcpl f g \<sigma> l) (j-1)))
         "
proof -
 have 0: "intlen \<sigma> >0 \<longrightarrow> intlen l >0" 
   using assms gr_zeroI index_sequence_def by fastforce
 have 1: "intlen \<sigma> >0 \<longrightarrow> lastfirst (lcpl f g \<sigma> l)" 
    using "0" assms index_sequence_def lcpl_lfuse_lastfirst by blast
 have 2: "intlen \<sigma>>0 \<longrightarrow> 
         (\<forall> j \<le> intlen (lcpl f g \<sigma> l). intlen(nth (lcpl f g \<sigma> l) j) > 0) "
   by (simp add: assms lcpl_intlen_nth_gr_zero)
 have 3: " intlen \<sigma>>0 \<longrightarrow> 
           intlen (addzero (lsum (lcpl f g \<sigma> l) 0)) =
           intlen (lcpl f g \<sigma> l) +1
          " 
          by (metis "2" One_nat_def Suc_eq_plus1 interval_intlen_gr_zero interval_nth_zero_intfirst 
         le_imp_less_or_eq lsum_addzero_intlen)
 have 4: " intlen \<sigma> > 0 \<longrightarrow>
           (nth (filt (lfuse ((lcpl f g \<sigma> l))) (addzero(lsum ((lcpl f g \<sigma> l)) 0))) 0) =
             intfirst(intfirst (lcpl f g \<sigma> l)) "
    by (simp add: "1" "2" filt_lfuse_lsum_3) 
 have 5: "intlen \<sigma> > 0 \<longrightarrow> 
          j \<le> intlen (addzero (lsum (lcpl f g \<sigma> l) 0)) \<and> j>0 \<longrightarrow>
          (nth (filt (lfuse ((lcpl f g \<sigma> l))) (addzero(lsum ((lcpl f g \<sigma> l)) 0))) (j))
           = intlast(nth (lcpl f g \<sigma> l) (j-1))"
   by (simp add: "1" "2" filt_lfuse_lsum_3)
 from 4 5 show ?thesis using assms(4) assms(5) by blast
qed   

lemma lcpl_lsum_nth_a:
 assumes "index_sequence 0 l"
         " (nth l (intlen l)) = intlen \<sigma> " 
         " (\<forall> i<intlen l. (sub (nth l i) (nth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)"
         " intlen \<sigma>>0 "
         " j \<le> intlen l " 
 shows  " (nth (filt (lfuse ((lcpl f g \<sigma> l))) (addzero(lsum ((lcpl f g \<sigma> l)) 0))) j) =
          (nth l j)"
proof -
 have 1: "intlen \<sigma>>0 \<longrightarrow> intlen l>0 "
  using assms gr_zeroI index_sequence_def by fastforce 
 have 2: "intlen \<sigma>>0 \<longrightarrow> intlen (addzero (lsum (lcpl f g \<sigma> l) 0)) = intlen l"
    by (simp add: assms lcpl_lsum_intlen) 
 have 3: "intlen \<sigma>>0 \<longrightarrow> 
          j \<le> intlen l \<longrightarrow>
            ( j=0 \<longrightarrow> 
          (nth (filt (lfuse ((lcpl f g \<sigma> l))) (addzero(lsum ((lcpl f g \<sigma> l)) 0))) j) =
           intfirst(intfirst (lcpl f g \<sigma> l)) )
          \<and>
          (j>0 \<longrightarrow> (nth (filt (lfuse ((lcpl f g \<sigma> l))) (addzero(lsum ((lcpl f g \<sigma> l)) 0))) j) = 
           intlast(nth (lcpl f g \<sigma> l) (j-1)))
          " 
     by (metis "2" assms lcpl_lsum_nth)
 have 4: "intlen \<sigma>>0 \<longrightarrow> 
          j \<le> intlen l \<and> j=0 \<longrightarrow> intfirst(intfirst (lcpl f g \<sigma> l)) = (nth l j) " 
     using assms lcpl_intfirst[of "l" "\<sigma>" "f" "g"]  
     by (metis "1" index_sequence_def interval_nth_zero_intfirst)
 have 5: "intlen \<sigma>>0 \<longrightarrow> 
          j \<le> intlen l \<and> j>0 \<longrightarrow> j-1 \<le> intlen (lcpl f g \<sigma> l)"
      using assms by (metis "1" diff_le_mono index_sequence_def lcpl_intlen)
 have 6: "intlen \<sigma>>0 \<longrightarrow> 
          j \<le> intlen l \<and> j>0 \<longrightarrow> intlast(nth (lcpl f g \<sigma> l) (j-1)) = (nth l j)"
    using lcpl_intlast_nth 
     by (metis "5" One_nat_def Suc_pred assms)
 show ?thesis 
 using "3" "4" "6" using assms(4) assms(5) by auto
qed

lemma lcpl_filt_lfuse_lsum:
 assumes "index_sequence 0 l"
         " (nth l (intlen l)) = intlen \<sigma>"
         " (\<forall> i<intlen l. (sub (nth l i) (nth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)"
         " intlen \<sigma>>0 " 
 shows  " (filt (lfuse ((lcpl f g \<sigma> l))) (addzero(lsum ((lcpl f g \<sigma> l)) 0))) = l"
using assms 
proof -
 have 0: "intlen \<sigma>>0 \<longrightarrow>   intlen l > 0" 
   using assms gr_zeroI index_sequence_def by fastforce 
 have 1: " intlen \<sigma>>0 \<longrightarrow>  
          intlen (filt (lfuse ((lcpl f g \<sigma> l))) (addzero(lsum ((lcpl f g \<sigma> l)) 0))) = intlen l"
   by (simp add: assms filt_intlen lcpl_lsum_intlen)
 have 2: "intlen \<sigma>>0 \<longrightarrow>  
           (\<forall> j. j\<le>intlen l \<longrightarrow>
          (nth (filt (lfuse ((lcpl f g \<sigma> l))) (addzero(lsum ((lcpl f g \<sigma> l)) 0))) j) =
          (nth l j))  "
     by (simp add: assms lcpl_lsum_nth_a)
 from 1 2 show ?thesis by (simp add: assms(4) interval_eq_nth_eq) 
qed


lemma lcpl_lfuse_filt_power:
 assumes "index_sequence 0 l"
         " (nth l (intlen l)) = intlen \<sigma> "  
         " (\<forall> i<intlen l. (sub (nth l i) (nth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)"
         "  intlen \<sigma> >0 " 
 shows   "  powerinterval g (filt \<sigma> (lfuse (lcpl f g \<sigma> l))) (addzero (lsum (lcpl f g \<sigma> l) 0))"
proof -
 have 0: "intlen \<sigma> > 0 \<longrightarrow> intlen l >0 "
 using assms gr_zeroI index_sequence_def by fastforce 
 have 01: "(\<forall> i<intlen l. 
            g (filt (sub (nth l i) (nth l (Suc i)) \<sigma>) (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>))))"
   using assms cpl_projection by auto
 have 02: "intlen \<sigma> > 0 \<longrightarrow> (\<forall> i<intlen l. g (filt \<sigma> (nth (lcpl f g \<sigma> l) i)) )"
   using "0" assms index_sequence_def lcpl_lfuse_filt_power_help by blast
 have 03 : "powerinterval g (filt \<sigma> (lfuse (lcpl f g \<sigma> l))) (addzero (lsum (lcpl f g \<sigma> l) 0)) =
           ( \<forall>i<intlen (addzero (lsum (lcpl f g \<sigma> l) 0)).
             g (sub (nth (addzero (lsum (lcpl f g \<sigma> l) 0)) i) 
                    (nth (addzero (lsum (lcpl f g \<sigma> l) 0)) (Suc i))
                    (filt \<sigma> (lfuse (lcpl f g \<sigma> l)))))" 
        by (simp add: powerinterval_def)
  have 04: "intlen \<sigma> >0 \<longrightarrow>  intlen (addzero (lsum (lcpl f g \<sigma> l) 0)) = intlen l"
    by (simp add: assms lcpl_lsum_intlen) 
  have 05: "intlen \<sigma> >0 \<longrightarrow>
           (\<forall> i<intlen l. 
              (filt \<sigma> (nth (lcpl f g \<sigma> l) i))  =
              (sub (nth (addzero (lsum (lcpl f g \<sigma> l) 0)) i) 
                   (nth (addzero (lsum (lcpl f g \<sigma> l) 0)) (Suc i))
                   (filt \<sigma> (lfuse (lcpl f g \<sigma> l)))) )
      " 
   proof -
    have 06: "intlen \<sigma> >0 \<longrightarrow> 
              (\<forall> i<intlen l. intlen (filt \<sigma> (nth (lcpl f g \<sigma> l) i)) =
                    intlen (nth (lcpl f g \<sigma> l) i))"
      using filt_intlen by blast
    have 07: "intlen \<sigma> >0 \<longrightarrow> 
              (\<forall> i<intlen l. intlen (nth (lcpl f g \<sigma> l) i) =
                intlen (map (shift (nth l i)) (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)))) 
              " 
        using assms by (metis index_sequence_def lcpl_nth)
    have 08: "intlen \<sigma> >0 \<longrightarrow> 
              (\<forall> i<intlen l. 
                intlen (map (shift (nth l i)) (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)))  =
                intlen (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)))
              " 
         using interval_intlen_map by blast
    have 09: "intlen \<sigma> >0 \<longrightarrow> 
              (\<forall>i<intlen l. 
               intlen (sub (nth (addzero (lsum (lcpl f g \<sigma> l) 0)) i)
                           (nth (addzero (lsum (lcpl f g \<sigma> l) 0)) (Suc i))
                           (filt \<sigma> (lfuse (lcpl f g \<sigma> l)))) =
               (nth (addzero (lsum (lcpl f g \<sigma> l) 0)) (Suc i)) -  
               (nth (addzero (lsum (lcpl f g \<sigma> l) 0)) i) )
             " 
             by (metis "04" assms filt_intlen interval_intlen_sub lcpl_lsum_less_th_equal 
                  lcpl_intlen_nth_gr_zero le_eq_less_or_eq lsum_addzero_nth_leq_Suc)
    have 10: "intlen \<sigma> >0 \<longrightarrow> 
               (\<forall> i<intlen l. (nth (addzero (lsum (lcpl f g \<sigma> l) 0)) (Suc i)) =
                 (nth (0\<odot>(lsum (lcpl f g \<sigma> l) 0)) (Suc i)))  
              " 
        using "04" addzero_def by auto
    have 11: "intlen \<sigma> >0 \<longrightarrow> 
               (\<forall> i<intlen l. (nth (0\<odot>(lsum (lcpl f g \<sigma> l) 0)) (Suc i)) =
                 (\<Sum>k::nat= 0..(i). intlen(nth (lcpl f g \<sigma> l) k)) )
              "   using "04" 
        proof simp_all
          assume "0 < intlen \<sigma> \<longrightarrow> intlen (addzero (lsum (lcpl f g \<sigma> l) 0)) = intlen l"
          show   " 0 < intlen \<sigma> \<longrightarrow>
                  (\<forall>i<intlen l. nth (lsum (lcpl f g \<sigma> l) 0) i = 
                   (\<Sum>k = 0..i. intlen (nth (lcpl f g \<sigma> l) k))) "
           proof 
            assume " 0 < intlen \<sigma>"
            show " \<forall>i<intlen l. nth (lsum (lcpl f g \<sigma> l) 0) i = 
                   (\<Sum>k = 0..i. intlen (nth (lcpl f g \<sigma> l) k)) " 
            proof 
             fix i
             show "i < intlen l \<longrightarrow>
                   nth (lsum (lcpl f g \<sigma> l) 0) i = (\<Sum>k = 0..i. intlen (nth (lcpl f g \<sigma> l) k)) " 
             by (metis (no_types, lifting) "0" One_nat_def Suc_leI Suc_le_mono Suc_pred 
                 add.left_neutral assms(1) assms(4) index_sequence_def lcpl_intlen lsum_nth sum.cong)
            qed
           qed
        qed              
    have 12: "intlen \<sigma> >0 \<longrightarrow> 
               (\<forall> i<intlen l. (nth (addzero (lsum (lcpl f g \<sigma> l) 0)) i) =
                 (nth (0\<odot>(lsum (lcpl f g \<sigma> l) 0)) i) )
              "      
       using "04" addzero_def by auto
    have 13: "intlen \<sigma> >0 \<longrightarrow> 
               (\<forall> i<intlen l. (nth (0\<odot>(lsum (lcpl f g \<sigma> l) 0)) i) =
               (case i of 0 \<Rightarrow> 0
                 | Suc j \<Rightarrow> (\<Sum>k::nat= 0..(j). intlen(nth (lcpl f g \<sigma> l) k)) ))    
              " using "11" by (simp_all add: Nitpick.case_nat_unfold)
    have 14: "intlen \<sigma> >0 \<longrightarrow> 
               (\<forall> i<intlen l. 
              (nth (addzero (lsum (lcpl f g \<sigma> l) 0)) (Suc i)) -  
              (nth (addzero (lsum (lcpl f g \<sigma> l) 0)) i) =
              (\<Sum>k::nat= 0..(i). intlen(nth (lcpl f g \<sigma> l) k)) -
              (case i of 0 \<Rightarrow> 0
                 | Suc j \<Rightarrow> (\<Sum>k::nat= 0..(j). intlen(nth (lcpl f g \<sigma> l) k)) ) )
             " 
      using "10" "11" "12" "13" by auto
    have 15: "intlen \<sigma> >0 \<longrightarrow> 
               (\<forall> i<intlen l.  
                (\<Sum>k::nat= 0..(i). intlen(nth (lcpl f g \<sigma> l) k)) -
              (case i of 0 \<Rightarrow> 0
                 | Suc j \<Rightarrow> (\<Sum>k::nat= 0..(j). intlen(nth (lcpl f g \<sigma> l) k)) ) =
                (case i of 0 \<Rightarrow> intlen(nth (lcpl f g \<sigma> l) 0)
                  | Suc j \<Rightarrow> (\<Sum>k::nat= 0..(Suc j). intlen(nth (lcpl f g \<sigma> l) k)) -
                             (\<Sum>k::nat= 0..(j). intlen(nth (lcpl f g \<sigma> l) k)) ))
             " 
        by (simp add: Nitpick.case_nat_unfold)
     have 16: "intlen \<sigma> >0 \<longrightarrow> 
               (\<forall> i<intlen l.  
                 (case i of 0 \<Rightarrow> intlen(nth (lcpl f g \<sigma> l) 0)
                  | Suc j \<Rightarrow> (\<Sum>k::nat= 0..(Suc j). intlen(nth (lcpl f g \<sigma> l) k)) -
                             (\<Sum>k::nat= 0..(j). intlen(nth (lcpl f g \<sigma> l) k)) ) = 
                 (case i of 0 \<Rightarrow> intlen(nth (lcpl f g \<sigma> l) 0)
                  | Suc j \<Rightarrow> intlen(nth (lcpl f g \<sigma> l) (Suc j))  ) )
              "
      by (metis (no_types, lifting) Nitpick.case_nat_unfold add_diff_cancel_left' 
           sum.atLeast0_atMost_Suc)
    have 17: "intlen \<sigma> >0 \<longrightarrow> 
               (\<forall> i<intlen l. 
                  (case i of 0 \<Rightarrow> intlen(nth (lcpl f g \<sigma> l) 0)
                  | Suc j \<Rightarrow> intlen(nth (lcpl f g \<sigma> l) (Suc j))  ) =
                  intlen(nth (lcpl f g \<sigma> l) i))
             " 
      by (simp add: Nitpick.case_nat_unfold)
    have 18: "intlen \<sigma> >0 \<longrightarrow> 
              (\<forall> i<intlen l. intlen (filt \<sigma> (nth (lcpl f g \<sigma> l) i)) =
                intlen (sub (nth (addzero (lsum (lcpl f g \<sigma> l) 0)) i) 
                            (nth (addzero (lsum (lcpl f g \<sigma> l) 0)) (Suc i))
                            (filt \<sigma> (lfuse (lcpl f g \<sigma> l)))) )
             "   
       by (simp add: "09" "14" "15" "16" "17" filt_intlen)
    have 19: "intlen \<sigma> >0 \<longrightarrow> 
               (\<forall> i<intlen l. 
                (\<forall> j\<le> intlen(nth (lcpl f g \<sigma> l) i).
                 (nth (addzero (lsum (lcpl f g \<sigma> l) 0)) (Suc i)) \<le> 
                  intlen (filt \<sigma> (lfuse (lcpl f g \<sigma> l)))))
             "
             by (simp add: "04" assms filt_intlen lcpl_lsum_less_th_equal)
    have 22: " intlen \<sigma> >0 \<longrightarrow>  lastfirst (lcpl f g \<sigma> l)"
      using "0" assms index_sequence_def lcpl_lfuse_lastfirst by blast
   have 23: " intlen \<sigma> >0 \<longrightarrow>  (\<forall> j \<le> intlen (lcpl f g \<sigma> l). intlen(nth (lcpl f g \<sigma> l) j) > 0)"
     by (simp add: assms lcpl_intlen_nth_gr_zero)
    have 190: "intlen \<sigma> >0 \<longrightarrow> 
                (\<forall> i<intlen l. 
                   (\<forall> j\<le> intlen(nth (lcpl f g \<sigma> l) i).
                      (nth (addzero (lsum (lcpl f g \<sigma> l) 0)) i) \<le> 
                      (nth (addzero (lsum (lcpl f g \<sigma> l) 0)) (Suc i))
                   ))
               " 
       by (simp add: "04" "23" less_imp_le_nat lsum_addzero_nth_leq_Suc)
    have 20: "intlen \<sigma> >0 \<longrightarrow> 
               (\<forall> i<intlen l. 
                 (\<forall> j\<le> intlen(nth (lcpl f g \<sigma> l) i). 
                   (nth (sub (nth (addzero (lsum (lcpl f g \<sigma> l) 0)) i) 
                             (nth (addzero (lsum (lcpl f g \<sigma> l) 0)) (Suc i))
                             (filt \<sigma> (lfuse (lcpl f g \<sigma> l)))) j) =
                   (nth (filt \<sigma> (lfuse (lcpl f g \<sigma> l))) 
                        ((nth (addzero (lsum (lcpl f g \<sigma> l) 0)) i) +j) ) ) ) 
              " 
    by (simp add: "14" "15" "16" "17" "19" "190")
   have 21: "intlen \<sigma> >0 \<longrightarrow> 
               (\<forall> i<intlen l. 
                 (\<forall> j\<le> intlen(nth (lcpl f g \<sigma> l) i).
                  (nth (filt \<sigma> (lfuse (lcpl f g \<sigma> l))) 
                       ((nth (addzero (lsum (lcpl f g \<sigma> l) 0)) i) +j) ) =
                  (nth \<sigma> (nth (lfuse (lcpl f g \<sigma> l)) 
                              ((nth (addzero (lsum (lcpl f g \<sigma> l) 0)) i) +j) )) ) )
            " 
       by (simp add: filt_map interval_nth_map)
   have 24: "intlen \<sigma> >0 \<longrightarrow>
             (\<forall> i\<le>intlen (lcpl f g \<sigma> l). 
                 (\<forall> j\<le> intlen(nth (lcpl f g \<sigma> l) i). 
                  ( (nth (lfuse (lcpl f g \<sigma> l)) ((nth (addzero (lsum (lcpl f g \<sigma> l) 0)) i) +j) )) =
                  ( (nth (nth (lcpl f g \<sigma> l) i) j))  ))
            " 
     by (simp add: "22" "23" lsum_lfuse_nth_lsum_nth)
   have 241: "intlen \<sigma> >0 \<longrightarrow> intlen (lcpl f g \<sigma> l) = intlen l -1"
     using "0" assms index_sequence_def lcpl_intlen by blast 
   have 25: " intlen \<sigma> >0 \<longrightarrow> 
               (\<forall> i<intlen l. 
                 (\<forall> j\<le> intlen(nth (lcpl f g \<sigma> l) i). 
                  (nth \<sigma> (nth (lfuse (lcpl f g \<sigma> l)) 
                              ((nth (addzero (lsum (lcpl f g \<sigma> l) 0)) i) +j) )) =
                  (nth (filt \<sigma> (nth (lcpl f g \<sigma> l) i)) j) ))
            " 
            by (simp add: "06" "24" "241" assms filt_nth)
    have 26: "intlen \<sigma> >0 \<longrightarrow> 
               (\<forall> i<intlen l. 
                 (\<forall> j\<le> intlen(nth (lcpl f g \<sigma> l) i). 
                  (nth (filt \<sigma> (nth (lcpl f g \<sigma> l) i)) j) =
                  (nth (sub (nth (addzero (lsum (lcpl f g \<sigma> l) 0)) i)
                            (nth (addzero (lsum (lcpl f g \<sigma> l) 0)) (Suc i))
                            (filt \<sigma> (lfuse (lcpl f g \<sigma> l)))) j) ) )
             "
     by (simp add: "20" "21" "25")
   from 18 26 show ?thesis 
   by (simp add: filt_intlen interval_eq_nth_eq)
   qed
 show ?thesis 
 using "02" "03" "04" "05" by (simp add: assms(4))
qed

lemma lcpl_lfuse_filt_intlen:
 assumes "index_sequence 0 l"
         " (nth l (intlen l)) = intlen \<sigma>" 
         " (\<forall> i<intlen l. (sub (nth l i) (nth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)"
         " intlen \<sigma> >0 " 
 shows   " (nth (addzero (lsum (lcpl f g \<sigma> l) 0)) (intlen(addzero (lsum (lcpl f g \<sigma> l) 0)))) =
          intlen (filt \<sigma> (lfuse (lcpl f g \<sigma> l))) "
proof -
 have 0: "intlen \<sigma> > 0 \<longrightarrow> intlen l > 0"
    using assms gr_zeroI index_sequence_def by fastforce 
 have 1: "intlen \<sigma> >0 \<longrightarrow>
          intlen (filt \<sigma> (lfuse (lcpl f g \<sigma> l))) = intlen ( (lfuse (lcpl f g \<sigma> l)))  " 
   using filt_intlen by blast
 have 2: "intlen \<sigma> >0 \<longrightarrow> 
           intlen ( (lfuse (lcpl f g \<sigma> l))) = 
            (\<Sum>k::nat= 0..(intlen (lcpl f g \<sigma> l)). intlen(nth (lcpl f g \<sigma> l) k))"
   using "0" assms index_sequence_def interval_lfuse_intlen lcpl_lfuse_lastfirst by blast 
 have 3: "intlen \<sigma> >0 \<longrightarrow> 
      intlast( addzero (lsum (lcpl f g \<sigma> l) 0)) = intlast( (lsum (lcpl f g \<sigma> l) 0))  "
    using lsum_addzero_intlast by blast
 have 4: "intlen \<sigma> >0 \<longrightarrow> 
           intlast( (lsum (lcpl f g \<sigma> l) 0)) = 
          (\<Sum>k::nat= 0..(intlen (lcpl f g \<sigma> l)). intlen(nth (lcpl f g \<sigma> l) k))"
   by (metis add_cancel_right_left lsum_intlast)
 show ?thesis using "1" "2" "3" "4" by (simp add: assms(4))
qed




subsection \<open>Soundness of Projection Axioms\<close>

subsubsection \<open>PJ1\<close>

lemma PJ1sem:
 "(\<sigma> \<Turnstile> f \<triangle> ( g \<or> h) \<longrightarrow> f \<triangle> g \<or> f \<triangle> h)"
by (simp add: projection_d_def) blast

lemma PJ1sema:
 "(\<sigma> \<Turnstile> f \<triangle> ( g \<or> h) = (f \<triangle> g \<or> f \<triangle> h))"
by (simp add: projection_d_def) blast

subsubsection \<open>PJ2\<close>

lemma PJ2sem:
 "(\<sigma> \<Turnstile> f \<triangle> empty = empty)"
proof auto
  show "(\<sigma> \<Turnstile> f \<triangle> empty) \<Longrightarrow> \<sigma> \<Turnstile> empty " 
  unfolding projection_d_def empty_defs index_sequence_def  
  by (metis filt.simps(2)  interval_intlen_cons_1 neq0_conv) 
  show "\<sigma> \<Turnstile> empty \<Longrightarrow> (\<sigma> \<Turnstile> f \<triangle> empty)" 
  unfolding projection_d_def empty_defs index_sequence_def powerinterval_def 
   by (metis Interval.nth.simps(1) filt.simps(1) index_sequence_def intlen.simps(1) 
       not_less_zero )
qed

subsubsection \<open>PJ3\<close>

lemma PJ3help:
 "sub 0 (intlen \<sigma>) \<sigma> = \<sigma>"
by (simp add: interval_sub_zero_prefix)

lemma PJ3help1:
assumes "f \<sigma> \<and> 0 < intlen \<sigma>"
shows   " (\<exists>l. index_sequence 0 l \<and>
        nth l (intlen l) = intlen \<sigma> \<and>
        (\<forall>i<intlen l. f (sub (nth l i) (nth l (Suc i)) \<sigma>)) \<and>
        (\<exists>s1. intlen s1 = intlen l \<and>
      (\<forall>i\<le>intlen s1. nth s1 i = nth \<sigma> (nth l i)) \<and> intlen s1 = Suc 0))"
proof -
 have 1: "index_sequence 0 \<langle>0,intlen \<sigma>\<rangle> " 
   by (simp add: assms index_sequence_def)
 have 2: "nth \<langle>0,intlen \<sigma>\<rangle> (intlen \<langle>0,intlen \<sigma>\<rangle>) = intlen \<sigma> " 
   by auto
 have 3: "(\<forall>i<intlen \<langle>0,intlen \<sigma>\<rangle>. f (sub (nth \<langle>0,intlen \<sigma>\<rangle> i) (nth \<langle>0,intlen \<sigma>\<rangle> (Suc i)) \<sigma>)) "     
   by (simp add: PJ3help assms)
 have 4: "intlen \<langle>nth \<sigma> (0), nth \<sigma> (intlen \<sigma>)\<rangle> = intlen \<langle>0,intlen \<sigma>\<rangle>" 
    by simp
 have 5: "(\<forall>i\<le>intlen  \<langle>nth \<sigma> (0), nth \<sigma> (intlen \<sigma>)\<rangle>. 
      nth  \<langle>nth \<sigma> (0), nth \<sigma> (intlen \<sigma>)\<rangle> i = nth \<sigma> (nth \<langle>0,intlen \<sigma>\<rangle> i)) "
   using antisym_conv2 by fastforce 
 have 6: "intlen \<langle>nth \<sigma> (0), nth \<sigma> (intlen \<sigma>)\<rangle> = Suc 0 " 
   by simp
 show ?thesis 
 using "1" "2" "3" "4" "5" "6" by blast
qed


lemma PJ3sem:
"(\<sigma> \<Turnstile> f \<triangle> skip = (f \<and> more))"
proof -
 have 1: "(\<sigma> \<Turnstile> f \<triangle> skip) \<Longrightarrow> (\<sigma> \<Turnstile> f \<and> more)"
   by (metis (mono_tags, lifting) One_nat_def PJ3help cpl_projection filt_expand index_sequence_def 
       more_defs powerinterval_def skip_defs unl_lift2 zero_less_one)
 have 2: "(\<sigma> \<Turnstile> f \<and> more) \<Longrightarrow> (\<sigma> \<Turnstile> f \<triangle> skip)"
   by (simp add:  projection_d_def skip_defs more_defs powerinterval_def,
       metis PJ3help1  filt_intlen)
 show ?thesis 
 using "1" "2" unl_lift2 by blast
qed


subsubsection \<open>PJ4\<close>

lemma PJ4semchaina:
assumes "(\<sigma> \<Turnstile> f \<triangle> (g;h))"
 shows  "(\<sigma> \<Turnstile> (f \<triangle> g) ; (f \<triangle> h))"
proof -
 have 1: "(\<sigma> \<Turnstile> f \<triangle> (g;h))" 
  using assms by auto
 have 2: "(\<exists>l. index_sequence 0 l \<and>
         nth l (intlen l) = intlen \<sigma> \<and>
         powerinterval f \<sigma> l \<and>
         (\<exists>n\<le>intlen (filt \<sigma> l). g (prefix n (filt \<sigma> l)) \<and> h (suffix n (filt \<sigma> l)))) "
  by (metis assms chop_defs projection_d_def)
  obtain l where 3: "index_sequence 0 l \<and>
         nth l (intlen l) = intlen \<sigma> \<and>
         powerinterval f \<sigma> l \<and>
         (\<exists>n\<le>intlen (filt \<sigma> l). g (prefix n (filt \<sigma> l)) \<and> h (suffix n (filt \<sigma> l)))" 
    using "2" by auto
 have 4: "index_sequence 0 l "  
   using "3" by auto
 have 5: "powerinterval f \<sigma> l" 
   using "3" by auto
 have 6: "nth l (intlen l) = intlen \<sigma>"
   using "3" by auto
 have 7: "(\<exists>n\<le>intlen (filt \<sigma> l). g (prefix n (filt \<sigma> l)) \<and> h (suffix n (filt \<sigma> l)))" 
   using "3" by auto
  obtain n where 8: "n\<le>intlen (filt \<sigma> l) \<and> g (prefix n (filt \<sigma> l)) \<and> h (suffix n (filt \<sigma> l)) " 
  using "7" by auto
 have 9: "n\<le>intlen (filt \<sigma> l) " 
  using "8" by auto
 have 10: " g (prefix n (filt \<sigma> l))" 
  using "8" by auto
 have 11: "h (suffix n (filt \<sigma> l))" 
  using "8" by auto
 have 12: "index_sequence 0 (prefix n l)" 
   by (metis "4" "8" filt_intlen interval_idx_split)
 have 13: "index_sequence (nth l n) (suffix n l)"
   by (metis "4" "9" filt_intlen interval_idx_split) 
 have 14: "index_sequence 0 ((map (shiftm (nth l n))  (suffix n l)))" 
  using "13" interval_idx_shiftm by blast
 have 15: "g  (filt \<sigma> (prefix n l))"
   by (metis "8" filt_intlen filt_prefix) 
 have 16: "h (filt \<sigma> (suffix n l))"
   by (metis "11" "9" filt_intlen filt_suffix) 
 have 17: "g  (filt (prefix (nth l n) \<sigma>) (prefix n l))"
   by (metis (no_types, lifting) "12" "15" "4" "6" "8" filt_intlen filt_prefix_idx 
       interval_idx_less_equal interval_intlast_prefix interval_nth_intlen_intlast order_refl)
 have 18: "h (filt (suffix (nth l n) \<sigma>) ((map (shiftm (nth l n))  (suffix n l))))" 
    proof -
     have 181: "intlen((filt \<sigma> (suffix n l))) = 
                intlen(filt (suffix (nth l n) \<sigma>) ((map (shiftm (nth l n))  (suffix n l)))) " 
        by (simp add: filt_intlen)
     have 182: "(\<forall>j\<le> intlen((filt \<sigma> (suffix n l))).
                 (nth (filt \<sigma> (suffix n l)) j) =
                 (nth \<sigma> ((nth l (n+j)))) 
                )" 
          by (metis "9" filt_intlen filt_map interval_nth_map interval_nth_suffix 
              interval_suffix_length_good)
     have 183: "(\<forall>j\<le> intlen((filt \<sigma> (suffix n l))).
                  (nth (filt (suffix (nth l n) \<sigma>) ((map (shiftm (nth l n))  (suffix n l)))) j) =
                  (nth (suffix (nth l n) \<sigma>) (nth (map (shiftm (nth l n))  (suffix n l)) j))
                )  "
             by (simp add: filt_map interval_nth_map) 
     have 184: "(nth l n) \<le> intlen \<sigma>"
          by (metis "4" "6" "9" filt_intlen interval_idx_less_equal order_refl)
     have 185: "(\<forall>j\<le> intlen((filt \<sigma> (suffix n l))). 
                 (nth (map (shiftm (nth l n))  (suffix n l)) j) =
                  (nth (suffix n l) j) - (nth l n)
                ) "
            by (metis "4" "6" "9" filt_intlen interval_nth_suffix interval_idx_shiftm_suffix_nth 
                interval_suffix_length_good) 
     have 186: "(\<forall>j\<le> intlen((filt \<sigma> (suffix n l))). 
                   (nth (suffix n l) j) - (nth l n) = (nth l (j+n)) - (nth l n)
                )"
           by (metis "9" add.commute filt_intlen interval_nth_suffix interval_suffix_length_good) 
     have 187: "(\<forall>j\<le> intlen((filt \<sigma> (suffix n l))).
                   (nth l (j+n)) - (nth l n) \<le> intlen \<sigma> - (nth l n)
                )"
          by (metis "4" "6" "9" add.commute diff_le_mono eq_imp_le filt_intlen 
             interval_idx_less_equal interval_suffix_length_good nat_add_left_cancel_le 
             ordered_cancel_comm_monoid_diff_class.add_diff_inverse)
     have 188: "(\<forall>j\<le> intlen((filt \<sigma> (suffix n l))).
                 (nth (suffix (nth l n) \<sigma>) (nth (map (shiftm (nth l n))  (suffix n l)) j)) =
                 (nth \<sigma> ((nth (map (shiftm (nth l n))  (suffix n l))  j) + (nth l n) ))
                ) "
           using interval_nth_suffix 
           by (simp add: "184" "185" "186" "187" add.commute) 
     have 189: "(\<forall>j\<le> intlen((filt \<sigma> (suffix n l))). 
                 (nth \<sigma> ((nth (map (shiftm (nth l n))  (suffix n l))  j) + (nth l n) )) =
                 (nth \<sigma> ((nth (suffix n l) j)) )
                ) "
         by (metis "13" "185" filt_intlen interval_idx_greater 
             ordered_cancel_comm_monoid_diff_class.le_imp_diff_is_add)
     have 190: "(\<forall>j\<le> intlen((filt \<sigma> (suffix n l))).
                 (nth (filt \<sigma> (suffix n l)) j) = 
                 (nth (filt (suffix (nth l n) \<sigma>) ((map (shiftm (nth l n))  (suffix n l)))) j)
                )"
           using "183" "188" "189" filt_expand by fastforce
     show ?thesis 
     by (metis "16" "181" "190" interval_eq_nth_eq)
    qed
 have 19: "powerinterval f (prefix (nth l n) \<sigma>) (prefix n l)"
   by (metis "4" "5" "6" "8" filt_intlen powerinterval_splita)
 have 20: "powerinterval f (suffix (nth l n) \<sigma>) ((map (shiftm (nth l n))  (suffix n l)))" 
   by (metis "4" "5" "6" "9" filt_intlen powerinterval_split)
 have 21: "(nth l n) \<le> intlen \<sigma> "
    by (metis "3" "9" filt_intlen interval_idx_less_equal order_refl)
 have 22: "nth (prefix n l) (intlen (prefix n l)) = (nth l n)"
   by (metis "8" filt_intlen interval_intlast_prefix interval_nth_intlen_intlast) 
 have 23: " nth ((map (shiftm (nth l n))  (suffix n l)))
                (intlen ((map (shiftm (nth l n))  (suffix n l)))) = intlen \<sigma> -(nth l n)" 
   by (metis "4" "6" "9" eq_imp_le filt_intlen interval_intlen_map interval_idx_shiftm_suffix_nth 
        interval_suffix_length_good ordered_cancel_comm_monoid_diff_class.add_diff_inverse)
 have 24: "l = fuse (prefix n l) 
                    (map (shift (nth l n) ) ((map (shiftm (nth l n))  (suffix n l))))"
   using interval_fuse_prefix_suffix 
   by (metis "13" "14" "8" filt_intlen interval_lsk_ls)
 have 25: "(\<exists>l1. index_sequence 0 l1 \<and>
             Interval.nth l1 (intlen l1) = intlen (prefix (nth l n) \<sigma>) \<and>
             powerinterval f (prefix (nth l n) \<sigma>) l1 \<and> g (filt (prefix (nth l n) \<sigma>) l1))"
   by (metis "12" "17" "19" "21" "22" interval_prefix_length_good) 
 have 26: "(\<exists>l2. index_sequence 0 l2 \<and> 
             Interval.nth l2 (intlen l2) = intlen (suffix (nth l n) \<sigma>) \<and>
             powerinterval f (suffix (nth l n) \<sigma>) l2 \<and> h (filt (suffix (nth l n) \<sigma>) l2))"
   using "18" "14" "20" "23" "21" by auto
 have 27: "(prefix (nth l n) \<sigma>) \<Turnstile> f\<triangle>g"
     by (metis "25" projection_d_def)
 have 28: " (suffix (nth l n) \<sigma>) \<Turnstile> f\<triangle>h"  
    by (metis "26" projection_d_def)
  show ?thesis 
  using "21" "27" "28" chop_defs by auto
qed

lemma PJ4semchainb:
assumes "(\<sigma> \<Turnstile> (f \<triangle> g) ; (f \<triangle> h))"
 shows  "(\<sigma> \<Turnstile> f \<triangle> (g;h))"
proof -
 have 1: "(\<exists>n\<le>intlen \<sigma>.
        (\<exists>l. index_sequence 0 l \<and>
             Interval.nth l (intlen l) = intlen (prefix n \<sigma>) \<and>
             powerinterval f (prefix n \<sigma>) l \<and> g (filt (prefix n \<sigma>) l)) \<and>
        (\<exists>l. index_sequence 0 l \<and>
             Interval.nth l (intlen l) = intlen (suffix n \<sigma>) \<and>
             powerinterval f (suffix n \<sigma>) l \<and> h (filt (suffix n \<sigma>) l))) " 
    using assms by (metis chop_defs cpl_projection)
 obtain cp where 2: "cp\<le>intlen \<sigma> \<and> 
        (\<exists>l. index_sequence 0 l \<and>
             Interval.nth l (intlen l) = intlen (prefix cp \<sigma>) \<and>
             powerinterval f (prefix cp \<sigma>) l \<and> g (filt (prefix cp \<sigma>) l)) \<and>
        (\<exists>l. index_sequence 0 l \<and>
             Interval.nth l (intlen l) = intlen (suffix cp \<sigma>) \<and>
             powerinterval f (suffix cp \<sigma>) l \<and> h (filt (suffix cp \<sigma>) l))"
   using "1" by auto
 have 3: "cp\<le>intlen \<sigma>"
  using "2" by auto
 have 4: "(\<exists>l. index_sequence 0 l \<and>
             Interval.nth l (intlen l) = intlen (prefix cp \<sigma>) \<and>
             powerinterval f (prefix cp \<sigma>) l \<and> g (filt (prefix cp \<sigma>) l))" 
  using "2" by auto
 obtain l1 where 5: "index_sequence 0 l1 \<and>
             Interval.nth l1 (intlen l1) = intlen (prefix cp \<sigma>) \<and>
             powerinterval f (prefix cp \<sigma>) l1 \<and> g (filt (prefix cp \<sigma>) l1)"
   using "4" by auto
 have 6: "index_sequence 0 l1"  
   using "5" by auto
 have 7: "Interval.nth l1 (intlen l1) = intlen (prefix cp \<sigma>)"  
   using "5" by auto
 have 8: "powerinterval f (prefix cp \<sigma>) l1" 
    using "5" by auto
 have 9: "g (filt (prefix cp \<sigma>) l1)"
   using "5" by auto
 have 10: "(\<exists>l. index_sequence 0 l \<and>
             Interval.nth l (intlen l) = intlen (suffix cp \<sigma>) \<and>
             powerinterval f (suffix cp \<sigma>) l \<and> h (filt (suffix cp \<sigma>) l))" 
   using "2" by auto
 obtain l2 where 11: "index_sequence 0 l2 \<and>
             Interval.nth l2 (intlen l2) = intlen (suffix cp \<sigma>) \<and>
             powerinterval f (suffix cp \<sigma>) l2 \<and> h (filt (suffix cp \<sigma>) l2)"  
   using "10" by auto
 have 12: "index_sequence 0 l2" 
   using "11" by auto
 have 13: "Interval.nth l2 (intlen l2) = intlen (suffix cp \<sigma>)"
     using "11" by auto
 have 14: "powerinterval f (suffix cp \<sigma>) l2" 
    using "11" by auto
 have 15: "h (filt (suffix cp \<sigma>) l2)" 
    using "11" by auto
 have 16: "index_sequence 0 (fuse l1 (map (shift cp) l2))"
    by (metis "11" "12" "2" "5" "6" eq_imp_le interval_idx_fuse_idx interval_prefix_length_good 
        interval_suffix_length_good)
 have 17: "intlast l1 = intfirst (map (shift cp) l2)"
    by (metis "12" "13" "3" "6" "7" interval_idx_fuse_intfirst_intlast interval_prefix_length_good 
        interval_suffix_length_good)
 have 18: "nth (fuse l1 (map (shift cp) l2)) (intlen l1) = cp"
    by (metis "17" "3" "7" eq_imp_le interval_fuse_intlen_a interval_fuse_nth 
        interval_prefix_length_good le_add1)
 have 19: "intlast (fuse l1 (map (shift cp) l2))  = intlen \<sigma> " 
    proof -
      have 191: "intlast (fuse l1 (map (shift cp) l2)) = intlast (map (shift cp) l2)" 
         by (metis "17" eq_imp_le interval_fuse_nth_a interval_fuse_intlen_a 
           interval_nth_intlen_intlast)
      have 192: "intlast (map (shift cp) l2) = intlast l2 + cp"
        by (metis Interval.shift_def interval_intlen_map interval_nth_intlen_intlast 
             interval_nth_map) 
      have 193: "intlast l2 + cp = intlen \<sigma>" 
         by (metis "13" "3" Nat.le_imp_diff_is_add interval_nth_intlen_intlast 
              interval_suffix_length_good)
      show ?thesis using "191" "192" "193" by auto
     qed
 have 20: "powerinterval f \<sigma> (fuse l1 (map (shift cp) l2))" 
   using powerinterval_fuse[of "l1" "(l2)" "cp" "\<sigma>" "f"]  
   using "12" "13" "14" "3" "5" by auto
 have 21: "\<sigma> = fuse (prefix cp \<sigma>) (suffix cp \<sigma>)"
     by (simp add: "3" interval_fuse_prefix_suffix) 
 have 22: "nth ((fuse l1 (map (shift cp) l2))) (intlen l1) = cp"
     using "18" by blast
 have 23: " (prefix (intlen l1) (filt \<sigma> (fuse l1 (map (shift cp) l2)) )) =
            (filt (prefix cp \<sigma>) l1)"
    by (metis "17" "2" "5" filt_prefix filt_prefix_idx interval_fuse_intlen_a 
        interval_prefix_fuse interval_prefix_length_good le_add1)
 have 24: "g (prefix (intlen l1) (filt \<sigma> (fuse l1 (map (shift cp) l2)) ))"
   by (simp add: "23" "9")    
 have 25: "(suffix (intlen l1) (filt \<sigma> (fuse l1 (map (shift cp) l2)) )) =
           (filt (suffix cp \<sigma>) l2)" 
   by (metis "12" "13" "17" "2" "23" filt_intlen filt_suffix filt_suffix_idx 
       interval_pref_intlen_bound interval_suffix_fuse interval_suffix_length_good)
 have 26: "intlen l1 \<le> intlen (filt \<sigma> (fuse l1 (map (shift cp) l2)))"
         by (metis "23" filt_intlen interval_pref_intlen_bound)
 have 27: "(\<exists>l. index_sequence 0 l \<and>
         Interval.nth l (intlen l) = intlen \<sigma> \<and>
         powerinterval f \<sigma> l \<and>
         (\<exists>n\<le>intlen (filt \<sigma> l). g (prefix n (filt \<sigma> l)) \<and> h (suffix n (filt \<sigma> l))))"
    by (metis "11" "16" "19" "20" "24" "25" "26" interval_nth_intlen_intlast)      
 show ?thesis
 by (metis "27" interval_chop_fuse interval_fuse_prefix_suffix interval_intlast_intfirst 
     projection_d_def)
qed 

lemma PJ4sem:
"(\<sigma> \<Turnstile> f \<triangle> (g;h) = (f \<triangle> g) ; (f \<triangle> h))"
using PJ4semchaina PJ4semchainb unl_lift2 by blast



subsubsection \<open>PJ5\<close>

lemma PJ5sem:
 "(\<sigma> \<Turnstile> f \<triangle> init(g) \<longrightarrow> init(g))"
by (simp add: projection_d_def init_defs)
   ( metis filt_nth filt_intlen index_sequence_def interval_intlen_gr_zero )

subsubsection \<open>PJ6\<close>

lemma PJ6help1:
 assumes "index_sequence 0 l"
         " (nth l (intlen l)) = (intlen \<sigma>)"
 shows   "(\<forall> i. 0\<le> i \<and> i<intlen l \<longrightarrow> intlen (sub (nth l i) (nth l (Suc i)) \<sigma>)
          = (nth l (Suc i)) - (nth l i)) "
proof 
  fix i
  show "0 \<le> i \<and> i < intlen l \<longrightarrow>
         intlen (sub (nth l i) (nth l (Suc i)) \<sigma>) = nth l (Suc i) - nth l i"
  using assms 
  by (simp add: index_sequence_def sub_def) 
     (metis Suc_lessI assms(1) interval_idx_less_last_1 le_diff_iff le_eq_less_or_eq min.orderE)  
qed


lemma PJ6help2:
assumes " index_sequence 0 l"
        " nth l (intlen l) = intlen \<sigma> \<and>
         (\<forall>i<intlen l. nth l (Suc i) - nth l i = Suc 0) "
shows   " (\<forall>i\<le>intlen l. nth l i = i)"
proof 
 fix i
 show " i\<le>intlen l \<longrightarrow> nth l i = i"
 proof
  (induct i)
  case 0
  then show ?case using assms index_sequence_def by blast
  next
  case (Suc i)
  then show ?case 
  by (metis One_nat_def Suc_eq_plus1 Suc_leD Suc_le_lessD assms interval_idx_expand  
      le_add_diff_inverse2 plus_1_eq_Suc)
  qed
qed


lemma PJ6help3:
assumes " index_sequence 0 l"
        " nth l (intlen l) = intlen \<sigma>"
        " (\<forall>i\<le>intlen l. nth l i = i) "
shows   " (\<forall>i<intlen l. nth l (Suc i) - nth l i = Suc 0)
          "
proof
 fix i
 show "i<intlen l \<longrightarrow> nth l (Suc i) - nth l i = Suc 0"
 by (simp add: assms)
qed


lemma PJ6help4:
 " (\<exists>l. index_sequence 0 l \<and> l = [0..\<le>intlen \<sigma>] \<and>
         Interval.nth l (intlen l) = intlen \<sigma> \<and>
         (\<forall>i\<le>intlen l. nth l i = i) \<and> 
         (\<exists>s1. intlen s1 = intlen l \<and> (\<forall>i\<le>intlen s1. nth s1 i = nth \<sigma> (i)) \<and> s1 = \<sigma>))
 "
by (simp add: index_sequence_def upt_length upt_nth)



lemma PJ6help5:
 "(\<exists>l. index_sequence 0 l \<and>
         Interval.nth l (intlen l) = intlen \<sigma> \<and>
         (\<forall>i<intlen l. intlen (sub (nth l i) (nth l (Suc i)) \<sigma>) = Suc 0) \<and>
         (\<exists>s1. intlen s1 = intlen l \<and> (\<forall>i\<le>intlen s1. nth s1 i = nth \<sigma> (Interval.nth l i)) \<and> g s1))
   = g \<sigma> "
proof -
 have "(\<exists>l. index_sequence 0 l \<and>
         Interval.nth l (intlen l) = intlen \<sigma> \<and>
         (\<forall>i<intlen l. intlen (sub (nth l i) (nth l (Suc i)) \<sigma>) = Suc 0) \<and>
         (\<exists>s1. intlen s1 = intlen l \<and> (\<forall>i\<le>intlen s1. nth s1 i = nth \<sigma> (nth l i)) \<and> g s1))
       = 
       (\<exists>l. index_sequence 0 l \<and>
         Interval.nth l (intlen l) = intlen \<sigma> \<and>
         (\<forall>i<intlen l. nth l (Suc i) - nth l i = Suc 0) \<and>
         (\<exists>s1. intlen s1 = intlen l \<and> (\<forall>i\<le>intlen s1. nth s1 i = nth \<sigma> (Interval.nth l i)) \<and> g s1))"
     using PJ6help1 by (metis zero_order(1))
  also have " ... = 
            (\<exists>l. index_sequence 0 l \<and>
         Interval.nth l (intlen l) = intlen \<sigma> \<and>
         (\<forall>i\<le>intlen l. nth l i = i) \<and>
         (\<exists>s1. intlen s1 = intlen l \<and> (\<forall>i\<le>intlen s1. nth s1 i = nth \<sigma> (nth l i)) \<and> g s1))
            "
       using PJ6help2 PJ6help3 by blast
 also have " ... =
             (\<exists>l. index_sequence 0 l \<and>
         Interval.nth l (intlen l) = intlen \<sigma> \<and>
         (\<forall>i\<le>intlen l. nth l i = i) \<and>
         (\<exists>s1. intlen s1 = intlen l \<and> (\<forall>i\<le>intlen s1. nth s1 i = nth \<sigma> (i)) \<and> g s1))
           "
       by metis
 also have " ... =
           (\<exists>l. index_sequence 0 l \<and>
         Interval.nth l (intlen l) = intlen \<sigma> \<and>
         (\<forall>i\<le>intlen l. nth l i = i) \<and> 
         (\<exists>s1. intlen s1 = intlen l \<and> (\<forall>i\<le>intlen s1. nth s1 i = nth \<sigma> (i)) \<and> s1 = \<sigma> \<and> g \<sigma>))
           "
    by (metis interval_eq_nth_eq le_eq_less_or_eq) 
 also have " ... =
             g \<sigma>
           " 
   using PJ6help4 by blast
 finally show "(\<exists>l. index_sequence 0 l \<and>
         Interval.nth l (intlen l) = intlen \<sigma> \<and>
         (\<forall>i<intlen l. intlen (sub (nth l i) (nth l (Suc i)) \<sigma>) = Suc 0) \<and>
         (\<exists>s1. intlen s1 = intlen l \<and> (\<forall>i\<le>intlen s1. nth s1 i = nth \<sigma> (Interval.nth l i)) \<and> g s1))
   = g \<sigma>
              " .
qed

lemma PJ6sem:
"(\<sigma> \<Turnstile> skip \<triangle> g = g)"
proof -
 have 1: "(\<sigma> \<Turnstile> skip \<triangle> g = g) =
          ((\<exists>l. index_sequence 0 l \<and> nth l (intlen l) = intlen \<sigma> \<and>
         (\<forall>i<intlen l.  (sub (nth l i) (nth l (Suc i)) \<sigma>) \<Turnstile> skip) \<and> g (filt \<sigma> l)) =
         g \<sigma>)" 
       by (simp add: projection_d_def powerinterval_def)
 have 2: "(\<exists>l. index_sequence 0 l \<and> nth l (intlen l) = intlen \<sigma> \<and>
         (\<forall>i<intlen l.  (sub (nth l i) (nth l (Suc i)) \<sigma>) \<Turnstile> skip) \<and> g (filt \<sigma> l)) =
         (\<exists>l s1 . index_sequence 0 l \<and> nth l (intlen l) = intlen \<sigma> \<and>
         (\<forall>i<intlen l.  (sub (nth l i) (nth l (Suc i)) \<sigma>) \<Turnstile> skip) \<and> 
          intlen s1 = intlen l \<and>
          (\<forall> i\<le> intlen s1. (nth s1 i) = (nth \<sigma> (nth l i))) \<and>
         g s1)" 
     using filt_expand by metis
  have 3: "(\<exists>l s1 . index_sequence 0 l \<and> nth l (intlen l) = intlen \<sigma> \<and>
         (\<forall>i<intlen l.  (sub (nth l i) (nth l (Suc i)) \<sigma>) \<Turnstile> skip) \<and> 
          intlen s1 = intlen l \<and>
          (\<forall> i\<le> intlen s1. (nth s1 i) = (nth \<sigma> (nth l i))) \<and>
         g s1) =
         (\<exists>l. index_sequence 0 l \<and> nth l (intlen l) = intlen \<sigma> \<and>
        (\<forall>i<intlen l. intlen (sub (nth l i) (nth l (Suc i)) \<sigma>) = Suc 0) \<and>
        (\<exists>s1. intlen s1 = intlen l \<and> (\<forall>i\<le>intlen s1. nth s1 i = nth \<sigma> (nth l i)) \<and> g s1)) "
   by (simp add: skip_defs)
  have 4: "(\<exists>l. index_sequence 0 l \<and> nth l (intlen l) = intlen \<sigma> \<and>
        (\<forall>i<intlen l. intlen (sub (nth l i) (nth l (Suc i)) \<sigma>) = Suc 0) \<and>
        (\<exists>s1. intlen s1 = intlen l \<and> (\<forall>i\<le>intlen s1. nth s1 i = nth \<sigma> (nth l i)) \<and> g s1)) =
        (g \<sigma>) "
    by (simp add: PJ6help5)
  from 1 2 3 4 show ?thesis 
  by simp
qed


subsubsection \<open>PJ7\<close>


lemma PJemptyImp:
assumes " intlen \<sigma> = 0"
 shows  "(\<sigma> \<Turnstile> (f \<triangle> g) = g)"  
using assms 
by (simp add: projection_d_def index_sequence_def powerinterval_def,
    auto, 
    metis filt.simps(1) interval_st_intlen lessI not_less_zero old.nat.exhaust,
    metis Interval.nth.simps(1) filt.simps(1) interval_suffix_intlast interval_suffix_zero 
    intlen.simps(1) not_less_zero)

lemma PJ7empty:
 assumes " intlen \<sigma> = 0"
 shows   "(\<sigma> \<Turnstile> f \<triangle> (g \<triangle> h) = (f \<triangle> g) \<triangle> h)"
proof -
 have 1: "(\<sigma> \<Turnstile> f \<triangle> (g \<triangle> h) = (g \<triangle> h)) "
   using PJemptyImp assms by blast
 have 2: "(\<sigma> \<Turnstile> (g \<triangle> h) = h)"
   using PJemptyImp assms by blast
 have 3: "(\<sigma> \<Turnstile> (f \<triangle> g) \<triangle> h = h) "
   using PJemptyImp assms by blast
 from 1 2 3 show ?thesis by simp
qed

lemma PJ7helpchain1a_help_1:
 assumes "index_sequence 0 l"
         "(nth l (intlen l)) = intlen \<sigma> "  
         " (\<forall> i<intlen l. (sub (nth l i) (nth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)"
         " intlen \<sigma> > 0 " 
 shows   " index_sequence 0  (lfuse ( (lcpl f g \<sigma> l)))"
proof -
 have 0:  "intlen \<sigma> > 0 \<longrightarrow> intlen l >0"
    using assms gr_zeroI index_sequence_def by fastforce
 have 01: "intfirst l = 0"
   using assms index_sequence_def by auto
 have 02: "lastfirst (lcpl f g \<sigma> l)"
    using assms lcpl_lfuse_lastfirst by (simp add: projection_d_def) 
     (metis "0" "01" assms(3) interval_nth_zero_intfirst lcpl_lfuse_lastfirst)
 have 1: "intlen \<sigma> > 0 \<longrightarrow> intfirst (lfuse  (lcpl f g \<sigma> l)) = 0 "
   using assms 0  01  02 
     lcpl_intfirst[of "l" "\<sigma>" "f" "g" ]  
    by (metis (mono_tags, lifting)  interval_lastfirst_lfuse interval_nth_zero_intfirst  )
 from 1 0 show ?thesis using assms lcpl_lfuse_idx[of "l" "\<sigma>" "f" "g"] by(simp add: projection_d_def)  
qed

lemma PJ7helpchain1a_help_2:
 assumes "index_sequence 0 l"
         " (nth l (intlen l)) = intlen \<sigma> "
         " (\<forall> i<intlen l. (sub (nth l i) (nth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)"
         " intlen \<sigma> > 0 " 
 shows   " powerinterval f \<sigma> (lfuse ( (lcpl f g \<sigma> l)))"
proof -
 have 1: "(\<forall> i<intlen l. 
             powerinterval f (sub (nth l i) (nth l (Suc i)) \<sigma>) 
                             (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)))"
   using assms cpl_projection by blast 
 have 2: "\<forall>i<intlen l.
           \<forall>ia<intlen (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)).
            f (sub (nth (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)) ia)
                   (nth (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)) (Suc ia))
                   (sub (nth l i) (nth l (Suc i)) \<sigma>))" 
   using "1" by (simp add: powerinterval_def)
 have 3: "\<forall>i<intlen l.
           \<forall>ia<intlen (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)).
              (sub (nth (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)) ia)
                   (nth (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)) (Suc ia))
                   (sub (nth l i) (nth l (Suc i)) \<sigma>)) =
            (sub (nth (map (shift (nth l i)) (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>))) ia)
                  (nth (map (shift (nth l i)) (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>))) (Suc ia))
                  \<sigma>)"
   proof -
    have 30: "\<forall>i<intlen l.
               \<forall>ia<intlen (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)).
                \<forall>j\<le>intlen (sub (nth (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)) ia)
                   (nth (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)) (Suc ia))
                   (sub (nth l i) (nth l (Suc i)) \<sigma>)).
                  (nth (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)) (Suc ia)) \<le> 
                  (nth l (Suc i)) - (nth l (i)) "
         using assms by (metis add.commute cpl_projection interval_idx_expand interval_intlen_sub
              plus_1_eq_Suc)
    have 31: "\<forall>i<intlen l.
               \<forall>ia<intlen (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)).
                intlen (sub (nth (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)) ia)
                            (nth (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)) (Suc ia))
                            (sub (nth l i) (nth l (Suc i)) \<sigma>)) =
                intlen (sub (nth (map (shift (nth l i)) 
                                 (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>))) ia)
                            (nth (map (shift (nth l i)) 
                                 (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>))) (Suc ia))
                  \<sigma>)"
         using assms by (auto simp add: shift_def interval_nth_map ) 
          (metis (no_types, lifting) PJ6help1 add.commute cpl_projection interval_idx_expand 
             interval_sub_sub_1 le_add1 le_add_same_cancel1 plus_1_eq_Suc)
    have 32: "\<forall>i<intlen l.
               \<forall>ia<intlen (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)).
                \<forall>j\<le>intlen (sub (nth (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)) ia)
                   (nth (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)) (Suc ia))
                   (sub (nth l i) (nth l (Suc i)) \<sigma>)).
                  nth (sub (nth (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)) ia)
                           (nth (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)) (Suc ia))
                           (sub (nth l i) (nth l (Suc i)) \<sigma>)) j =
                  nth (sub (nth (map (shift (nth l i)) 
                                (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>))) ia)
                           (nth (map (shift (nth l i)) 
                                (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>))) (Suc ia))
                  \<sigma>) j"
              using assms by (simp add: shift_def interval_nth_map cpl_projection  )   
                 (metis "30" add.commute interval_idx_expand interval_sub_sub_1 plus_1_eq_Suc) 
    show ?thesis using "31" "32" interval_eq_nth_eq by (simp add: interval_eq_nth_eq)
   qed
  have 4: "\<forall>i<intlen l.
           \<forall>ia<intlen (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>)).
            f (sub (nth (map (shift (nth l i)) (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>))) ia)
                  (nth (map (shift (nth l i)) (cpl f g (sub (nth l i) (nth l (Suc i)) \<sigma>))) (Suc ia))
                  \<sigma>)" 
       using "2" "3" by auto
  have 5: "intlen(lcpl f g \<sigma> l) = intlen l -1"
      using assms index_sequence_def lcpl_intlen lcpl_intlen_zero by fastforce
  have 6: "intlen \<sigma> >0 \<longrightarrow> intlen l >0" 
      using assms gr_zeroI index_sequence_def by fastforce
  have 7: "\<forall>i< intlen l.
            \<forall>ia<intlen ((nth (lcpl f g \<sigma> l) i)).
             f (sub (nth (nth (lcpl f g \<sigma> l) i) ia)
                    (nth (nth (lcpl f g \<sigma> l) i) (Suc ia))
                    \<sigma>)" 
      using assms by (metis (no_types, lifting) "4" index_sequence_def interval_intlen_map lcpl_nth)
  have 8: "intlen \<sigma> >0 \<longrightarrow>
            (\<forall>i\<le> intlen(lcpl f g \<sigma> l).
              \<forall>ia<intlen ((nth (lcpl f g \<sigma> l) i)).
               f (sub (nth (nth (lcpl f g \<sigma> l) i) ia)
                      (nth (nth (lcpl f g \<sigma> l) i) (Suc ia))
                      \<sigma>))" 
     by (metis "5" "6" "7" One_nat_def Suc_pred le_imp_less_Suc)
  have 9: "intlen \<sigma> >0 \<longrightarrow> 
            powerinterval f \<sigma> (lfuse ( (lcpl f g \<sigma> l))) =
            (\<forall>i<intlen (lfuse (lcpl f g \<sigma> l)).
              f (sub (nth (lfuse (lcpl f g \<sigma> l)) i) (nth (lfuse (lcpl f g \<sigma> l)) (Suc i)) \<sigma>))"
   by (simp add: powerinterval_def)
  have 10: "intlen \<sigma> >0 \<longrightarrow> lastfirst (lcpl f g \<sigma> l)"
    using "6" assms index_sequence_def lcpl_lfuse_lastfirst by blast
  have 11: "intlen \<sigma> >0 \<longrightarrow>
             (\<forall> j \<le> intlen (lcpl f g \<sigma> l). intlen(nth (lcpl f g \<sigma> l) j) > 0)
             " 
      by (simp add: assms lcpl_intlen_nth_gr_zero)
  have 12: "intlen \<sigma> >0 \<longrightarrow>
             (\<forall>i\<le> intlen(lcpl f g \<sigma> l).
               (\<forall>ia<intlen ((nth (lcpl f g \<sigma> l) i)).
                 f (sub (nth (nth (lcpl f g \<sigma> l) i) ia)
                        (nth (nth (lcpl f g \<sigma> l) i) (Suc ia))
                        \<sigma>))) =
                  (\<forall>j<intlen (lfuse (lcpl f g \<sigma> l)).
                 f (sub (nth (lfuse (lcpl f g \<sigma> l)) j) 
                        (nth (lfuse (lcpl f g \<sigma> l)) (Suc j)) 
                        \<sigma>))
              " 
     using interval_lfuse_split[of "(lcpl f g \<sigma> l)" "f" "\<sigma>"] "10" "11" by auto
   show ?thesis using "12" "8" "9" using assms(4) by blast
qed    

lemma PJ7helpchain1a_help_3:
 assumes "index_sequence 0 l"
         " (nth l (intlen l)) = intlen \<sigma>" 
         " (\<forall> i<intlen l. (sub (nth l i) (nth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g) " 
         " h (filt \<sigma> l) "
         " intlen \<sigma> >0 " 
 shows   " intlast (lfuse (lcpl f g \<sigma> l)) = intlen \<sigma>"
proof -
 have 0: "intlen \<sigma> > 0 \<longrightarrow> intlen l >0"
    using assms gr_zeroI index_sequence_def by fastforce
 have 1: "intlen \<sigma> >0 \<longrightarrow>
            intlast (lfuse (lcpl f g \<sigma> l)) = intlast(intlast ( (lcpl f g \<sigma> l)))" 
      using assms 
      using "0" index_sequence_def interval_lastfirst_lfuse_intlast lcpl_lfuse_lastfirst by blast
 have 2: "intlen \<sigma> >0 \<longrightarrow>
         intlast ( (lcpl f g \<sigma> l))
          = (nth (lcpl f g \<sigma> l) (intlen l -1)) " 
       using assms by (simp add: "0" index_sequence_def lcpl_intlen)
 have 3: "intlen \<sigma> >0 \<longrightarrow>
          (nth (lcpl f g \<sigma> l) (intlen l -1)) =
           (map (shift (nth l (intlen l -1))) 
                (cpl f g (sub (nth l (intlen l -1)) (nth l (Suc (intlen l -1))) \<sigma>)))
         " 
       using assms by (simp add: "0" index_sequence_def lcpl_nth)
 have 4: "intlen \<sigma> >0 \<longrightarrow>
          intlast( ( (cpl f g (sub (nth l (intlen l -1)) (nth l (Suc (intlen l -1))) \<sigma>)))) =
           intlen ((sub (nth l (intlen l -1)) (nth l (Suc (intlen l -1))) \<sigma>))
          " using "0" assms 
          by (metis cpl_projection diff_less interval_nth_intlen_intlast zero_less_one)         
 have 5: "intlen \<sigma> >0 \<longrightarrow>
         intlen ((sub (nth l (intlen l -1)) (nth l (Suc (intlen l -1))) \<sigma>)) =
            (nth l (Suc (intlen l -1))) - (nth l (intlen l -1))     
        "  
     using "0" PJ6help1 assms diff_less zero_less_one by blast 
 have 6: "intlen \<sigma> >0 \<longrightarrow>
          intlast ( (map (shift (nth l (intlen l -1))) 
                         (cpl f g (sub (nth l (intlen l -1)) (nth l (Suc (intlen l -1))) \<sigma>)))) =
             (shift (nth l (intlen l -1))) 
              (intlast (cpl f g (sub (nth l (intlen l -1)) (nth l (Suc (intlen l -1))) \<sigma>)))
         "      
   by (metis interval_intlen_map interval_nth_intlen_intlast interval_nth_map)
 have 7: "intlen \<sigma> >0 \<longrightarrow> 
          (shift (nth l (intlen l -1))) 
           (intlast (cpl f g (sub (nth l (intlen l -1)) (nth l (Suc (intlen l -1))) \<sigma>))) =
          (shift (nth l (intlen l -1))) ((nth l (Suc (intlen l -1))) - (nth l (intlen l -1)))
          " 
      using "4" "5" by auto
 have 8: "intlen \<sigma> >0 \<longrightarrow> 
          (shift (nth l (intlen l -1))) ((nth l (Suc (intlen l -1))) - (nth l (intlen l -1))) =
           ((nth l (Suc (intlen l -1))) - (nth l (intlen l -1))) + (nth l (intlen l -1))
         " 
    by (simp add: Interval.shift_def)
 have 9: "intlen \<sigma> >0 \<longrightarrow>  
          ((nth l (Suc (intlen l -1))) - (nth l (intlen l -1))) + (nth l (intlen l -1)) 
          = (nth l (Suc (intlen l -1)))
         " 
    using assms 
    by (metis (no_types, lifting) "0" "2" "3" "6" "7" "8" Suc_pred' index_sequence_def 
        lcpl_last_last lessI)
 have 10: "intlen \<sigma> >0 \<longrightarrow> (nth l (Suc (intlen l -1))) = intlen \<sigma> " 
     by (simp add: "0" assms)
  show ?thesis 
  using "1" "10" "2" "3" "6" "7" "8" "9" assms(5) by presburger
qed

lemma PJ7helpchain1a_help_4:
 assumes "index_sequence 0 l"
         " (nth l (intlen l)) = intlen \<sigma>" 
         " (\<forall> i<intlen l. (sub (nth l i) (nth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g) " 
         "  h (filt \<sigma> l) "
         " intlen \<sigma> >0 "
 shows   " ((filt \<sigma> (lfuse (lcpl f g \<sigma> l))) \<Turnstile> g \<triangle> h) "
proof -
 have 0: "intlen \<sigma> > 0 \<longrightarrow> intlen l >0"
    using assms gr_zeroI index_sequence_def by fastforce
 have 1: "intlen \<sigma> >0 \<longrightarrow> index_sequence 0 (addzero (lsum (lcpl f g \<sigma> l) 0)) "
    by (simp add: assms lcpl_intlen_nth_gr_zero lsum_addzero_idx)
 have 2: "intlen \<sigma> > 0 \<longrightarrow> 
          (nth (addzero (lsum (lcpl f g \<sigma> l) 0)) (intlen(addzero (lsum (lcpl f g \<sigma> l) 0)))) =
          intlen (filt \<sigma> (lfuse (lcpl f g \<sigma> l)))"
  by (simp add: assms lcpl_lfuse_filt_intlen)
 have 3: "intlen \<sigma> >0 \<longrightarrow>  
          powerinterval g (filt \<sigma> (lfuse (lcpl f g \<sigma> l))) (addzero (lsum (lcpl f g \<sigma> l) 0))
              " 
    by (simp add: assms lcpl_lfuse_filt_power)
 have 4: "intlen \<sigma> >0 \<longrightarrow>
           h (filt (filt \<sigma> (lfuse (lcpl f g \<sigma> l))) (addzero (lsum (lcpl f g \<sigma> l) 0))) "
   by (simp add: assms filt_map_filt lcpl_filt_lfuse_lsum)
show ?thesis by (metis "1" "2" "3" "4" assms(5) projection_d_def)
qed


lemma PJ7helpchain1a:
assumes "intlen \<sigma> > 0"
        "(\<sigma> \<Turnstile> (f \<triangle> g) \<triangle> h)" 
shows " (\<sigma> \<Turnstile> f \<triangle> (g \<triangle> h))"
proof -
 have 1: "intlen \<sigma> > 0"
  using assms by auto
 have 2: "(\<exists>l. index_sequence 0 l \<and> nth l (intlen l) = intlen \<sigma> \<and>
       powerinterval (LIFT(f \<triangle> g)) \<sigma> l \<and>
         h (filt \<sigma> l)) "  
 using assms using cpl_projection by blast
 obtain l where 3: "index_sequence 0 l \<and> nth l (intlen l) = intlen \<sigma> \<and>
       powerinterval (LIFT(f \<triangle> g)) \<sigma> l \<and>
         h (filt \<sigma> l)"
  using "2" by blast
 have 4: "index_sequence 0 l \<and> nth l (intlen l) = intlen \<sigma> \<and>
            (\<forall> i <intlen l. (sub (nth l i) (nth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g) \<and>
         h (filt \<sigma> l)" 
 using "3"  by (simp add: powerinterval_def) 
 have 6: "intlen l>0"
    using "1" "4" gr0I index_sequence_def by force 
 have 7: "index_sequence 0  (lfuse (lcpl f g \<sigma> l))" 
 by (metis (no_types, lifting) "1" "4" PJ7helpchain1a_help_1)
 have 8: "powerinterval f \<sigma> (lfuse (lcpl f g \<sigma> l))"   
  using "4" "6" PJ7helpchain1a_help_2 assms index_sequence_def by auto
 have 9: "(nth (lfuse (lcpl f g \<sigma> l)) (intlen (lfuse (lcpl f g \<sigma> l)))) = intlen \<sigma> " 
  by (metis "1" "4" interval_nth_intlen_intlast PJ7helpchain1a_help_3)
 have 10: "((filt \<sigma> (lfuse (lcpl f g \<sigma> l))) \<Turnstile> g \<triangle> h) " 
   by (simp add: "1" "4" "6" PJ7helpchain1a_help_4) 
show ?thesis 
using "10" "7" "8" "9" by (metis projection_d_def)
qed
       

lemma PJ7helpchain1b:
assumes "intlen \<sigma> > 0"
        "(\<sigma> \<Turnstile> f \<triangle> (g \<triangle> h))" 
shows " (\<sigma> \<Turnstile> (f \<triangle> g) \<triangle> h)"
proof -
 have 1: "intlen \<sigma> > 0" 
  using assms by auto
 have 2: "(\<exists>l. index_sequence 0 l \<and> nth l (intlen l) = intlen \<sigma> \<and>
         powerinterval f \<sigma> l \<and>
         ((filt \<sigma> l) \<Turnstile> g \<triangle> h) ) " 
  using assms by (simp add: projection_d_def)
 obtain l where 3: "index_sequence 0 l \<and> nth l (intlen l) = intlen \<sigma> \<and>
         powerinterval f \<sigma> l \<and>
         ((filt \<sigma> l) \<Turnstile> g \<triangle> h)" 
  using "2" by blast
 have 4: "(\<exists> la. index_sequence 0 la \<and> nth la (intlen la) = intlen(filt \<sigma> l) \<and>
           powerinterval g (filt \<sigma> l) la \<and>
           ((filt (filt \<sigma> l) la) \<Turnstile> h))" 
  using "3" using cpl_projection by blast
 obtain la where 5: "index_sequence 0 la \<and> nth la (intlen la) = intlen(filt \<sigma> l) \<and>
           powerinterval g (filt \<sigma> l) la \<and>
           ((filt (filt \<sigma> l) la) \<Turnstile> h) " 
  using "4" by blast
 have 6: "intlen l >0 "
   using "1" "3" gr0I index_sequence_def by force
 have 7: "intlen(filt \<sigma> l) = intlen l "
    by (simp add: filt_intlen)
 have 8: "(filt (filt \<sigma> l) la) = (filt \<sigma> (filt l la)) " 
   using filt_map_filt by blast
 have 9: "(nth (filt l la) (intlen (filt l la))) = intlen \<sigma> "
  by (metis "3" "5" filt_expand order_refl)
 have 10: "intlen la >0"
   using "5" "6" "7" gr0I index_sequence_def by force
 have 11: "intlen (filt l la) > 0  "
    by (simp add: "10" filt_intlen)
 have 12: "index_sequence 0 (filt l la) "
    proof - 
     have 111: "nth (filt l la) 0 = 0"
       by (metis "3" "5" filt_nth index_sequence_def interval_intlen_gr_zero)
     have 112: "intlen(filt l la) = intlen la"
        using filt_expand by blast
     have 113: "(\<forall> i< intlen la. (nth (filt l la) i) = (nth l (nth la i))) "
      by (simp add: filt_map interval_nth_map)
     have 114: "(\<forall> i< intlen la. (nth (filt l la) (Suc i)) = (nth l (nth la (Suc i)))) "
       by (simp add: filt_map interval_nth_map)
     have 115: "(\<forall> i< intlen la. (nth l (nth la i)) < (nth l (nth la (Suc i))) ) "
       by (metis "3" "5" "7" Suc_lessI interval_idx_less_than interval_idx_less_last_1 lessI 
           less_imp_le_nat)
     show ?thesis by (simp add: "111" "113" "114" "115" filt_intlen index_sequence_def)
    qed
  have 20: "powerinterval (LIFT(f\<triangle>g)) \<sigma> (filt l la)"
    proof -
     have 201: "powerinterval (LIFT(f\<triangle>g)) \<sigma> (filt l la) =
               (\<forall> i< intlen la. (sub (nth l (nth la i)) (nth l (nth la (Suc i))) \<sigma>) \<Turnstile> f \<triangle> g)  "
       by (simp add: filt_map interval_nth_map powerinterval_def)
     have 202: "(\<forall> i< intlen la. (sub (nth l (nth la i)) (nth l (nth la (Suc i))) \<sigma>) \<Turnstile> f \<triangle> g) =
                (\<forall> i< intlen la. 
                  (\<exists> ll. index_sequence 0 ll 
                  \<and> intlast ll = intlen (sub (nth l (nth la i)) (nth l (nth la (Suc i))) \<sigma>) \<and>
                    powerinterval f (sub (nth l (nth la i)) (nth l (nth la (Suc i))) \<sigma>) ll \<and>
                    ((filt (sub (nth l (nth la i)) (nth l (nth la (Suc i))) \<sigma>) ll) \<Turnstile> g)   
                     ))
                "
       by (simp add: projection_d_def)
      have 203: "(\<forall> i< intlen la.  intlen (sub (nth l (nth la i)) (nth l (nth la (Suc i))) \<sigma>) =
                    (nth l (nth la (Suc i))) - (nth l (nth la i)))  "
           by (metis "3" "5" "7" Suc_leI Suc_lessI interval_idx_less_equal interval_idx_less_last_1 
               interval_intlen_sub lessI less_imp_le_nat)
      have 2041: "(\<forall> i< intlen la.  
                      (nth la (Suc i)) \<le> intlen l 
                )
                "
       using "5" "7" Suc_lessI interval_idx_less_last_1 by fastforce 
      have 204: "(\<forall> i< intlen la.  
                   intlen (map (shiftm (nth l (nth la i))) (sub (nth la i) (nth la (Suc i)) l)) =
                     (nth la (Suc i)) - (nth la i))
                " 
                by (simp add: "5" PJ6help1 filt_intlen)
      have 205: "(\<forall> i< intlen la.  
                  (\<forall> j \<le> (nth la (Suc i)) - (nth la i). 
                     (nth (sub (nth la i) (nth la (Suc i)) l) j) =
                      (nth l ((nth la i) +j))
                  )
                )           
                " 
        using "2041" "5" index_sequence_def interval_nth_sub order.strict_implies_order by blast
      have 2060: "(\<forall> i< intlen la.  
                    (nth la i) \<le> (nth la (Suc i)) )
                 " 
           using "5" interval_idx_expand by fastforce
      have 206: "(\<forall> i< intlen la.  
                   (\<forall> j \<le> (nth la (Suc i)) - (nth la i).
                      (nth l (nth la i)) \<le> (nth l ((nth la i) +j))
                   )
                 )
                "  
             using "2041" "3" "2060" interval_idx_less_equal by fastforce
     have 207: " (\<forall> i< intlen la.  
                   (\<forall> j \<le> (nth la (Suc i)) - (nth la i).
                    (nth (map (shiftm (nth l (nth la i))) (sub (nth la i) (nth la (Suc i)) l)) j ) =
                      (nth l ((nth la i) +j)) - (nth l (nth la i))
                    ))
                " 
                by (simp add: "205" interval_nth_map shiftm_def)
     have 208: "(\<forall> i< intlen la.  
                   (nth (map (shiftm (nth l (nth la i))) (sub (nth la i) (nth la (Suc i)) l)) 0) =0 
              ) "
              by (simp add: "207")
     have 209: "(\<forall> i< intlen la.  
                   intlast (map (shiftm (nth l (nth la i))) (sub (nth la i) (nth la (Suc i)) l)) =
                   (nth l (nth la (Suc i))) - (nth l (nth la i))
                ) 
               "   
             using "204" "207" "5" "2060" by simp_all        
     have 210: "(\<forall> i< intlen la. 
              index_sequence 0 (map (shiftm (nth l (nth la i))) (sub (nth la i) (nth la (Suc i)) l))
                ) 
               " 
               by (metis "3" "5" "7" add.commute interval_idx_expand interval_idx_shiftm 
                   interval_idx_sub plus_1_eq_Suc)
     have 211: "(\<forall> i< intlen la. 
                 powerinterval f (sub (nth l (nth la i)) (nth l (nth la (Suc i))) \<sigma>) 
                       (map (shiftm (nth l (nth la i))) (sub (nth la i) (nth la (Suc i)) l))
                ) 
              " 
       proof -
        have 2111: "(\<forall> i< intlen la. 
                     powerinterval f (sub (nth l (nth la i)) (nth l (nth la (Suc i))) \<sigma>) 
                                     (map (shiftm (nth l (nth la i))) 
                                     (sub (nth la i) (nth la (Suc i)) l))
                    )  = 
                    (\<forall>i<intlen la.
                     (\<forall>ia<intlen (sub (nth la i) (nth la (Suc i)) l).
                       f (sub (nth (map (shiftm (nth l (nth la i)))
                                        (sub (nth la i) (nth la (Suc i)) l)) ia)
                              (nth (map (shiftm (nth l (Interval.nth la i))) 
                                        (sub (nth la i) (nth la (Suc i)) l)) (Suc ia))
                              (sub (nth l (nth la i)) (nth l (nth la (Suc i))) \<sigma>) ) 
                     ) 
                     ) " 
        by (simp add: powerinterval_def)
       have 2112: " ... = 
                    (\<forall>i<intlen la.
                     (\<forall>ia<  (nth la (Suc i)) - (nth la i).
                       f (sub (nth (map (shiftm (nth l (nth la i))) 
                                        (sub (nth la i) (nth la (Suc i)) l)) ia)
                              (nth (map (shiftm (nth l (Interval.nth la i))) 
                                        (sub (nth la i) (nth la (Suc i)) l)) (Suc ia))
                              (sub (nth l (nth la i)) (nth l (nth la (Suc i))) \<sigma>))
                     )
                    )"
               using "204" by auto
       have 2113: "... = 
                    (\<forall>i<intlen la.
                     (\<forall>ia<  (nth la (Suc i)) - (nth la i).
                        f (sub  ((nth l ((nth la i) +ia)) - (nth l (nth la i)))   
                                ((nth l ((nth la i) +(Suc ia))) - (nth l (nth la i)))   
                                (sub (nth l (nth la i)) (nth l (nth la (Suc i))) \<sigma>))
                     )
                    )" 
               using "207" by auto
       have 2114: "
                    (\<forall>i<intlen la.
                      (let m = (nth la (Suc i)); n= (nth la i) in
                       (\<forall>ia<  m - n. 
                         (nth l (n +ia))  \<le> ((nth l (n+ (Suc ia))) ) 
                       )
                      )
                     )" 
                using "2041" "3" "2060" interval_idx_less_equal by simp_all fastforce
         have 2115: "
                    (\<forall>i<intlen la.
                      (let m = (nth la (Suc i)); n= (nth la i) in
                       (\<forall>ia<  m - n. 
                         (nth l (n +ia)) - (nth l n) \<le> ((nth l ((Suc (n+ia)))) - (nth l n)) 
                       )
                      )
                     ) "
           by (metis "2114" add_Suc_right diff_le_mono)
        have 2116: "
                    (\<forall>i<intlen la.
                      (let m = (nth la (Suc i)); n= (nth la i) in
                       (\<forall>ia<  m - n. 
                        ((nth l ((Suc (n+ia)))) - (nth l n))  \<le> (nth l m) - (nth l n) 
                       )
                      )
                     ) "
            by (metis "3" "5" "7" Interval.interval_idx_expand Suc_leI add.commute diff_le_mono 
            interval_idx_less_equal less_diff_conv plus_1_eq_Suc)  
      have 2117: "(\<forall>i<intlen la.
                      (let m = (nth la (Suc i)); n= (nth la i) in 
                        (nth l m) \<le> intlen \<sigma> )) "
         by (metis "12" "9" Suc_lessI filt_nth filt_intlen interval_idx_less_last_1  
                   less_or_eq_imp_le)
      have 2118: "(\<forall>i<intlen la.
                    (let m = (nth la (Suc i)); n= (nth la i) in
                     (\<forall>ia<  m - n.
                        (n+(Suc ia)) \<le> intlen l )
                     ) 
                  ) "
              using "5" "7" Interval.interval_idx_expand less_diff_conv by fastforce
      have 21190: "(\<forall>i<intlen la.
                    (let m = (nth la (Suc i)); n= (nth la i) in
                     (\<forall>ia<  m - n. 
                       ((nth l (n +ia)) - (nth l n)) \<le> ((nth l (n+(Suc ia))) - (nth l n))
                     ) ))
                   " 
            by (meson "2114" diff_le_mono)
      have 21191: "(\<forall>i<intlen la.
                    (let m = (nth la (Suc i)); n= (nth la i) in
                     (\<forall>ia<  m - n. 
                      ((nth l (n+(Suc ia))) - (nth l n)) \<le>
                       intlen (sub (nth l n) (nth l m) \<sigma>)
                      )))
                  " 
          by (simp add: "203" "2116")
      have 2119: "(\<forall>i<intlen la.
                    (let m = (nth la (Suc i)); n= (nth la i) in
                     (\<forall>ia<  m - n. 
                        intlen (sub  ((nth l (n +ia)) - (nth l n))   
                                     ((nth l (n+(Suc ia))) - (nth l n))   
                                     (sub (nth l n) (nth l m) \<sigma>)) =
                        ((nth l (n+(Suc ia)))) - ((nth l (n +ia))) 
                     )
                    )
                   )
                   "  
            by (metis "203" "206" "2115" "2116"  Nat.diff_diff_eq Suc_leI add_Suc_right 
                interval_intlen_sub less_imp_le_nat)
      have 2120: "(\<forall>i<intlen la.
                    (let m = (nth la (Suc i)); n= (nth la i) in
                     (\<forall>ia<  m - n.
                      (\<forall>j \<le> ((nth l (n+(Suc ia)))) - ((nth l (n +ia))).
                         nth  (sub  ((nth l (n +ia)) - (nth l n))   
                                    ((nth l (n+(Suc ia))) - (nth l n))   
                                    (sub (nth l n) (nth l m) \<sigma>)) j =
                         nth (sub (nth l n) (nth l m) \<sigma>) (((nth l (n +ia)) - (nth l n))+j) 
               )   )))  " 
               by (metis "2119" "21190" "21191" interval_intlen_sub interval_nth_sub)
      have 212111: "(\<forall>i<intlen la.
                     (let m = (nth la (Suc i)); n= (nth la i) in
                      (\<forall>ia<  m - n.
                           ((nth l (n +ia))) \<le> ((nth l (n+(Suc ia))))
                      )
                     )
                    )
                 "
               using "2114" by blast
      have 212112: "(\<forall>i<intlen la.
                     (let m = (nth la (Suc i)); n= (nth la i) in
                      (\<forall>ia<  m - n.
                           ((nth l (n))) \<le> ((nth l (n+(ia))))
                      )
                     )
                    )" 
          by (simp add: "206")
      have 212113: "(\<forall>i<intlen la.
                     (let m = (nth la (Suc i)); n= (nth la i) in
                      (\<forall>ia<  m - n.
                       (\<forall> j \<le> ((nth l (n+(Suc ia)))) - ((nth l (n +ia))). 
                         ((nth l (n +ia)) )+j \<le> (nth l m) 
                       )
                      )
                     )
                    )" 
           using "212111"  "212112"  "206" "2060" "2116"  
           unfolding Let_def by fastforce
      have 21211: "(\<forall>i<intlen la.
                     (let m = (nth la (Suc i)); n= (nth la i) in
                      (\<forall>ia<  m - n.
                       (\<forall> j \<le> ((nth l (n+(Suc ia)))) - ((nth l (n +ia))). 
                         ((nth l (n +ia)) - (nth l n))+j \<le> (nth l m) - (nth l n)
                       )
                      )
                     )
                    )
                 " 
       by (metis "212112" "212113" Nat.add_diff_assoc2 diff_le_mono)
      have 2121: "(\<forall>i<intlen la.
                    (let m = (nth la (Suc i)); n= (nth la i) in
                     (\<forall>ia<  m - n.
                      (\<forall> j \<le> ((nth l (n+(Suc ia)))) - ((nth l (n +ia))). 
                        (nth (sub (nth l n) (nth l m) \<sigma>) (((nth l (n +ia)) - (nth l n))+j) ) =
                        (nth \<sigma> ((nth l n) + (((nth l (n +ia)) - (nth l n))+j))   ) 
                      ) 
                     ) 
                    ) 
                   ) 
                 " 
              by (metis "206" "2117" "21211" add_diff_inverse_nat interval_nth_sub nat_diff_split 
                  not_less_zero order_refl)
      have 2122: "(\<forall>i<intlen la.
                    (let m = (nth la (Suc i)); n= (nth la i) in
                     (\<forall>ia<  m - n.
                      (\<forall> j \<le> ((nth l (n+(Suc ia)))) - ((nth l (n +ia))). 
                       (nth \<sigma> ((nth l n) + (((nth l (n +ia)) - (nth l n))+j))   ) =
                       (nth \<sigma> (  (((nth l (n +ia)))+j))   ) 
                      ) 
                     ) 
                    ) 
                   )   " 
            by (metis "206" add.assoc le_add_diff_inverse less_imp_le_nat)
      have 2123:  "(\<forall>i<intlen la.
                     (let m = (nth la (Suc i)); n= (nth la i) in
                      (\<forall>ia<  m - n.
                       (\<forall> j \<le> ((nth l (n+(Suc ia)))) - ((nth l (n +ia))).
                         nth (sub (nth l (n +ia)) ((nth l (n+(Suc ia)))) \<sigma>) j =
                         (nth \<sigma> (  (((nth l (n +ia)))+j)) )
                       )
                      )
                     )
                    )
                  " 
                  by (metis "2114" "2118" "3" eq_imp_le interval_idx_less_equal interval_nth_sub)
      have 2124: "(\<forall>i<intlen la.
                    (let m = (nth la (Suc i)); n= (nth la i) in
                     (\<forall>ia<  m - n. 
                       intlen (sub (nth l (n +ia)) ((nth l (n+(Suc ia)))) \<sigma>) =
                       ((nth l (n+(Suc ia)))) - ((nth l (n +ia)))
                     )
                    )
                   )
                 " 
              by (metis "2118" "3" Suc_eq_plus1 Suc_le_lessD add_Suc_right interval_idx_expand 
                 interval_intlen_sub)
    have 2125: "(\<forall>i<intlen la.
                  (let m = (nth la (Suc i)); n= (nth la i) in
                    (\<forall>ia<  m - n.
                      (\<forall> j \<le> ((nth l (n+(Suc ia)))) - ((nth l (n +ia))).
                         nth  (sub  ((nth l (n +ia)) - (nth l n))   
                                    ((nth l (n+(Suc ia))) - (nth l n))   
                                    (sub (nth l n) (nth l m) \<sigma>)) j =
                        (nth (sub (nth l (n +ia)) ((nth l (n+(Suc ia)))) \<sigma>) j) 
                      )
                     )
                    ) 
                   )
                  " 
     by (metis "2120" "2121" "2122" "2123")  
    have 2126: "(\<forall>i<intlen la.
                   (let m = (nth la (Suc i)); n= (nth la i) in
                     (\<forall>ia<  m - n. 
                       intlen (sub  ((nth l (n +ia)) - (nth l n))   
                                    ((nth l (n+(Suc ia))) - (nth l n))   
                                    (sub (nth l n) (nth l m) \<sigma>)) =
                       intlen (sub (nth l (n +ia)) ((nth l (n+(Suc ia)))) \<sigma>)
                     )
                    )
                  ) "
       by (metis "2119" "2124")
    have 2127: "(\<forall>i<intlen la.
                   (let m = (nth la (Suc i)); n= (nth la i) in
                     (\<forall>ia<  m - n. 
                      (sub  ((nth l (n +ia)) - (nth l n))   
                            ((nth l (n+(Suc ia))) - (nth l n))   
                            (sub (nth l n) (nth l m) \<sigma>)) =
                      (sub (nth l (n +ia)) ((nth l (n+(Suc ia)))) \<sigma>) 
                     )
                   )
                 )
              " 
        using interval_eq_nth_eq 2125 2126 
        by (metis "2119")
    have 2128: "(\<forall>i<intlen la.
                   (let m = (nth la (Suc i)); n= (nth la i) in
                     (\<forall>ia<  m - n. 
                      f (sub (nth l (n +ia)) ((nth l (n+(Suc ia)))) \<sigma>) 
                     )
                    )
                  )
                 " 
           by (metis "2041" "3" add.commute add_Suc_right less_diff_conv less_le_trans 
                powerinterval_def)
    show ?thesis 
    by (metis "2111" "2112" "2113" "2127" "2128")
   qed
   have 220: "(\<forall> i< intlen la. 
                 ((filt (sub (nth l (nth la i)) (nth l (nth la (Suc i))) \<sigma>)
                        (map (shiftm (nth l (nth la i))) (sub (nth la i) (nth la (Suc i)) l)) )
                   \<Turnstile> g)    
                )     
               " 
         proof -
          have 2201: "(\<forall> i< intlen la. 
                        intlen (filt (sub (nth l (nth la i)) (nth l (nth la (Suc i))) \<sigma>)
                                     (map (shiftm (nth l (nth la i)))
                                          (sub (nth la i) (nth la (Suc i)) l)) ) =
                      intlen (map (shiftm (nth l (nth la i))) (sub (nth la i) (nth la (Suc i)) l)) 
                    )
                   " 
               using filt_intlen by blast
          have 2202: "(\<forall> i< intlen la.
                         intlen (map (shiftm (nth l (nth la i))) 
                                     (sub (nth la i) (nth la (Suc i)) l)) =
                            (nth la (Suc i)) - (nth la i)
                      )
                     "
            using "204" by blast         
          have 2203: "(\<forall> i< intlen la.  
                       (\<forall> j\<le> (nth la (Suc i)) - (nth la i) .
                         nth ((filt (sub (nth l (nth la i)) (nth l (nth la (Suc i))) \<sigma>)
                                    (map (shiftm (nth l (nth la i))) 
                                         (sub (nth la i) (nth la (Suc i)) l)) ) ) j =
                           nth (sub (nth l (nth la i)) (nth l (nth la (Suc i))) \<sigma>) 
                               (nth (map (shiftm (nth l (nth la i))) 
                                         (sub (nth la i) (nth la (Suc i)) l)) j) 
                        ))
                      " 
                by (simp add: filt_map interval_nth_map)
          have 2204: "(\<forall> i< intlen la.  
                       (\<forall> j\<le> (nth la (Suc i)) - (nth la i) .
                          nth (sub (nth l (nth la i)) (nth l (nth la (Suc i))) \<sigma>) 
                               (nth (map (shiftm (nth l (nth la i))) 
                                         (sub (nth la i) (nth la (Suc i)) l)) j) =
                          nth (sub (nth l (nth la i)) (nth l (nth la (Suc i))) \<sigma>) 
                              ((nth (sub (nth la i) (nth la (Suc i)) l) j) -(nth l (nth la i)))
                       )) 
                     " 
              by (simp add: interval_nth_map shiftm_def)   
          have 2205: "(\<forall> i< intlen la.  
                       (
                          (nth l (nth la i)) \<le> (nth l (nth la (Suc i))) \<and>
                          (nth l (nth la (Suc i))) \<le> intlen \<sigma> 
                       ))
                     "  
                     by (metis "12" "9" add.commute filt_intlen filt_map interval_idx_expand 
                          interval_nth_map  plus_1_eq_Suc)
          have 2206 : "(\<forall> i< intlen la.  
                       (\<forall> j\<le> (nth la (Suc i)) - (nth la i) . 
                         (nth (sub (nth la i) (nth la (Suc i)) l) j) =
                          (nth l ((nth la i)+j))
                       ))
                       "
            using "205" by blast 
          have 2207: "(\<forall> i< intlen la.  
                       (\<forall> j\<le> (nth la (Suc i)) - (nth la i) .
                        ((nth l (nth la i)) + ((nth (sub (nth la i) (nth la (Suc i)) l) j)
                                               -(nth l (nth la i)))) =
                           (nth l ((nth la i)+j))
                       ))
                      "  
              by (simp add: "206" "2206")      
          have 2208: "(\<forall> i< intlen la.  
                       (\<forall> j\<le> (nth la (Suc i)) - (nth la i) .
                         nth (sub (nth l (nth la i)) (nth l (nth la (Suc i))) \<sigma>) 
                             ((nth (sub (nth la i) (nth la (Suc i)) l) j) -(nth l (nth la i))) =
                         nth \<sigma>  (nth l ((nth la i)+j))          
                       ))
                      " 
               by (metis (no_types, lifting) "2041" "206" "2060" "2205" "3" Nat.le_diff_conv2
                    add.commute interval_idx_less_equal interval_nth_sub le_add_diff_inverse2)
          have 2209: "(\<forall> i< intlen la.  
                       (\<forall> j\<le> (nth la (Suc i)) - (nth la i) .
                          (nth (filt \<sigma> l) ((nth la i)+j)) = (nth \<sigma> (nth l ((nth la i)+j)))
                       ))
                      "
           by (simp add: filt_map interval_nth_map)
         have 2210: "(\<forall> i< intlen la. 
                        intlen  (sub (nth la i) (nth la (Suc i)) (filt \<sigma> l)) =
                          (nth la (Suc i)) - (nth la i) 
                     )
                    " 
                    using "5" PJ6help1 by blast
         have 2211: "(\<forall> i< intlen la. 
                      (\<forall> j\<le> (nth la (Suc i)) - (nth la i) .
                       (nth (sub (nth la i) (nth la (Suc i)) (filt \<sigma> l)) j) =
                        (nth \<sigma> (nth l ((nth la i)+j)))
                      ))
                    " 
                    using "2209" "5" interval_idx_expand interval_nth_sub by fastforce
         have 2212: "(\<forall> i< intlen la. 
                      (\<forall> j\<le> (nth la (Suc i)) - (nth la i) .
                             nth ((filt (sub (nth l (nth la i)) (nth l (nth la (Suc i))) \<sigma>)
                                    (map (shiftm (nth l (nth la i))) 
                                         (sub (nth la i) (nth la (Suc i)) l)) ) ) j =
                     (nth (sub (nth la i) (nth la (Suc i)) (filt \<sigma> l)) j)
                 ))
                "  
               by (simp add: "2203" "2204" "2208" "2211")
         have 2213: "(\<forall> i< intlen la. 
                 intlen ((filt (sub (nth l (nth la i)) (nth l (nth la (Suc i))) \<sigma>)
                               (map (shiftm (nth l (nth la i))) 
                                    (sub (nth la i) (nth la (Suc i)) l)) ) ) =
                 intlen (sub (nth la i) (nth la (Suc i)) (filt \<sigma> l))
                )
               " 
               by (metis "2202" "2210" filt_intlen)    
         have 2214: "(\<forall> i< intlen la. 
                       (filt (sub (nth l (nth la i)) (nth l (nth la (Suc i))) \<sigma>) 
                             (map (shiftm (nth l (nth la i))) (sub (nth la i) (nth la (Suc i)) l))) =
                       (sub (nth la i) (nth la (Suc i)) (filt \<sigma> l))
                     ) 
                    " 
         using interval_eq_nth_eq 2212 2213 using "2201" "2202" by fastforce  
         have 2215: "(\<forall> i< intlen la. 
                   (sub (nth la i) (nth la (Suc i)) (filt \<sigma> l)) \<Turnstile> g
                ) "
          using "5" powerinterval_def by blast  
      show ?thesis by (simp add: "2214" "2215") 
     qed      
    show ?thesis 
    using "201" "202" "203" "209" "210" "211" "220" by metis
   qed
  show ?thesis 
  by (metis "12" "20" "5" "8" "9" projection_d_def)          
qed      



lemma PJ7sem:
"(\<sigma> \<Turnstile> f \<triangle> (g \<triangle> h) = (f \<triangle> g) \<triangle> h)  "
proof -
 have 1: " intlen \<sigma> >0 \<longrightarrow> (\<sigma> \<Turnstile> f \<triangle> (g \<triangle> h) = (f \<triangle> g) \<triangle> h)"
  using PJ7helpchain1a PJ7helpchain1b unl_lift2 by blast
 have 2: " intlen \<sigma> = 0 \<longrightarrow> (\<sigma> \<Turnstile> f \<triangle> (g \<triangle> h) = (f \<triangle> g) \<triangle> h)"
   using PJ7empty by blast
 from 1 2 show ?thesis by auto
qed


subsubsection \<open>PJ8\<close>

lemma PJ8semhelp:
 assumes "index_sequence 0 l" 
         " Interval.nth l (intlen l) = intlen \<sigma> "
         " (\<forall>n na. na + n \<le> intlen \<sigma> \<longrightarrow> f (sub n (n+ na) \<sigma>) \<longrightarrow> g (sub n (n + na) \<sigma>)) "
 shows "
         (\<forall>i<intlen l. f (sub (Interval.nth l i) (Interval.nth l (Suc i)) \<sigma>)
                       \<longrightarrow> g (sub (Interval.nth l i) (Interval.nth l (Suc i)) \<sigma>)
         )"
by (metis add.commute assms interval_idx_expand 
    ordered_cancel_comm_monoid_diff_class.add_diff_inverse plus_1_eq_Suc)


lemma PJ8sem:
"(\<sigma> \<Turnstile> ba(f \<longrightarrow> g) \<longrightarrow> (f \<triangle> h)  \<longrightarrow> (g \<triangle> h))"
using PJ8semhelp by (simp add: projection_d_def ba_defs powerinterval_def) blast


subsubsection \<open>PJ9\<close>

lemma PJ9sem:
"(\<sigma> \<Turnstile> f \<nabla> ( g \<longrightarrow> h) \<longrightarrow> f \<triangle> g  \<longrightarrow> f \<triangle> h)"
by (simp add: uprojection_d_def projection_d_def, metis)



subsection \<open>Axioms\<close>

lemma BpGen:
 assumes "\<turnstile> f"
 shows   "\<turnstile> bp f"
using assms 
by (simp add: bp_d_def uprojection_d_def projection_d_def Valid_def)

lemma PJ1:
 "\<turnstile> f \<triangle> ( g \<or> h) \<longrightarrow> f \<triangle> g \<or> f \<triangle> h"
using PJ1sem Valid_def by blast

lemma PJ2:
 "\<turnstile> f \<triangle> empty = empty"
 using PJ2sem Valid_def by blast

lemma PJ3:
 "\<turnstile> f \<triangle> skip = (f \<and> more)" 
using PJ3sem Valid_def by blast

lemma PJ4:
 "\<turnstile> f \<triangle> (g;h) = (f \<triangle> g) ; ( f \<triangle> h)"
using PJ4sem Valid_def by blast

lemma PJ5:
 "\<turnstile> f \<triangle> init(g) \<longrightarrow> init(g)"
using PJ5sem Valid_def by blast

lemma PJ6:
 "\<turnstile> skip \<triangle> g = g"
using PJ6sem Valid_def by blast

lemma PJ7:
 "\<turnstile> f \<triangle> (g \<triangle> h) = (f \<triangle> g) \<triangle> h"
using PJ7sem Valid_def by blast

lemma PJ8:
 "\<turnstile> ba(f \<longrightarrow> g) \<longrightarrow> (f \<triangle> h)  \<longrightarrow> (g \<triangle> h)"
using PJ8sem Valid_def by blast

lemma PJ9:
 "\<turnstile> f \<nabla> (g \<longrightarrow> h) \<longrightarrow> f \<triangle> g  \<longrightarrow> f \<triangle> h"
using PJ9sem Valid_def by blast


subsection \<open>Time Reversal\<close>


lemma filt_intapp:
 "filt w (l \<ominus> \<langle>x\<rangle>) = (filt w l) \<ominus> \<langle>(nth w x)\<rangle>"
proof 
 (induct l)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a l)
 then show ?case 
 by simp
 qed



lemma filt_rev:
assumes "\<forall> i \<le> intlen l. (nth l i) \<le> intlen w"
 shows "(filt (intrev w) l) = (intrev (filt w (map (\<lambda>x. intlen w -x) (intrev l)) )) " 
using assms
proof 
  (induct l)
  case (St x)
  then show ?case 
    proof -
     have 01: "filt (intrev w) \<langle>x\<rangle> = \<langle> nth (intrev w) x \<rangle>"
       by simp
     have 02: "\<langle> nth (intrev w) x \<rangle> = \<langle> nth w (intlen w -x) \<rangle>"  
       using St.prems interval_intrev_nth by auto
     have 03: "intrev (filt w (interval.map ((-) (intlen w)) \<langle>x\<rangle>)) = \<langle> nth w (intlen w -x) \<rangle>" 
       by simp
     from 01 02 03 show ?thesis by auto
    qed
  next
  case (Cons x1a l)
  then show ?case 
     proof -
     have 1: "filt (intrev w) (x1a \<odot> l) =
               (nth (intrev w) x1a) \<odot> (filt (intrev w) l) " 
           by simp
     have 2: "(nth (intrev w) x1a) = (nth w (intlen w - x1a) )"
         using Cons.prems interval_intrev_nth by fastforce
     have 3: "intrev (filt w (x1a \<odot> l))  = intrev ((nth w x1a) \<odot> (filt w l)) "
       by simp
     have 4: "intrev (filt w (map ((-) (intlen w)) (intrev (x1a \<odot> l)))) =
              intrev (filt w  (map ((-) (intlen w)) ( (intrev l) \<ominus> \<langle> x1a \<rangle> )))  
                "
       by simp
     have 5: "intrev (filt w  (map ((-) (intlen w)) ( (intrev l) \<ominus> \<langle> x1a \<rangle> ))) =
              intrev ( (filt w  (map ((-) (intlen w)) ( (intrev l)))) \<ominus> \<langle> nth w (intlen w -x1a) \<rangle>)"
        by (simp add: filt_intapp) 
     have 6: "intrev ( (filt w  (map ((-) (intlen w)) ( (intrev l)))) \<ominus> \<langle> nth w (intlen w -x1a) \<rangle>) =
              (nth w (intlen w -x1a)) \<odot> intrev( (filt w  (map ((-) (intlen w)) ( (intrev l)))))  "
        by auto 
     have 7: "\<forall> i \<le> intlen l. (nth l i) \<le> intlen w"
         using local.Cons(2) by auto
     have 8 : "(filt (intrev w) l) = intrev(filt w (interval.map ((-) (intlen w)) (intrev l)))"
        using "7" Cons.hyps by blast
     show ?thesis 
     using "2" "5" "8" by auto
     qed
qed

lemma ProjectionRevsema:
 assumes "(\<sigma> \<Turnstile> (f \<triangle> g)\<^sup>r)"
 shows   "  (\<sigma> \<Turnstile> (f\<^sup>r) \<triangle> (g\<^sup>r))"
proof -
 have 1: " (\<sigma> \<Turnstile> (f \<triangle> g)\<^sup>r)" 
  using assms by auto
 have 2: "\<exists>l. index_sequence 0 l \<and> Interval.nth l (intlen l) = intlen \<sigma> \<and> 
          powerinterval f (intrev \<sigma>) l \<and> g (filt (intrev \<sigma>) l)"
  using "1" by (simp add: projection_d_def reverse_d_def)
 obtain l where 3: "index_sequence 0 l \<and> Interval.nth l (intlen l) = intlen \<sigma> \<and> 
          powerinterval f (intrev \<sigma>) l \<and> g (filt (intrev \<sigma>) l)" 
  using "2" by auto
 have 4: "index_sequence 0 l" 
  using "3" by auto
 have 5: "Interval.nth l (intlen l) = intlen \<sigma> "
  using "3" by auto
 have 6: "powerinterval f (intrev \<sigma>) l" 
   using "3" by auto
 have 7: "g (filt (intrev \<sigma>) l)" 
    using "3" by auto
 have 8: "\<forall> i \<le> intlen l. (nth l i) \<le> intlen \<sigma>"  
  using "4" "5" interval_idx_less_last_1 le_eq_less_or_eq by fastforce
 have 9: "g ( intrev(filt \<sigma> (map (\<lambda>x. intlen \<sigma> -x) (intrev l)) ))" 
  using "7" "8" filt_rev by fastforce
 have 10: "nth (map ((-) (intlen \<sigma>)) (intrev l)) 0 = 0"
      by (metis "5" diff_self_eq_0 interval_intfirst_intrev interval_nth_intlen_intlast 
          interval_nth_map interval_nth_zero_intfirst)
 have 11: "(\<forall>n<intlen l.
             nth (map ((-) (intlen \<sigma>)) (intrev l)) n
             < nth (map ((-) (intlen \<sigma>)) (intrev l)) (Suc n))" 
      by (simp add: interval_nth_map interval_intrev_nth) 
          (metis (no_types, lifting) "4" "5" Suc_diff_Suc Suc_less_eq diff_less_Suc diff_less_mono2 
           index_sequence_def interval_idx_less_last_1)
 have 12: "index_sequence 0 (map (\<lambda>x. intlen \<sigma> -x) (intrev l))" 
   by (simp add: "10" "11" index_sequence_def)
 have 13: "intlen (map (\<lambda>x. intlen \<sigma> -x) (intrev l)) = intlen l"
    by auto
 have 14: "nth (map ((-) (intlen \<sigma>)) (intrev l)) (intlen (map ((-) (intlen \<sigma>)) (intrev l))) = 
           intlen \<sigma>" 
   by (metis "4" diff_zero index_sequence_def interval_intlast_intrev interval_intlen_map 
       interval_nth_intlen_intlast interval_nth_map interval_nth_zero_intfirst)
 have 15: "\<forall>i<intlen l.
           (nth (map ((-) (intlen \<sigma>)) (intrev l)) i) \<le>
           (nth (map ((-) (intlen \<sigma>)) (intrev l)) (Suc i))
          " 
       by (simp add: "11" less_imp_le_nat)
 have 16: "\<forall>i<intlen l. 
            (nth (map ((-) (intlen \<sigma>)) (intrev l)) (Suc i)) \<le> intlen \<sigma>
          " 
     by (simp add: interval_nth_map)
 have 17: "\<forall>i<intlen l.
           intrev
               (sub (nth (map ((-) (intlen \<sigma>)) (intrev l)) i)
                    (nth (map ((-) (intlen \<sigma>)) (intrev l)) (Suc i)) \<sigma>) =
              sub (intlen \<sigma> - (nth (map ((-) (intlen \<sigma>)) (intrev l)) (Suc i)))
                  (intlen \<sigma> - (nth (map ((-) (intlen \<sigma>)) (intrev l)) i))
                  (intrev \<sigma>)  
          " 
      using interval_intrev_sub 
      using "15" "16" by blast
  have 18: "\<forall>i<intlen l.
             intlen \<sigma> - (nth (map ((-) (intlen \<sigma>)) (intrev l)) (Suc i)) =
             (nth (intrev l) (Suc i)) 
           " 
     by (metis "10" "12" "13" "14" Suc_leI diff_diff_cancel interval_idx_less_than 
         interval_intlen_gr_zero interval_nth_map less_le zero_less_Suc zero_less_diff)
  have 19: "\<forall>i<intlen l.
             intlen \<sigma> - (nth (map ((-) (intlen \<sigma>)) (intrev l)) (i)) =
             (nth (intrev l) (i)) " 
   by (metis "12" "13" "5" diff_diff_cancel gr_zeroI interval_idx_greater_first 
       interval_intfirst_intrev interval_intlast_prefix interval_nth_map interval_nth_zero_intfirst 
       interval_prefix_intlen order.order_iff_strict zero_less_diff)
  have 20: "\<forall>i<intlen l.
              ( sub (intlen \<sigma> - (nth (map ((-) (intlen \<sigma>)) (intrev l)) (Suc i)))
                  (intlen \<sigma> - (nth (map ((-) (intlen \<sigma>)) (intrev l)) i))
                  (intrev \<sigma>)) =
              ( sub (nth (intrev l) (Suc i)) 
                  (nth (intrev l) (i)) 
                  (intrev \<sigma>))
           " 
    using "18" "19" by simp
  have 21: "\<forall>i<intlen l.
             ( sub (nth (intrev l) (Suc i)) 
                  (nth (intrev l) (i)) 
                  (intrev \<sigma>)) =
            ( sub (nth l (intlen l -  (Suc i))) 
                  (nth l (intlen l - i))  
                  (intrev \<sigma>))
             " 
      by (simp add: interval_intrev_nth)
  have 22: "\<forall>i<intlen l.
             f ( sub (nth l (intlen l -  (Suc i))) 
                  (nth l (intlen l - i))  
                  (intrev \<sigma>))
            " 
       by (metis "6" Suc_diff_Suc Suc_less_eq diff_less_Suc powerinterval_def)
  have 24: "\<forall>i<intlen l.
            f (intrev
               (sub (nth (map ((-) (intlen \<sigma>)) (intrev l)) i)
                    (nth (map ((-) (intlen \<sigma>)) (intrev l)) (Suc i)) \<sigma>))"
       by (simp add: "17" "20" "21" "22")
  have 25: "powerinterval (\<lambda>s. f (intrev s)) \<sigma> (map ((-) (intlen \<sigma>)) (intrev l))" 
    by (simp add: "24" powerinterval_def)     
  have 26: "(\<exists>l. index_sequence 0 l \<and> 
                 nth l (intlen l) = intlen \<sigma> \<and> 
                 powerinterval (\<lambda>s. f (intrev s)) \<sigma> l \<and> g (intrev(filt \<sigma> l)))" 
  using "12" "14" "25" "9" by blast    
  from 26 show ?thesis
  by (simp add: projection_d_def reverse_d_def)
qed


lemma ProjectionRevsemb:
 assumes "  (\<sigma> \<Turnstile> (f\<^sup>r) \<triangle> (g\<^sup>r))"
 shows   "(\<sigma> \<Turnstile> (f \<triangle> g)\<^sup>r)"
proof -
 have 1: "(\<exists>l. index_sequence 0 l \<and> 
                 nth l (intlen l) = intlen \<sigma> \<and> 
                 powerinterval (\<lambda>s. f (intrev s)) \<sigma> l \<and> g (intrev(filt \<sigma> l)))"
  using assms by (simp add:  projection_d_def reverse_d_def)
 obtain l where 2: "index_sequence 0 l \<and> 
                 nth l (intlen l) = intlen \<sigma> \<and> 
                 powerinterval (\<lambda>s. f (intrev s)) \<sigma> l \<and> g (intrev(filt \<sigma> l))" 
 using "1" by auto
 have 3: "index_sequence 0 l" 
  using "2" by auto
 have 4: "nth l (intlen l) = intlen \<sigma>" 
  using "2" by auto
 have 5: "powerinterval (\<lambda>s. f (intrev s)) \<sigma> l"
  using "2" by auto
 have 6: "g (intrev(filt \<sigma> l))"
  using "2" by auto
 have 7: "intlen (map ((-) (intlen \<sigma>)) (intrev l)) = intlen l"
   by simp 
 have 8: "\<forall> i \<le> intlen l. (nth (map ((-) (intlen \<sigma>)) (intrev l)) i) \<le> intlen \<sigma>"
   by (simp add: interval_nth_map)
 have 9: "g (filt (intrev \<sigma>) (map ((-) (intlen \<sigma>)) (intrev l)))" 
  by (metis "2" filt_rev interval_idx_less_equal interval_intrev_intlen interval_rev_rev_ident 
      le_refl)
 have 10: "nth (map ((-) (intlen \<sigma>)) (intrev l)) (intlen (map ((-) (intlen \<sigma>)) (intrev l))) = intlen \<sigma>" 
   by (metis "2" diff_zero index_sequence_def interval_intlast_intrev interval_intlen_map 
       interval_nth_intlen_intlast interval_nth_map interval_nth_zero_intfirst)
 have 11: "nth (map ((-) (intlen \<sigma>)) (intrev l)) 0 = 0"
      by (metis "4" diff_self_eq_0 interval_intfirst_intrev interval_nth_intlen_intlast 
          interval_nth_map interval_nth_zero_intfirst)
 have 12: "(\<forall>n<intlen l.
             nth (map ((-) (intlen \<sigma>)) (intrev l)) n
             < nth (map ((-) (intlen \<sigma>)) (intrev l)) (Suc n))" 
    by (simp add: interval_nth_map interval_intrev_nth)
      (metis (no_types, lifting) "2" Suc_diff_Suc Suc_less_eq diff_less_Suc diff_less_mono2 
       index_sequence_def interval_idx_less_last_1)
 have 13: "index_sequence 0 (map (\<lambda>x. intlen \<sigma> -x) (intrev l))"
   by (simp add: "11" "12" index_sequence_def)
 have 14: "\<forall>i<intlen l. f (intrev (sub (nth l i) (nth l (Suc i)) \<sigma>))"
  using "5" by (simp add:powerinterval_def)
 have 15: "\<forall>i<intlen l. (nth l (Suc i)) \<le> intlen \<sigma>"
    using "2" interval_idx_expand by fastforce
 have 16: "\<forall>i<intlen l. (nth l i) \<le> (nth l (Suc i))" 
    using "2" interval_idx_expand by fastforce
 have 17: "\<forall>i<intlen l. 
            f (sub ((intlen \<sigma>)-(nth l (Suc i))) ((intlen \<sigma>) - (nth l i)) (intrev \<sigma>))  
          " 
    using "14" 
    by (simp add: interval_intrev_sub "15" "16")
 have 18: "\<forall>i<intlen l.
            (nth (map ((-) (intlen \<sigma>)) (intrev l)) i) =
            intlen \<sigma> - (nth (intrev l) i) 
          " 
  using interval_nth_map by blast
 have 19: "\<forall>i<intlen l.
           (nth (map ((-) (intlen \<sigma>)) (intrev l)) (Suc i)) =
            intlen \<sigma> - (nth (intrev l) (Suc i)) 
          " 
   using interval_nth_map by blast
 have 20: "\<forall>i<intlen l.
            (nth (intrev l) i) = (nth l (intlen l - i))
          " 
     by (simp add: interval_intrev_nth)
 have 21: "\<forall>i<intlen l.
            (nth (intrev l) (Suc i)) = (nth l (intlen l - (Suc i)))" 
   by (simp add: interval_intrev_nth)
 have 22: "\<forall>i<intlen l.
           f (sub (intlen \<sigma> - (nth l (intlen l - i))) 
                   (intlen \<sigma> - (nth l (intlen l - (Suc i))) )
                   (intrev \<sigma>))
          " 
   by (metis "17" Suc_diff_Suc Suc_less_eq diff_less_Suc)
 have 23: "\<forall>i<intlen l.
            f (sub (intlen \<sigma> - (nth (intrev l) i)) 
                   (intlen \<sigma> - (nth (intrev l) (Suc i)) )
                   (intrev \<sigma>))
          " 
   by (simp add: "20" "21" "22")
 have 24: "\<forall>i<intlen l.
            f (sub (nth (map ((-) (intlen \<sigma>)) (intrev l)) i)
                   (nth (map ((-) (intlen \<sigma>)) (intrev l)) (Suc i)) (intrev \<sigma>))"
  by (simp add: "23" interval_nth_map)
 have 25: "powerinterval f (intrev \<sigma>) (map ((-) (intlen \<sigma>)) (intrev l))" 
   by (simp add: "24" powerinterval_def)
 have 26: "\<exists>l. index_sequence 0 l \<and> Interval.nth l (intlen l) = intlen \<sigma> \<and> 
          powerinterval f (intrev \<sigma>) l \<and> g (filt (intrev \<sigma>) l)" 
   using "10" "13" "25" "9" by blast
 from 26 show ?thesis by (simp add: projection_d_def reverse_d_def)
qed



lemma ProjectionRev:
 "\<turnstile> (f \<triangle> g)\<^sup>r = f\<^sup>r \<triangle> g\<^sup>r"
using ProjectionRevsema ProjectionRevsemb unl_lift2 by blast



subsection \<open>Theorems\<close>

subsubsection \<open>Projection\<close>

lemma PowerProjLen:
 "\<turnstile> f \<triangle> len n = power (f \<and> more) n"
proof 
  (induct n)
  case 0
  then show ?case by (metis PJ2 len_d_def pow_0)
  next
  case (Suc n)
  then show ?case 
  by (metis PJ3 PJ4 inteq_reflection len_d_def pow_Suc)
  qed  

lemma ProjLenExist:
 "\<turnstile> f \<triangle> (\<exists>n. len n) = (\<exists> n. f \<triangle> len n)"
  by (simp add: Valid_def projection_d_def, blast)

lemma PowerProjLenExist:
 "\<turnstile> (\<exists> n. f \<triangle> len n) = (\<exists> n. power (f \<and> more) n)"
using PowerProjLen by (simp add: Valid_def PowerProjLen, blast)


lemma RightProjImpProj:
 assumes "\<turnstile> g1 \<longrightarrow> g2"
 shows   "\<turnstile> f \<triangle> g1 \<longrightarrow> f \<triangle> g2"
using assms 
by (simp add: Valid_def projection_d_def, blast)

lemma LeftProjImpProj:
 assumes "\<turnstile> f1 \<longrightarrow> f2"
 shows   "\<turnstile> f1 \<triangle> g \<longrightarrow> f2 \<triangle> g"
using assms 
by (simp add: Valid_def projection_d_def powerinterval_def, blast)

lemma RightProjEqvProj:
 assumes "\<turnstile> g1 = g2"
 shows   "\<turnstile> f \<triangle> g1 = f \<triangle> g2"
using assms 
by (metis (mono_tags, lifting) Valid_def inteq_reflection unl_lift2)

lemma LeftProjEqvProj:
 assumes "\<turnstile> f1 = f2"
 shows   "\<turnstile> f1 \<triangle> g = f2 \<triangle> g"
using assms 
by (metis (mono_tags, lifting) Valid_def inteq_reflection unl_lift2)


lemma ProjTrueEqvChopstar:
 "\<turnstile> f \<triangle> #True = f\<^sup>\<star>"
proof -
 have 1: "\<turnstile>  #True = ( (\<exists>n. len n))"
   by (simp add: Valid_def len_defs)
 have 2: "\<turnstile> f \<triangle> #True = f \<triangle> (\<exists>n. len n) " 
   using "1" RightProjEqvProj by blast
 have 3: "\<turnstile> f \<triangle> (\<exists>n. len n) = (\<exists> n. power (f \<and> more) n)" 
  using PowerProjLenExist ProjLenExist by fastforce
 have 4: "\<turnstile> (\<exists> n. power (f \<and> more) n) = f\<^sup>\<star>" 
  by (simp add: chopstar_d_def powerstar_d_def)
 show ?thesis using "2" "3" "4" by fastforce
qed

lemma ProjChopstarEqvChopstarProj:
 "\<turnstile> f \<triangle> (g\<^sup>\<star>) = (f \<triangle> g)\<^sup>\<star>"
proof -
 have 1: "\<turnstile> f \<triangle> (g\<^sup>\<star>) = f \<triangle> (g \<triangle> #True)"
  by (metis ProjTrueEqvChopstar RightProjEqvProj inteq_reflection) 
 have 2: "\<turnstile> f \<triangle> (g \<triangle> #True) = (f \<triangle> g) \<triangle> #True"
  by (simp add: PJ7) 
 have 3: "\<turnstile> (f \<triangle> g) \<triangle> #True = (f \<triangle> g)\<^sup>\<star>" 
    by (simp add: ProjTrueEqvChopstar)
 show ?thesis using "1" "2" "3" by fastforce
qed  


lemma ProjAndImp:
 " \<turnstile> f \<triangle> (g1 \<and> g2) \<longrightarrow> f \<triangle> g1 \<and> f \<triangle> g2  "
by (meson Prop12 RightProjImpProj int_iffD1 lift_and_com)

lemma ProjOrDist:
 "\<turnstile> #True \<triangle> (f \<or>  g) = ( #True \<triangle> f  \<or>  #True \<triangle> g) "
using PJ1sema by blast

lemma StateImportProj:
 "\<turnstile> ((init w) \<and> f \<triangle> g) = f \<triangle> ((init w) \<and> g)"
by (auto simp add: Valid_def init_defs projection_d_def filt_nth index_sequence_def) 

lemma ProjStateAndNextEqvStateAndMoreChopProj:
" \<turnstile> f \<triangle> ( (init w) \<and> \<circle> g) = ((init w) \<and> (f \<and> more);(f\<triangle> g))"
proof -
 have 2: "\<turnstile> (f \<and> more);(f\<triangle> g) = f \<triangle> \<circle> g"  
   by (metis PJ3 PJ4 inteq_reflection next_d_def)   
 have 3: "\<turnstile> f \<triangle> ( (init w) ) \<longrightarrow> init w "
   by (simp add: PJ5)
 have 4: "\<turnstile> (init w \<and>   f \<triangle> \<circle> g) = f \<triangle> ( (init w) \<and> \<circle> g) "
  by (simp add: StateImportProj)
 have 5: "\<turnstile> f \<triangle> ( (init w) \<and> \<circle> g) \<longrightarrow> ((init w) \<and> (f \<and> more);(f\<triangle> g))"
    using "2" "3" ProjAndImp by fastforce
  from 5 4 show ?thesis using "2" by fastforce
qed

lemma ProjNext:
 "\<turnstile> f \<triangle> \<circle> g = (f \<and> more);(f\<triangle> g)"
by (metis PJ3 PJ4 inteq_reflection next_d_def)

lemma ProjWnext:
 "\<turnstile> f \<triangle> (wnext g) = (empty \<or> (f \<and> more);(f\<triangle> g))"
proof -
 have 1: "\<turnstile> f \<triangle> (wnext g) = f \<triangle> (empty \<or> \<circle> g)"
  by (simp add: RightProjEqvProj WnextEqvEmptyOrNext)
 have 2: "\<turnstile> f \<triangle> (empty \<or> \<circle> g) = (empty \<or> f \<triangle> (\<circle> g)) "
   using PJ1sema PJ2 by fastforce
 have 3: "\<turnstile> f \<triangle> (\<circle> g) = (f \<and> more);(f\<triangle> g)"
   by (metis PJ3 PJ4 inteq_reflection next_d_def)
 show ?thesis 
 using "1" "2" "3" by fastforce
qed


lemma ProjIntro:
 assumes "\<turnstile>   f \<and>  more  \<longrightarrow> (g \<and>  more ); f "
 shows   "\<turnstile> f \<longrightarrow> g \<triangle> #True "
using assms CSIntro ProjTrueEqvChopstar by force 
    

lemma RightBoxStateImportProj:
 "\<turnstile> \<box>(init w) \<and> f \<triangle> g \<longrightarrow> f \<triangle> (\<box>(init w) \<and> g)"
by (simp add: Valid_def always_defs init_defs projection_d_def)
   ( metis diff_zero filt_expand interval_idx_bound_1 interval_intlen_gr_zero 
    interval_suffix_length_good)

lemma LeftBoxStateImportProjhelp:
 "(\<forall>n\<le>intlen wa. w \<langle>Interval.nth wa n\<rangle>) \<and>
         (\<exists>l. Interval.nth l 0 = 0 \<and>
              (\<forall>n<intlen l. Interval.nth l n < Interval.nth l (Suc n)) \<and>
              Interval.nth l (intlen l) = intlen wa \<and>
              (\<forall>i<intlen l. f (sub (Interval.nth l i) (Interval.nth l (Suc i)) wa)) \<and>
              g (filt wa l)) \<longrightarrow>
         (\<exists>l. Interval.nth l 0 = 0 \<and>
              (\<forall>n<intlen l. Interval.nth l n < Interval.nth l (Suc n)) \<and>
              Interval.nth l (intlen l) = intlen wa \<and>
              (\<forall>i<intlen l.
                  f (sub (Interval.nth l i) (Interval.nth l (Suc i)) wa) \<and>
                  (\<forall>n\<le>intlen (sub (Interval.nth l i) (Interval.nth l (Suc i)) wa).
                      w \<langle>Interval.nth (sub (Interval.nth l i) (Interval.nth l (Suc i)) wa) n\<rangle>)) \<and>
              g (filt wa l))"
proof 
 assume 0: "(\<forall>n\<le>intlen wa. w \<langle>Interval.nth wa n\<rangle>) \<and>
    (\<exists>l. Interval.nth l 0 = 0 \<and>
         (\<forall>n<intlen l. Interval.nth l n < Interval.nth l (Suc n)) \<and>
         Interval.nth l (intlen l) = intlen wa \<and>
         (\<forall>i<intlen l. f (sub (Interval.nth l i) (Interval.nth l (Suc i)) wa)) \<and>
         g (filt wa l))" 
 show   "\<exists>l. Interval.nth l 0 = 0 \<and>
        (\<forall>n<intlen l. Interval.nth l n < Interval.nth l (Suc n)) \<and>
        Interval.nth l (intlen l) = intlen wa \<and>
        (\<forall>i<intlen l.
            f (sub (Interval.nth l i) (Interval.nth l (Suc i)) wa) \<and>
            (\<forall>n\<le>intlen (sub (Interval.nth l i) (Interval.nth l (Suc i)) wa).
                w \<langle>Interval.nth (sub (Interval.nth l i) (Interval.nth l (Suc i)) wa) n\<rangle>)) \<and>
        g (filt wa l)"
  proof -
   have 1: "(\<forall>n\<le>intlen wa. w \<langle>Interval.nth wa n\<rangle>) "
    using 0 by auto
   have 2: "(\<exists>l. Interval.nth l 0 = 0 \<and>
         (\<forall>n<intlen l. Interval.nth l n < Interval.nth l (Suc n)) \<and>
         Interval.nth l (intlen l) = intlen wa \<and>
         (\<forall>i<intlen l. f (sub (Interval.nth l i) (Interval.nth l (Suc i)) wa)) \<and>
         g (filt wa l))"
     using 0 by auto
   obtain l where 3: "Interval.nth l 0 = 0 \<and>
         (\<forall>n<intlen l. Interval.nth l n < Interval.nth l (Suc n)) \<and>
         Interval.nth l (intlen l) = intlen wa \<and>
         (\<forall>i<intlen l. f (sub (Interval.nth l i) (Interval.nth l (Suc i)) wa)) \<and>
         g (filt wa l)"
      using "2" by auto
   have 4: "Interval.nth l 0 = 0" 
     using "3" by auto
   have 5: "(\<forall>n<intlen l. Interval.nth l n < Interval.nth l (Suc n))"  
     using "3" by auto
   have 6: "Interval.nth l (intlen l) = intlen wa" 
      using "3" by auto
   have 7: "(\<forall>i<intlen l. f (sub (Interval.nth l i) (Interval.nth l (Suc i)) wa))" 
      using "3" by auto
   have 8: "g (filt wa l)"
      using "3" by auto
   have 9: "(\<forall>i<intlen l.
            f (sub (Interval.nth l i) (Interval.nth l (Suc i)) wa) \<and>
            (\<forall>n\<le>(nth l (Suc i)) -(nth l i).
                w \<langle>nth wa ((nth l i)+n)\<rangle>))" 
   by (metis "1" "7" interval_nth_last_stutter nat_le_iff_add nat_le_linear)
   have 10: "(\<forall>i<intlen l. 
               intlen (sub (Interval.nth l i) (Interval.nth l (Suc i)) wa) =
                (nth l (Suc i)) -(nth l i) ) " 
       by (simp add: "3" PJ6help1 index_sequence_def)
   have 11: "(\<forall>i<intlen l.
              (\<forall>n\<le>(nth l (Suc i)) -(nth l i).
                Interval.nth (sub (Interval.nth l i) (Interval.nth l (Suc i)) wa) n =
                 nth wa ((nth l i)+n) )) "
       using "3" index_sequence_def interval_idx_expand by fastforce
  have 12: "(\<forall>i<intlen l.
            f (sub (Interval.nth l i) (Interval.nth l (Suc i)) wa) \<and>
            (\<forall>n\<le>intlen (sub (Interval.nth l i) (Interval.nth l (Suc i)) wa).
                w \<langle>Interval.nth (sub (Interval.nth l i) (Interval.nth l (Suc i)) wa) n\<rangle>))"
    using "9" "10" "11" by simp 
  show ?thesis 
  using "12" "3" by blast
  qed
qed
 
lemma LeftBoxStateImportProj:
 "\<turnstile> \<box>(init w) \<and> f \<triangle> g \<longrightarrow> (f \<and> \<box> (init w)) \<triangle> g"
using LeftBoxStateImportProjhelp
 by (simp add: index_sequence_def Valid_def always_defs init_defs projection_d_def powerinterval_def)
 blast

subsubsection \<open>dp and bp\<close>

lemma NotDpEqvBpNot:
 "\<turnstile> (\<not>(dp f)) =  bp (\<not> f)"
by (simp add: bp_d_def dp_d_def uprojection_d_def)


lemma NotBpEqvDpNot:
 "\<turnstile> (\<not>(bp f)) = dp(\<not> f)"
by (simp add: bp_d_def dp_d_def uprojection_d_def)


lemma NowImpDp:
 "\<turnstile> f \<longrightarrow> dp f"
proof -
 have 1: "\<turnstile>  (skip \<longrightarrow> #True)"
  by simp
 have 2: "\<turnstile> ba(skip \<longrightarrow> #True)"
  using "1" by (simp add: BaGen)
 have 3: "\<turnstile> ba(skip \<longrightarrow> #True) \<longrightarrow> (skip \<triangle> f \<longrightarrow> #True \<triangle> f)"
  using PJ8 by blast 
 have 4: "\<turnstile> (skip \<triangle> f \<longrightarrow> #True \<triangle> f)" 
  using "2" "3" MP by blast
 show ?thesis 
 by (metis "4" PJ6 dp_d_def inteq_reflection)
qed 

lemma BpElim:
 "\<turnstile> bp f \<longrightarrow> f"
proof -
 have 1: "\<turnstile> \<not> f \<longrightarrow> dp (\<not> f)"
   by (simp add: NowImpDp) 
 hence 2: "\<turnstile> \<not>(dp (\<not> f)) \<longrightarrow> f"
   by auto
 from 2 show ?thesis 
 by (simp add: bp_d_def dp_d_def uprojection_d_def)
qed

lemma BpImpDpImpDp:
  "\<turnstile> bp (f \<longrightarrow> g) \<longrightarrow>  dp  f \<longrightarrow>  dp  g "
proof -
 have 1: "\<turnstile> bp (f\<longrightarrow> g)  \<longrightarrow> (#True \<triangle> f) \<longrightarrow> (#True \<triangle> g) "  
   by (simp add: PJ9 bp_d_def)
 from 1 show ?thesis   by (simp add: dp_d_def)
qed



lemma BpContraPosImpDist:
  "\<turnstile> bp (\<not>  g \<longrightarrow> \<not>  f) \<longrightarrow> (bp  f) \<longrightarrow> (bp  g) "
proof -
  have  1: "\<turnstile> bp (\<not>  g \<longrightarrow> \<not>  f) \<longrightarrow> ( dp (\<not>  g)) \<longrightarrow> ( dp (\<not>  f)) " 
     by (rule BpImpDpImpDp)
  hence 2: "\<turnstile> bp (\<not>  g \<longrightarrow> \<not>  f) \<longrightarrow> (\<not> ( dp (\<not>  f))) \<longrightarrow> (\<not> ( dp (\<not>  g))) " by auto
  from 2 show ?thesis 
  by (simp add: bp_d_def dp_d_def uprojection_d_def)
qed


lemma BpImpDist:
 "\<turnstile> bp (f\<longrightarrow> g) \<longrightarrow> (bp  f) \<longrightarrow> (bp  g) "
proof -
 have  1: "\<turnstile> (f \<longrightarrow> g) \<longrightarrow> (\<not>  g \<longrightarrow> \<not>  f) " by auto
 hence 2: "\<turnstile> \<not> (\<not>  g \<longrightarrow> \<not>  f) \<longrightarrow> \<not> (f \<longrightarrow> g) " by auto
 hence 3: "\<turnstile> bp (\<not> (\<not>  g \<longrightarrow> \<not>  f) \<longrightarrow> \<not> (f \<longrightarrow> g)) " by (rule BpGen)
 have  4: "\<turnstile> bp (\<not> (\<not>  g \<longrightarrow> \<not>  f) \<longrightarrow> \<not> (f \<longrightarrow> g))
             \<longrightarrow>  
             bp (f \<longrightarrow> g) \<longrightarrow> bp (\<not>  g \<longrightarrow> \<not>  f) " by (rule BpContraPosImpDist)
 have  5: "\<turnstile> bp (f \<longrightarrow> g) \<longrightarrow> bp (\<not>  g \<longrightarrow> \<not>  f) " using 3 4 MP by blast
 have  6: "\<turnstile> bp (\<not>  g \<longrightarrow> \<not>  f) \<longrightarrow> (bp  f) \<longrightarrow> (bp  g) " by (rule BpContraPosImpDist)
 from 5 6 show ?thesis  using lift_imp_trans by blast 
qed

lemma DpImpDpRule:
 assumes "\<turnstile> f\<longrightarrow> g "
 shows   "\<turnstile>  dp  f \<longrightarrow>  dp  g "
proof -
 have  1: "\<turnstile> f\<longrightarrow> g " using assms by auto
 hence 2: "\<turnstile> #True \<triangle> f \<longrightarrow> #True \<triangle> g " 
   by (metis BpGen MP PJ9 bp_d_def)
 from 2 show ?thesis by (simp add: dp_d_def)
qed

lemma BpImpBpRule:
 assumes "\<turnstile> f\<longrightarrow> g"
 shows   "\<turnstile> bp  f \<longrightarrow> bp  g "
proof -
 have  1: "\<turnstile> f\<longrightarrow> g" using assms by auto
 hence 2: "\<turnstile> \<not>  g \<longrightarrow> \<not>  f " by auto
 hence 3: "\<turnstile>  dp (\<not>  g) \<longrightarrow>  dp (\<not>  f) " by (rule DpImpDpRule)
 hence 4: "\<turnstile> \<not> ( dp (\<not>  f)) \<longrightarrow> \<not> ( dp (\<not>  g)) " by auto
 from 4 show ?thesis 
   by (meson BpGen BpImpDist MP assms)
qed

lemma DpEqvDpRule:
 assumes "\<turnstile> f = g "
 shows   "\<turnstile>  dp f =  dp g "
proof -
 have  1: "\<turnstile> f = g " using assms by auto
 hence 2: "\<turnstile> #True \<triangle> f =  #True \<triangle> g " 
    using RightProjEqvProj by blast
 from 2 show ?thesis by (simp add: dp_d_def)
qed

lemma BpEqvBpRule:
 assumes "\<turnstile> f = g "
 shows   "\<turnstile>  bp f =  bp  g "
proof -
 have  1: "\<turnstile> f = g " using assms by auto
 hence 2: "\<turnstile> (\<not>  f) = (\<not>  g) " by auto
 hence 3: "\<turnstile>  dp (\<not>  f) =  dp (\<not>  g) " by (rule DpEqvDpRule)
 hence 4: "\<turnstile> (\<not>  (dp (\<not>  f))) = (\<not> ( dp (\<not>  g))) " by auto
 from 4 show ?thesis 
 by (metis BpImpBpRule assms int_iffD1 int_iffI inteq_reflection)
qed


lemma DpState:
 "\<turnstile> dp (init w) = (init w) "
by (metis NowImpDp PJ5 dp_d_def int_iffI)

lemma StateEqvBp:
 "\<turnstile>   (init w) = bp  (init w) "
proof -
  have 1: "\<turnstile> (init w) \<longrightarrow> bp  (init w) " 
   by (metis (no_types, lifting) DiState DpState Initprop(2) StateEqvBi bi_d_def bp_d_def dp_d_def 
       int_iffD1 inteq_reflection uprojection_d_def)
  have 2: "\<turnstile> bp  (init w) \<longrightarrow> (init w) " by (rule BpElim)
  from 1 2 show ?thesis by fastforce
qed

lemma DpDpEqvDp:
 "\<turnstile> dp (dp f) = dp f"
proof -
 have 2: "\<turnstile> #True \<triangle> ( #True \<triangle> f) = (#True \<triangle> #True) \<triangle> f" 
   by (simp add: PJ7)
 have 3: "\<turnstile> (#True \<triangle> #True) = #True"
    by (metis DpState Initprop(4) dp_d_def int_eq_true inteq_reflection) 
 show ?thesis by (metis "2" "3" dp_d_def inteq_reflection)
qed


lemma BpBpEqvBp:
 "\<turnstile> bp (bp f) = bp f"
proof -
 have 1: "\<turnstile> dp (dp (\<not> f)) = dp (\<not> f)"
   using DpDpEqvDp by blast 
 have 2: "\<turnstile> (\<not> (dp (dp (\<not> f)))) = (\<not> (dp (\<not> f)))"
   using "1" by auto
 have 3: "\<turnstile> (\<not> (dp (\<not> f))) = bp f"
    by (simp add: bp_d_def dp_d_def uprojection_d_def)
 have 4: "\<turnstile> (\<not> (dp (dp (\<not> f)))) = bp (bp f)"
    by (simp add: bp_d_def dp_d_def uprojection_d_def) 
 from 2 3 4 show ?thesis 
 by fastforce
qed



lemma DpOrEqv:
 "\<turnstile>    dp (f \<or>  g) =  (dp  f \<or>   dp  g) " 
proof -
 have 1: "\<turnstile> #True \<triangle> (f\<or>  g) = ( #True \<triangle> f  \<or>  #True \<triangle> g) " 
  using ProjOrDist by auto
 from 1 show ?thesis by (simp add: dp_d_def)
qed

lemma BpAndEqv:
 "\<turnstile> bp(f \<and> g) = (bp f \<and> bp g)"
proof -
 have 1: "\<turnstile> dp ((\<not> f) \<or>  (\<not>g)) =  (dp  (\<not>f) \<or>   dp  (\<not>g))"
   using DpOrEqv by auto
 hence 2: "\<turnstile> (\<not> (dp ((\<not> f) \<or>  (\<not>g)))) =  (\<not>(dp  (\<not>f) \<or>   dp  (\<not>g)))  "
   by auto
 have 3: "\<turnstile> (\<not> (dp ((\<not> f) \<or>  (\<not>g)))) = bp ( \<not>((\<not> f) \<or>  (\<not>g)))"
       using NotDpEqvBpNot by blast
 have 4: "\<turnstile> (\<not>((\<not> f) \<or>  (\<not>g))) =(f \<and> g)"
   by auto
 hence 5: "\<turnstile> bp(\<not>((\<not> f) \<or>  (\<not>g))) = bp(f \<and> g)"
    by (simp add: BpEqvBpRule)
 have 6: "\<turnstile> (\<not>(dp  (\<not>f) \<or>   dp  (\<not>g))) = ((\<not>(dp (\<not> f))) \<and> (\<not>(dp (\<not> g))))"
   by auto
 have 7: "\<turnstile> ((\<not>(dp (\<not> f))) \<and> (\<not>(dp (\<not> g)))) = (bp f \<and> bp g)"
    by (simp add: bp_d_def dp_d_def uprojection_d_def)
  show ?thesis 
  by (metis "2" "3" "4" "6" "7" inteq_reflection)
qed

lemma DpAndA:
 "\<turnstile>    dp (f \<and> g) \<longrightarrow>  dp  f "
proof -
 have 1: "\<turnstile> #True \<triangle> (f \<and> g) \<longrightarrow> #True \<triangle> f " 
   by (meson Prop12 RightProjImpProj int_iffD1 lift_and_com)
 from 1 show ?thesis by (simp add: dp_d_def)
qed

lemma BpOrA:
 "\<turnstile> bp f \<longrightarrow>   bp(f \<or> g)"
by (simp add: BpImpBpRule intI)

lemma BpOrB:
 "\<turnstile> bp g \<longrightarrow>   bp(f \<or> g)"
by (simp add: BpImpBpRule intI)

lemma BpOrImpOr:
 "\<turnstile> bp f \<or> bp g \<longrightarrow> bp(f \<or> g)"
using BpOrA BpOrB by fastforce


lemma DpAndB:
 "\<turnstile>    dp (f \<and> g) \<longrightarrow>  dp  g "
proof -
 have 1: "\<turnstile> #True \<triangle> (f \<and> g) \<longrightarrow> #True \<triangle> g " 
    by (meson Prop12 RightProjImpProj int_iffD2 lift_and_com)
 from 1 show ?thesis by (simp add: dp_d_def)
qed

lemma DpAndImpAnd:
 "\<turnstile>    dp (f \<and> g) \<longrightarrow>  dp  f \<and>  dp  g "  
proof -
 have 1: "\<turnstile>  dp (f \<and> g) \<longrightarrow>  dp  f " by (rule DpAndA)
 have 2: "\<turnstile>  dp (f \<and> g) \<longrightarrow>  dp  g " by (rule DpAndB)
 from 1 2 show ?thesis by fastforce
qed

lemma DpSkipEqvMore:
 "\<turnstile>    dp  skip  =  more "
proof -
 have 1: "\<turnstile> dp skip = #True \<triangle> skip"
  by (simp add: dp_d_def)
 have 2: "\<turnstile> #True \<triangle> skip = (#True \<and> more)"
 using PJ3 by blast
 have 3: "\<turnstile> (#True \<and> more) = more"
   by auto
 from 1 2 3 show ?thesis by fastforce
qed 

lemma DpMoreEqvMore:
 "\<turnstile>    dp  more  =  more "
by (metis DpDpEqvDp DpSkipEqvMore inteq_reflection)



lemma BpEmptyEqvEmpty:
 "\<turnstile> bp empty = empty"
by (metis DpMoreEqvMore NotDpEqvBpNot empty_d_def inteq_reflection)

lemma DpEmptyEqvEmpty:
 "\<turnstile>    dp  empty = empty"
proof -
 have 1: "\<turnstile> dp empty = #True \<triangle> empty"
  by (simp add: dp_d_def)
 have 2: "\<turnstile> #True \<triangle> empty = empty"
   by (simp add: PJ2)
 from 1 2 show ?thesis by fastforce
qed


lemma BpMoreEqvMore:
 "\<turnstile> bp more =more"
by (metis DpEmptyEqvEmpty NotDpEqvBpNot NotEmptyEqvMore inteq_reflection)

lemma NextDpImpDpNext:
 "\<turnstile>  \<circle> (dp f) \<longrightarrow> dp (\<circle> f)"
proof -
 have 1: "\<turnstile> dp(\<circle> f) = #True \<triangle> (skip;f) "
  by (simp add: dp_d_def next_d_def) 
 have 2: "\<turnstile> #True \<triangle> (skip;f) = (#True \<triangle> skip);(#True \<triangle> f)"
  by (simp add: PJ4)
 have 3: "\<turnstile> (#True \<triangle> skip) = (#True \<and> more)"
  using PJ3 by blast
 have 4: "\<turnstile> (#True \<and> more) = more"
   by auto
 have 5: "\<turnstile> skip;(#True \<triangle> f) \<longrightarrow> more;(#True \<triangle> f) "
  by (metis DpSkipEqvMore LeftChopImpChop NowImpDp inteq_reflection)
 show ?thesis  
 by (metis "2" "3" "4" "5" dp_d_def inteq_reflection next_d_def)
qed

lemma BoxStateImportBp:
 "\<turnstile> \<box>(init w)  \<longrightarrow> bp (\<box>(init w))"
by (simp add: Valid_def always_defs init_defs projection_d_def bp_d_def uprojection_d_def 
       powerinterval_def)
    (metis (mono_tags, lifting) filt_expand interval_idx_bound_1 interval_suffix_zero)

lemma BoxStateEqvBpBoxState:
 "\<turnstile> \<box> (init w) = bp(\<box> (init w))"
proof -
 have 1: "\<turnstile>  bp(\<box> (init w)) \<longrightarrow> \<box> (init w)" 
  by (simp add: BpElim)
 have 2: "\<turnstile>  bp(\<box> (init w)) = (\<not>(#True \<triangle> (\<not> \<box> (init w))))"
   by (simp add: bp_d_def uprojection_d_def)
 have 2: "\<turnstile> \<box>(init w) \<longrightarrow> dp (\<box>(init w))"
    by (metis NowImpDp)
 have 2: "\<turnstile> \<box> (init w) \<longrightarrow> bp(\<box> (init w))"
 using BoxStateImportBp by auto
 from 1 2 show ?thesis by fastforce
qed



end
